@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2008, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -42,9 +42,6 @@ If you want  to  debug  this  module,  put  a  '$:-'(trace).   directive
 somewhere.   The  tracer will work properly under boot compilation as it
 will use the C defined write predicate  to  print  goals  and  does  not
 attempt to call the Prolog defined trace interceptor.
-
-Please note that of version  3.3,  $:-   is  no  longer  an operator and
-therefore you have to use braces.
 */
 
 '$:-'(format('Loading boot file ...~n', [])).
@@ -59,13 +56,27 @@ therefore you have to use braces.
 		*          DIRECTIVES           *
 		*********************************/
 
-'$set_pattr'([], _) :- !.
-'$set_pattr'([H|T], Attr) :- !,		% ISO
-	'$set_pattr'(H, Attr),
-	'$set_pattr'(T, Attr).
-'$set_pattr'((A,B), Attr) :- !,		% ISO and traditional
-	'$set_pattr'(A, Attr),
-	'$set_pattr'(B, Attr).
+:- meta_predicate
+	dynamic(:),
+	multifile(:),
+	module_transparent(:),
+	discontiguous(:),
+	volatile(:),
+	thread_local(:),
+	noprofile(:),
+	'$iso'(:),
+	'$hide'(:).
+
+'$set_pattr'(_:X, _) :-
+	var(X),
+	throw(error(instantiation_error, _)).
+'$set_pattr'(_:[], _) :- !.
+'$set_pattr'(M:[H|T], Attr) :- !,		% ISO
+	'$set_pattr'(M:H, Attr),
+	'$set_pattr'(M:T, Attr).
+'$set_pattr'(M:(A,B), Attr) :- !,		% ISO and traditional
+	'$set_pattr'(M:A, Attr),
+	'$set_pattr'(M:B, Attr).
 '$set_pattr'(A, Attr) :-
 	'$set_predicate_attribute'(A, Attr, 1).
 
@@ -78,28 +89,12 @@ thread_local(Spec)	 :- '$set_pattr'(Spec, (thread_local)).
 noprofile(Spec)		 :- '$set_pattr'(Spec, (noprofile)).
 '$iso'(Spec)		 :- '$set_pattr'(Spec, (iso)).
 
-:- module_transparent
-	'$set_pattr'/2,
-	(dynamic)/1,
-	(multifile)/1,
-	(module_transparent)/1,
-	(discontiguous)/1,
-	(volatile)/1,
-	(thread_local)/1,
-	(noprofile)/1,
-	'$hide'/2,
-	'$iso'/1.
-
-
-		/********************************
-		*        TRACE BEHAVIOUR        *
-		*********************************/
-
-%	'$hide'(+Name, +Arity)
+%%	'$hide'(:PI)
+%
 %	Predicates protected this way are never visible in the tracer.
 
-'$hide'(Name, Arity) :-
-	'$set_predicate_attribute'(Name/Arity, trace, 0).
+'$hide'(Pred) :-
+	'$set_predicate_attribute'(Pred, trace, 0).
 
 
 		/********************************
@@ -112,39 +107,43 @@ noprofile(Spec)		 :- '$set_pattr'(Spec, (noprofile)).
 	      ignore/1,
 	      call_cleanup/2,
 	      call_cleanup/3,
-	      setup_and_call_cleanup/3,
-	      setup_and_call_cleanup/4,
+	      setup_call_cleanup/3,
+	      setup_call_catcher_cleanup/4,
 	      (^)/2)).
 
+:- meta_predicate
+	';'(0,0),
+	'|'(0,0),
+	','(0,0),
+	call(0),
+	call(1,?),
+	call(2,?,?),
+	call(3,?,?,?),
+	call(4,?,?,?,?),
+	call(5,?,?,?,?,?),
+	call(6,?,?,?,?,?,?),
+	call(7,?,?,?,?,?,?,?),
+	^(+,0),
+	not(0),
+	\+(0),
+	'->'(0,0),
+	'*->'(0,0),
+	once(0),
+	ignore(0),
+	block(+,0,-),
+	catch(0,?,0),
+	setup_call_cleanup(0,0,0),
+	setup_call_catcher_cleanup(0,0,?,0),
+	call_cleanup(0,0),
+	call_cleanup(0,?,0).
+
+/* Currently, meta_predicate is only supported upto arity 8
+*/
+
 :- module_transparent
-	';'/2,
-	'|'/2,
-	','/2,
-	call/1,
-	call/2,
-	call/3,
-	call/4,
-	call/5,
-	call/6,
-	call/7,
-	call/8,
 	call/9,
 	call/10,
-	call/11,
-	(^)/2,
-	(not)/1,
-	(\+)/1,
-	(->)/2,
-	(*->)/2,
-	once/1,
-	ignore/1,
-	block/3,
-	catch/3,
-	setup_and_call_cleanup/3,
-	setup_and_call_cleanup/4,
-	call_cleanup/2,
-	call_cleanup/3,
-	apply/2.
+	call/11.
 
 :- '$iso'((call/1, (\+)/1, once/1, (;)/2, (,)/2, (->)/2, catch/3)).
 
@@ -211,17 +210,16 @@ ignore(Goal) :-
 	Goal, !.
 ignore(_Goal).
 
-apply(Pred, Arguments) :-
-	'$apply'(Pred, Arguments).		% handled by the compiler
-
 _Var^Goal :-					% setof/3, bagof/3
 	Goal.
 
+false :-					% SICStus compatibility
+	fail.
+
 %	block/3, !/1, exit/2, fail/1
 %	`longjmp' like control-structures.  See manual.  The predicate
 %	system:block/3 is used by the VMI's I_CUT_BLOCK and B_EXIT.
-%	'$exit' and '$cut' are interpreted by the compiler/decompiler,
-%	just like '$apply'/2.
+%	'$exit' and '$cut' are interpreted by the compiler/decompiler.
 
 block(_Label, Goal, _RVal) :-
 	Goal.
@@ -243,26 +241,26 @@ fail(Label) :-
 catch(_Goal, _Catcher, _Recover) :-
 	'$catch'.
 
-%%	setup_and_call_cleanup(:Setup, :Goal, :Cleanup).
-%%	setup_and_call_cleanup(:Setup, :Goal, +Catcher, :Cleanup).
+%%	setup_call_cleanup(:Setup, :Goal, :Cleanup).
+%%	setup_call_catcher_cleanup(:Setup, :Goal, +Catcher, :Cleanup).
 %%	call_cleanup(:Goal, :Cleanup)
 %%	call_cleanup(:Goal, +Catcher, :Cleanup)
 %
-%	Call Cleanup as Goal finished (deterministic success, failure,
+%	Call Cleanup once after Goal is finished (deterministic success, failure,
 %	exception or cut).  '$call_cleanup' translated to I_CALLCLEANUP.
 
-setup_and_call_cleanup(Setup, _Goal, _Catcher, _Cleanup) :-
+setup_call_catcher_cleanup(Setup, _Goal, _Catcher, _Cleanup) :-
 	Setup,
 	'$call_cleanup'.
 
-setup_and_call_cleanup(Setup, Goal, Cleanup) :-
-	setup_and_call_cleanup(Setup, Goal, _Catcher, Cleanup).
+setup_call_cleanup(Setup, Goal, Cleanup) :-
+	setup_call_catcher_cleanup(Setup, Goal, _Catcher, Cleanup).
 
 call_cleanup(Goal, Cleanup) :-
-	setup_and_call_cleanup(true, Goal, _Catcher, Cleanup).
+	setup_call_catcher_cleanup(true, Goal, _Catcher, Cleanup).
 
 call_cleanup(Goal, Catcher, Cleanup) :-
-	setup_and_call_cleanup(true, Goal, Catcher, Cleanup).
+	setup_call_catcher_cleanup(true, Goal, Catcher, Cleanup).
 
 
 		/********************************
@@ -311,13 +309,15 @@ call_cleanup(Goal, Catcher, Cleanup) :-
 	;   Level = silent
 	),
 	print_message(Level, autoload(Module:Name/Arity, Library)),
+	flag('$compiling', OldComp, database),
 	(   Module == LoadModule
 	->  ensure_loaded(Module:Library)
-	;   (   '$c_current_predicate'(_, LoadModule:Head)
+	;   (   '$get_predicate_attribute'(LoadModule:Head, defined, 1)
 	    ->	Module:import(LoadModule:Head)
 	    ;	use_module(Module:Library, [Name/Arity])
 	    )
 	),
+	flag('$compiling', _, OldComp),
 	flag('$autoloading', _, Old),
 	'$c_current_predicate'(_, Module:Head).
 
@@ -327,7 +327,7 @@ call_cleanup(Goal, Catcher, Cleanup) :-
 	;   true
 	).
 
-:- '$hide'('$calleventhook', 1).
+:- '$hide'('$calleventhook'/1).
 
 %	 handle debugger 'w', 'p' and <N> depth options.
 
@@ -376,12 +376,12 @@ call_cleanup(Goal, Catcher, Cleanup) :-
 	    ;	print_message(help, query(confirm)),
 		fail
 	    ).
-	    
+
 :- dynamic
 	user:portray/1.
 :- multifile
 	user:portray/1.
-	
+
 
 		 /*******************************
 		 *	 FILE_SEARCH_PATH	*
@@ -436,7 +436,7 @@ expand_file_search_path(Spec, Expanded) :-
 	'$make_path'(Exp1, Base, Expanded).
 '$expand_file_search_path'(A/B, Expanded, _, _) :- !,
 	'$make_path_from_slash'(A/B, Parts, []),
-	concat_atom(Parts, Expanded).
+	atomic_list_concat(Parts, Expanded).
 '$expand_file_search_path'(Spec, Spec, _, _) :-
 	atomic(Spec).
 
@@ -446,7 +446,7 @@ expand_file_search_path(Spec, Expanded) :-
 '$make_path'(Dir, File, Path) :-
 	'$make_path_from_slash'(File, Parts, []),
 	'$append'([Dir, /], Parts, AllParts),
-	'$concat_atom'(AllParts, Path).
+	atomic_list_concat(AllParts, Path).
 
 
 %%	'$make_path_from_slash'(+SlashPath, -Parts, ?Tail) is det.
@@ -466,7 +466,7 @@ expand_file_search_path(Spec, Expanded) :-
 
 %%	absolute_file_name(+Term, +Args, -AbsoluteFile) is nondet.
 %%	absolute_file_name(+Term, -AbsoluteFile, +Args) is nondet.
-%	
+%
 %	Translate path-specifier into a full   path-name. This predicate
 %	originates from Quintus was introduced  in SWI-Prolog very early
 %	and  has  re-appeared  in  SICStus  3.9.0,  where  they  changed
@@ -573,7 +573,7 @@ user:prolog_file_type(Ext,	executable) :-
 	'$absolute_file_name'(Extended, FullName).
 '$dochk_file'(File, Exts, Cond, FullName) :-
 	'$relative_to'(Cond, source, Dir),
-	'$concat_atom'([Dir, /, File], AbsFile),
+	atomic_list_concat([Dir, /, File], AbsFile),
 	'$extend_file'(AbsFile, Exts, Extended),
 	'$file_condition'(Cond, Extended), !,
 	'$absolute_file_name'(Extended, FullName).
@@ -583,7 +583,7 @@ user:prolog_file_type(Ext,	executable) :-
 	'$absolute_file_name'(Extended, FullName).
 
 %	'$relative_to'(+Condition, +Default, -Dir)
-%	
+%
 %	Determine the directory to work from.  This can be specified
 %	explicitely using one or more relative_to(FileOrDir) options
 %	or implicitely relative to the working directory or current
@@ -633,10 +633,10 @@ user:prolog_file_type(Ext,	executable) :-
 '$search_message'(_).
 
 
-	
+
 %	'$file_condition'(+Condition, +Path)
 %
-%	Verify Path satisfies Condition. 
+%	Verify Path satisfies Condition.
 
 '$file_condition'([], _) :- !.
 '$file_condition'([H|T], File) :- !,
@@ -660,7 +660,7 @@ user:prolog_file_type(Ext,	executable) :-
 	'$ensure_extensions'(Exts, File, Fs),
 	'$list_to_set'(Fs, FsSet),
 	'$member'(FileEx, FsSet).
-	
+
 '$ensure_extensions'([], _, []).
 '$ensure_extensions'([E|E0], F, [FE|E1]) :-
 	file_name_extension(F, E, FE),
@@ -668,7 +668,7 @@ user:prolog_file_type(Ext,	executable) :-
 
 '$list_to_set'([], []).
 '$list_to_set'([H|T], R) :-
-	memberchk(H, T), !, 
+	memberchk(H, T), !,
 	'$list_to_set'(T, R).
 '$list_to_set'([H|T], [H|R]) :-
 	'$list_to_set'(T, R).
@@ -725,8 +725,8 @@ compiling :-
 	       flag('$directive', database, database)
 	   ).
 
-:- module_transparent
-	'$ifcompiling'/1.
+:- meta_predicate
+	'$ifcompiling'(0).
 
 '$ifcompiling'(_) :-
 	flag('$compiling', database, database), !.
@@ -740,33 +740,38 @@ compiling :-
 preprocessor(Old, New) :-
 	flag('$preprocessor', Old, New).
 
-'$set_encoding'(default, _) :- !.
-'$set_encoding'(Encoding, Stream) :-
-	set_stream(Stream, encoding(Encoding)).
+'$set_encoding'(Stream, Options) :-
+	memberchk(encoding(Enc), Options),
+	Enc \== default,
+	set_stream(Stream, encoding(Enc)).
+'$set_encoding'(_, _).
 
-'$open_source'(stream(Id, In), Enc, In, Goal) :- !,
+
+%%	'$open_source'(+Spec, -In, :Goal, +Options) is semidet.
+
+'$open_source'(stream(Id, In), In, Goal, Options) :- !,
 	'$push_input_context',
-	'$set_encoding'(Enc, In),
+	'$set_encoding'(In, Options),
 	set_stream(In, file_name(Id)),
 	set_stream(In, record_position(true)),
 	'$open_source_call'(Id, In, Goal, True),
 	'$pop_input_context',
 	True == yes.
-'$open_source'(File, Enc, In, Goal) :-
+'$open_source'(File, In, Goal, Options) :-
 	preprocessor(none, none), !,
 	'$push_input_context',
 	open(File, read, In),
-	'$set_encoding'(Enc, In),
+	'$set_encoding'(In, Options),
 	'$open_source_call'(File, In, Goal, True),
 	close(In),
 	'$pop_input_context',
 	True == yes.
-'$open_source'(File, Enc, In, Goal) :-
+'$open_source'(File, In, Goal, Options) :-
 	preprocessor(Pre, Pre),
 	(   '$substitute_atom'('%f', File, Pre, Command)
 	->  '$push_input_context',
 	    open(pipe(Command), read, In),
-	    '$set_encoding'(Enc, In),
+	    '$set_encoding'(In, Options),
 	    '$open_source_call'(File, In, Goal, True),
 	    close(In),
 	    '$pop_input_context',
@@ -801,7 +806,7 @@ preprocessor(Old, New) :-
 	sub_atom(In, B, _, A, From),
 	sub_atom(In, 0, B, _, Before),
 	sub_atom(In, _, A, 0, After),
-	concat_atom([Before, To, After], Out).
+	atomic_list_concat([Before, To, After], Out).
 
 
 		 /*******************************
@@ -827,20 +832,19 @@ preprocessor(Old, New) :-
 		*       LOAD PREDICATES         *
 		*********************************/
 
-:- module_transparent
-	ensure_loaded/1,
-	'.'/2,
-	consult/1,
-	use_module/1,
-	use_module/2,
-	reexport/1,
-	reexport/2,
-	'$load_file'/3,
-	load_files/1,
-	load_files/2.
+:- meta_predicate
+	ensure_loaded(:),
+	[:|+],
+	consult(:),
+	use_module(:),
+	use_module(:, +),
+	reexport(:),
+	reexport(:, +),
+	load_files(:),
+	load_files(:, +).
 
 %	ensure_loaded(+File|+ListOfFiles)
-%	
+%
 %	Load specified files, provided they where not loaded before. If the
 %	file is a module file import the public predicates into the context
 %	module.
@@ -849,7 +853,7 @@ ensure_loaded(Files) :-
 	load_files(Files, [if(not_loaded)]).
 
 %	use_module(+File|+ListOfFiles)
-%	
+%
 %	Very similar to ensure_loaded/1, but insists on the loaded file to
 %	be a module file. If the file is already imported, but the public
 %	predicates are not yet imported into the context module, then do
@@ -861,7 +865,7 @@ use_module(Files) :-
 			  ]).
 
 %	use_module(+File, +ImportList)
-%	
+%
 %	As use_module/1, but takes only one file argument and imports only
 %	the specified predicates rather than all public predicates.
 
@@ -895,35 +899,35 @@ reexport(File, Import) :-
 
 [X] :- !,
 	consult(X).
-[F|R] :-
-	consult([F|R]).
+[M:F|R] :-
+	consult(M:[F|R]).
 [].
 
-consult(X) :-
+consult(M:X) :-
 	X == user, !,
 	flag('$user_consult', N, N+1),
 	NN is N + 1,
 	atom_concat('user://', NN, Id),
-	load_files(Id, [stream(user_input)]).
+	load_files(M:Id, [stream(user_input)]).
 consult(List) :-
 	load_files(List).
 
-%	'$consult_goal'(+Path, -Goal)
+%%	'$consult_goal'(+Path, -Goal)
 %
 %	Determine how to load the indicated file
 
 '$consult_goal'(Path, Goal) :-
-	(   file_name_extension(_, Ext, Path),
-	    user:prolog_file_type(Ext, qlf)
-	->  Goal = '$qload_file'
-	;   Goal = '$consult_file'
+	file_name_extension(_, Ext, Path),
+	(   user:prolog_file_type(Ext, qlf)
+	->  Goal = system:'$qload_file'
+	;   Goal = system:'$consult_file'
 	).
 
-%	load_files(+File, +Options)
-%	
+%%	load_files(:File, +Options)
+%
 %	Common entry for all the consult derivates.  File is the raw user
 %	specified file specification, possibly tagged with the module.
-%	
+%
 %	`Options' is a list of additional options.  Defined values are
 %
 %	    verbose		Print statistics on user channel
@@ -932,9 +936,8 @@ consult(List) :-
 
 load_files(Files) :-
 	load_files(Files, []).
-load_files(Files, Options) :-
-	strip_module(Files, Module, TheFiles),
-        with_mutex('$load', '$load_files'(TheFiles, Module, Options)).
+load_files(Module:Files, Options) :-
+        with_mutex('$load', '$load_files'(Files, Module, Options)).
 
 '$load_files'(Id, Module, Options) :-	% load_files(foo, [stream(In)])
 	memberchk(stream(_), Options), !,
@@ -961,9 +964,13 @@ load_files(Files, Options) :-
 
 
 '$get_option'(Term, Options, Default) :-
-	(   memberchk(Term, Options)
-	->  true
-	;   arg(1, Term, Default)
+	arg(1, Term, Value),
+	functor(Term, Name, 1),
+	functor(Gen, Name, 1),
+	arg(1, Gen, GVal),
+	(   memberchk(Gen, Options)
+	->  Value = GVal
+	;   Value = Default
 	).
 
 
@@ -984,7 +991,7 @@ load_files(Files, Options) :-
 %
 %	Return the QLF file if it exists.  Might check for modification
 %	time, version, etc.
-%	
+%
 %	If the user-specification specified a prolog file, do not
 %	replace this with a .qlf file.
 
@@ -1016,7 +1023,7 @@ load_files(Files, Options) :-
 '$load_file'(File, Module, Options) :-
 	statistics(heapused, OldHeap),
 	statistics(cputime, OldTime),
- 
+
 	(   memberchk(stream(FromStream), Options)
 	->  true
 	;   absolute_file_name(File,
@@ -1025,10 +1032,9 @@ load_files(Files, Options) :-
 			       ],
 			       FullFile)
 	),
-	    
+
 	'$get_option'(imports(Import), Options, all),
 	'$get_option'(reexport(Reexport), Options, false),
-	'$get_option'(must_be_module(IsModule), Options, false),
 	current_prolog_flag(verbose_load, DefVerbose),
 	'$negate'(DefVerbose, DefSilent),
 	'$get_option'(silent(Silent), Options, DefSilent),
@@ -1037,7 +1043,7 @@ load_files(Files, Options) :-
 	'$get_option'(if(If), Options, true),
 	'$get_option'(autoload(Autoload), Options, false),
 	'$get_option'(derived_from(DerivedFrom), Options, -),
-	'$get_option'(encoding(Encoding), Options, default),
+
 	current_prolog_flag(generate_debug_info, DebugInfo),
 
 	(   Autoload == false
@@ -1072,14 +1078,16 @@ load_files(Files, Options) :-
 			     load_file(start(Level,
 					     file(File, Absolute)))),
 	    (   nonvar(FromStream),
-	        '$consult_file'(stream(Absolute, FromStream), Encoding,
-				Module, IsModule, Action, LM)
+		(   '$get_option'(format(qlf), Options, source)
+		->  set_stream(FromStream, file_name(Absolute)),
+		    '$qload_stream'(FromStream, Module, Action, LM, Options)
+		;   '$consult_file'(stream(Absolute, FromStream),
+				    Module, Action, LM, Options)
+		)
 	    ->	true
 	    ;   var(FromStream),
 		'$consult_goal'(Absolute, Goal),
-		call(Goal,
-		     Absolute, Encoding,
-		     Module, IsModule, Action, LM)
+		call(Goal, Absolute, Module, Action, LM, Options)
 	    ->  true
 	    ;   print_message(error, load_file(failed(File))),
 		fail
@@ -1130,19 +1138,23 @@ load_files(Files, Options) :-
 	print_message(error, E),
 	fail.
 
-%%	'$consult_file'(+Path, +Encoding, +Module, +IsModule, -Action, -LoadedIn)
+%%	'$consult_file'(+Path, +Module, -Action, -LoadedIn, +Options)
+%
+%	Called  from  '$load_file'/3  using   the    goal   returned  by
+%	'$consult_goal'/2. This means that the  calling conventions must
+%	be kept synchronous with '$qload_file'/6.
 
-'$consult_file'(Absolute, Enc, Module, IsModule, What, LM) :-
+'$consult_file'(Absolute, Module, What, LM, Options) :-
 	'$set_source_module'(Module, Module), !, % same module
-	'$consult_file_2'(Absolute, Enc, Module, IsModule, What, LM).
-'$consult_file'(Absolute, Enc, Module, IsModule, What, LM) :-
+	'$consult_file_2'(Absolute, Module, What, LM, Options).
+'$consult_file'(Absolute, Module, What, LM, Options) :-
 	'$set_source_module'(OldModule, Module),
 	'$ifcompiling'('$qlf_start_sub_module'(Module)),
-        '$consult_file_2'(Absolute, Enc, Module, IsModule, What, LM),
+        '$consult_file_2'(Absolute, Module, What, LM, Options),
 	'$ifcompiling'('$qlf_end_part'),
 	'$set_source_module'(_, OldModule).
 
-'$consult_file_2'(Absolute, Enc, Module, IsModule, What, LM) :-
+'$consult_file_2'(Absolute, Module, What, LM, Options) :-
 	'$set_source_module'(OldModule, Module),	% Inform C we start loading
 	'$load_id'(Absolute, Id),
 	'$start_consult'(Id),
@@ -1153,10 +1165,11 @@ load_files(Files, Options) :-
 	),
 	'$assert_load_context_module'(Id, OldModule),
 
-	
+
 	'$save_lex_state'(LexState),
-	'$open_source'(Absolute, Enc, In,
-		     '$load_file'(In, Id, IsModule, LM)),
+	'$open_source'(Absolute, In,
+		       '$load_file'(In, Id, LM, Options),
+		       Options),
 	'$restore_lex_state'(LexState),
 	'$set_source_module'(_, OldModule).	% Restore old module
 
@@ -1168,17 +1181,21 @@ load_files(Files, Options) :-
 %	Switch the currently expected dialect.
 
 '$set_dialect'(Old, New) :-
-	Old == New, !,
-	(   '$current_load_dialect'(Dialect)
+	'$my_dialect'(Old),
+	(   Old == New
 	->  true
-	;   Dialect = swi
+	;   retractall('$current_load_dialect'(_)),
+	    (	New == swi
+	    ->	true
+	    ;	assert('$current_load_dialect'(New))
+	    )
+	).
+
+'$my_dialect'(Current) :-
+	(   '$current_load_dialect'(Dialect)
+	->  Current = Dialect
+	;   Current = swi
 	).
-'$set_dialect'(Old, Dialect) :-
-	(   retract('$current_load_dialect'(Old))
-	->  true
-	;   Old = swi
-	),
-	assert('$current_load_dialect'(Dialect)).
 
 '$save_lex_state'(lexstate(Style, Dialect)) :-
 	'$style_check'(Style, Style),
@@ -1201,9 +1218,9 @@ load_files(Files, Options) :-
 	).
 
 %%	'$load_context_module'(+File, -Module)
-%	
+%
 %	Record the module a file was loaded from (see make/0)
-%	
+%
 %	@tbd	Should also know which predicates are imported!
 
 :- dynamic
@@ -1216,34 +1233,37 @@ load_files(Files, Options) :-
 	).
 
 
-%   '$load_file'(+In, +Path, +IsModule, -Module)
+%   '$load_file'(+In, +Path, -Module, +Options)
 %
 %   '$load_file'/4 does the actual loading.
 
-'$load_file'(In, File, IsModule, Module) :-
+'$load_file'(In, File, Module, Options) :-
 	(   peek_char(In, #)
 	->  skip(In, 10)
 	;   true
 	),
 	'$read_clause'(In, First),
-	'$load_file'(First, In, File, IsModule, Module).
+	'$load_file'(First, In, File, Module, Options).
 
 
-'$load_file'((?- module(Module, Public)), In, File, _, Module) :- !,
-	'$load_module'(Module, Public, In, File).
-'$load_file'((:- module(Module, Public)), In, File, _, Module) :- !,
-	'$load_module'(Module, Public, In, File).
+'$load_file'((?- module(Module, Public)), In, File, Module, Options) :- !,
+	'$load_module'(Module, Public, In, File, Options).
+'$load_file'((:- module(Module, Public)), In, File, Module, Options) :- !,
+	'$load_module'(Module, Public, In, File, Options).
 '$load_file'((?- module(Module, Public)),
-	     In, File, _, Module) :- !,
-	'$load_module'(Module, Public, In, File).
+	     In, File, Module, Options) :- !,
+	'$load_module'(Module, Public, In, File, Options).
 '$load_file'((:- module(Module, Public)),
-	     In, File, _, Module) :- !,
-	'$load_module'(Module, Public, In, File).
-'$load_file'(_, _, File, true, _) :- !,
+	     In, File, Module, Options) :- !,
+	'$load_module'(Module, Public, In, File, Options).
+'$load_file'(_, _, File, _, Options) :-
+	'$get_option'(must_be_module(true), Options, false), !,
 	throw(error(domain_error(module_file, File), _)).
-'$load_file'(end_of_file, _, _, _, Module) :- !, 	% empty file
-	'$set_source_module'(Module, Module).
-'$load_file'(FirstClause, In, File, false, Module) :- !,
+'$load_file'(end_of_file, _, File, Module, _) :- !, 	% empty file
+	'$set_source_module'(Module, Module),
+	'$ifcompiling'('$qlf_start_file'(File)),
+	'$ifcompiling'('$qlf_end_part').
+'$load_file'(FirstClause, In, File, Module, _Options) :- !,
 	'$set_source_module'(Module, Module),
 	'$ifcompiling'('$qlf_start_file'(File)),
 	ignore('$consult_clause'(FirstClause, File)),
@@ -1253,20 +1273,41 @@ load_files(Files, Options) :-
 '$reserved_module'(system).
 '$reserved_module'(user).
 
-'$load_module'(Reserved, _, _, _) :-
+%%	'$load_module'(+Module, +Public, +Stream, +File, +Options)
+%
+%	Options processed:
+%
+%		* redefine_module(Action)
+%		Action is one of =true=, =false= or =ask=.
+%
+%	Redefining a module by loading another file must be more subtle.
+%	Verify the compatibility of the interface could be one example.
+
+'$load_module'(Reserved, _, _, _, _) :-
 	'$reserved_module'(Reserved), !,
 	throw(error(permission_error(load, module, Reserved), _)).
-'$load_module'(Module, Public, In, File) :-
+'$load_module'(Module, Public, In, File, Options) :-
 	'$set_source_module'(OldModule, OldModule),
 	source_location(_File, Line),
-	'$declare_module'(Module, File, Line),
+	'$get_option'(redefine_module(Action), Options, false),
+	'$declare_module'(Module, File, Line, Action),
 	'$export_list'(Public, Module, Ops),
 	'$ifcompiling'('$qlf_start_module'(Module)),
 	'$export_ops'(Ops, Module, File),
 	'$consult_stream'(In, File),
-	Module:'$check_export',
+	'$check_export'(Module),
 	'$ifcompiling'('$qlf_end_part').
 
+'$check_export'(Module) :-
+	'$undefined_export'(Module, UndefList),
+	(   '$member'(Undef, UndefList),
+	    strip_module(Undef, _, Local),
+	    print_message(error,
+			  undefined_export(Module, Local)),
+	    fail
+	;   true
+	).
+
 
 %%	'$import_list'(+TargetModule, +FromModule, +Import, +Reexport) is det.
 %
@@ -1278,11 +1319,11 @@ load_files(Files, Options) :-
 	var(Var), !,
 	throw(error(instantitation_error, _)).
 '$import_list'(Target, Source, all, Reexport) :- !,
-	export_list(Source, Import),
+	'$module_property'(Source, exports(Import)),
 	'$import_ops'(Target, Source),
 	'$import_list'(Target, Source, Import, Reexport).
 '$import_list'(Target, Source, except(Spec), Reexport) :- !,
-	export_list(Source, Export),
+	'$module_property'(Source, exports(Export)),
 	(   is_list(Spec)
 	->  true
 	;   throw(error(type_error(list, Spec), _))
@@ -1360,7 +1401,7 @@ load_files(Files, Options) :-
 	'$list_to_conj'(T, Rest).
 
 %%	'$import_ops'(+Target, +Source)
-%	
+%
 %	Import the operators export from Source into the module table of
 %	Target.
 
@@ -1374,7 +1415,7 @@ load_files(Files, Options) :-
 
 
 %%	'$export_list'(+Declarations, +Module, -Ops)
-%	
+%
 %	Handle the export list of the module declaration for Module
 %	associated to File.
 
@@ -1419,7 +1460,7 @@ load_files(Files, Options) :-
 		  fail
 	      )), !.
 
-	      
+
 '$consult_stream2'(In, File) :-
 	repeat,
 	    '$read_clause'(In, Clause),
@@ -1433,34 +1474,42 @@ load_files(Files, Options) :-
 	       E,
 	       '$print_message_fail'(E)).
 
-'$execute_directive'(include(File), F) :- !,
+%%	'$execute_directive'(:Goal, +F) is det.
+%
+%	Execute the argument of :- or ?- while loading a file.
+
+'$execute_directive'(Goal, F) :-
+	expand_goal(Goal, Goal1),
+	'$execute_directive_2'(Goal1, F).
+
+'$execute_directive_2'(include(File), F) :- !,
 	'$expand_include'(File, F).
-'$execute_directive'(encoding(Encoding), F) :- !,
+'$execute_directive_2'(encoding(Encoding), F) :- !,
 	source_location(F, _),
 	'$load_input'(F, S),
 	set_stream(S, encoding(Encoding)).
-'$execute_directive'(ISO, F) :-
+'$execute_directive_2'(ISO, F) :-
 	'$expand_directive'(ISO, Normal), !,
 	'$execute_directive'(Normal, F).
-'$execute_directive'(Goal, _) :-
+'$execute_directive_2'(Goal, _) :-
 	\+ flag('$compiling', database, database), !,
 	'$add_directive_wic2'(Goal, Type),
 	(   Type == call		% suspend compiling into .qlf file
 	->  flag('$compiling', Old, database),
 	    flag('$directive', OldDir, Old),
-	    call_cleanup('$execute_directive2'(Goal),
+	    call_cleanup('$execute_directive_3'(Goal),
 			 (   flag('$compiling', _, Old),
 			     flag('$directive', _, OldDir)
 			 ))
-	;   '$execute_directive2'(Goal)
+	;   '$execute_directive_3'(Goal)
 	).
-'$execute_directive'(Goal, _) :-
-	'$execute_directive2'(Goal).
+'$execute_directive_2'(Goal, _) :-
+	'$execute_directive_3'(Goal).
 
-'$execute_directive2'(Goal) :-
+'$execute_directive_3'(Goal) :-
 	'$set_source_module'(Module, Module),
 	catch(Module:Goal, Term, '$exception_in_directive'(Term)), !.
-'$execute_directive2'(Goal) :-
+'$execute_directive_3'(Goal) :-
 	'$set_source_module'(Module, Module),
 	print_message(warning, goal_failed(directive, Module:Goal)),
 	fail.
@@ -1507,7 +1556,7 @@ load_files(Files, Options) :-
 	->  true
 	;   print_message(error, mixed_directive(Goal))
 	).
-	
+
 '$common_goal_type'((A,B), Type) :- !,
 	'$common_goal_type'(A, Type),
 	'$common_goal_type'(B, Type).
@@ -1672,7 +1721,7 @@ compile_aux_clauses(Clauses) :-
 '$if_expansion'((:- else), []) :-
 	(   retract('$include_code'(X))
 	->  (   X == true
-	    ->  X2 = false 
+	    ->  X2 = false
 	    ;   X == false
 	    ->	X2 = true
 	    ;	X2 = X
@@ -1689,7 +1738,8 @@ compile_aux_clauses(Clauses) :-
 
 '$eval_if'(G) :-
 	expand_goal(G, G2),
-	G2.
+	'$set_source_module'(Module, Module),
+	Module:G2.
 
 
 		 /*******************************
@@ -1816,10 +1866,13 @@ expand_goal(A, B) :-
 '$do_expand_body'(V^G, V^EG) :- !,
         '$do_expand_body'(G, EG).
 '$do_expand_body'(M:G, M:EG) :-
-	atom(M),
-	'$set_source_module'(Old, M),
-	call_cleanup('$do_expand_body'(G, EG),
-		     '$set_source_module'(_, Old)), !.
+	atom(M), !,
+	(   M == system			% or should we define this:
+	->  EG = G			% system:goal_expansion(X,X)
+	;   '$set_source_module'(Old, M),
+	    call_cleanup('$do_expand_body'(G, EG),
+			 '$set_source_module'(_, Old))
+	).
 '$do_expand_body'(A, A).
 
 %	Delete extraneous true's that result from goal_expansion(..., true)
@@ -1999,7 +2052,7 @@ then the call p([a], [a]) will succeed, which is quite definitely wrong.
 
 
 %	'$extend'(+Head, +Extra1, +Extra2, -NewHead)
-%	
+%
 %	Extend Head with two more arguments (on behalf DCG compilation).
 %	The solution below is one option. Using   =..  and append is the
 %	alternative. In the current version (5.3.2), the =.. is actually
@@ -2053,9 +2106,9 @@ then the call p([a], [a]) will succeed, which is quite definitely wrong.
 	'$copy_args'(I2, Arity, Old, New).
 '$copy_args'(_, _, _, _).
 
-:- module_transparent
-	phrase/2,
-	phrase/3.
+:- meta_predicate
+	phrase(2, ?),
+	phrase(2, ?, ?).
 :- noprofile((phrase/2,
 	      phrase/3)).
 
@@ -2138,7 +2191,7 @@ saved state.
 	'$reverse'(List1, [Head|List2], List3).
 
 '$delete'([], _, []) :- !.
-'$delete'([Elem|Tail], Elem, Result) :- !, 
+'$delete'([Elem|Tail], Elem, Result) :- !,
 	'$delete'(Tail, Elem, Result).
 '$delete'([Head|Tail], Elem, [Head|Rest]) :-
 	'$delete'(Tail, Elem, Rest).
@@ -2167,14 +2220,13 @@ halt :-
 	halt(0).
 
 
-:- module_transparent
-	at_halt/1.
+:- meta_predicate
+	at_halt(0).
 :- dynamic
 	'$at_halt'/1.
 
-at_halt(Spec) :-
-	strip_module(Spec, Module, Goal),
-	asserta(system:'$at_halt'(Module:Goal)).
+at_halt(Goal) :-
+	asserta('$at_halt'(Goal)).
 
 '$run_at_halt' :-
 	(   '$at_halt'(Goal),
@@ -2188,11 +2240,11 @@ at_halt(Spec) :-
 		*      LOAD OTHER MODULES       *
 		*********************************/
 
-:- module_transparent
-	'$load_wic_files'/2,
-	'$load_additional_boot_files'/0.
+:- meta_predicate
+	'$load_wic_files'(:).
 
-'$load_wic_files'(Module, Files) :-
+'$load_wic_files'(Files) :-
+	Files = Module:_,
 	'$execute_directive'('$set_source_module'(OldM, Module), []),
 	'$save_lex_state'(LexState),
 	'$style_check'(_, 2'1111),
@@ -2203,12 +2255,17 @@ at_halt(Spec) :-
 	flag('$compiling', _, OldC).
 
 
+%%	'$load_additional_boot_files' is det.
+%
+%	Called from compileFileList() in pl-wic.c.   Gets the files from
+%	"-c file ..." and loads them into the module user.
+
 '$load_additional_boot_files' :-
 	current_prolog_flag(argv, Argv),
 	'$get_files_argv'(Argv, Files),
 	(   Files \== []
 	->  format('Loading additional boot files~n'),
-	    '$load_wic_files'(user, Files),
+	    '$load_wic_files'(user:Files),
 	    format('additional boot files loaded~n')
 	;   true
         ).
@@ -2217,10 +2274,10 @@ at_halt(Spec) :-
        source_location(File, _Line),
        file_directory_name(File, Dir),
        atom_concat(Dir, '/load.pl', LoadFile),
-       '$load_wic_files'(system, [LoadFile]),
+       '$load_wic_files'(system:[LoadFile]),
        (   current_prolog_flag(windows, true)
        ->  atom_concat(Dir, '/menu.pl', MenuFile),
-	   '$load_wic_files'(system, [MenuFile])
+	   '$load_wic_files'(system:[MenuFile])
        ;   true
        ),
        format('SWI-Prolog boot files loaded~n', []),
