@@ -78,7 +78,7 @@
 	->  asserta(loaded_init_file(system))
 	;   current_prolog_flag(home, Home),
 	    file_name_extension(Base, rc, Name),
-	    concat_atom([Home, '/', Name], File),
+	    atomic_list_concat([Home, '/', Name], File),
 	    access_file(File, read),
 	    asserta(loaded_init_file(system)),
 	    load_files(user:File, [silent(true)]), !
@@ -111,18 +111,19 @@
 		 *	 AT_INITIALISATION	*
 		 *******************************/
 
-:- module_transparent
-	at_initialization/1,
-	(initialization)/1.
+:- meta_predicate
+	at_initialization(0),
+	initialization(0).
 :- dynamic
 	'$at_initialization'/1.
 
-at_initialization(Spec) :-
-	strip_module(Spec, Module, Goal),
-	'$toplevel':assert('$at_initialization'(Module:Goal)).
+%%	at_initialization(:Goal)
+%
+%	Runs Goal after loading a saved state.
+
+at_initialization(Goal) :-
+	assert('$at_initialization'(Goal)).
 
-'$run_at_initialization' :-
-	\+ current_prolog_flag(saved_program, true), !.
 '$run_at_initialization' :-
 	(   '$at_initialization'(Goal),
 	    (   catch(Goal, E,
@@ -135,33 +136,38 @@ at_initialization(Spec) :-
 	),
 	'$thread_init'.
 
-%	initialization(+Goal)
+%%	initialization(+Goal)
 %
-%	Runs `Goal' both a load and initialization time.
+%	Runs `Goal' immediately and after loading a saved state.
 
 initialization(Goal) :-
 	at_initialization(Goal),
-	Goal.
+	call(Goal).
 
 
 		 /*******************************
 		 *     THREAD INITIALIZATION	*
 		 *******************************/
 
-:- module_transparent
-	(thread_initialization)/1.
+:- meta_predicate
+	thread_initialization(0).
 :- dynamic
 	'$at_thread_initialization'/1.
 
-thread_initialization(Spec) :-
-	strip_module(Spec, Module, Goal),
-	'$toplevel':assert('$at_thread_initialization'(Module:Goal)),
-	Spec.
+%%	thread_initialization(:Goal)
+%
+%	Run Goal now and everytime a new thread is created.
+
+thread_initialization(Goal) :-
+	assert('$at_thread_initialization'(Goal)),
+	call(Goal), !.
 
 '$thread_init' :-
 	(   '$at_thread_initialization'(Goal),
-	    Goal,
-	    fail
+	    (	call(Goal)
+	    ->	fail
+	    ;	fail
+	    )
 	;   true
 	).
 
@@ -228,7 +234,7 @@ path_sep -->
 		 *******************************/
 
 %%	set_associated_file
-%	
+%
 %	If SWI-Prolog is started as <exe> <file>.<ext>, where <ext> is
 %	the extension registered for associated files, set the Prolog
 %	flag associated_file, switch to the directory holding the file
@@ -259,7 +265,7 @@ set_associated_file.
 
 %%	start_pldoc
 %
-%	If the option  =|--pldoc[=port]|=  is   given,  load  the PldDoc
+%	If the option  =|--pldoc[=port]|=  is   given,  load  the  PlDoc
 %	system.
 
 start_pldoc :-
@@ -282,7 +288,7 @@ start_pldoc.
 
 
 %%	load_associated_file
-%	
+%
 %	Load  the  file-name  set  by   set_associated_file/0  from  the
 %	commandline arguments. Note the expand(false) to avoid expanding
 %	special characters in the filename.
@@ -341,8 +347,8 @@ initialise_prolog :-
 	start_pldoc,
 	'$load_script_file',
 	load_associated_file,
-	'$option'(goal, GoalAtom, GoalAtom), 
-	term_to_atom(Goal, GoalAtom), 
+	'$option'(goal, GoalAtom, GoalAtom),
+	term_to_atom(Goal, GoalAtom),
 	(   Goal == '$welcome'
 	->  flag('$banner_goal', TheGoal, TheGoal)
 	;   TheGoal = Goal
@@ -350,24 +356,24 @@ initialise_prolog :-
 	ignore(user:TheGoal).
 
 '$abort' :-
-	see(user), 
-	tell(user), 
-	flag('$break_level', _, 0), 
+	see(user),
+	tell(user),
+	flag('$break_level', _, 0),
 	flag('$compilation_level', _, 0),
 	'$calleventhook'(abort),
 	print_message(informational, '$aborted'),
 	'$toplevel'.
 
 '$break' :-
-	flag('$break_level', Old, Old+1), 
-	flag('$break_level', New, New), 
+	flag('$break_level', Old, Old+1),
+	flag('$break_level', New, New),
 	print_message(informational, break(enter(New))),
 	'$runtoplevel',
 	print_message(informational, break(exit(New))),
 	flag('$break_level', _, Old), !.
 
-:- '$hide'('$toplevel', 0).			% avoid in the GUI stacktrace
-:- '$hide'('$abort', 0).			% same after an abort
+:- '$hide'('$toplevel'/0).		% avoid in the GUI stacktrace
+:- '$hide'('$abort'/0).			% same after an abort
 
 '$toplevel' :-
 	'$runtoplevel',
@@ -380,7 +386,7 @@ initialise_prolog :-
 %	crashing in a loop?
 
 '$runtoplevel' :-
-	'$option'(toplevel, TopLevelAtom, TopLevelAtom), 
+	'$option'(toplevel, TopLevelAtom, TopLevelAtom),
 	catch(term_to_atom(TopLevel, TopLevelAtom), E,
 	      (print_message(error, E),
 	       halt(1))),
@@ -401,10 +407,10 @@ initialise_prolog :-
 		*********************************/
 
 prolog :-
-	flag('$tracing', _, off), 
-	flag('$break_level', BreakLev, BreakLev), 
-	repeat, 
-	    (   '$module'(TypeIn, TypeIn), 
+	flag('$tracing', _, off),
+	flag('$break_level', BreakLev, BreakLev),
+	repeat,
+	    (   '$module'(TypeIn, TypeIn),
 		(   stream_property(user_input, tty(true))
 		->  '$system_prompt'(TypeIn, BreakLev, Prompt),
 		    prompt(Old, '|    ')
@@ -433,9 +439,12 @@ read_query(Prompt, Goal, Bindings) :-
 	       ->  fail
 	       ;   throw(E)
 	       ))),
-	atom_concat(Line, '.', CompleteLine),
 	(   current_predicate(_, user:rl_add_history(_))
-	->  call(user:rl_add_history(CompleteLine))
+	->  format(atom(CompleteLine), '~W~W',
+		   [ Line, [partial(true)],
+		     '.', [partial(true)]
+		   ]),
+	    call(user:rl_add_history(CompleteLine))
 	;   true
 	),
 	catch(atom_to_term(Line, Goal, Bindings), E,
@@ -445,8 +454,8 @@ read_query(Prompt, Goal, Bindings) :-
 	'$save_history'(Line).
 read_query(Prompt, Goal, Bindings) :-
 	seeing(Old), see(user_input),
-	(   read_history(h, '!h', 
-			 [trace, end_of_file], 
+	(   read_history(h, '!h',
+			 [trace, end_of_file],
 			 Prompt, Goal, Bindings)
 	->  see(Old)
 	;   see(Old),
@@ -464,7 +473,7 @@ clean_history_prompt_chars([], []).
 clean_history_prompt_chars(['%', !|T], T) :- !.
 clean_history_prompt_chars([H|T0], [H|T]) :-
 	clean_history_prompt_chars(T0, T).
- 
+
 delete_leading_blanks([' '|T0], T) :- !,
 	delete_leading_blanks(T0, T).
 delete_leading_blanks(L, L).
@@ -513,7 +522,7 @@ set_default_history :-
 	;    '$substitute'("%d", [], P2, P3)
 	),
 	atom_chars(Prompt, P3).
-	
+
 '$substitute'(From, T, Old, New) :-
 	phrase(subst_chars(T), T0),
 	'$append'(Pre, S0, Old),
@@ -521,7 +530,7 @@ set_default_history :-
 	'$append'(Pre, T0, S1),
 	'$append'(S1, Post, New), !.
 '$substitute'(_, _, Old, Old).
-	
+
 subst_chars([]) -->
 	[].
 subst_chars([H|T]) -->
@@ -546,23 +555,22 @@ subst_chars([H|T]) -->
 '$execute'(end_of_file, _) :- !,
 	print_message(query, query(eof)).
 '$execute'(Goal, Bindings) :-
-	'$module'(TypeIn, TypeIn), 
+	'$module'(TypeIn, TypeIn),
 	expand_goal(Goal, Expanded),
-	TypeIn:'$dwim_correct_goal'(Expanded, Bindings, Corrected), !, 
+	'$dwim_correct_goal'(TypeIn:Expanded, Bindings, Corrected), !,
 	'$execute_goal'(Corrected, Bindings).
 '$execute'(_, _) :-
-	notrace, 
+	notrace,
 	print_message(query, query(no)),
 	fail.
 
 '$execute_goal'(trace, []) :-
-	trace, 
+	trace,
 	print_message(query, query(yes)), !,
 	fail.
 '$execute_goal'(Goal, Bindings) :-
-	'$module'(TypeIn, TypeIn),
-	print_message(silent, toplevel_goal(TypeIn:Goal, Bindings)),
-	'$execute_goal2'(TypeIn:Goal, Bindings).
+	print_message(silent, toplevel_goal(Goal, Bindings)),
+	'$execute_goal2'(Goal, Bindings).
 
 '$execute_goal2'(Goal, Bindings) :-
 	Goal,
@@ -575,12 +583,12 @@ subst_chars([H|T]) -->
 	     fail
 	).
 '$execute_goal2'(_, _) :-
-	notrace, 
+	notrace,
 	print_message(query, query(no)),
 	fail.
 
 %%	write_bindings(+Bindings, +Deterministic)
-%	
+%
 %	Write   bindings   resulting   from   a     query.    The   flag
 %	prompt_alternatives_on determines whether the   user is prompted
 %	for alternatives. =groundness= gives   the  classical behaviour,
@@ -601,7 +609,7 @@ write_bindings(Bindings, Det) :-
 
 write_bindings2([], Residuals, _) :-
 	current_prolog_flag(prompt_alternatives_on, groundness), !,
-	print_message(query, query(yes, Residuals)).
+	print_message(query, query(yes(Residuals))).
 write_bindings2(Bindings, Residuals, true) :-
 	current_prolog_flag(prompt_alternatives_on, determinism), !,
 	print_message(query, query(yes(Bindings, Residuals))).
@@ -623,23 +631,34 @@ write_bindings2(Bindings, Residuals, _Det) :-
 %	Omit unneeded module qualifiers  from   QGoals  relative  to the
 %	given module TypeIn.
 
+
 omit_qualifiers([], _, []).
 omit_qualifiers([Goal0|Goals0], TypeIn, [Goal|Goals]) :-
 	omit_qualifier(Goal0, TypeIn, Goal),
 	omit_qualifiers(Goals0, TypeIn, Goals).
 
 omit_qualifier(M:G0, TypeIn, G) :-
-	predicate_property(TypeIn:G0, imported_from(M)), !,
-	omit_meta_qualifiers(G0, TypeIn, G).
-omit_qualifier(_:G0, TypeIn, G) :-
-	predicate_property(G0, built_in), !,
+	M == TypeIn, !,
 	omit_meta_qualifiers(G0, TypeIn, G).
+omit_qualifier(M:G0, TypeIn, G) :-
+	predicate_property(TypeIn:G0, imported_from(M)),
+	\+ predicate_property(G0, transparent), !,
+	G0 = G.
+omit_qualifier(_:G0, _, G) :-
+	predicate_property(G0, built_in),
+	\+ predicate_property(G0, transparent), !,
+	G0 = G.
 omit_qualifier(M:G0, _, M:G) :-
 	atom(M), !,
 	omit_meta_qualifiers(G0, M, G).
 omit_qualifier(G0, TypeIn, G) :-
 	omit_meta_qualifiers(G0, TypeIn, G).
 
+omit_meta_qualifiers(V, _, V) :-
+	var(V), !.
+omit_meta_qualifiers((QA,QB), TypeIn, (A,B)) :- !,
+	omit_qualifier(QA, TypeIn, A),
+	omit_qualifier(QB, TypeIn, B).
 omit_meta_qualifiers(freeze(V, QGoal), TypeIn, freeze(V, Goal)) :-
 	callable(QGoal), !,
 	omit_qualifier(QGoal, TypeIn, Goal).
@@ -650,7 +669,7 @@ omit_meta_qualifiers(G, _, G).
 
 
 %%	bind_vars(+Bindings)
-%	
+%
 %	Bind variables to '$VAR'(Name), so they are printed by the names
 %	used in the query. Note that by   binding  in the reverse order,
 %	variables bound to one another come out in the natural order.
@@ -664,7 +683,7 @@ bind_vars([Name=Var|T]) :-
 	).
 
 %%	filter_bindings(+Bindings0, -Bindings)
-%	
+%
 %	Remove bindings that must not be printed.
 
 filter_bindings([], []).
