@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -35,11 +35,14 @@
 	    '$home'/1,
 	    '$argv'/1,
 	    '$strip_module'/3,
+	    '$declare_module'/3,
 	    displayq/1,
 	    displayq/2,
 	    sformat/2,			% -String, +Fmt
 	    sformat/3,			% -String, +Fmt, +Args
 	    concat/3,
+	    concat_atom/2,		% +List, -Atom
+	    concat_atom/3,		% +List, +Sep, -Atom
 	    read_variables/2,
 	    read_variables/3,
 	    feature/2,
@@ -49,6 +52,7 @@
 	    write_ln/1,
 	    proper_list/1,
 	    free_variables/2,		% +Term, -Variables
+	    hash_term/2,		% +Term, -Hash
 	    checklist/2,		% :Goal, +List
 	    sublist/3,			% :Goal, +List, -Sublist
 	    convert_time/2,		% +Stamp, -String
@@ -58,7 +62,13 @@
 	    current_mutex/3,		% ?Mutex, ?Owner, ?Count
 	    message_queue_size/2,	% +Queue, -TermsWaiting
 	    lock_predicate/2,		% +Name, +Arity
-	    unlock_predicate/2		% +Name, +Arity
+	    unlock_predicate/2,		% +Name, +Arity
+	    current_module/2,		% ?Module, ?File
+	    export_list/2,		% +Module, -Exports
+	    setup_and_call_cleanup/3,	% :Setup, :Goal, :Cleanup
+	    setup_and_call_cleanup/4,	% :Setup, :Goal, ?Catcher, :Cleanup
+	    merge/3,			% +List1, +List2, -Union
+	    merge_set/3			% +Set1, +Set2, -Union
 	  ]).
 
 /** <module> Backward compatibility
@@ -84,9 +94,9 @@ compatibility predicates.
 
 %%	displayq(@Term) is det.
 %%	displayq(+Stream, @Term) is det.
-%	
+%
 %	Write term ignoring operators and quote atoms.
-%	
+%
 %	@deprecated Use write_term/3 or write_canonical/2.
 
 displayq(Term) :-
@@ -114,6 +124,31 @@ sformat(String, Format, Arguments) :-
 concat(A, B, C) :-
 	atom_concat(A, B, C).
 
+%%	concat_atom(+List, -Atom) is det.
+%
+%	Concatenate a list of atomic values to an atom.
+%
+%	@deprecated Use atomic_list_concat/2 as proposed by the prolog
+%		    commons initiative.
+
+concat_atom([A, B], C) :- !,
+	atom_concat(A, B, C).
+concat_atom(L, Atom) :-
+	atomic_list_concat(L, Atom).
+
+
+%%	concat_atom(+List, +Seperator, -Atom) is det.
+%
+%	Concatenate a list of atomic values to an atom, inserting Seperator
+%	between each consecutive elements.
+%
+%	@deprecated Use atomic_list_concat/3 as proposed by the prolog
+%		    commons initiative.
+
+concat_atom(L, Sep, Atom) :-
+	atomic_list_concat(L, Sep, Atom).
+
+
 %%	read_variables(-Term, -Bindings) is det.
 %%	read_variables(+In:stream, -Term, -Bindings) is det.
 %
@@ -141,7 +176,7 @@ set_feature(Key, Value) :-
 %%	substring(+String, +Offset, +Length, -Sub)
 %
 %	Predecessor of sub_string using 1-based Offset.
-%	
+%
 %	@deprecated Use sub_string/5.
 
 substring(String, Offset, Length, Sub) :-
@@ -167,24 +202,34 @@ write_ln(X) :-
 %	Old SWI-Prolog predicate to check for a list that really ends
 %	in a [].  There is not much use for the quick is_list, as in
 %	most cases you want to process the list element-by-element anyway.
-%	
+%
 %	@deprecated Use ISO is_list/1.
 
 proper_list(List) :-
 	is_list(List).
 
 %%	free_variables(+Term, -Variables)
-%	
+%
 %	Return  a  list  of  unbound  variables    in   Term.  The  name
 %	term_variables/2 is more widely used.
-%	
+%
 %	@deprecated Use term_variables/2.
 
 free_variables(Term, Variables) :-
 	term_variables(Term, Variables).
 
+%%	hash_term(+Term, -Hash) is det.
+%
+%	If Term is ground, Hash is unified to an integer representing
+%	a hash for Term.  Otherwise Hash is left unbound.
+%
+%	@deprecated Use term_hash/2.
+
+hash_term(Term, Hash) :-
+	term_hash(Term, Hash).
+
 %%	checklist(:Goal, +List)
-%	
+%
 %	@deprecated Use maplist/2
 
 :- module_transparent
@@ -195,29 +240,29 @@ checklist(Goal, List) :-
 	maplist(Goal, List).
 
 %%	sublist(:Goal, +List1, ?List2)
-%	
+%
 %	Succeeds if List2 unifies with a list holding those terms for wich
 %	call(Goal, Elem) succeeds.
-%	
+%
 %	@deprecated Use include/3 from library(apply)
 %	@compat	DEC10 library
 
 sublist(_, [], []) :- !.
 sublist(Goal, [H|T], Sub) :-
-	call(Goal, H), !, 
-	Sub = [H|R], 
+	call(Goal, H), !,
+	Sub = [H|R],
 	sublist(Goal, T, R).
 sublist(Goal, [_|T], R) :-
 	sublist(Goal, T, R).
 
 
 %%	strip_module(+Term, -Module, -Plain)
-%	
+%
 %	This used to be an internal predicate.  It was added to the XPCE
 %	compatibility library without $ and  since   then  used  at many
 %	places. From 5.4.1 onwards strip_module/3 is  built-in and the $
 %	variation is added here for compatibility.
-%	
+%
 %	@deprecated Use strip_module/3.
 
 :- module_transparent
@@ -226,13 +271,22 @@ sublist(Goal, [_|T], R) :-
 '$strip_module'(Term, Module, Plain) :-
 	strip_module(Term, Module, Plain).
 
+
+%%	'$declare_module'(Module, File, Line)
+%
+%	Used in triple20 particle library. Should use a public interface
+
+'$declare_module'(Module, File, Line) :-
+	'$declare_module'(Module, File, Line, false).
+
+
 %%	convert_time(+Stamp, -String)
 %
 %	Convert  a time-stamp as  obtained though get_time/1 into a  textual
 %	representation  using the C-library function ctime().  The  value is
 %	returned  as a  SWI-Prolog string object  (see section  4.23).   See
 %	also convert_time/8.
-%	
+%
 %	@deprecated Use format_time/3.
 
 
@@ -249,7 +303,7 @@ convert_time(Stamp, String) :-
 %	milliseconds  (0--999).  Note that the latter might not  be accurate
 %	or  might always be 0, depending  on the timing capabilities of  the
 %	system.  See also convert_time/2.
-%	
+%
 %	@deprecated Use stamp_date_time/3.
 
 convert_time(Stamp, Y, Mon, Day, Hour, Min, Sec, MilliSec) :-
@@ -265,7 +319,7 @@ convert_time(Stamp, Y, Mon, Day, Hour, Min, Sec, MilliSec) :-
 %
 %	Used to be generated by DCG.  Some people appear to be using in
 %	in normal code too.
-%	
+%
 %	@deprecated Do not use in normal code; DCG no longer generates it.
 
 'C'([H|T], H, T).
@@ -304,7 +358,7 @@ map_mutex_status(locked(Owner, Count), Owner, Count).
 %%	message_queue_size(+Queue, -Size) is det.
 %
 %	True if Queue holds Size terms.
-%	
+%
 %	@deprecated Please use message_queue_property(Queue, Size)
 
 message_queue_size(Queue, Size) :-
@@ -325,3 +379,75 @@ lock_predicate(Name, Arity) :-
 unlock_predicate(Name, Arity) :-
 	unlock_predicate(Name/Arity).
 
+%%	current_module(?Module, ?File) is nondet.
+%
+%	True if Module is a module loaded from File.
+%
+%	@deprecated Use module_property(Module, file(File))
+
+current_module(Module, File) :-
+	module_property(Module, file(File)).
+
+%%	export_list(+Module, -List) is det.
+%
+%	Module exports the predicates of List.
+%
+%	@deprecated Use module_property(Module, exports(List))
+
+export_list(Module, List) :-
+	module_property(Module, exports(List)).
+
+%%	setup_and_call_cleanup(:Setup, :Goal, :Cleanup).
+%
+%	Call Cleanup once after Goal is finished.
+%
+%	@deprecated Use setup_call_cleanup/3.
+
+:- module_transparent setup_and_call_cleanup/3, setup_and_call_cleanup/4.
+
+setup_and_call_cleanup(Setup, Goal, Cleanup) :-
+	setup_call_cleanup(Setup, Goal, Cleanup).
+
+%%	setup_and_call_cleanup(:Setup, :Goal, Catcher, :Cleanup).
+%
+%	Call Cleanup once after Goal is finished, with Catcher
+%       unified to the reason
+%
+%	@deprecated Use setup_call_cleanup/3.
+
+setup_and_call_cleanup(Setup, Goal, Catcher, Cleanup) :-
+	setup_call_catcher_cleanup(Setup, Goal, Catcher,Cleanup).
+
+%%	merge_set(+Set1, +Set2, -Set3)
+%
+%	Merge the ordered sets Set1 and  Set2   into  a  new ordered set
+%	without duplicates.
+%
+%	@deprecated	New code should use ord_union/3 from
+%			library(ordsets)
+
+merge_set([], L, L) :- !.
+merge_set(L, [], L) :- !.
+merge_set([H1|T1], [H2|T2], [H1|R]) :- H1 @< H2, !, merge_set(T1, [H2|T2], R).
+merge_set([H1|T1], [H2|T2], [H2|R]) :- H1 @> H2, !, merge_set([H1|T1], T2, R).
+merge_set([H1|T1], [H2|T2], [H1|R]) :- H1 == H2,    merge_set(T1, T2, R).
+
+
+%%	merge(+List1, +List2, -List3)
+%
+%	Merge the ordered sets List1 and List2 into a new ordered  list.
+%	Duplicates are not removed and their order is maintained.
+%
+%	@deprecated	The name of this predicate is far too general for
+%			a rather specific function.
+
+merge([], L, L) :- !.
+merge(L, [], L) :- !.
+merge([H1|T1], [H2|T2], [H|R]) :-
+	(   H1 @=< H2
+	->  H = H1,
+	    merge(T1, [H2|T2], R)
+	;   H = H2,
+	    merge([H1|T1], T2, R)
+	).
+
