@@ -1,4 +1,4 @@
-/*  $Id: debug.pl,v 1.15 2007/05/11 10:01:46 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -41,6 +41,8 @@
 	  ]).
 
 :- meta_predicate(assertion(:)).
+/*:- use_module(library(prolog_stack)).*/ % We use the autoloader if needed
+
 :- set_prolog_flag(generate_debug_info, false).
 
 :- dynamic
@@ -117,7 +119,7 @@ debug_topic(Topic) :-
 	).
 
 %%	list_debug_topics is det.
-%	
+%
 %	List currently known debug topics and their setting.
 
 list_debug_topics :-
@@ -155,7 +157,7 @@ print_debug(_, Format, Args) :-
 		 *******************************/
 
 %%	assertion(:Goal) is det.
-%	
+%
 %	Acts similar to C assert() macro.  It has no effect of Goal
 %	succeeds.  If Goal fails it prints a message, a stack-trace
 %	and finally traps the debugger.
@@ -171,11 +173,11 @@ assertion(G) :-
 assertion_failed.
 
 %%	assume(:Goal) is det.
-%	
+%
 %	Acts similar to C assert() macro.  It has no effect of Goal
 %	succeeds.  If Goal fails it prints a message, a stack-trace
 %	and finally traps the debugger.
-%	
+%
 %	@deprecated	Use assertion/1 in new code.
 
 		 /*******************************
