@@ -1,11 +1,11 @@
-/*  $Id: qsave.pl,v 1.38 2006/08/10 13:45:47 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -137,7 +137,7 @@ make_header(RC, _, Options) :-
 	current_prolog_flag(executable, Executable),
 	'$rc_append_file'(RC, '$header', '$rc', none, Executable).
 make_header(RC, SaveClass, _Options) :-
-	current_prolog_flag(unix, true),
+	current_prolog_flag(unix, true), !,
 	current_prolog_flag(executable, Executable),
 	'$rc_open'(RC, '$header', '$rc', write, Fd),
 	format(Fd, '#!/bin/sh~n', []),
@@ -169,7 +169,7 @@ convert_option(Stack, Val, NewVal) :-	% stack-sizes are in K-bytes
 convert_option(_, Val, Val).
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Save the options in the $options resource.   The home directory is saved
+Save the options in the '$options' resource.   The home directory is saved
 for development saves, so it keeps refering to the development home.
 
 The script-file (-s script) is not saved at all. I think this is fine to
@@ -214,7 +214,7 @@ save_resources(RC, _SaveClass) :-
 
 mkrcname(user, Name, Name) :- !.
 mkrcname(M, Name, RcName) :-
-	concat_atom([M, :, Name], RcName).
+	atomic_list_concat([M, :, Name], RcName).
 
 save_resource(RC, Name, Class, FileSpec) :-
 	absolute_file_name(FileSpec,
@@ -323,21 +323,25 @@ save_module(M, SaveClass) :-
 	;   '$qlf_end_part',
 	    feedback('~n', [])
 	).
-	
-pred_attrib(dynamic,       P, '$set_predicate_attribute'(P, dynamic,       1)).
-pred_attrib(volatile,      P, '$set_predicate_attribute'(P, volatile,      1)).
-pred_attrib(thread_local,  P, '$set_predicate_attribute'(P, thread_local,  1)).
-pred_attrib(multifile,     P, '$set_predicate_attribute'(P, multifile,     1)).
-pred_attrib(transparent,   P, '$set_predicate_attribute'(P, transparent,   1)).
-pred_attrib(discontiguous, P, '$set_predicate_attribute'(P, discontiguous, 1)).
-pred_attrib(notrace,       P, '$set_predicate_attribute'(P, trace,         0)).
-pred_attrib(show_childs,   P, '$set_predicate_attribute'(P, hide_childs,   0)).
-pred_attrib(indexed(Term), P, M:index(Term)) :-
-	strip_module(P, M, _).
 
-predicate_attribute(P, Attribute) :-
-	pred_attrib(Attribute, P, _),
-	predicate_property(P, Attribute).
+pred_attrib(indexed(Term), Head, index(M:Term)) :- !,
+	    strip_module(Head, M, _).
+pred_attrib(meta_predicate(Term), Head, meta_predicate(M:Term)) :- !,
+	    strip_module(Head, M, _).
+pred_attrib(Attrib, Head, '$set_predicate_attribute'(M:Name/Arity, AttName, Val)) :-
+	attrib_name(Attrib, AttName, Val),
+	strip_module(Head, M, Term),
+	functor(Term, Name, Arity).
+
+attrib_name(dynamic,	   dynamic,	  1).
+attrib_name(volatile,	   volatile,	  1).
+attrib_name(thread_local,  thread_local,  1).
+attrib_name(multifile,	   multifile,	  1).
+attrib_name(transparent,   transparent,	  1).
+attrib_name(discontiguous, discontiguous, 1).
+attrib_name(notrace,	   trace,	  0).
+attrib_name(show_childs,   hide_childs,	  0).
+
 
 save_attribute(P, Attribute) :-
 	pred_attrib(Attribute, P, D),
@@ -351,12 +355,12 @@ save_attribute(P, Attribute) :-
 	feedback('(~w) ', [Attribute]).
 
 save_attributes(P) :-
-	(   predicate_attribute(P, Attribute),
+	(   predicate_property(P, Attribute),
 	    save_attribute(P, Attribute),
 	    fail
 	;   true
 	).
-	    
+
 %	Save status of the unknown flag
 
 save_unknown(M) :-
@@ -417,10 +421,22 @@ save_imports :-
 	    \+ default_import(M, H, I),
 	    functor(H, F, A),
 	    feedback('~t~8|~w:~w/~d <-- ~w~n', [M, F, A, I]),
-	    '$add_directive_wic'(M:import(I:H)),
+	    '$add_directive_wic'(qsave:restore_import(M, I, F/A)),
 	    fail
 	;   true
-	).	    
+	).
+
+%%	restore_import(+TargetModule, +SourceModule, +PI) is det.
+%
+%	Restore import relation. This notably   deals  with imports from
+%	the module =user=, avoiding a message  that the predicate is not
+%	exported.
+
+restore_import(To, user, PI) :- !,
+	export(user:PI),
+	To:import(user:PI).
+restore_import(To, From, PI) :-
+	To:import(From:PI).
 
 		 /*******************************
 		 *	      FEATURES		*
@@ -558,7 +574,7 @@ option(List, Name/_Default, Value, Rest) :- % goal(Goal)
 	Term =.. [Name, Value],
 	select(Term, List, Rest), !.
 option(List, _Name/Default, Default, List).
-	
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Option checking and exception generation.  This should be in a library!
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
@@ -609,7 +625,7 @@ check_type(ground, V) :-
 	ground(V).
 check_type(bool, true).
 check_type(bool, false).
-	
+
 		 /*******************************
 		 *	      MESSAGES		*
 		 *******************************/
