@@ -42,7 +42,7 @@
 	    current_setting/1,		% Module:Name
 	    setting_property/2,		% ?Setting, ?Property
 	    list_settings/0,
-	    
+
 	    convert_setting_text/3	% +Type, +Text, -Value
 	  ]).
 :- use_module(library(error)).
@@ -79,14 +79,14 @@ access, loading and saving of settings.
 */
 
 :- dynamic
-	value/3,			% Name, Module, Value
-	default/3,			% Name, Module, Value
+	st_value/3,			% Name, Module, Value
+	st_default/3,			% Name, Module, Value
 	local_file/1.			% Path
 
 :- multifile
 	current_setting/6.		% Name, Module, Type, Default, Comment, Source
 
-:- module_transparent
+:- meta_predicate
 	setting(:, +, +, +),
 	setting(:, ?),
 	set_setting(:, +),
@@ -96,7 +96,7 @@ access, loading and saving of settings.
 
 curr_setting(Name, Module, Type, Default, Comment) :-
 	current_setting(Name, Module, Type, Default0, Comment, _Src),
-	(   default(Name, Module, Default1)
+	(   st_default(Name, Module, Default1)
 	->  Default = Default1
 	;   Default = Default0
 	).
@@ -107,16 +107,18 @@ curr_setting(Name, Module, Type, Default, Comment) :-
 %	type. Default is the value before  it is modified. Default refer
 %	to environment variables  and  use   arithmetic  expressions  as
 %	defined by eval_default/4.
-%	
+%
 %	@param Name	Name of the setting (an atom)
 %	@param Type	Type for setting.  One of =any= or a type defined
 %			by must_be/2.
 %	@param Default  Default value for the setting.
 %	@param Comment	Atom containing a (short) descriptive note.
 
-	
-setting(_Name, _Type, _Default, _Comment) :-
-	throw(error(context_error(only_in_directive), _)).
+
+setting(Name, Type, Default, Comment) :-
+	throw(error(context_error(nodirective,
+				  setting(Name, Type, Default, Comment)),
+		    _)).
 
 :- multifile
 	user:term_expansion/2.
@@ -138,7 +140,7 @@ user:term_expansion((:- setting(QName, Type, Default, Comment)),
 	->  Expanded = settings:current_setting(Name, Module, Type, Default,
 						CommentAtom, File:Line)
 	).
-	
+
 to_atom(Atom, Atom) :-
 	atom(Atom), !.
 to_atom(String, Atom) :-
@@ -147,13 +149,13 @@ to_atom(String, Atom) :-
 %%	setting(:Name, ?Value) is nondet.
 %
 %	True if Name is a currently defined setting with Value.
-%	
+%
 %	@error	existence_error(setting, Name)
 
 setting(QName, Value) :-
 	strip_module(QName, Module, Name),
 	(   ground(Name)
-	->  (   value(Name, Module, Value0)
+	->  (   st_value(Name, Module, Value0)
 	    ->  Value = Value0
 	    ;   curr_setting(Name, Module, Type, Default, _)
 	    ->	eval_default(Default, Module, Type, Value)
@@ -164,35 +166,42 @@ setting(QName, Value) :-
 	).
 
 
+:- dynamic
+	setting_cache/3.
+:- volatile
+	setting_cache/3.
+
+%%	clear_setting_cache is det.
+%
+%	Clear the cache for evaluation of default values.
+
+clear_setting_cache :-
+	retractall(setting_cache(_,_,_)).
+
 %%	eval_default(+Default, +Module, +Type, -Value) is det.
 %
 %	Convert the settings default value. The notation allows for some
 %	`function-style' notations to make the library more generic:
-%	
+%
 %		* env(Name)
 %		Get value from the given environment variable. The value
 %		is handed to convert_setting_text/3 to convert the
 %		textual representation into a Prolog term.  Raises an
 %		existence_error of the variable is not defined.
-%		
+%
 %		* env(Name, Default)
 %		As env(Name), but uses the value Default if the variable
 %		is not defined.
-%		
+%
 %		* setting(Name)
 %		Ask the value of another setting.
-%		
+%
 %		* Expression
 %		If Type is numeric, evaluate the expression.  env(Var)
 %		evaluates to the value of an environment variable.
 %		If Type is =atom=, concatenate A+B+....  Elements of the
 %		expression can be env(Name).
 
-:- dynamic
-	setting_cache/3.
-:- volatile
-	setting_cache/3.
-
 :- multifile
 	eval_default/3.			% +Default, +Type, -Value
 
@@ -236,7 +245,7 @@ eval_default(Expr, _, Type, Value) :-
 	Value = Val.
 eval_default(A+B, Module, atom, Value) :- !,
 	phrase(expr_to_list(A+B, Module), L),
-	concat_atom(L, Val),
+	atomic_list_concat(L, Val),
 	assert(setting_cache(A+B, atom, Val)),
 	Value = Val.
 eval_default(List, Module, list(Type), Value) :- !,
@@ -280,7 +289,7 @@ expr_to_list(setting(Name), Module) --> !,
 	[ Value ].
 expr_to_list(A, _) -->
 	[A].
-	  
+
 %%	env(+Name:atom, -Value:number) is det.
 %%	env(+Name:atom, +Default:number, -Value:number) is det
 %
@@ -322,7 +331,7 @@ numeric_type(between(L,_), Type) :-
 %	broadcasts the event below.
 %
 %		settings(changed(Module:Name, Old, New))
-%	
+%
 %	@error	existence_error(setting, Name)
 %	@error  type_error(Type, Value)
 
@@ -333,22 +342,23 @@ set_setting(QName, Value) :-
 	    eval_default(Default0, Module, Type, Default)
 	->  (   Value == Default
 	    ->	retract_setting(Module:Name)
-	    ;	value(Name, Module, Value)
+	    ;	st_value(Name, Module, Value)
 	    ->	true
 	    ;	check_type(Type, Value)
 	    ->	setting(Module:Name, Old),
 	        retract_setting(Module:Name),
 	        assert_setting(Module:Name, Value),
-		broadcast(settings(changed(Module:Name, Old, Value)))
+		broadcast(settings(changed(Module:Name, Old, Value))),
+		clear_setting_cache	% might influence dependent settings.
 	    )
 	;   existence_error(setting, Name)
 	).
 
 retract_setting(Module:Name) :-
-	retractall(value(Name, Module, _)).
+	retractall(st_value(Name, Module, _)).
 
 assert_setting(Module:Name, Value) :-
-	assert(value(Name, Module, Value)).
+	assert(st_value(Name, Module, Value)).
 
 %%	restore_setting(:Name) is det.
 %
@@ -359,7 +369,7 @@ assert_setting(Module:Name, Value) :-
 restore_setting(QName) :-
 	strip_module(QName, Module, Name),
 	must_be(atom, Name),
-	(   value(Name, Module, Old)
+	(   st_value(Name, Module, Old)
 	->  retract_setting(Module:Name),
 	    setting(Module:Name, Value),
 	    (	Old \== Value
@@ -380,11 +390,11 @@ set_setting_default(QName, Default) :-
 	strip_module(QName, Module, Name),
 	must_be(atom, Name),
 	(   current_setting(Name, Module, Type, Default0, _Comment, _Src)
-	->  retractall(settings:default(Name, Module, _)),
+	->  retractall(settings:st_default(Name, Module, _)),
 	    retract_setting(Module:Name),
 	    (   Default == Default0
 	    ->	true
-	    ;	assert(settings:default(Name, Module, Default))
+	    ;	assert(settings:st_default(Name, Module, Default))
 	    ),
 	    eval_default(Default, Module, Type, Value),
 	    set_setting(Module:Name, Value)
@@ -414,7 +424,7 @@ check_type(Type, Term) :-
 %
 %	Load local settings from File. Succeeds  if File does not exist,
 %	setting the default save-file to File.  Options are:
-%	
+%
 %	  * undefined(+Action)
 %	  Define how to handle settings that are not defined.  When
 %	  =error=, an error is printed and the setting is ignored.
@@ -432,7 +442,8 @@ load_settings(File, Options) :-
 	assert(local_file(Path)),
 	open(Path, read, In, [encoding(utf8)]),
 	read_setting(In, T0),
-	call_cleanup(load_settings(T0, In, Options), close(In)).
+	call_cleanup(load_settings(T0, In, Options), close(In)),
+	clear_setting_cache.
 load_settings(File, _) :-
 	absolute_file_name(File, Path,
 			   [ access(write),
@@ -463,14 +474,14 @@ store_setting(setting(Module:Name, Value), _) :-
 	(   Value == Default
 	->  true
 	;   check_type(Type, Value)
-	->  retractall(value(Name, Module, _)),
-	    assert(value(Name, Module, Value)),
+	->  retractall(st_value(Name, Module, _)),
+	    assert(st_value(Name, Module, Value)),
 	    broadcast(settings(changed(Module:Name, Default, Value)))
 	).
 store_setting(setting(Module:Name, Value), Options) :- !,
 	(   option(undefined(load), Options, load)
-	->  retractall(value(Name, Module, _)),
-	    assert(value(Name, Module, Value))
+	->  retractall(st_value(Name, Module, _)),
+	    assert(st_value(Name, Module, Value))
 	;   existence_error(setting, Module:Name)
 	).
 store_setting(Term, _) :-
@@ -508,7 +519,7 @@ write_setting_header(Out) :-
 
 save_setting(Out, Module:Name) :-
 	curr_setting(Name, Module, Type, Default, Comment),
-	(   value(Name, Module, Value),
+	(   st_value(Name, Module, Value),
 	    \+ ( eval_default(Default, Module, Type, DefValue),
 		 debug(setting, '~w <-> ~w~n', [DefValue, Value]),
 	         DefValue =@= Value
@@ -533,7 +544,7 @@ current_setting(Module:Name) :-
 %%	setting_property(?Setting, ?Property) is nondet.
 %
 %	Query currently defined settings.  Property is one of
-%	
+%
 %		* comment(-Atom)
 %		* type(-Type)
 %		Type of the setting.
@@ -587,7 +598,7 @@ list_setting(Module:Name) :-
 %	Converts from textual form to  Prolog   Value.  Used  to convert
 %	values obtained from the environment.  Public to provide support
 %	in user-interfaces to this library.
-%	
+%
 %	@error	type_error(Type, Value)
 
 :- multifile
