@@ -127,6 +127,7 @@ leave the details to this function.
 #include <malloc.h>
 #endif
 
+
 #ifndef __WINDOWS__
 #define closesocket(n) close((n))	/* same on Unix */
 #endif
@@ -227,12 +228,27 @@ typedef struct _plsocket
       size_t written;
       size_t size;			/* buffer size */
     } write;
+    struct
+    { int bytes;			/* byte count */
+      void *buffer;			/* the buffer */
+      size_t size;			/* buffer size */
+      int flags;
+      struct sockaddr *from;
+      socklen_t *fromlen;
+    } recvfrom;
+    struct
+    { int bytes;			/* byte count */
+      void *buffer;			/* the buffer */
+      size_t size;			/* buffer size */
+      int flags;
+      const struct sockaddr *to;
+      socklen_t tolen;
+    } sendto;
   } rdata;
 #endif
 } plsocket;
 
 static plsocket *allocSocket(SOCKET socket);
-static plsocket *nbio_to_plsocket(nbio_sock_t socket);
 #ifdef __WINDOWS__
 static plsocket *lookupOSSocket(SOCKET socket);
 static const char *WinSockError(unsigned long eno);
@@ -295,6 +311,8 @@ request_name(nbio_request request)
     case REQ_CONNECT: return "req_connect";
     case REQ_READ:    return "req_read";
     case REQ_WRITE:   return "req_write";
+    case REQ_RECVFROM:return "req_recvfrom";
+    case REQ_SENDTO:  return "req_sendto";
     default:	      return "req_???";
   }
 }
@@ -395,7 +413,7 @@ waitRequest(plsocket *s)
 
   for(;;)
   { MSG msg;
-  
+
     if ( PL_handle_signals() < 0 )
     { DEBUG(1, Sdprintf("[%d]: Exception\n", PL_thread_self()));
       return FALSE;
@@ -439,7 +457,7 @@ nbio_wait(nbio_sock_t socket, nbio_request request)
 
   for(;;)
   { MSG msg;
-  
+
     if ( PL_handle_signals() < 0 )
     { DEBUG(1, Sdprintf("[%d]: Exception\n", PL_thread_self()));
       return -1;
@@ -464,7 +482,7 @@ nbio_wait(nbio_sock_t socket, nbio_request request)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-nbio_select() selects using a set of socket streams. 
+nbio_select() selects using a set of socket streams.
 
 NOTE: The Windows versions uses our   nbio_sock_t abstraction, while the
 other version uses  the  raw  Unix   file  descriptors  referencing  the
@@ -527,7 +545,7 @@ nbio_select(int n,
     { DEBUG(1, Sdprintf("[%d]: Exception\n", PL_thread_self()));
       return -1;
     }
-    
+
     for(ready=0, i=0, s=sockets; i<n; i++, s++)
     { if ( *s && (*s)->done )
       { ready++;
@@ -588,7 +606,7 @@ placeRequest(plsocket *s, nbio_request request)
 		    PL_thread_self(), s->thread,
 		    request_name(request), (int)s->socket));
 
-  return TRUE; 
+  return TRUE;
 }
 
 static int
@@ -612,7 +630,7 @@ doRequest(plsocket *s)
 		     (struct sockaddr*)&s->rdata.connect.addr,
 		     (int)s->rdata.connect.addrlen) )
 	{ s->error = WSAGetLastError();
-	  
+
 	  switch(s->error)
 	  { case WSAEWOULDBLOCK:
 	    case WSAEINVAL:
@@ -647,7 +665,7 @@ doRequest(plsocket *s)
 
 	if ( slave == SOCKET_ERROR )
 	{ s->error = WSAGetLastError();
-	  
+
 	  DEBUG(2, Sdprintf("Accept(%d): %s\n",
 			    (int)s->socket, WinSockError(s->error)));
 
@@ -699,6 +717,36 @@ doRequest(plsocket *s)
 	}
       }
       break;
+    case REQ_RECVFROM:
+      if ( s->w32_flags & (FD_READ|FD_CLOSE) )
+      { s->w32_flags &= ~FD_READ;
+
+	if ( true(s, SOCK_WAITING) )
+	{ doneRequest(s);
+	  break;
+	}
+
+	s->rdata.recvfrom.bytes =
+		recvfrom(s->socket,
+			 s->rdata.recvfrom.buffer,
+			 (int)s->rdata.recvfrom.size,
+			 s->rdata.recvfrom.flags,
+			 s->rdata.recvfrom.from,
+			 s->rdata.recvfrom.fromlen);
+
+	if ( s->rdata.recvfrom.bytes < 0 )
+	{ s->error = WSAGetLastError();
+
+	  if ( s->error != WSAEWOULDBLOCK )
+	  { DEBUG(1, Sdprintf("Error recvfrom from %d: %s\n",
+			      s->socket, WinSockError(s->error)));
+	    doneRequest(s);
+	  }
+	} else
+	{ doneRequest(s);
+	}
+      }
+      break;
     case REQ_WRITE:
       if ( s->w32_flags & FD_WRITE )
       { int n;
@@ -727,13 +775,46 @@ doRequest(plsocket *s)
 	  doneRequest(s);
 	} else
 	  s->error = 0;
-	  
+
 	s->rdata.write.written += n;
 	if ( s->rdata.write.written >= s->rdata.write.size )
 	{ s->rdata.write.bytes = (int)s->rdata.write.written;
 	  doneRequest(s);
 	}
       }
+    case REQ_SENDTO:
+      if ( s->w32_flags & FD_WRITE )
+      { int n;
+
+	s->w32_flags &= ~FD_WRITE;
+
+	if ( true(s, SOCK_WAITING) )
+	{ doneRequest(s);
+	  break;
+	}
+
+	DEBUG(2, Sdprintf("sendto() %d bytes\n", s->rdata.write.size));
+	n = sendto(s->socket,
+		   s->rdata.sendto.buffer,
+		   s->rdata.sendto.size,
+		   s->rdata.sendto.flags,
+		   s->rdata.sendto.to,
+		   s->rdata.sendto.tolen);
+	DEBUG(2, Sdprintf("Wrote %d bytes\n", n));
+	if ( n < 0 )
+	{ s->error = WSAGetLastError();
+	  s->rdata.write.bytes = n;
+	  DEBUG(1, Sdprintf("[%d]: send(%d, %d bytes): %s\n",
+			    PL_thread_self(), s->socket,
+			    s->rdata.sendto.size,
+			    WinSockError(s->error)));
+	  doneRequest(s);
+	} else
+	  s->error = 0;
+
+	s->rdata.sendto.bytes = n;
+	doneRequest(s);
+      }
   }
 
   return TRUE;
@@ -774,14 +855,14 @@ socket_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
     if ( evt&FD_CLOSE )
     { DEBUG(1,
 	    { char *nm = event_name(evt);
-	      Sdprintf("WM_SOCKET on %d: ev=(%s); err=%s\n", 
+	      Sdprintf("WM_SOCKET on %d: ev=(%s); err=%s\n",
 		       (int)sock, nm, err ? WinSockError(err) : "none");
 	      free(nm);
 	    });
     } else
     { DEBUG(3,
 	    { char *nm = event_name(evt);
-	      Sdprintf("WM_SOCKET on %d: ev=(%s); err=%s\n", 
+	      Sdprintf("WM_SOCKET on %d: ev=(%s); err=%s\n",
 		       (int)sock, nm, err ? WinSockError(err) : "none");
 	      free(nm);
 	    });
@@ -819,9 +900,15 @@ socket_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	  case REQ_READ:
 	    s->rdata.read.bytes = -1;
 	    break;
+	  case REQ_RECVFROM:
+	    s->rdata.recvfrom.bytes = -1;
+	    break;
 	  case REQ_WRITE:
 	    s->rdata.write.bytes = -1;
 	    break;
+	  case REQ_SENDTO:
+	    s->rdata.sendto.bytes = -1;
+	    break;
 	}
 	doneRequest(s);
       } else if ( s->socket >= 0 )
@@ -959,7 +1046,7 @@ wait_socket(plsocket *s)
       FD_SET(fd, &rfds);
       tv.tv_sec = 0;
       tv.tv_usec = 250000;
-      
+
       select(fd+1, &rfds, NULL, NULL, &tv);
       return TRUE;
     } else
@@ -1012,7 +1099,7 @@ int
 nbio_select(int n,
 	    fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
 	    struct timeval *timeout)
-{ return select(n, readfds, writefds, exceptfds, timeout); 
+{ return select(n, readfds, writefds, exceptfds, timeout);
 }
 
 #endif /*__WINDOWS__*/
@@ -1064,7 +1151,7 @@ lookupOSSocket(SOCKET socket)
 
 
 static plsocket *
-nbio_to_plsocket_raw(nbio_sock_t socket)
+nbio_to_plsocket_nolock(nbio_sock_t socket)
 { plsocket *p;
 
   if ( socket < 0 || (size_t)socket >= socks_allocated )
@@ -1072,9 +1159,7 @@ nbio_to_plsocket_raw(nbio_sock_t socket)
     return NULL;
   }
 
-  LOCK();
   p = sockets[socket];
-  UNLOCK();
 
   if ( !p || p->magic != SOCK_MAGIC )
   { DEBUG(1, Sdprintf("Invalid NBIO socket: %d\n", socket));
@@ -1086,7 +1171,25 @@ nbio_to_plsocket_raw(nbio_sock_t socket)
 }
 
 
+SOCKET
+plsocket_handle(plsocket_ptr pls)
+{ return pls->socket;
+}
+
+
 static plsocket *
+nbio_to_plsocket_raw(nbio_sock_t socket)
+{ plsocket *s;
+
+  LOCK();
+  s = nbio_to_plsocket_nolock(socket);
+  UNLOCK();
+
+  return s;
+}
+
+
+plsocket *
 nbio_to_plsocket(nbio_sock_t socket)
 { plsocket *p;
 
@@ -1108,7 +1211,7 @@ SOCKET
 nbio_fd(nbio_sock_t socket)
 { plsocket *p;
 
-  if ( !(p=nbio_to_plsocket(socket)) )
+  if ( !(p=nbio_to_plsocket_nolock(socket)) )
     return -1;
 
   return p->socket;
@@ -1161,7 +1264,7 @@ allocSocket(SOCKET socket)
   UNLOCK();
 
   assert(i<socks_allocated);
-  
+
   memset(p, 0, sizeof(*p));
   p->id     = (int)i;			/* place in the array */
   p->socket = socket;
@@ -1238,7 +1341,7 @@ WinSockError(unsigned long eno)
   static HMODULE netmsg = 0;
   static int netmsg_loaded = FALSE;
   unsigned long flags = (FORMAT_MESSAGE_FROM_SYSTEM|
-			 FORMAT_MESSAGE_IGNORE_INSERTS); 
+			 FORMAT_MESSAGE_IGNORE_INSERTS);
 
   if ( !netmsg_loaded )
   { netmsg_loaded = TRUE;
@@ -1246,7 +1349,7 @@ WinSockError(unsigned long eno)
     if ( !netmsg )
       Sdprintf("failed to load netmsg.dll\n");
     else
-      Sdprintf("Loaded netmsg.dll as %p\n", netmsg); 
+      Sdprintf("Loaded netmsg.dll as %p\n", netmsg);
   }
 
   if ( netmsg )
@@ -1438,12 +1541,12 @@ nbio_init(const char *module)
   }
   initialised = TRUE;
 
-  FUNCTOR_module2 = PL_new_functor(PL_new_atom(":"), 2);
-  FUNCTOR_ip4     = PL_new_functor(PL_new_atom("ip"), 4);
-  FUNCTOR_ip1     = PL_new_functor(PL_new_atom("ip"), 1);
-  ATOM_any	  = PL_new_atom("any");
-  ATOM_broadcast  = PL_new_atom("broadcast");
-  ATOM_loopback   = PL_new_atom("loopback");
+  FUNCTOR_module2  = PL_new_functor(PL_new_atom(":"), 2);
+  FUNCTOR_ip4	   = PL_new_functor(PL_new_atom("ip"), 4);
+  FUNCTOR_ip1	   = PL_new_functor(PL_new_atom("ip"), 1);
+  ATOM_any	   = PL_new_atom("any");
+  ATOM_broadcast   = PL_new_atom("broadcast");
+  ATOM_loopback	   = PL_new_atom("loopback");
 
 #ifdef __WINDOWS__
 { WSADATA WSAData;
@@ -1468,7 +1571,7 @@ nbio_init(const char *module)
   UNLOCK();
   return TRUE;
 }
-	
+
 
 int
 nbio_cleanup(void)
@@ -1493,7 +1596,7 @@ nbio_sock_t
 nbio_socket(int domain, int type, int protocol)
 { SOCKET sock;
   plsocket *s;
-	
+
   assert(initialised);
 
   if ( (sock = socket(domain, type , protocol)) < 0)
@@ -1550,7 +1653,6 @@ nbio_closesocket(nbio_sock_t socket)
   return 0;
 }
 
-
 int
 nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 { plsocket *s;
@@ -1570,7 +1672,7 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
     { int val = va_arg(args, int);
 
       if( setsockopt(s->socket, SOL_SOCKET, SO_REUSEADDR,
-		     (const char *)&val, sizeof(val)) == -1)
+		     (const char *)&val, sizeof(val)) == -1 )
       { nbio_error(h_errno, TCP_HERRNO);
 	rc = -1;
       } else
@@ -1578,11 +1680,31 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 
       break;
     }
+    case TCP_NO_DELAY:
+#ifdef TCP_NODELAY
+    { int val = va_arg(args, int);
+
+#ifndef IPPROTO_TCP			/* Is this correct? */
+#define IPPROTO_TCP SOL_SOCKET
+#endif
+      if ( setsockopt(s->socket, IPPROTO_TCP, TCP_NODELAY,
+		      (const char *)&val, sizeof(val)) == -1 )
+      { nbio_error(h_errno, TCP_HERRNO);
+	rc = -1;
+      } else
+      { rc = 0;
+      }
+
+      break;
+    }
+#else
+    rc = -2;				/* not implemented */
+#endif
     case UDP_BROADCAST:
     { int val = va_arg(args, int);
 
-      if( setsockopt(s->socket, SOL_SOCKET, SO_BROADCAST,
-		     (const char *)&val, sizeof(val)) == -1)
+      if ( setsockopt(s->socket, SOL_SOCKET, SO_BROADCAST,
+		     (const char *)&val, sizeof(val)) == -1 )
       { nbio_error(h_errno, TCP_HERRNO);
 	rc = -1;
       } else
@@ -1592,7 +1714,7 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
     }
     case TCP_DISPATCH:
     { int val = va_arg(args, int);
-      
+
       if ( val )
 	set(s, SOCK_DISPATCH);
       else
@@ -1636,12 +1758,12 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 int
 nbio_get_flags(nbio_sock_t socket)
 { plsocket *s;
-  
+
   if ( !(s = nbio_to_plsocket(socket)) )
     return -1;
 
   return s->flags;
-} 
+}
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1654,11 +1776,12 @@ nbio_get_port(term_t Port, int *port)
 
   if ( PL_get_atom_chars(Port, &name) )
   { struct servent *service;
-    
+
     if ( !(service = getservbyname(name, "tcp")) )
-      return nbio_error(errno, TCP_ERRNO);
+      return pl_error(NULL, 0, NULL, ERR_EXISTENCE, "service", Port);
 
     *port = ntohs(service->s_port);
+    DEBUG(1, Sdprintf("Service %s at port %d\n", name, *port));
     return TRUE;
   }
 
@@ -1677,7 +1800,7 @@ or the name of a registered port (e.g. 'smtp').
 int
 nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr)
 { int		  port;
-	
+
   addr->sin_family = AF_INET;
   addr->sin_addr.s_addr = INADDR_ANY;
 
@@ -1792,10 +1915,10 @@ nbio_connect(nbio_sock_t socket,
 	     const struct sockaddr *serv_addr,
 	     size_t addrlen)
 { plsocket *s;
-       
+
   if ( !(s = nbio_to_plsocket(socket)) )
     return -1;
-  
+
 #ifdef __WINDOWS__
   if ( connect(s->socket, serv_addr, (int)addrlen) )
   { s->error = WSAGetLastError();
@@ -1837,7 +1960,7 @@ nbio_sock_t
 nbio_accept(nbio_sock_t master, struct sockaddr *addr, socklen_t *addrlen)
 { SOCKET slave;
   plsocket *m, *s;
-	
+
   if ( !(m = nbio_to_plsocket(master)) )
     return -1;
 
@@ -1890,7 +2013,7 @@ nbio_accept(nbio_sock_t master, struct sockaddr *addr, socklen_t *addrlen)
   }
 
 #endif /*__WINDOWS__*/
-  
+
   s = allocSocket(slave);
   s->flags |= SOCK_ACCEPT;
 #ifndef __WINDOWS__
@@ -2074,7 +2197,6 @@ nbio_close_input(nbio_sock_t socket)
 
   DEBUG(2, Sdprintf("[%d]: nbio_close_input(%d, flags=0x%x)\n",
 		    PL_thread_self(), socket, s->flags));
-  s->input = NULL;
   s->flags &= ~SOCK_INSTREAM;
 #ifdef __WINDOWS__
   if ( false(s, SOCK_LISTEN) )
@@ -2087,11 +2209,13 @@ nbio_close_input(nbio_sock_t socket)
     { DEBUG(1, Sdprintf("shutdown(%d=%d, SD_RECEIVE) failed: %s\n",
 			socket, s->socket,
 			WinSockError(WSAGetLastError())));
+      Sseterr(s->input, SIO_FERR, WinSockError(WSAGetLastError()));
+      rc = -1;
     }
-    rc = -1;
   }
 #endif
 
+  s->input = NULL;
   if ( !(s->flags & (SOCK_INSTREAM|SOCK_OUTSTREAM)) )
     return freeSocket(s);
 
@@ -2115,25 +2239,23 @@ nbio_close_output(nbio_sock_t socket)
     SOCKET sock;
 #endif
 
-    s->output = NULL;
     s->flags &= ~SOCK_OUTSTREAM;
 #if __WINDOWS__
     if ( (sock=s->socket) < 0 )
     { s->error = WSAECONNRESET;
       rc = -1;
     } else if ( shutdown(sock, SD_SEND) == SOCKET_ERROR )
-    {
+    { const char *msg;
+#ifdef __WINDOWS__
+      msg = WinSockError(WSAGetLastError());
+#else
+      msg = strerror(errno);
+#endif
+      Sseterr(s->output, SIO_FERR, msg);
 #ifdef O_DEBUG
       if ( debugging )
-      { const char *msg;
-#ifdef __WINDOWS__
-	msg = WinSockError(WSAGetLastError());
-#else
-        msg = strerror(errno);
-#endif
-        Sdprintf("shutdown(%d=%d, SD_SEND) failed: %s\n",
+	Sdprintf("shutdown(%d=%d, SD_SEND) failed: %s\n",
 		 socket, s->socket, msg);
-      }
 #endif
       rc = -1;
     }
@@ -2141,9 +2263,137 @@ nbio_close_output(nbio_sock_t socket)
   }
 
   DEBUG(3, Sdprintf("%d->flags = 0x%x\n", socket, s->flags));
+  s->output = NULL;
   if ( !(s->flags & (SOCK_INSTREAM|SOCK_OUTSTREAM)) )
     return freeSocket(s);
 
   return rc;
 }
 
+
+		 /*******************************
+		 *	    UDP SUPPORT		*
+		 *******************************/
+
+ssize_t
+nbio_recvfrom(int socket, void *buf, size_t bufSize, int flags,
+	     struct sockaddr *from, socklen_t *fromlen)
+{ plsocket *s;
+  int n;
+
+  if ( !(s = nbio_to_plsocket(socket)) )
+    return -1;
+
+#ifdef __WINDOWS__
+
+  DEBUG(3, Sdprintf("[%d] recvfrom from socket %d\n",
+		    PL_thread_self(), socket));
+
+  n = recvfrom(s->socket, buf, (int)bufSize, flags, from, fromlen);
+  if ( n < 0 )
+  { int wsaerrno;
+
+    if ( (wsaerrno=WSAGetLastError()) == WSAEWOULDBLOCK )
+    { s->rdata.recvfrom.buffer  = buf;
+      s->rdata.recvfrom.size    = bufSize;
+      s->rdata.recvfrom.flags   = flags;
+      s->rdata.recvfrom.from    = from;
+      s->rdata.recvfrom.fromlen = fromlen;
+      placeRequest(s, REQ_RECVFROM);
+      if ( !waitRequest(s) )
+      { errno = EPLEXCEPTION;
+	return -1;
+      }
+      n = s->rdata.recvfrom.bytes;
+    }
+
+    if ( n < 0 )
+    { s->error = wsaerrno;
+      errno = EIO;
+    }
+  } else
+  { s->error = 0;
+  }
+
+#else /*__WINDOWS__*/
+
+  for(;;)
+  { if ( (flags & MSG_DONTWAIT) == 0 && !wait_socket(s) )
+    { errno = EPLEXCEPTION;
+      return -1;
+    }
+
+    n = recvfrom(s->socket, buf, bufSize, flags, from, fromlen);
+
+    if ( n == -1 && need_retry(errno) )
+    { if ( PL_handle_signals() < 0 )
+      { errno = EPLEXCEPTION;
+	return -1;
+      }
+
+      if((flags & MSG_DONTWAIT) != 0)
+        break;
+
+      continue;
+    }
+
+    break;
+  }
+
+#endif /*__WINDOWS__*/
+
+  return n;
+}
+
+
+ssize_t
+nbio_sendto(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
+	    const struct sockaddr *to, socklen_t tolen)
+{ plsocket *s;
+#ifdef __WINDOWS__
+  ssize_t n;
+#endif
+
+  if ( !(s = nbio_to_plsocket(socket)) )
+    return -1;
+
+#ifdef __WINDOWS__
+  DEBUG(3, Sdprintf("[%d] sending %d bytes using socket %d\n",
+		    PL_thread_self(), (int)bufSize, socket));
+
+  n = sendto(s->socket, buf, (int)bufSize, flags, to, tolen);
+  if ( n < 0 )
+  { int error = WSAGetLastError();
+
+    if ( error == WSAEWOULDBLOCK )
+      goto wouldblock;
+
+    DEBUG(1, Sdprintf("[%d]: sendto(%d, %d bytes): %s\n",
+		      PL_thread_self(), s->socket, (int)bufSize,
+		      WinSockError(error)));
+
+    s->error = error;
+    return -1;
+  } else
+    return n;
+
+wouldblock:
+  s->rdata.sendto.buffer  = buf;
+  s->rdata.sendto.size    = bufSize;
+  s->rdata.sendto.bytes   = 0;
+  s->rdata.sendto.flags   = flags;
+  s->rdata.sendto.to      = to;
+  s->rdata.sendto.tolen   = tolen;
+  placeRequest(s, REQ_SENDTO);
+  if ( !waitRequest(s) )
+  { errno = EPLEXCEPTION;		/* handled Prolog signal */
+    return -1;
+  }
+
+  return s->rdata.sendto.bytes;
+
+#else /*__WINDOWS__*/
+
+  return sendto(s->socket, buf, bufSize, flags, to, tolen);
+#endif /*__WINDOWS__*/
+}
