@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -60,7 +60,7 @@ with winmm.lib
 
 #include <windows.h>
 #include <sys/timeb.h>
-#if (_MSC_VER < 1400) 
+#if (_MSC_VER < 1400)
 typedef DWORD DWORD_PTR;
 #endif
 
@@ -131,10 +131,10 @@ print_trace (void)
   size_t size;
   char **strings;
   size_t i;
-     
+
   size = backtrace(array, sizeof(array)/sizeof(void *));
   strings = backtrace_symbols(array, size);
-     
+
 #ifdef _REENTRANT
   Sdprintf("on_alarm() Prolog-context [thread %d]:\n", PL_thread_self());
 #else
@@ -143,12 +143,12 @@ print_trace (void)
   PL_action(PL_ACTION_BACKTRACE, 3);
 
   Sdprintf("on_alarm() C-context:\n");
-  
+
   for(i = 0; i < size; i++)
   { if ( !strstr(strings[i], "checkData") )
       Sdprintf("\t[%d] %s\n", i, strings[i]);
   }
-       
+
   free(strings);
 }
 #endif /*BACKTRACE*/
@@ -168,7 +168,7 @@ The module contains three implementations:
 
 	* Windows
 	The Windows versions is based on multimedia timer objects.
-	
+
 	* Unix setitimer (single threaded)
 	This implementation uses setitimer() and SIGALRM.  Whenever
 	something is changed, re_schedule() is called to set the
@@ -287,14 +287,14 @@ static schedule the_schedule;		/* the schedule */
 int signal_function_set = FALSE;	/* signal function is set */
 static handler_t signal_function;	/* Current signal function */
 
-static void uninstallEvent(Event ev);
+static int removeEvent(Event ev);
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Allocate the event, maintaining a time-sorted list of scheduled events.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static Event
-allocEvent(struct timeval *at)
+allocEvent()
 { Event ev = malloc(sizeof(*ev));
 
   if ( !ev )
@@ -303,7 +303,6 @@ allocEvent(struct timeval *at)
   }
 
   memset(ev, 0, sizeof(*ev));
-  ev->at = *at;
   ev->magic = EV_MAGIC;
 
   return ev;
@@ -311,6 +310,24 @@ allocEvent(struct timeval *at)
 
 
 static void
+setTimeEvent(Event ev, double t)
+{ struct timeval tv;
+
+  gettimeofday(&tv, NULL);
+  tv.tv_usec += (long)((t-floor(t))*1000000);
+  tv.tv_sec  += (long)t;
+  if ( tv.tv_usec >= 1000000 )
+  { tv.tv_usec -= 1000000;
+    tv.tv_sec++;
+  }
+
+  ev->at = tv;
+  ev->time = t;
+}
+
+
+
+static int
 insertEvent(Event ev)
 { schedule *sched = TheSchedule();
   Event e;
@@ -320,6 +337,9 @@ insertEvent(Event ev)
   for(e = sched->first; e; e = e->next)
   { struct timeval d;
 
+    if ( e == ev )
+      return ERR_PERMISSION;		/* already scheduled */
+
     d.tv_sec  = ev->at.tv_sec  - e->at.tv_sec;
     d.tv_usec = ev->at.tv_usec - e->at.tv_usec;
     if ( d.tv_usec < 0 )
@@ -331,13 +351,14 @@ insertEvent(Event ev)
     { ev->next = e;
       ev->previous = e->previous;
       if ( e->previous )
-	e->previous->next = ev;
+      { e->previous->next = ev;
+      } else
+      { assert(sched->first == e);
+	sched->first = ev;
+      }
       e->previous = ev;
 
-      if ( sched->first == e )		/* allocated as first */
-	sched->first = ev;
-
-      return;
+      return TRUE;
     } else
     { if ( e->next )
 	continue;
@@ -345,16 +366,17 @@ insertEvent(Event ev)
       ev->previous = e;			/* end of the list */
       e->next = ev;
 
-      return;
+      return TRUE;
     }
   }
 
-  sched->first = ev;
+  sched->first = ev;			/* the very first one */
+  return TRUE;
 }
 
 
 static void
-freeEvent(Event ev)
+unlinkEvent(Event ev)
 { schedule *sched = TheSchedule();
 
   if ( sched->scheduled == ev )
@@ -367,6 +389,12 @@ freeEvent(Event ev)
 
   if ( ev->next )
     ev->next->previous = ev->previous;
+}
+
+
+static void
+freeEvent(Event ev)
+{ unlinkEvent(ev);
 
   if ( ev->goal )
     PL_erase(ev->goal);
@@ -383,7 +411,7 @@ callEvent(Event ev)
 { term_t goal = PL_new_term_ref();
 
   ev->flags |= EV_DONE;
-    
+
   PL_recorded(ev->goal, goal);
   PL_call_predicate(ev->module,
 		    PL_Q_PASS_EXCEPTION,
@@ -395,7 +423,7 @@ callEvent(Event ev)
 
 static void
 cleanupHandler()
-{ 
+{
 #ifndef __WINDOWS__
   struct itimerval v;
 
@@ -427,7 +455,7 @@ cleanup()
 
   for(ev=sched->first; ev; ev = next)
   { next = ev->next;
-    uninstallEvent(ev);
+    removeEvent(ev);
   }
 
   cleanupHandler();
@@ -470,13 +498,18 @@ callTimer(UINT id, UINT msg, DWORD_PTR dwuser, DWORD_PTR dw1, DWORD_PTR dw2)
 static int
 installEvent(Event ev)
 { MMRESULT rval;
+  int rc;
 
-  insertEvent(ev);
+  LOCK();
+  rc = insertEvent(ev);
+  UNLOCK();
+  if ( rc != TRUE )
+    return rc;
 
   rval = timeSetEvent((int)(ev->time*1000),
 		      50,			/* resolution (milliseconds) */
 		      callTimer,
-		      (DWORD)ev,
+		      (DWORD_PTR)ev,
 		      TIME_ONESHOT);
 
   if ( rval )
@@ -485,14 +518,36 @@ installEvent(Event ev)
 
     return TRUE;
   }
-    
-  return pl_error(NULL, 0, NULL, ERR_RESOURCE, "no_timers");
+
+  return ERR_RESOURCE;
 }
 
 
-static void
+static int
 uninstallEvent(Event ev)
-{ if ( TheSchedule()->scheduled == ev )
+{ LOCK();
+
+  if ( TheSchedule()->scheduled == ev )
+    ev->flags = EV_DONE;
+
+  if ( ev->mmid )
+  { timeKillEvent(ev->mmid);
+    ev->mmid = 0;
+  }
+
+  ev->flags &= ~(EV_DONE|EV_FIRED);
+
+  UNLOCK();
+
+  return TRUE;
+}
+
+
+static int
+removeEvent(Event ev)
+{ LOCK();
+
+  if ( TheSchedule()->scheduled == ev )
     ev->flags |= EV_DONE;
 
   if ( ev->mmid )
@@ -501,6 +556,9 @@ uninstallEvent(Event ev)
   }
 
   freeEvent(ev);
+  UNLOCK();
+
+  return TRUE;
 }
 
 
@@ -614,7 +672,9 @@ on_alarm(int sig)
 
 static int
 installEvent(Event ev)
-{ LOCK();
+{ int rc;
+
+  LOCK();
 
   ev->thread_id = pthread_self();
 #ifdef O_DEBUG
@@ -638,22 +698,40 @@ installEvent(Event ev)
     scheduler_running = TRUE;
   }
 
-  insertEvent(ev);
-  pthread_cond_signal(&cond);
+  if ( (rc=insertEvent(ev)) == TRUE )
+    pthread_cond_signal(&cond);
+
   UNLOCK();
 
+  return rc;
+}
+
+
+static int
+uninstallEvent(Event ev)
+{ LOCK();
+  if ( TheSchedule()->scheduled == ev )
+    ev->flags |= EV_DONE;
+  unlinkEvent(ev);
+  pthread_cond_signal(&cond);
+  ev->flags &= ~(EV_FIRED|EV_DONE);
+  UNLOCK();
+
+
   return TRUE;
 }
 
 
-static void
-uninstallEvent(Event ev)
+static int
+removeEvent(Event ev)
 { LOCK();
   if ( TheSchedule()->scheduled == ev )
     ev->flags |= EV_DONE;
   freeEvent(ev);
   pthread_cond_signal(&cond);
   UNLOCK();
+
+  return TRUE;
 }
 
 
@@ -664,7 +742,7 @@ re_schedule()
 { struct itimerval v;
   Event ev;
   schedule *sched = TheSchedule();
-  
+
   for(ev=sched->first; ev; ev = ev->next)
   { struct timeval now;
     struct timeval left;
@@ -686,7 +764,7 @@ re_schedule()
 
       callEvent(ev);		/* Time has passed.  What about exceptions? */
 
-      continue;		
+      continue;
     }
 
     sched->scheduled = ev;	/* This is the scheduled one */
@@ -737,21 +815,38 @@ on_alarm(int sig)
 
 static int
 installEvent(Event ev)
-{ insertEvent(ev);
-  re_schedule();
+{ int rc;
 
-  return TRUE;
+  if ( (rc=insertEvent(ev)) == TRUE )
+    re_schedule();
+
+  return rc;
 }
 
 
-static void
+static int
 uninstallEvent(Event ev)
 { if ( TheSchedule()->scheduled == ev )
   { ev->flags |= EV_DONE;
     re_schedule();
   }
 
+  ev->flags &= ~(EV_DONE|EV_FIRED);
+
+  return TRUE;
+}
+
+
+static int
+removeEvent(Event ev)
+{ if ( TheSchedule()->scheduled == ev )
+  { ev->flags |= EV_DONE;
+    re_schedule();
+  }
+
   freeEvent(ev);
+
+  return TRUE;
 }
 
 #endif /*SHARED_TABLE*/
@@ -762,6 +857,20 @@ uninstallEvent(Event ev)
 		 *	PROLOG CONNECTION	*
 		 *******************************/
 
+int
+alarm_error(term_t alarm, int err)
+{ switch(err)
+  { case ERR_RESOURCE:
+      return pl_error(NULL, 0, NULL, ERR_RESOURCE, "timers");
+    case ERR_PERMISSION:
+      return pl_error(NULL, 0, "already installed", ERR_PERMISSION,
+		      alarm, "install", "alarm");
+    default:
+      assert(0);
+  }
+}
+
+
 static int
 unify_timer(term_t t, Event ev)
 { if ( !PL_is_variable(t) )
@@ -814,7 +923,7 @@ alarm4(term_t time, term_t callable, term_t id, term_t options)
   struct timeval tv;
   module_t m = NULL;
   unsigned long flags = 0L;
-    
+
   if ( options )
   { term_t tail = PL_copy_term_ref(options);
     term_t head = PL_new_term_ref();
@@ -854,7 +963,7 @@ alarm4(term_t time, term_t callable, term_t id, term_t options)
   if ( !PL_get_float(time, &t) )
     return pl_error(NULL, 0, NULL, ERR_ARGTYPE, 1,
 		    time, "number");
-		    
+
   gettimeofday(&tv, NULL);
   tv.tv_usec += (long)((t-floor(t))*1000000);
   tv.tv_sec  += (long)t;
@@ -863,11 +972,11 @@ alarm4(term_t time, term_t callable, term_t id, term_t options)
     tv.tv_sec++;
   }
 
-  if ( !(ev = allocEvent(&tv)) )
+  if ( !(ev = allocEvent()) )
     return FALSE;
-  ev->time = t;
+  setTimeEvent(ev, t);
   if ( !unify_timer(id, ev) )
-  { freeEvent(ev);
+  { freeEvent(ev);			/* not linked: no need to lock */
     return FALSE;
   }
 
@@ -877,9 +986,11 @@ alarm4(term_t time, term_t callable, term_t id, term_t options)
   ev->goal = PL_record(callable);
 
   if ( !(ev->flags & EV_NOINSTALL) )
-  { if ( !installEvent(ev) )
-    { freeEvent(ev);
-      return FALSE;
+  { int rc;
+
+    if ( (rc=installEvent(ev)) != TRUE )
+    { freeEvent(ev);			/* not linked: no need to lock */
+      return alarm_error(id, rc);
     }
   }
 
@@ -896,19 +1007,50 @@ alarm3(term_t time, term_t callable, term_t id)
 static foreign_t
 install_alarm(term_t alarm)
 { Event ev = NULL;
+  int rc;
 
   if ( !get_timer(alarm, &ev) )
     return FALSE;
 
-  if ( !installEvent(ev) )
-  { freeEvent(ev);
-    return FALSE;
-  }
+  if ( (rc=installEvent(ev)) != TRUE )
+    return alarm_error(alarm, rc);
 
   return TRUE;
 }
 
 
+static foreign_t
+install_alarm2(term_t alarm, term_t time)
+{ Event ev = NULL;
+  double t;
+  int rc;
+
+  if ( !get_timer(alarm, &ev) )
+    return FALSE;
+
+  if ( !PL_get_float(time, &t) )
+    return pl_error(NULL, 0, NULL, ERR_ARGTYPE, 1,
+		    time, "number");
+
+  setTimeEvent(ev, t);
+  if ( (rc=installEvent(ev)) != TRUE )
+    return alarm_error(alarm, rc);
+
+  return TRUE;
+}
+
+
+static foreign_t
+uninstall_alarm(term_t alarm)
+{ Event ev = NULL;
+
+  if ( !get_timer(alarm, &ev) )
+    return FALSE;
+
+  return uninstallEvent(ev);
+}
+
+
 static foreign_t
 remove_alarm(term_t alarm)
 { Event ev = NULL;
@@ -916,9 +1058,7 @@ remove_alarm(term_t alarm)
   if ( !get_timer(alarm, &ev) )
     return FALSE;
 
-  uninstallEvent(ev);
-
-  return TRUE;
+  return removeEvent(ev);
 }
 
 
@@ -956,7 +1096,7 @@ current_alarms(term_t time, term_t goal, term_t id, term_t status,
       s = ATOM_next;
     else
       s = ATOM_scheduled;
-    
+
     if ( !PL_unify_atom(status, s) )
       goto nomatch;
 
@@ -973,7 +1113,7 @@ current_alarms(term_t time, term_t goal, term_t id, term_t status,
 
     if ( !unify_timer(id, ev) )
       goto nomatch;
-      
+
     PL_discard_foreign_frame(fid);
 
     PL_put_float(av+0, at);		/* time */
@@ -1021,7 +1161,9 @@ install()
   PL_register_foreign("alarm",          4, alarm4,         PL_FA_TRANSPARENT);
   PL_register_foreign("alarm",          3, alarm3,         PL_FA_TRANSPARENT);
   PL_register_foreign("remove_alarm",   1, remove_alarm,   0);
+  PL_register_foreign("uninstall_alarm",1, uninstall_alarm,0);
   PL_register_foreign("install_alarm",  1, install_alarm,  0);
+  PL_register_foreign("install_alarm",  2, install_alarm2, 0);
   PL_register_foreign("remove_alarm_notrace",1, remove_alarm,   PL_FA_NOTRACE);
   PL_register_foreign("current_alarms", 5, current_alarms, 0);
 #ifdef O_DEBUG
