@@ -1,11 +1,11 @@
-/*  $Id: SWI-cpp.h,v 1.10 2005/12/15 18:17:09 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -22,6 +22,9 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+#ifndef _SWI_CPP_H
+#define _SWI_CPP_H
+
 #include <SWI-Prolog.h>
 #include <string.h>
 #ifndef __APPLE__
@@ -29,8 +32,8 @@
 #endif
 
 #ifdef __BORLANDC__
-#define __inline inline 
-#endif 
+#define __inline inline
+#endif
 
 #ifdef _MSC_VER			/* MSVC doesn't know throw doesn't return */
 #define PL_THROWN(value)	return value;
@@ -97,7 +100,7 @@ public:
   PlTerm(term_t t)
   { ref = t;
   }
-  
+
 					/* C --> PlTerm */
   PlTerm(const char *text)
   { ref = PL_new_term_ref();
@@ -237,7 +240,7 @@ public:
 class PlCompound : public PlTerm
 {
 public:
-  
+
   PlCompound(const char *text);
   PlCompound(const char *functor, const PlTermv &args);
 };
@@ -257,7 +260,7 @@ public:
 
 
 class PlCodeList : public PlTerm
-{ 
+{
 public:
 
   PlCodeList(const char *text) : PlTerm()
@@ -353,7 +356,7 @@ public:
 class PlTail : public PlTerm
 {
 public:
-  
+
   PlTail(const PlTerm &l)
   { if ( PL_is_variable(l.ref) || PL_is_list(l.ref) )
       ref = PL_copy_term_ref(l.ref);
@@ -397,21 +400,28 @@ public:
 class PlRegister
 {
 public:
-  
-  PlRegister(const char *name, int arity,
+
+  PlRegister(const char *module, const char *name, int arity,
 	    foreign_t (f)(term_t t0, int a, control_t ctx))
-  { PL_register_foreign(name, arity, (void *)f, PL_FA_VARARGS);
+  { PL_register_foreign_in_module(module, name, arity, (void *)f, PL_FA_VARARGS);
   }
 
-  PlRegister(const char *name, foreign_t (*f)(PlTerm a0))
-  { PL_register_foreign(name, 1, (void *)f, 0);
+  PlRegister(const char *module, const char *name, foreign_t (*f)(PlTerm a0))
+  { PL_register_foreign_in_module(module, name, 1, (void *)f, 0);
   }
-  PlRegister(const char *name, foreign_t (*f)(PlTerm a0, PlTerm a1))
-  { PL_register_foreign(name, 2, (void *)f, 0);
+  PlRegister(const char *module, const char *name, foreign_t (*f)(PlTerm a0, PlTerm a1))
+  { PL_register_foreign_in_module(module, name, 2, (void *)f, 0);
   }
-  PlRegister(const char *name, foreign_t (*f)(PlTerm a0, PlTerm a1, PlTerm a2))
-  { PL_register_foreign(name, 3, (void *)f, 0);
+  PlRegister(const char *module, const char *name, foreign_t (*f)(PlTerm a0, PlTerm a1, PlTerm a2))
+  { PL_register_foreign_in_module(module, name, 3, (void *)f, 0);
   }
+
+  // for non-deterministic calls
+  PlRegister(const char *module, const char *name, int arity,
+             foreign_t (f)(term_t t0, int a, control_t ctx), short flags)
+  { PL_register_foreign_in_module(module, name, arity, (void *)f, flags);
+  }
+
 };
 
 
@@ -445,13 +455,19 @@ public:
 
   PlQuery(const char *name, const PlTermv &av)
   { predicate_t p = PL_predicate(name, av.size, "user");
-    
+
     qid = PL_open_query((module_t)0, PL_Q_CATCH_EXCEPTION, p, av.a0);
   }
   PlQuery(const char *module, const char *name, const PlTermv &av)
-  { predicate_t p = PL_predicate(name, av.size, module);
-    
-    qid = PL_open_query((module_t)0, PL_Q_CATCH_EXCEPTION, p, av.a0);
+  { atom_t ma = PL_new_atom(module);
+    atom_t na = PL_new_atom(name);
+    module_t m = PL_new_module(ma);
+    predicate_t p = PL_pred(PL_new_functor(na, av.size), m);
+
+    PL_unregister_atom(ma);
+    PL_unregister_atom(na);
+
+    qid = PL_open_query(m, PL_Q_CATCH_EXCEPTION, p, av.a0);
   }
 
   ~PlQuery()
@@ -515,7 +531,7 @@ __inline PlTerm::operator char *(void) const
 
 __inline PlTerm::operator long(void) const
 { long v;
-  
+
   if ( PL_get_long(ref, &v) )
     return v;
 
@@ -525,7 +541,7 @@ __inline PlTerm::operator long(void) const
 
 __inline PlTerm::operator int(void) const
 { int v;
-  
+
   if ( PL_get_integer(ref, &v) )
     return v;
 
@@ -535,7 +551,7 @@ __inline PlTerm::operator int(void) const
 
 __inline PlTerm::operator double(void) const
 { double v;
-  
+
   if ( PL_get_float(ref, &v) )
     return v;
 
@@ -545,7 +561,7 @@ __inline PlTerm::operator double(void) const
 
 __inline PlTerm::operator PlAtom(void) const
 { atom_t v;
-  
+
   if ( PL_get_atom(ref, &v) )
     return PlAtom(v);
 
@@ -590,7 +606,7 @@ __inline int
 PlTerm::arity()
 { atom_t name;
   int arity;
-  
+
   if ( PL_get_name_arity(ref, &name, &arity) )
     return arity;
 
@@ -603,7 +619,7 @@ __inline const char *
 PlTerm::name()
 { atom_t name;
   int arity;
-  
+
   if ( PL_get_name_arity(ref, &name, &arity) )
     return PL_atom_chars(name);
 
@@ -681,7 +697,7 @@ __inline int PlTerm::operator ==(const char *s)
 
   if ( PL_get_chars(ref, &s0, CVT_ALL) )
     return strcmp(s0, s) == 0;
-  
+
   throw PlTypeError("text", ref);
   PL_THROWN(0);
 }
@@ -692,7 +708,7 @@ __inline int PlTerm::operator ==(const PlAtom &a)
 
   if ( PL_get_atom(ref, &v) )
     return v == a.handle;
-  
+
   throw PlTypeError("atom", ref);
   PL_THROWN(0);
 }
@@ -884,10 +900,8 @@ public:
 		 *     REGISTER PREDICATES	*
 		 *******************************/
 
-#ifdef PROLOG_MODULE
-#define PlPredName(name) PROLOG_MODULE ":" #name
-#else
-#define PlPredName(name) #name
+#ifndef PROLOG_MODULE
+#define PROLOG_MODULE (const char*)NULL
 #endif
 
 #define PREDICATE(name, arity) \
@@ -902,17 +916,37 @@ public:
 	  { return ex.plThrow(); \
 	  } \
 	} \
-	static PlRegister _x ## name ## __ ## arity(PlPredName(name), \
-						    arity, \
+	static PlRegister _x ## name ## __ ## arity(PROLOG_MODULE, #name, arity, \
 					    _pl_ ## name ## __ ## arity); \
 	static foreign_t pl_ ## name ## __ ## arity(PlTermv _av)
-#define A1  _av[0]        
-#define A2  _av[1]        
-#define A3  _av[2]        
-#define A4  _av[3]        
-#define A5  _av[4]        
-#define A6  _av[5]        
-#define A7  _av[6]        
-#define A8  _av[7]        
-#define A9  _av[8]        
-#define A10 _av[9]        
+
+
+#define PREDICATE_NONDET(name, arity)          \
+	static foreign_t \
+	pl_ ## name ## __ ## arity(PlTermv _av, foreign_t handle);       \
+	static foreign_t \
+	_pl_ ## name ## __ ## arity(term_t t0, int a, control_t c) \
+	{ try \
+	  { \
+	    return pl_ ## name ## __ ## arity(PlTermv(arity, t0), (foreign_t)c); \
+	  } catch ( PlException &ex ) \
+	  { return ex.plThrow(); \
+	  } \
+	} \
+        static PlRegister _x ## name ## __ ## arity(PROLOG_MODULE, #name, arity, \
+                                                    _pl_ ## name ## __ ## arity, \
+                                                    PL_FA_NONDETERMINISTIC | PL_FA_VARARGS); \
+	static foreign_t pl_ ## name ## __ ## arity(PlTermv _av, foreign_t handle)
+
+#define A1  _av[0]
+#define A2  _av[1]
+#define A3  _av[2]
+#define A4  _av[3]
+#define A5  _av[4]
+#define A6  _av[5]
+#define A7  _av[6]
+#define A8  _av[7]
+#define A9  _av[8]
+#define A10 _av[9]
+
+#endif /*_SWI_CPP_H*/
