@@ -1,11 +1,11 @@
-/*  $Id: pl.pl,v 1.33 2006/08/29 10:25:57 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -46,6 +46,16 @@
 #(defitem(Label),	[ html('<DT>'), Label,
 			  html('<DD class="defbody">')
 			]).
+#(predtag(Value),	[ html('<span class="pred-tag">'), Value,
+			  html('</span>')
+			]).
+
+		 /*******************************
+		 *	   ENVIRONMENTS		*
+		 *******************************/
+
+list_command(tags,       _, html('<DL>'), html('</DL>')).
+
 
 		 /*******************************
 		 *	    INDEX HACK		*
@@ -61,7 +71,7 @@ cmd(=, index, nospace('\=')).
 cmd(spaces({X}), html(Spaces)) :-
 	atom_number(X, N),
 	n_list(N, '&nbsp;', L),
-	concat_atom(L, Spaces).
+	atomic_list_concat(L, Spaces).
 cmd(hrule, html('<HR>')).
 cmd(bug({TeX}), #footnote(bug, +TeX)).
 cmd(fileext({Ext}), #code(Text)) :-
@@ -84,6 +94,7 @@ cmd(exam(		{A1}), #code(+A1)).
 cmd(jargon(		{A1}), #em(+A1)).
 cmd(chr(		{A1}), #code(+A1)).
 cmd(const(		{A1}), #code(+A1)).
+cmd(resource(		{A1}), #code(A1)).
 cmd(key(		{A1}), #code(+A1)).
 cmd(plflag(		{A1}), #code(+A1)).
 cmd(module(		{A1}), #code(+A1)).
@@ -101,7 +112,7 @@ cmd(versionshort,	    _, nospace(Version)) :-
 	Major is V // 10000,
 	Minor is (V // 100) mod 100,
 	Patch is V mod 100,
-	concat_atom([Major, Minor, Patch], '.', Version).
+	atomic_list_concat([Major, Minor, Patch], '.', Version).
 cmd(bnfor, '|').
 cmd(bnfmeta({Meta}), [nospace('<'), #var(+Meta), nospace('>')]).
 cmd(argoption({RawName}, {ArgName}),
@@ -112,9 +123,17 @@ cmd(predref({RawName}, {Arity}), #lref(pred, RefName, Text)) :-
 	clean_name(RawName, Name),
 	predicate_refname(Name, Arity, RefName),
 	sformat(Text, '~w/~w', [Name, Arity]).
+cmd(dcgref({RawName}, {DCGArity}), #lref(pred, RefName, Text)) :-
+	clean_name(RawName, Name),
+	atom_number(DCGArity, ArityInt),
+	Arity is ArityInt + 2,
+	predicate_refname(Name, Arity, RefName),
+	sformat(Text, '~w/~w', [Name, Arity]).
 cmd(nopredref({RawName}, {Arity}), Text) :-
 	clean_name(RawName, Name),
 	sformat(Text, '~w/~w', [Name, Arity]).
+cmd(prologflag({Name}), #lref(flag, RefName, Name)) :-
+	atom_concat('flag:', Name, RefName).
 cmd(functor({RawName}, {Arity}), Text) :-
 	clean_name(RawName, Name),
 	sformat(Text, '~w/~w', [Name, Arity]).
@@ -134,35 +153,46 @@ cmd(definition({Tag}),
     #defitem(#b(+Tag))).
 cmd('DCG'(A,B,C), X) :-
 	cmd(predicate(A,B,C), X).
-cmd(predicate({RawName}, {'0'}, {_}),
-    #defitem(pubdef, #label(RefName, #strong(Name)))) :-
+cmd(predicate(A, {RawName}, {'0'}, {_}),
+    #defitem(pubdef, Content)) :-
+	pred_tag(A, Content, [#label(RefName, #strong(Name))]),
 	clean_name(RawName, Name),
 	sformat(RefName, '~w/0', [Name]),
 	add_to_index(RefName, +RefName).
-cmd(predicate({RawName}, {Arity}, {Args}),
+cmd(predicate(A, {RawName}, {Arity}, {Args}),
+    #defitem(pubdef, Content)) :-
+	pred_tag(A, Content,
+		 [#label(RefName, [#strong(Name), #embrace(#var(+Args))])]),
+	clean_name(RawName, Name),
+	sformat(RefName, '~w/~w', [Name, Arity]),
+	add_to_index(RefName, +RefName).
+cmd(dcg(A, {RawName}, {'0'}, {_}),
+    #defitem(pubdef, Content)) :-
+	pred_tag(A, Content, [#label(RefName, #strong(Name)), #code(//)]),
+	clean_name(RawName, Name),
+	sformat(RefName, '~w/0', [Name]),
+	add_to_index(RefName, +RefName).
+cmd(dcg(A, {RawName}, {Arity}, {Args}),
+    #defitem(pubdef, Content)) :-
+	pred_tag(A, Content,
+		 [ #label(RefName,
+			  [ #strong(Name), #embrace(#var(+Args))
+			  ]),
+		   #code(//)
+		 ]),
+	clean_name(RawName, Name),
+	sformat(RefName, '~w/~w', [Name, Arity]),
+	add_to_index(RefName, +RefName).
+cmd(directive({RawName}, {'0'}, {_}),
     #defitem(pubdef, #label(RefName,
-			    [ #strong(Name), #embrace(#var(+Args))
-			    ]))) :-
+			    [ ':- ', #strong(Name)
+			    ]))) :- !,
 	clean_name(RawName, Name),
-	sformat(RefName, '~w/~w', [Name, Arity]),
-	add_to_index(RefName, +RefName).
-cmd(dcg({RawName}, {'0'}, {_}),
-    #defitem(pubdef, [#label(RefName, #strong(Name)), #code(-->)])) :-
-	clean_name(RawName, Name),
-	sformat(RefName, '~w/0', [Name]),
-	add_to_index(RefName, +RefName).
-cmd(dcg({RawName}, {Arity}, {Args}),
-    #defitem(pubdef, [#label(RefName,
-			     [ #strong(Name), #embrace(#var(+Args))
-			     ]),
-	      #code(-->)
-	     ])) :-
-	clean_name(RawName, Name),
-	sformat(RefName, '~w/~w', [Name, Arity]),
+	sformat(RefName, '~w/~w', [Name, 0]),
 	add_to_index(RefName, +RefName).
 cmd(directive({RawName}, {Arity}, {Args}),
     #defitem(pubdef, #label(RefName,
-			    [ #strong(Name), #embrace(#var(Args))
+			    [ ':- ', #strong(Name), #embrace(#var(Args))
 			    ]))) :-
 	clean_name(RawName, Name),
 	sformat(RefName, '~w/~w', [Name, Arity]),
@@ -182,15 +212,24 @@ cmd(cmacro({RType}, {Name}, {Args}),
 			    ]))) :-
 	sformat(RefName, '~w()', [Name]),
 	add_to_index(RefName, +RefName).
-cmd(prefixop({RawName}, {Arg}),
-    #defitem(pubdef, #label(RefName, [#strong(Name), ' ', #var(Arg)]))) :-
+cmd(resitem({Resource}),
+    #defitem(pubdef, #label(Resource,
+			    [ #strong(Resource)
+			    ]))) :-
+	add_to_index(Resource, +Resource).
+cmd(prefixop(A, {RawName}, {Arg}),
+    #defitem(pubdef, Content)) :-
+	pred_tag(A, Content,
+		 #label(RefName, [#strong(Name), ' ', #var(Arg)])),
 	clean_name(RawName, Name),
 	predicate_refname(Name, 1, RefName),
 	add_to_index(RefName, +RefName).
-cmd(infixop({RawName}, {Arg1}, {Arg2}),
-    #defitem(pubdef, #label(RefName,
-		    [ #var(Arg1), ' ', #strong(Name), ' ', #var(Arg2)
-		    ]))) :-
+cmd(infixop(A, {RawName}, {Arg1}, {Arg2}),
+    #defitem(pubdef, Content)) :-
+	pred_tag(A, Content,
+		 #label(RefName,
+			[ #var(Arg1), ' ', #strong(Name), ' ', #var(Arg2)
+			])),
 	clean_name(RawName, Name),
 	predicate_refname(Name, 2, RefName),
 	add_to_index(RefName, +RefName).
@@ -200,8 +239,11 @@ cmd(constitem({Name}), #defitem(#label(RefName, #strong(+Name)))) :-
 cmd(termitem({Name}, {[]}), #defitem(#strong(+Name))).
 cmd(termitem({Name}, {Arg}),
     #defitem([#strong(+Name), #embrace(#var(+Arg))])).
+cmd(infixtermitem({Name}, {Left}, {Right}),
+    #defitem([#var(+Left), ' ', #strong(+Name), ' ', #var(+Right)])).
 cmd(prologflagitem({Name}, {Type}, {Access}),
-    #defitem([#strong(Name), #embrace([#var(Type)|Change])])) :-
+    #defitem(pubdef, #label(RefName, [#strong(Name), #embrace([#var(Type)|Change])]))) :-
+	atom_concat('flag:', Name, RefName),
 	(   Access == r
 	->  Change = []
 	;   Change = nospace(', changeable')
@@ -235,13 +277,13 @@ cmd(menuitem({Name}, {[]}),
     #defitem(#label(RefName, #strong(+Name)))) :-
 	clean_name(Name, RefName0),
 	atom_concat('menu:', RefName0, RefName),
-	concat_atom(Name, ' ', Atom),
+	atomic_list_concat(Name, ' ', Atom),
 	add_to_index(Atom, +RefName).
 cmd(menuitem({Name}, {Arg}),
     #defitem([#label(RefName, #strong(+Name)), ' ', #embrace(#var(+Arg))])) :-
 	clean_name(Name, RefName0),
 	atom_concat('menu:', RefName0, RefName),
-	concat_atom(Name, ' ', Atom),
+	atomic_list_concat(Name, ' ', Atom),
 	add_to_index(Atom, +RefName).
 cmd(escapeitem({Name}), #defitem(#code([nospace('\\'), +Name]))).
 cmd(ttdef({Def}), #defitem(#code(+Def))).
@@ -286,7 +328,7 @@ cmd(class({Name}),              #lref(Label, Name)) :-
 cmd(menuref({A1}),            #lref(RefName, Name)) :-
 	clean_name(A1, RefName0),
 	atom_concat('menu:', RefName0, RefName),
-	concat_atom(A1, ' ', Name),
+	atomic_list_concat(A1, ' ', Name),
         add_to_index(Name).
 
 % Glossary support
@@ -321,6 +363,9 @@ select_csym([_|T0], T) :-
 	select_csym(T0, T).
 
 
+pred_tag([], L, L).
+pred_tag([Value], [#predtag(#embrace("[]", +Value))|L], L).
+
 		 /*******************************
 		 *	     GLOSSARY		*
 		 *******************************/
@@ -354,91 +399,109 @@ canonical_char(X, X).
 
 remove_trailing_spaces([], []).
 remove_trailing_spaces([0' |T], []) :-
-	maplist(=(0' ), T), !.
+	maplist(=(0' ), T), !.		% '
 remove_trailing_spaces([H|T0], [H|T]) :-
 	remove_trailing_spaces(T0, T).
 
+
+		 /*******************************
+		 *	   PlDoc KEYWORDS	*
+		 *******************************/
+
+cmd(tag({Tag}),
+    [ html('<DT><B>'), +Tag, html('</B><DD>') ]).
+cmd(mtag({Tag}),
+    [ html('<DT><B>'), +Tag, html('</B><DD>') ]).
+
+cmd(param({Param}, {Description}),
+    [ html('<TR><TD><VAR>'), +Param, html('</VAR></TD>'),
+      html('<TD class="argdescr"> -'), +Description, html('<TD></TR>')
+    ]).
+
+
 		 /*******************************
 		 *               C		*
 		 *******************************/
 
+cmd(backslash, #code(\)).
+cmd(bsl, #code(\)).
 cmd(Cmd, HTML) :-
-	special(Cmd, Atom),
+	urldef(Cmd, Atom), !,
 	HTML = #code(Atom).
 
 
-special('Sexe', '#!').
-special('Scgt', #>).
-special('Scge', #>=).
-special('Sclt', #<).
-special('Scle', #=<).
-special('Scne', #\=).
-special('Sceq', '#=').
-special('Sequiv', #<=>).
-special('Slimpl', #<=).
-special('Srimpl', #=>).
-special('Scine', #=\=).
-special('Scieq', #=:=).
-special('Scut', !).
-special('Scomma',  (,)).
-special('Sifthen',  (->)).
-special('Ssoftcut',  (*->)).
-special('Sdot', '.').
-special('Ssemicolon',  (;)).
-special('Slt', <).
-special('Seq', =).
-special('Suniv', =..).
-special('Saeq', =:=).
-special('Sle', =<).
-special('Sel', <=).
-special('Sequal', ==).
-special('Sstructeq', =@=).
-special('Sstructneq', \=@=).
-special('Sane', =\=).
-special('Sgt', >).
-special('Sge', >=).
-special('Stlt', @<).
-special('Stle', @=<).
-special('Stgt', @>).
-special('Stge', @>=).
-special('Snot', \+).
-special('Sne', \=).
-special('Snequal', \==).
-special('Shat', ^).
-special('Sbar',  ('|')).
-special('Stimes', *).
-special('Spow', **).
-special('Splus', +).
-special('Sminus', -).
-special('Sdiv', /).
-special('Sidiv', //).
-special('Sand', /\).
-special('Slshift', <<).
-special('Srshift', >>).
-special('Sneg', \).
-special('Sesc', \).
-special('Sor', \/).
-special('Sdollar', $).
-special('Squest', ?).
-special('Smodule', :).
-special('Sneck',  (:-)).
-special('Sdirective',  (?-)).
-special('Sdcg',  (-->)).
-special('Scurl',  ({})).
-special('SxXX',  'xXX..\\').
-special('Bc', '\\c').
-special('Bn', '\\n').
-special('Br', '\\r').
-special('Bl', '\\l').
-special('BB', \\).
-special('Stilde', ~).
-special('Spercent', '%').
-special('Shash', #).
-special(backslash, \).
-special(bsl, \).
+		 /*******************************
+		 *    LATEX SPECIAL SEQUENCES	*
+		 *******************************/
+
+%	NOTE: This code is copied from doc_latex.pl from PlDoc.
+
+%%	urldef(?DefName, ?String)
+%
+%	True if \DefName is  a  urldef   for  String.  UrlDefs are LaTeX
+%	sequences that can be used to  represent strings with symbols in
+%	fragile environments. Whenever a word can   be  expressed with a
+%	urldef, we will  do  this  to   enhance  the  robustness  of the
+%	generated LaTeX code.
+
+:- dynamic
+	urldef/2,
+	urldefs_loaded/1.
+
+%%	load_urldefs.
+%%	load_urldefs(+File)
+%
+%	Load   =|\urldef|=   definitions   from    File   and   populate
+%	urldef_name/2. See =|pldoc.sty|= for details.
+
+load_urldefs :-
+	urldefs_loaded(_), !.
+load_urldefs :-
+	absolute_file_name(library('pldoc/pldoc.sty'), File,
+			   [ access(read) ]),
+	load_urldefs(File).
+
+load_urldefs(File) :-
+	urldefs_loaded(File), !.
+load_urldefs(File) :-
+	open(File, read, In),
+	call_cleanup((   read_line_to_codes(In, L0),
+			 process_urldefs(L0, In)),
+		     close(In)),
+	assert(urldefs_loaded(File)).
+
+process_urldefs(end_of_file, _) :- !.
+process_urldefs(Line, In) :-
+	(   phrase(urldef(Name, String), Line)
+	->  assert(urldef(Name, String))
+	;   true
+	),
+	read_line_to_codes(In, L2),
+	process_urldefs(L2, In).
+
+urldef(Name, String) -->
+	"\\urldef{\\", string(NameS), "}\\satom{", string(StringS), "}",
+	ws,
+	(   "%"
+	->  string(_)
+	;   []
+	),
+	eol, !,
+	{ atom_codes(Name, NameS),
+	  atom_codes(String, StringS)
+	}.
+
+ws --> [C], { C =< 32 }, !, ws.
+ws --> [].
+
+string([]) --> [].
+string([H|T]) --> [H], string(T).
+
+eol([],[]).
+
 
 clean_name([\Special], Out) :-
-	special(Special, Out), !.
+	urldef(Special, Out), !.
 clean_name([\tt, Out], Out) :- !.
 clean_name($(Out), Out) :- !.
 clean_name([Out], Out) :- !.
@@ -446,13 +509,13 @@ clean_name(X, X) :-
 	atomic(X), !.
 clean_name(L, Out) :-
 	maplist(clean_name, L, L2),
-	concat_atom(L2, Out).
-	
+	atomic_list_concat(L2, Out).
+
 predicate_refname(Symbol, Arity, Ref) :-
 	symbol_name(Symbol, Name), !,
-	concat_atom([Name, /, Arity], Ref).
+	atomic_list_concat([Name, /, Arity], Ref).
 predicate_refname(Name, Arity, Ref) :-
-	concat_atom([Name, /, Arity], Ref).
+	atomic_list_concat([Name, /, Arity], Ref).
 
 symbol_name('->',	send_arrow).
 symbol_name('<-',	get_arrow).
@@ -462,3 +525,5 @@ n_list(N, X, [X|T]) :-
 	N > 0,
 	N2 is N - 1,
 	n_list(N2, X, T).
+
+:- load_urldefs.
