@@ -1,11 +1,11 @@
-/*  $Id: odbc.c,v 1.75 2007/06/06 07:20:43 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -70,6 +70,14 @@ static int odbc_debuglevel = 0;
 #define SQLULEN SQLUINTEGER
 #endif
 
+#ifndef SQL_COPT_SS_MARS_ENABLED
+#define SQL_COPT_SS_MARS_ENABLED 1224
+#endif
+
+#ifndef SQL_MARS_ENABLED_YES
+#define SQL_MARS_ENABLED_YES (SQLPOINTER)1
+#endif
+
 #include <SWI-Stream.h>
 #include <SWI-Prolog.h>
 #include <stdio.h>
@@ -109,9 +117,11 @@ static time_t timegm(struct tm *tm);
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Work around bug in MS SQL Server that reports NumCols of SQLColumns()
 as 19, while there are only 12.  Grrr!
+
+This bug appears fixed now, so we'll remove the work-around
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define SQL_SERVER_BUG	1
+/*#define SQL_SERVER_BUG	1*/
 
 static atom_t    ATOM_row;		/* "row" */
 static atom_t    ATOM_informational;	/* "informational" */
@@ -163,7 +173,9 @@ static functor_t FUNCTOR_odbc_statement1; /* $odbc_statement(Id) */
 static functor_t FUNCTOR_odbc_connection1;
 static functor_t FUNCTOR_user1;
 static functor_t FUNCTOR_password1;
+static functor_t FUNCTOR_driver_string1;
 static functor_t FUNCTOR_alias1;
+static functor_t FUNCTOR_mars1;
 static functor_t FUNCTOR_open1;
 static functor_t FUNCTOR_auto_commit1;
 static functor_t FUNCTOR_types1;
@@ -241,7 +253,7 @@ typedef struct connection
   nulldef     *null;			/* Prolog null value */
   unsigned     flags;			/* general flags */
   int	       max_qualifier_lenght;	/* SQL_MAX_QUALIFIER_NAME_LEN */
-  int	       max_nogetdata;		/* handle as long field if larger */
+  SQLULEN      max_nogetdata;		/* handle as long field if larger */
   struct connection *next;		/* next in chain */
 } connection;
 
@@ -261,7 +273,7 @@ typedef struct
   unsigned     flags;			/* general flags */
   nulldef     *null;			/* Prolog null value */
   findall     *findall;			/* compiled code to create result */
-  int	       max_nogetdata;		/* handle as long field if larger */
+  SQLULEN      max_nogetdata;		/* handle as long field if larger */
   struct context *clones;		/* chain of clones */
 } context;
 
@@ -335,7 +347,7 @@ odbc_report(HENV henv, HDBC hdbc, HSTMT hstmt, RETCODE rc)
       PL_unify_term(msg, PL_FUNCTOR, FUNCTOR_odbc3,
 			   PL_CHARS,   state,
 		           PL_INTEGER, (long)native,
-		           PL_NCHARS,  msglen, message);
+		           PL_NCHARS,  (size_t)msglen, message);
       break;
     case SQL_INVALID_HANDLE:
       return PL_warning("ODBC INTERNAL ERROR: Invalid handle in error");
@@ -349,18 +361,18 @@ odbc_report(HENV henv, HDBC hdbc, HSTMT hstmt, RETCODE rc)
     { fid_t fid = PL_open_foreign_frame();
       predicate_t pred = PL_predicate("print_message", 2, "user");
       term_t av = PL_new_term_refs(2);
-      
+
       PL_put_atom(av+0, ATOM_informational);
       PL_put_term(av+1, msg);
-	
+
       PL_call_predicate(NULL, PL_Q_NORMAL, pred, av);
       PL_discard_foreign_frame(fid);
-	
+
       return TRUE;
     }
     case SQL_ERROR:
     { term_t ex = PL_new_term_ref();
-      
+
       PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
 		    PL_TERM, msg,
 		    PL_VARIABLE);
@@ -536,7 +548,7 @@ static int
 PL_get_typed_arg_ex(int i, term_t t, int (*func)(), const char *ex, void *ap)
 { term_t a = PL_new_term_ref();
 
-  if ( !PL_get_arg(i, t, a) ) 
+  if ( !PL_get_arg(i, t, a) )
     return type_error(t, "compound");
   if ( !(*func)(a, ap) )
     return type_error(a, ex);
@@ -551,7 +563,7 @@ static int
 PL_get_typed_arg(int i, term_t t, int (*func)(), void *ap)
 { term_t a = PL_new_term_ref();
 
-  if ( !PL_get_arg(i, t, a) ) 
+  if ( !PL_get_arg(i, t, a) )
     return FALSE;
   return (*func)(a, ap);
 }
@@ -599,7 +611,7 @@ formatted_string(term_t in, size_t *len, char **out)
       PL_free(*out);
     return FALSE;
   }
-  
+
   Sclose(fd);
   return TRUE;
 }
@@ -623,7 +635,7 @@ nulldef_spec(term_t t)
 
   if ( !(nd=odbc_malloc(sizeof(*nd))) )
     return NULL;
-  
+
   memset(nd, 0, sizeof(*nd));
 
   if ( PL_get_atom(t, &a) )
@@ -649,7 +661,7 @@ nulldef_spec(term_t t)
   } else
   { term:
     nd->nulltype = NULL_RECORD;
-    nd->nullvalue.record = PL_record(t);  
+    nd->nullvalue.record = PL_record(t);
   }
 
   nd->references = 1;
@@ -755,7 +767,7 @@ The current implementation is incomplete. It does not allow arguments of
 row(...) to be instantiated. Plain instantiation   can always be avoided
 using a proper SELECT statement. Potentionally   useful however would be
 the  translation  of   compound   terms,    especially   to   translates
-date/time/timestamp structures to a format for use by the application. 
+date/time/timestamp structures to a format for use by the application.
 
 The  statement  is  compiled  into  a    findall  statement,  a  set  of
 instructions that builds the target structure   from the row returned by
@@ -818,7 +830,7 @@ compile_arg(compile_info *info, term_t t)
       if ( true(info, CTX_PERSISTENT) )
 	PL_register_atom(val);
       break;
-    } 
+    }
     case PL_STRING:
     case PL_FLOAT:
       if ( true(info, CTX_PERSISTENT) )
@@ -877,7 +889,7 @@ compile_arg(compile_info *info, term_t t)
 
 static findall *
 compile_findall(term_t all, unsigned flags)
-{ compile_info info; 
+{ compile_info info;
   term_t t = PL_new_term_ref();
   atom_t a;
   findall *f;
@@ -901,7 +913,7 @@ compile_findall(term_t all, unsigned flags)
 
   PL_get_arg(1, all, t);
   compile_arg(&info, t);
-  
+
   if ( !(f = odbc_malloc(FND_SIZE(info.size))) )
     return NULL;
   f->references = 1;
@@ -1051,7 +1063,7 @@ find_connection(atom_t alias)
   { if ( c->alias == alias )
       return c;
   }
-  
+
 
   return NULL;
 }
@@ -1076,7 +1088,7 @@ alloc_connection(atom_t alias, atom_t dsn)
 
   if ( alias && find_connection(alias) )
     return NULL;			/* already existenting */
-  
+
   if ( !(c = odbc_malloc(sizeof(*c))) )
     return NULL;
   memset(c, 0, sizeof(*c));
@@ -1141,7 +1153,7 @@ get_connection(term_t tdsn, connection **cn)
     if ( !(c=find_connection(dsn)) )
       return existence_error(tdsn, "odbc_connection");
   }
-  
+
   *cn = c;
 
   return TRUE;
@@ -1179,7 +1191,9 @@ pl_odbc_connect(term_t tdsource, term_t cid, term_t options)
    const char *dsource;			/* odbc data source */
    char *uid = NULL;			/* user id */
    char *pwd = NULL;			/* password */
+   char *driver_string = NULL;			/* driver_string */
    atom_t alias = 0;			/* alias-name */
+   int mars = 0;			/* mars-value */
    atom_t open = 0;			/* open next connection */
    RETCODE rc;				/* result code for ODBC functions */
    HDBC hdbc;
@@ -1204,6 +1218,12 @@ pl_odbc_connect(term_t tdsource, term_t cid, term_t options)
      } else if ( PL_is_functor(head, FUNCTOR_alias1) )
      { if ( !get_atom_arg_ex(1, head, &alias) )
 	 return FALSE;
+     } else if ( PL_is_functor(head, FUNCTOR_driver_string1) )
+     { if ( !get_text_arg_ex(1, head, &driver_string) )
+	 return FALSE;
+     } else if ( PL_is_functor(head, FUNCTOR_mars1) )
+     { if ( !get_bool_arg_ex(1, head, &mars) )
+	 return FALSE;
      } else if ( PL_is_functor(head, FUNCTOR_open1) )
      { if ( !get_atom_arg_ex(1, head, &open) )
 	 return FALSE;
@@ -1253,10 +1273,39 @@ pl_odbc_connect(term_t tdsource, term_t cid, term_t options)
 
    if ( (rc=SQLAllocConnect(henv, &hdbc)) != SQL_SUCCESS )
      return odbc_report(henv, NULL, NULL, rc);
+
+   if ( mars )
+   { if ( (rc=SQLSetConnectAttr(hdbc,
+				SQL_COPT_SS_MARS_ENABLED,
+				SQL_MARS_ENABLED_YES,
+				SQL_IS_UINTEGER)) != SQL_SUCCESS )
+     return odbc_report(henv, NULL, NULL, rc);
+   }
+
    /* Connect to a data source. */
-   rc = SQLConnect(hdbc, (SQLCHAR *)dsource, SQL_NTS,
-                         (SQLCHAR *)uid,     SQL_NTS,
-                         (SQLCHAR *)pwd,     SQL_NTS);
+   if ( driver_string != NULL )
+   { if ( uid != NULL )
+     { return context_error(options, "Option incompatible with driver_string",
+			    "user");
+     } else if ( pwd != NULL )
+     { return context_error(options, "Option incompatible with driver_string",
+			    "password");
+     } else
+     { SQLCHAR connection_out[1025];	/* completed driver string */
+       SQLSMALLINT connection_out_len;
+
+       rc = SQLDriverConnect(hdbc,
+                             NULL, /* window handle */
+                             (SQLCHAR *)driver_string, SQL_NTS,
+                             connection_out, 1024,
+                             &connection_out_len,
+                             SQL_DRIVER_NOPROMPT);
+     }
+   } else
+   { rc = SQLConnect(hdbc, (SQLCHAR *)dsource, SQL_NTS,
+                           (SQLCHAR *)uid,     SQL_NTS,
+                           (SQLCHAR *)pwd,     SQL_NTS);
+   }
    if ( rc == SQL_ERROR )
      return odbc_report(henv, hdbc, NULL, rc);
    if ( rc != SQL_SUCCESS && !silent && !odbc_report(henv, hdbc, NULL, rc) )
@@ -1300,7 +1349,7 @@ odbc_disconnect(+Connection)
 static foreign_t
 pl_odbc_disconnect(term_t dsn)
 { connection *cn;
-  
+
   if ( !get_connection(dsn, &cn) )
     return FALSE;
 
@@ -1322,7 +1371,7 @@ odbc_current_connection(term_t cid, term_t dsn, control_t h)
       { if ( get_connection(cid, &cn) &&
 	     PL_unify_atom(dsn, cn->dsn) )
 	  return TRUE;
-	
+
 	return FALSE;
       }
 
@@ -1373,7 +1422,7 @@ odbc_set_connection(connection *cn, term_t option)
     if ( !get_atom_arg_ex(1, option, &val) )
       return FALSE;
     opt = SQL_ACCESS_MODE;
-	      
+
     if ( val == ATOM_read )
       optval = SQL_MODE_READ_ONLY;
     else if ( val == ATOM_update )
@@ -1386,7 +1435,7 @@ odbc_set_connection(connection *cn, term_t option)
     if ( !get_atom_arg_ex(1, option, &val) )
       return FALSE;
     opt = SQL_CURSOR_TYPE;
-    
+
     if ( val == ATOM_dynamic )
       optval = SQL_CURSOR_DYNAMIC;
     else if ( val == ATOM_forwards_only )
@@ -1402,7 +1451,7 @@ odbc_set_connection(connection *cn, term_t option)
 
     if ( !get_bool_arg_ex(1, option, &val) )
       return FALSE;
-    
+
     set(cn, CTX_SILENT);
 
     return TRUE;
@@ -1438,7 +1487,7 @@ pl_odbc_set_connection(term_t con, term_t option)
 
   if ( !get_connection(con, &cn) )
     return FALSE;
-  
+
   return odbc_set_connection(cn, option);
 }
 
@@ -1454,7 +1503,7 @@ typedef struct
   functor_t functor;
 } conn_option;
 
-static conn_option conn_option_list[] = 
+static conn_option conn_option_list[] =
 { { "database_name",	   SQL_DATABASE_NAME, text },
   { "dbms_name",           SQL_DBMS_NAME, text },
   { "dbms_version",        SQL_DBMS_VER, text },
@@ -1563,7 +1612,7 @@ odbc_end_transaction(term_t conn, term_t action)
   RETCODE rc;
   UWORD opt;
   atom_t a;
-  
+
 
   if ( !get_connection(conn, &cn) )
     return FALSE;
@@ -1576,7 +1625,7 @@ odbc_end_transaction(term_t conn, term_t action)
   { opt = SQL_ROLLBACK;
   } else
     return domain_error(action, "transaction");
-  
+
   if ( (rc=SQLTransact(henv, cn->hdbc, opt)) != SQL_SUCCESS )
     return odbc_report(henv, cn->hdbc, NULL, rc);
 
@@ -1826,7 +1875,7 @@ max_qualifier_lenght(connection *cn)
     { /*Sdprintf("SQL_MAX_QUALIFIER_NAME_LEN = %d\n", (int)len);*/
       cn->max_qualifier_lenght = (int)len; /* 0: unknown */
     } else
-    { odbc_report(henv, cn->hdbc, NULL, rc); 
+    { odbc_report(henv, cn->hdbc, NULL, rc);
       cn->max_qualifier_lenght = -1;
     }
 
@@ -1892,7 +1941,7 @@ prepare_result(context *ctxt)
 	PL_register_atom(ATOM_);
       }
     }
-	
+
     ptr_result->sqlTypeID = dataType;
     ptr_result->cTypeID = CvtSqlToCType(ctxt, dataType, ptr_result->plTypeID);
     if (ptr_result->cTypeID == CVNERR)
@@ -1902,7 +1951,7 @@ prepare_result(context *ctxt)
 
     DEBUG(1, Sdprintf("prepare_result(): column %d, "
 		      "sqlTypeID = %d, cTypeID = %d, "
-		      "columnSize = %d\n",
+		      "columnSize = %u\n",
 		      i, ptr_result->sqlTypeID, ptr_result->cTypeID,
 		      columnSize));
 
@@ -1925,35 +1974,35 @@ prepare_result(context *ctxt)
     switch (ptr_result->sqlTypeID)
     { case SQL_LONGVARCHAR:
       case SQL_LONGVARBINARY:
-      { if ( (int)columnSize > ctxt->max_nogetdata )
-	{ DEBUG(2,
-		Sdprintf("Wide SQL_LONGVAR* column %d: using SQLGetData()\n",
-			 i));
+      { if ( columnSize > ctxt->max_nogetdata || columnSize == 0 )
+	{ use_sql_get_data:
+	  DEBUG(2,
+		Sdprintf("Wide SQL_LONGVAR* column %d: using SQLGetData()\n", i));
 	  ptr_result->ptr_value = NULL;	/* handle using SQLGetData() */
 	  continue;
 	}
-	ptr_result->len_value = sizeof(char)*columnSize+1;
+	ptr_result->len_value = sizeof(char)*(columnSize+1);
 	goto bind;
       }
     }
-    
+
     switch (ptr_result->cTypeID)
     { case SQL_C_CHAR:
+	if ( columnSize == 0 )
+	  goto use_sql_get_data;
 	columnSize++;			/* one for decimal dot */
         /*FALLTHROUGH*/
       case SQL_C_BINARY:
-	if ( (int)columnSize > ctxt->max_nogetdata )
-	{ DEBUG(2,
-		Sdprintf("Wide SQL_C_BINARY column %d: using SQLGetData()\n",
-			 i));
-	  ptr_result->ptr_value = NULL;	/* handle using SQLGetData() */
-	  continue;
-	}
+	if ( columnSize > ctxt->max_nogetdata || columnSize == 0 )
+	  goto use_sql_get_data;
         ptr_result->len_value = sizeof(char)*columnSize+1;
 	break;
       case SQL_C_SLONG:
 	ptr_result->len_value = sizeof(SQLINTEGER);
 	break;
+      case SQL_C_SBIGINT:
+	ptr_result->len_value = sizeof(SQLBIGINT);
+	break;
       case SQL_C_DOUBLE:
 	ptr_result->len_value = sizeof(SQLDOUBLE);
 	break;
@@ -1970,7 +2019,7 @@ prepare_result(context *ctxt)
 	assert(0);
         return FALSE;			/* make compiler happy */
     }
-	
+
   bind:
     if ( ptr_result->len_value <= PARAM_BUFSIZE )
       ptr_result->ptr_value = (void *)ptr_result->buf;
@@ -2010,7 +2059,7 @@ odbc_row(context *ctxt, term_t trow)
     }
     set(ctxt, CTX_BOUND);
   }
-      
+
   if ( !ctxt->result )			/* not a SELECT statement */
   { SQLLEN rows;			/* was DWORD */
     int rval;
@@ -2023,9 +2072,9 @@ odbc_row(context *ctxt, term_t trow)
 			     PL_LONG, (long)rows);
     else
       rval = TRUE;
-    
+
     close_context(ctxt);
-    
+
     return rval;
   }
 
@@ -2036,7 +2085,7 @@ odbc_row(context *ctxt, term_t trow)
 
     for(;;)
     { ctxt->rc = SQLFetch(ctxt->hstmt);
-      
+
       switch(ctxt->rc)
       { case SQL_NO_DATA_FOUND:
 	  close_context(ctxt);
@@ -2049,14 +2098,14 @@ odbc_row(context *ctxt, term_t trow)
 	    return FALSE;
 	  }
       }
-      
+
       if ( !PL_unify_list(tail, head, tail) ||
 	   !put_findall(ctxt, tmp) ||
 	   !PL_unify(head, tmp) )
       { close_context(ctxt);
 	return FALSE;
       }
-    }      
+    }
   }
 
   local_trow = PL_new_term_ref();
@@ -2072,12 +2121,12 @@ odbc_row(context *ctxt, term_t trow)
     { close_context(ctxt);
       return FALSE;			/* with pending exception */
     }
-      
+
     if ( !PL_unify(trow, local_trow) )
     { PL_rewind_foreign_frame(fid);
       continue;
     }
-     
+
 					/* success! */
 					/* pre-fetch to get determinism */
     ctxt->rc = SQLFetch(ctxt->hstmt);
@@ -2172,8 +2221,12 @@ set_statement_options(context *ctxt, term_t options)
 	  return domain_error(a, "fetch");
 	}
       } else if ( PL_is_functor(head, FUNCTOR_wide_column_threshold1) )
-      { if ( !get_int_arg_ex(1, head, &ctxt->max_nogetdata) )
+      { int val;
+
+	if ( !get_int_arg_ex(1, head, &val) )
 	  return FALSE;
+
+	ctxt->max_nogetdata = val;
       } else
 	return domain_error(head, "odbc_option");
     }
@@ -2197,12 +2250,12 @@ pl_odbc_query(term_t dsn, term_t tquery, term_t trow, term_t options,
 { context *ctxt;
 
   switch( PL_foreign_control(handle) )
-  { case PL_FIRST_CALL:   
+  { case PL_FIRST_CALL:
     { connection *cn;
-  
+
       if ( !get_connection(dsn, &cn) )
 	return FALSE;
-         
+
       if ( !(ctxt = new_context(cn)) )
 	return FALSE;
       if ( !get_sql_text(ctxt, tquery) )
@@ -2292,7 +2345,7 @@ pl_odbc_column(term_t dsn, term_t db, term_t row, control_t handle)
 	  SQLColumns(ctxt->hstmt, NULL, 0, NULL, 0,
 		     (SQLCHAR*)s, (SWORD)len, NULL, 0),
 	  close_context(ctxt));
-      
+
       return odbc_row(ctxt, row);
     }
     case PL_REDO:
@@ -2338,7 +2391,7 @@ odbc_types(term_t dsn, term_t sqltype, term_t row, control_t handle)
       TRY(ctxt,
 	  SQLGetTypeInfo(ctxt->hstmt, type),
 	  close_context(ctxt));
-      
+
       return odbc_row(ctxt, row);
     }
     case PL_REDO:
@@ -2364,7 +2417,7 @@ odbc_data_sources(term_t list)
   RETCODE rc;
   term_t tail = PL_copy_term_ref(list);
   term_t head = PL_new_term_ref();
-  
+
   if ( !henv )
     SQLAllocEnv(&henv);		/* Allocate an environment handle */
 
@@ -2431,7 +2484,7 @@ typedef struct
   atom_t name;				/* Prolog name */
 } sqltypedef;
 
-static sqltypedef sqltypes[] = 
+static sqltypedef sqltypes[] =
 { { SQL_BIGINT,	       "bigint" },
   { SQL_BINARY,	       "binary" },
   { SQL_BIT,	       "bit" },
@@ -2471,7 +2524,7 @@ get_sqltype_from_atom(atom_t name, SWORD *type)
   return FALSE;
 }
 
-static sqltypedef pltypes[] = 
+static sqltypedef pltypes[] =
 { { SQL_PL_DEFAULT,    "default" },
   { SQL_PL_ATOM,       "atom" },
   { SQL_PL_STRING,     "string" },
@@ -2616,20 +2669,26 @@ declare_parameters(context *ctxt, term_t parms)
         vlenptr = &params->len_value;
 	break;
       case SQL_C_SLONG:
-	params->len_value = sizeof(long);
+	params->len_value = sizeof(SQLINTEGER);
+        vlenptr = &params->len_value;
+	break;
+      case SQL_C_SBIGINT:
+	params->len_value = sizeof(SQLBIGINT);
         vlenptr = &params->len_value;
 	break;
       case SQL_C_DOUBLE:
-	params->len_value = sizeof(double);
+	params->len_value = sizeof(SQLDOUBLE);
         vlenptr = &params->len_value;
 	break;
       case SQL_C_DATE:
+      case SQL_C_TYPE_DATE:
 	if ( !(params->ptr_value = odbc_malloc(sizeof(DATE_STRUCT))) )
 	  return FALSE;
         params->len_value = sizeof(DATE_STRUCT);
 	vlenptr = &params->len_value;
         break;
       case SQL_C_TIME:
+      case SQL_C_TYPE_TIME:
 	if ( !(params->ptr_value = odbc_malloc(sizeof(TIME_STRUCT))) )
 	  return FALSE;
         params->len_value = sizeof(TIME_STRUCT);
@@ -2641,6 +2700,9 @@ declare_parameters(context *ctxt, term_t parms)
         params->len_value = sizeof(SQL_TIMESTAMP_STRUCT);
 	vlenptr = &params->len_value;
         break;
+      default:
+	Sdprintf("declare_parameters(): cTypeID %d not supported\n",
+		 params->cTypeID);
     }
 
 
@@ -2654,7 +2716,7 @@ declare_parameters(context *ctxt, term_t parms)
 			       params->ptr_value,	/* rgbValue */
 			       0,			/* cbValueMax */
 			       vlenptr), 		/* pcbValue */
-	(void)0);		
+	(void)0);
   }
 
   return TRUE;
@@ -2877,11 +2939,27 @@ bind_parameters(context *ctxt, term_t parms)
 
     switch(prm->cTypeID)
     { case SQL_C_SLONG:
-	if ( PL_get_long(head, (long *)prm->ptr_value) )
-	  prm->len_value = sizeof(long);
-	else if ( !try_null(ctxt, prm, head, "integer") )
+      { long val;
+
+	if ( PL_get_long(head, &val) )
+	{ SQLINTEGER sqlval = val;
+	  memcpy(prm->ptr_value, &sqlval, sizeof(SQLINTEGER));
+	  prm->len_value = sizeof(SQLINTEGER);
+	} else if ( !try_null(ctxt, prm, head, "integer") )
 	  return FALSE;
         break;
+      }
+      case SQL_C_SBIGINT:
+      { int64_t val;
+
+	if ( PL_get_int64(head, &val) )
+	{ SQLBIGINT sqlval = val;
+	  memcpy(prm->ptr_value, &sqlval, sizeof(SQLBIGINT));
+	  prm->len_value = sizeof(SQLBIGINT);
+	} else if ( !try_null(ctxt, prm, head, "integer") )
+	  return FALSE;
+        break;
+      }
       case SQL_C_DOUBLE:
 	if ( PL_get_float(head, (double *)prm->ptr_value) )
 	  prm->len_value = sizeof(double);
@@ -3229,7 +3307,9 @@ install_odbc4pl()
    FUNCTOR_odbc_connection1	 = MKFUNCTOR("$odbc_connection", 1);
    FUNCTOR_user1		 = MKFUNCTOR("user", 1);
    FUNCTOR_password1		 = MKFUNCTOR("password", 1);
+   FUNCTOR_driver_string1		 = MKFUNCTOR("driver_string", 1);
    FUNCTOR_alias1		 = MKFUNCTOR("alias", 1);
+   FUNCTOR_mars1		 = MKFUNCTOR("mars", 1);
    FUNCTOR_open1		 = MKFUNCTOR("open", 1);
    FUNCTOR_auto_commit1		 = MKFUNCTOR("auto_commit", 1);
    FUNCTOR_types1		 = MKFUNCTOR("types", 1);
@@ -3308,7 +3388,7 @@ CvtSqlToCType(context *ctxt, SQLSMALLINT fSqlType, SQLSMALLINT plTypeID)
 	case SQL_WLONGVARCHAR:
 #endif
 	  return SQL_C_CHAR;
-    
+
 	case SQL_BINARY:
 	case SQL_VARBINARY:
 	case SQL_LONGVARBINARY:
@@ -3317,18 +3397,21 @@ CvtSqlToCType(context *ctxt, SQLSMALLINT fSqlType, SQLSMALLINT plTypeID)
 	case SQL_DECIMAL:
 	case SQL_NUMERIC:
 	  return SQL_C_CHAR;
-  
+
 	case SQL_REAL:
 	case SQL_FLOAT:
 	case SQL_DOUBLE:
 	  return SQL_C_DOUBLE;
-       
+
 	case SQL_BIT:
 	case SQL_TINYINT:
 	case SQL_SMALLINT:
 	case SQL_INTEGER:
 	  return SQL_C_SLONG;
-    
+
+	case SQL_BIGINT:		/* 64-bit integers */
+	  return SQL_C_SBIGINT;
+
 	case SQL_DATE:
 	case SQL_TYPE_DATE:
 	  return SQL_C_TYPE_DATE;
@@ -3358,6 +3441,8 @@ CvtSqlToCType(context *ctxt, SQLSMALLINT fSqlType, SQLSMALLINT plTypeID)
       switch(fSqlType)
       { case SQL_TIMESTAMP:
 	  return SQL_C_TIMESTAMP;
+	case SQL_BIGINT:		/* 64-bit integers */
+	  return SQL_C_SBIGINT;
 	default:
 	  return SQL_C_SLONG;
       }
@@ -3516,6 +3601,9 @@ pl_put_column(context *c, int nth, term_t col)
       case SQL_C_SLONG:
 	PL_put_integer(val,*(SQLINTEGER *)p->ptr_value);
 	break;
+      case SQL_C_SBIGINT:
+	PL_put_int64(val, *(SQLBIGINT *)p->ptr_value);
+	break;
       case SQL_C_DOUBLE:
 	PL_put_float(val,*(SQLDOUBLE *)p->ptr_value);
 	break;
@@ -3526,7 +3614,7 @@ pl_put_column(context *c, int nth, term_t col)
 	PL_put_integer(av+0, ds->year);
 	PL_put_integer(av+1, ds->month);
 	PL_put_integer(av+2, ds->day);
-	
+
 	PL_cons_functor_v(val, FUNCTOR_date3, av);
 	break;
       }
@@ -3537,11 +3625,11 @@ pl_put_column(context *c, int nth, term_t col)
 	PL_put_integer(av+0, ts->hour);
 	PL_put_integer(av+1, ts->minute);
 	PL_put_integer(av+2, ts->second);
-	
+
 	PL_cons_functor_v(val, FUNCTOR_time3, av);
 	break;
       }
-      case SQL_C_TIMESTAMP: 
+      case SQL_C_TIMESTAMP:
       { SQL_TIMESTAMP_STRUCT* ts = (SQL_TIMESTAMP_STRUCT*)p->ptr_value;
 
 	switch( p->plTypeID )
@@ -3556,7 +3644,7 @@ pl_put_column(context *c, int nth, term_t col)
 	    PL_put_integer(av+4, ts->minute);
 	    PL_put_integer(av+5, ts->second);
 	    PL_put_integer(av+6, ts->fraction);
-	
+
 	    PL_cons_functor_v(val, FUNCTOR_timestamp7, av);
 	    break;
 	  }
@@ -3623,7 +3711,7 @@ static int
 pl_put_row(term_t row, context *c)
 { term_t columns = PL_new_term_refs(c->NumCols);
   SQLSMALLINT i;
-   
+
   for (i=0; i<c->NumCols; i++)
   { if ( !pl_put_column(c, i, columns+i) )
       return FALSE;			/* with exception */
