@@ -1,11 +1,11 @@
-/*  $Id: doc_html.pl,v 1.22 2007/05/24 14:34:23 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2006, University of Amsterdam
+    Copyright (C): 2009, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -30,39 +30,54 @@
 */
 
 :- module(pldoc_html,
-	  [ doc_for_file/3,		% +FileSpec, +Out, +Options
+	  [ doc_for_file/2,		% +FileSpec, +Options
 	    doc_write_html/3,		% +Stream, +Title, +Term
-	    doc_for_wiki_file/3,	% +FileSpec, +Out, +Options
+	    doc_for_wiki_file/2,	% +FileSpec, +Options
 	    				% Support doc_index
 	    doc_page_dom/3,		% +Title, +Body, -DOM
 	    print_html_head/1,		% +Stream
-	    predref/3,			% +PI //
-	    predref/4,			% +PI, Options //
+	    predref//1,			% +PI //
+	    predref//2,			% +PI, Options //
 	    module_info/3,		% +File, +Options0, -Options
 	    doc_hide_private/3,		% +Doc0, -Doc, +Options
-	    edit_button/4,		% +File, +Options, //
-	    source_button/4,		% +File, +Options, //
-	    pred_edit_button/4,		% +PredInd, +Options, //
-	    object_edit_button/4,	% +Obj, +Options, //
-
+	    edit_button//2,		% +File, +Options, //
+	    source_button//2,		% +File, +Options, //
+	    pred_edit_button//2,		% +PredInd, +Options, //
+	    object_edit_button//2,	% +Obj, +Options, //
+	    object_source_button//2,	% +Obj, +Options, //
+					% Support other backends
+	    doc_file_objects/5,		% +FSpec, -File, -Objs, -FileOpts, +Opts
+	    existing_linked_file/2,	% +FileSpec, -Path
+	    doc_tag_title/2,		% +Tag, -Title
+	    pred_anchor_name/3,		% +Head, -PI, -Anchor
 	    private/2,			% +Obj, +Options
-	    file/3,			% +File, //
-	    image/3,			% +File, //
-	    tags/3,			% +Tags, //
-	    file_header/4,		% +File, +Options, //
-	    objects/4,			% +Objects, +Options, //
-	    object_ref/4,		% +Object, +Options, //
+	    is_pi/1,			% @Term
+	    is_op_type/2,		% +Atom, ?Type
+					% Output routines
+	    file//1,			% +File, //
+	    file//2,			% +File, +Options, //
+	    include//3,			% +File, +Type, +Options //
+	    tags//1,			% +Tags, //
+	    term//2,			% +Term, +Bindings, //
+	    file_header//2,		% +File, +Options, //
+	    objects//2,			% +Objects, +Options, //
+	    object_ref//2,		% +Object, +Options, //
 	    object_href/2,		% +Object, -URL
-	    object_page/4		% +Object, +Options, //
+	    object_page//2		% +Object, +Options, //
 	  ]).
 :- use_module(library(lists)).
 :- use_module(library(option)).
 :- use_module(library(url)).
 :- use_module(library(readutil)).
-:- use_module(library('http/html_write')).
+:- use_module(library(http/html_write)).
+:- use_module(library(http/http_dispatch)).
+:- use_module(library(http/http_path)).
+:- use_module(library(http/html_head)).
 :- use_module(library(doc_http)).
 :- use_module(library(debug)).
+:- use_module(library(apply)).
 :- use_module(doc_process).
+:- use_module(doc_man).
 :- use_module(doc_modes).
 :- use_module(doc_wiki).
 :- use_module(doc_search).
@@ -78,26 +93,48 @@ extracting module doc_wiki.pl into HTML+CSS.
 	in other backends.
 */
 
+		 /*******************************
+		 *	     RESOURCES		*
+		 *******************************/
+
+:- html_resource(pldoc_css,
+		 [ virtual(true),
+		   requires([ pldoc_resource('pldoc.css')
+			    ])
+		 ]).
+:- html_resource(pldoc_js,
+		 [ virtual(true),
+		   requires([ pldoc_resource('pldoc.js')
+			    ])
+		 ]).
+:- html_resource(pldoc,
+		 [ virtual(true),
+		   requires([ pldoc_css,
+			      pldoc_js
+			    ])
+		 ]).
+
+
 		 /*******************************
 		 *	 FILE PROCESSING	*
 		 *******************************/
 
-%%	doc_for_file(+File, +Out:stream, +Options) is det
+%%	doc_for_file(+File, +Options) is det
+%
+%	Write documentation for File as HTML.  Options:
 %
-%	Write documentation for File to Out in HTML.  Options:
-%	
 %		* public_only(+Bool)
 %		If =true= (default), only emit documentation for
 %		exported predicates.
-%	
+%
 %		* edit(Bool)
 %		If =true=, provide edit buttons. Default, these buttons
 %		are suppressed.
-%		
+%
 %	@param File	Prolog file specification.
 %	@param Out	Output stream
 
-doc_for_file(FileSpec, Out, Options) :-
+doc_for_file(FileSpec, Options) :-
 	absolute_file_name(FileSpec,
 			   [ file_type(prolog),
 			     access(read)
@@ -105,48 +142,61 @@ doc_for_file(FileSpec, Out, Options) :-
 			   File),
 	file_base_name(File, Base),
 	Title = Base,
-	doc_page_dom(Title, \prolog_file(FileSpec, Options), DOM),
-	phrase(html(DOM), Tokens),
-	print_html_head(Out),
-	print_html(Out, Tokens).
+	reply_html_page(title(Title),
+			\prolog_file(FileSpec, Options)).
 
 prolog_file(FileSpec, Options) -->
-	{ absolute_file_name(FileSpec,
-			     [ file_type(prolog),
-			       access(read)
-			     ],
-			     File),
-	  Pos = File:_Line,
-	  findall(doc(Obj,Pos,Comment),
-		  doc_comment(Obj, Pos, _, Comment), Objs0),
-	  module_info(File, ModuleOptions, Options),
-	  file_info(Objs0, Objs1, FileOptions, ModuleOptions),
-	  doc_hide_private(Objs1, Objs, ModuleOptions),
+	{ doc_file_objects(FileSpec, File, Objects, FileOptions, Options),
 	  b_setval(pldoc_file, File),	% TBD: delete?
 	  file_directory_name(File, Dir)
 	},
-	html([ \doc_links(Dir, FileOptions),
+	html([ \html_requires(pldoc),
+	       \doc_links(Dir, FileOptions),
 	       \file_header(File, FileOptions)
-	     | \objects(Objs, FileOptions)
+	     | \objects(Objects, FileOptions)
 	     ]),
-	undocumented(Objs, FileOptions).
-	  
+	undocumented(Objects, FileOptions).
+
+%%	doc_file_objects(+FileSpec, -File, -Objects, -FileOptions, +Options) is det.
+%
+%	Extracts  relevant  information  for  FileSpec  from  the  PlDoc
+%	database.  FileOptions contains:
+%
+%		* file(Title:string, Comment:string)
+%		* module(Module:atom)
+%		* public(Public:list(predicate_indicator)
+%
+%	Objects contains
+%
+%		* doc(PI:predicate_indicator, File:Line, Comment)
+%
+%	@param FileSpec File specification as used for load_files/2.
+%	@param File	Prolog canonical filename
+
+doc_file_objects(FileSpec, File, Objects, FileOptions, Options) :-
+	absolute_file_name(FileSpec,
+			   [ file_type(prolog),
+			     access(read)
+			   ],
+			   File),
+	Pos = File:_Line,
+	findall(doc(Obj,Pos,Comment),
+		doc_comment(Obj, Pos, _, Comment), Objs0),
+	module_info(File, ModuleOptions, Options),
+	file_info(Objs0, Objs1, FileOptions, ModuleOptions),
+	doc_hide_private(Objs1, Objects, ModuleOptions).
+
+
 %%	module_info(+File, -ModuleOptions, +OtherOptions) is det.
 %
 %	Add options module(Name),  public(Exports)   to  OtherOptions if
 %	File is a module file.
 
 module_info(File, [module(Module), public(Exports)|Options], Options) :-
-	current_module(Module, File), !,
-	export_list(Module, Public),
-	maplist(head_to_pi, Public, Exports).
+	module_property(Module, file(File)), !,
+	module_property(Module, exports(Exports)).
 module_info(_, Options, Options).
 
-head_to_pi(M:Head, M:PI) :- !,
-	head_to_pi(Head, PI).
-head_to_pi(Head, Name/Arity) :-
-	functor(Head, Name, Arity).
-	
 
 %%	doc_hide_private(+Objs, +Public, +Options)
 %
@@ -193,9 +243,8 @@ private(Module:PI, Options) :-
 	     eq_pi(PI, PI2)
 	   ).
 private(Module:PI, _Options) :-
-	export_list(Module, Public),
-	\+ ( member(Head, Public),
-	     head_to_pi(Head, PI2),
+	export_list(Module, Exports),
+	\+ ( member(PI2, Exports),
 	     eq_pi(PI, PI2)
 	   ).
 
@@ -218,7 +267,8 @@ file_header(File, Options) -->
 	},
 	file_title([Base, ' -- ', Title], File, Options),
 	{ is_structured_comment(Comment, Prefixes),
-	  indented_lines(Comment, Prefixes, Lines),
+	  string_to_list(Comment, Codes),
+	  indented_lines(Codes, Prefixes, Lines),
 	  section_comment_header(Lines, _Header, Lines1),
 	  wiki_lines_to_dom(Lines1, [], DOM)
 	},
@@ -254,17 +304,14 @@ file_title(Title, File, Options) -->
 
 reload_button(Base, Options) -->
 	{ option(edit(true), Options), !,
-	  option(public_only(Public), Options, true),
-	  format(string(HREF), '~w?reload=true&public_only=~w', 
-		 [Base, Public]),
-	  doc_server_root(Root)
+	  option(public_only(Public), Options, true)
 	},
-	html(a(href=HREF,
+	html(a(href(Base+[reload(true), public_only(Public)]),
 	       img([ %class(icon),
 		     height=24,
 		     alt('Reload'),
 		     style('padding-top:4px; border:0;'),
-		     src(Root+'reload.gif')
+		     src(location_by_id(pldoc_resource)+'reload.gif')
 		   ]))).
 reload_button(_, _) -->
 	[].
@@ -277,19 +324,17 @@ reload_button(_, _) -->
 
 edit_button(File, Options) -->
 	{ option(edit(true), Options), !,
-	  option(button_height(H), Options, 24),
-	  www_form_encode(File, Enc),
-	  doc_server_root(Root),
-	  format(string(HREF), '~wedit?file=~w', [Root, Enc]),
-	  format(string(OnClick), 'HTTPrequest("~w")', [HREF])
+	  option(button_height(H), Options, 24)
 	},
-	html(a([ onClick(OnClick),
+	html(a([ onClick('HTTPrequest(\'' +
+			 location_by_id(pldoc_edit) + [file(File)] +
+			 '\')'),
 		 onMouseOver('window.status=\'Edit file\'; return true;')
 	       ],
 	       img([ height(H),
 		     alt(edit),
 		     style('border:0'),
-		     src(Root+'edit.gif')
+		     src(location_by_id(pldoc_resource)+'edit.gif')
 		 ]))).
 edit_button(_, _) -->
 	[].
@@ -303,22 +348,21 @@ zoom_button(_, Options) -->
 	{ option(files(_Map), Options) }, !.	% generating files
 zoom_button(Base, Options) -->
 	{   (   option(public_only(true), Options, true)
-	    ->  format(string(HREF), '~w?public_only=false', [Base]),
-		Zoom = 'zoomin.gif',
-		Alt = 'Show all'
-	    ;   format(string(HREF), '~w?public_only=true', [Base]),
-		Zoom = 'zoomout.gif',
-		Alt = 'Show public'
-	    ),
-	    doc_server_root(Root)
+	    ->  Zoom = 'zoomin.gif',
+		Alt = 'Show all',
+		PublicOnly = false
+	    ;   Zoom = 'zoomout.gif',
+		Alt = 'Show public',
+		PublicOnly = true
+	    )
 	},
-	html(a(href=HREF,
+	html(a(href(Base+[public_only(PublicOnly)]),
 	       img([ height=24,
 		     alt(Alt),
 		     style('padding-top:4px; border:0;'),
-		     src(Root+Zoom)
+		     src(location_by_id(pldoc_resource)+Zoom)
 		   ]))).
-	
+
 
 %%	source_button(+File, +Options)// is det.
 %
@@ -328,20 +372,18 @@ source_button(_File, Options) -->
 	{ option(files(_Map), Options) }, !.	% generating files
 source_button(File, Options) -->
 	{ (   is_absolute_file_name(File)
-	  ->  doc_file_href(File, HREF0),
-	      atom_concat(HREF0, '?source=true', HREF)
-	  ;   format(string(HREF), '~w?source=true', [File])
+	  ->  doc_file_href(File, HREF0)
+	  ;   HREF0 = File
 	  ),
-	  option(button_height(H), Options, 24),
-	  doc_server_root(Root)
+	  option(button_height(H), Options, 24)
 	},
-	html(a(href=HREF,
+	html(a(href(HREF0+[source(true)]),
 	       img([ height(H),
 		     alt('Show source'),
 		     style('padding-top:4px; border:0;'),
-		     src(Root+'source.gif')
+		     src(location_by_id(pldoc_resource)+'source.gif')
 		   ]))).
-	
+
 
 %%	objects(+Objects:list, +Options)// is det.
 %
@@ -364,10 +406,12 @@ object(Obj, Mode0, Mode, Options) -->
 	object(Obj, Pos, Comment, Mode0, Mode, Options).
 
 object(Obj, Pos, Comment, Mode0, Mode, Options) -->
-	{ pi(Obj), !,
+	{ is_pi(Obj), !,
 	  is_structured_comment(Comment, Prefixes),
-	  indented_lines(Comment, Prefixes, Lines),
-	  process_modes(Lines, Pos, Modes, Args, Lines1),
+	  string_to_list(Comment, Codes),
+	  indented_lines(Codes, Prefixes, Lines),
+	  strip_module(user:Obj, Module, _),
+	  process_modes(Lines, Module, Pos, Modes, Args, Lines1),
 	  (   private(Obj, Options)
 	  ->  Class = privdef		% private definition
 	  ;   Class = pubdef		% public definition
@@ -387,7 +431,7 @@ object([Obj|_Same], Pos, Comment, Mode0, Mode, Options) --> !,
 object(Obj, _Pos, _Comment, Mode, Mode, _Options) -->
 	{ debug(pldoc, 'Skipped ~p', [Obj]) },
 	[].
-	
+
 
 %%	need_mode(+Mode:atom, +Stack:list, -NewStack:list)// is det.
 %
@@ -402,7 +446,7 @@ need_mode(Mode, Stack, Stack) -->
 need_mode(Mode, Stack, Rest) -->
 	{ memberchk(Mode, Stack)
 	}, !,
-	pop_mode(Mode, Stack, Rest).	
+	pop_mode(Mode, Stack, Rest).
 need_mode(Mode, Stack, [Mode|Stack]) --> !,
 	html_begin(Mode).
 
@@ -438,14 +482,14 @@ undocumented_predicates([], _) -->
 undocumented_predicates([H|T], Options) -->
 	undocumented_pred(H, Options),
 	undocumented_predicates(T, Options).
-		
+
 undocumented_pred(Name/Arity, Options) -->
 	{ functor(Head, Name, Arity) },
 	html(dt(class=undoc, \pred_mode(Head, [], _, Options))).
-		
+
 select_undocumented([], _, _, []).
 select_undocumented([PI|T0], M, Objs, [PI|T]) :-
-	pi(PI),
+	is_pi(PI),
 	\+ in_doc(M:PI, Objs),
 	select_undocumented(T0, M, Objs, T).
 select_undocumented([_|T0], M, Objs, T) :-
@@ -473,14 +517,17 @@ eq_pi(Name/A, Name//DCGA) :-
 eq_pi(Name//DCGA, Name/A) :-
 	A =:= DCGA+2.
 
-%%	pi(@Term) is semidet.
+%%	is_pi(@Term) is semidet.
 %
 %	True if Term is a predicate indicator.
 
-pi(_:PI) :- !,
-	pi(PI).
-pi(_/_).
-pi(_//_).
+is_pi(Var) :-
+	var(Var), !,
+	fail.
+is_pi(_:PI) :- !,
+	is_pi(PI).
+is_pi(_/_).
+is_pi(_//_).
 
 
 		 /*******************************
@@ -490,26 +537,30 @@ pi(_//_).
 %%	object_page(+Obj, +Options)// is det.
 %
 %	Generate an HTML page describing Obj.  The top presents the file
-%	the object is documented in and a search-form.
+%	the object is documented in and a search-form.  Options:
+%
+%	    * header(+Boolean)
+%	    Show the navigation and search header.
 
 object_page(Obj, Options) -->
 	prolog:doc_object_page(Obj, Options).
 object_page(Obj, Options) -->
-	{ doc_comment(Obj, File:_Line, _Summary, _Comment),
-	  doc_server_root(Root),
-	  format(string(FileRef), '~wdoc~w', [Root, File])
+	{ doc_comment(Obj, File:_Line, _Summary, _Comment)
 	},
-	html([ div(class(navhdr),
-		   [ span(style('float:left'), a(href(FileRef), File)),
-		     span(style('float:right'), \search_form(Options)),
-		     br(clear(both))
-		   ]),
+	html([ \html_requires(pldoc),
+	       \object_page_header(File, Options),
 	       \objects([Obj], Options)
 	     ]).
 
-	
-
-
+object_page_header(File, Options) -->
+	{ option(header(true), Options, true) }, !,
+	html(div(class(navhdr),
+		 [ div(class(jump),
+		       a(href(location_by_id(pldoc_doc)+File), File)),
+		   div(class(search), \search_form(Options)),
+		   br(clear(right))
+		 ])).
+object_page_header(_, _) --> [].
 
 
 
@@ -533,13 +584,12 @@ doc_write_html(Out, Title, Doc) :-
 %	links to the style-sheet and javaScript files.
 
 doc_page_dom(Title, Body, DOM) :-
-	doc_server_root(Root),
 	DOM = html([ head([ title(Title),
 			    link([ rel(stylesheet),
 				   type('text/css'),
-				   href('pldoc.css')
+				   href(location_by_id(pldoc_resource)+'pldoc.css')
 				 ]),
-			    script([ src(Root+'pldoc.js'),
+			    script([ src(location_by_id(pldoc_resource)+'pldoc.js'),
 				     type('text/javascript')
 				   ], [])
 			  ]),
@@ -562,34 +612,43 @@ print_html_head(Out) :-
 %%	tags(+Tags)// is det.
 %
 %	Emit the @tag tags of a description. Tags is produced by tags/3.
-%	
+%
 %	@see combine_tags/2.
 
 tags(Tags) -->
 	html(dl(class=tags, Tags)).
 
-%%	tag(+Tag, +Value)// is det.
+%%	tag(+Tag, +Values:list)// is det.
 %
-%	Called from \tag(Name, Value) terms produced by doc_wiki.pl.
+%	Called from \tag(Name, Values) terms produced by doc_wiki.pl.
 
-tag(Tag, Value) -->
-	{   tag_title(Tag, Title)
-	->  true
-	;   Title = Tag
+tag(Tag, Values) -->
+	{   doc_tag_title(Tag, Title),
+	    atom_concat('keyword-', Tag, Class)
 	},
-	{   tag_class(Tag, Class)
-	->  true
-	;   Class = tag
-	},
-	html([dt(class=Class, Title), dd(Value)]).
+	html([ dt(class=Class, Title),
+	       \tag_values(Values, Class)
+	     ]).
 
-tag_title(compat, 'Compatibility:').
-tag_title(tbd,    'To be done:').
-tag_title(see,    'See also:').
+tag_values([], _) -->
+	[].
+tag_values([H|T], Class) -->
+	html(dd(class=Class, ['- '|H])),
+	tag_values(T, Class).
 
-tag_class(tbd, 		warn).
-tag_class(bug, 		error).
-tag_class(deprecated,	warn).
+
+%%	doc_tag_title(+Tag, -Title) is det.
+%
+%	Title is the name to use for Tag in the generated documentation.
+
+doc_tag_title(Tag, Title) :-
+	tag_title(Tag, Title), !.
+doc_tag_title(Tag, Tag).
+
+tag_title(compat, 'Compatibility').
+tag_title(tbd,    'To be done').
+tag_title(see,    'See also').
+tag_title(error,  'Errors').
 
 %%	params(+Params:list) is det.
 %
@@ -617,7 +676,8 @@ param(param(Name,Descr)) -->
 		 *******************************/
 
 section(Type, Title) -->
-	{ wiki_string_to_dom(Title, [], Content0),
+	{ string_to_list(Title, Codes),
+	  wiki_codes_to_dom(Codes, [], Content0),
 	  strip_leading_par(Content0, Content),
 	  make_section(Type, Content, HTML)
 	},
@@ -634,7 +694,7 @@ make_section(section, Title, h1(class=section, Title)).
 %%	pred_dt(+Modes, +Class, Options)// is det.
 %
 %	Emit the predicate header.
-%	
+%
 %	@param Modes	List as returned by process_modes/5.
 
 pred_dt(Modes, Class, Options) -->
@@ -649,7 +709,7 @@ pred_dt([H|T], Class, Done0, Done, Options) -->
 
 
 pred_mode(mode(Head,Vars), Done0, Done, Options) --> !,
-	{ bind_vars(Vars) },
+	{ bind_vars(Head, Vars) },
 	pred_mode(Head, Done0, Done, Options).
 pred_mode(Head is Det, Done0, Done, Options) --> !,
 	anchored_pred_head(Head, Done0, Done, Options),
@@ -657,19 +717,39 @@ pred_mode(Head is Det, Done0, Done, Options) --> !,
 pred_mode(Head, Done0, Done, Options) -->
 	anchored_pred_head(Head, Done0, Done, Options).
 
+bind_vars(Term, Bindings) :-
+	bind_vars(Bindings),
+	anon_vars(Term).
+
 bind_vars([]).
 bind_vars([Name=Var|T]) :-
 	Var = '$VAR'(Name),
 	bind_vars(T).
 
+%%	anon_vars(+Term) is det.
+%
+%	Bind remaining variables in Term to '$VAR'('_'), so they are
+%	printed as '_'.
+
+anon_vars(Var) :-
+	var(Var), !,
+	Var = '$VAR'('_').
+anon_vars(Term) :-
+	compound(Term), !,
+	Term =.. [_|Args],
+	maplist(anon_vars, Args).
+anon_vars(_).
+
+
 anchored_pred_head(Head, Done0, Done, Options) -->
-	{ anchor_name(Head, PI, Name)
-	},
+	{ pred_anchor_name(Head, PI, Name) },
 	(   { memberchk(PI, Done0) }
 	->  { Done = Done0 },
 	    pred_head(Head)
-	;   pred_edit_button(Head, Options),
-	    html(a(name=Name, \pred_head(Head))),
+	;   html([ span(style('float:right'),
+			\pred_edit_button(Head, Options)),
+		   a(name=Name, \pred_head(Head))
+		 ]),
 	    { Done = [PI|Done0] }
 	).
 
@@ -679,51 +759,33 @@ anchored_pred_head(Head, Done0, Done, Options) -->
 
 pred_edit_button(_, Options) -->
 	{ \+ option(edit(true), Options) }, !.
-pred_edit_button(M:PI, Options) --> 
-	{ atom(M) }, !,
-	pred_edit_button(PI, [module(M)|Options]).
-pred_edit_button(//(Head), Options) --> !,
-	{ functor(Head, Name, Arity),
-	  PredArity is Arity + 2
-	},
-	pred_edit_button(Name/PredArity, Options).
-pred_edit_button(Name//Arity, Options) --> !,
-	{ PredArity is Arity + 2
-	},
-	pred_edit_button(Name/PredArity, Options).	  
-pred_edit_button(Name/Arity, Options) -->
+pred_edit_button(PI0, Options0) -->
+	{ canonise_predref(PI0, PI, Options0, Options) },
+	pred_edit_button2(PI, Options).
+
+pred_edit_button2(Name/Arity, Options) -->
 	{ functor(Head, Name, Arity),
 	  option(module(M), Options, _),
 	  \+ ( current_module(M),
 	       source_file(M:Head, _File)
 	     )
 	}, !.
-pred_edit_button(Name/Arity, Options) -->
-	{ doc_server_root(Root),
-	  www_form_encode(Name, QName),
-	  (   option(module(M), Options)
-	  ->  www_form_encode(M, QM),
-	      format(string(OnClick),
-		     'HTTPrequest("~wedit?name=~w&arity=~w&module=~w")',
-		     [Root, QName, Arity, QM])
-	  ;   format(string(OnClick),
-		     'HTTPrequest("~wedit?name=~w&arity=~w")',
-		     [Root, QName, Arity])
+pred_edit_button2(Name/Arity, Options) -->
+	{ (   option(module(M), Options)
+	  ->  Extra = [module(M)]
+	  ;   Extra = []
 	  )
 	},
-	html(span(style('float:right'),
-		  a([ onClick(OnClick)
-		    ],
-		    img([ height=14,
-			  style('border:0;'),
-			  src(Root+'edit.gif')
-			])))).
-pred_edit_button(_/_, _) --> !,
+	html(a([ onClick('HTTPrequest(\'' +
+			 location_by_id(pldoc_edit)+[name(Name),arity(Arity)|Extra] +
+			 '\')')
+	       ],
+	       img([ height=12,
+		     style('border:0;'),
+		     src(location_by_id(pldoc_resource)+'edit.gif')
+		   ]))).
+pred_edit_button2(_, _) --> !,
 	[].
-pred_edit_button(Head, Options) -->
-	{ functor(Head, Name, Arity)
-	},
-	pred_edit_button(Name/Arity, Options).
 
 
 %%	object_edit_button(+Object, +Options)// is det.
@@ -733,12 +795,63 @@ pred_edit_button(Head, Options) -->
 object_edit_button(_, Options) -->
 	{ \+ option(edit(true), Options) }, !.
 object_edit_button(PI, Options) -->
-	{ pi(PI) }, !,
+	{ is_pi(PI) }, !,
 	pred_edit_button(PI, Options).
 object_edit_button(_, _) -->
 	[].
 
 
+%%	pred_source_button(+PredIndicator, +Options)// is det.
+%
+%	Create a button for viewing the source of a predicate.
+
+pred_source_button(PI0, Options0) -->
+	{ canonise_predref(PI0, PI, Options0, Options),
+	  option(module(M), Options, _),
+	  pred_source_href(PI, M, HREF), !
+	},
+	html(a([ href(HREF)
+	       ],
+	       img([ height=12,
+		     style('border:0;'),
+		     src(location_by_id(pldoc_resource)+'source.gif')
+		   ]))).
+pred_source_button(_, _) -->
+	[].
+
+
+%%	object_source_button(+Object, +Options)// is det.
+%
+%	Create a button	for showing the source of Object.
+
+object_source_button(PI, Options) -->
+	{ is_pi(PI) }, !,
+	pred_source_button(PI, Options).
+object_source_button(_, _) -->
+	[].
+
+
+%%	canonise_predref(+PredRef, -PI:Name/Arity, +Options0, -Options) is det.
+%
+%	Canonise a predicate reference. A   possible module qualifier is
+%	added as module(M) to Options.
+
+canonise_predref(M:PI0, PI, Options0, [module(M),Options]) :- !,
+	canonise_predref(PI0, PI, Options0, Options).
+canonise_predref(//(Head), PI, Options0, Options) :-	!,
+	functor(Head, Name, Arity),
+	PredArity is Arity + 2,
+	canonise_predref(Name/PredArity, PI, Options0, Options).
+canonise_predref(Name//Arity, PI, Options0, Options) :- !,
+	PredArity is Arity + 2,
+	canonise_predref(Name/PredArity, PI, Options0, Options).
+canonise_predref(PI, PI, Options, Options) :-
+	PI = _/_, !.
+canonise_predref(Head, PI, Options0, Options) :-
+	functor(Head, Name, Arity),
+	canonise_predref(Name/Arity, PI, Options0, Options).
+
+
 %%	pred_head(+Term) is det.
 %
 %	Emit a predicate head. The functor is  typeset as a =span= using
@@ -752,28 +865,25 @@ pred_head(Head) -->
 	html(b(class=pred, Head)).
 pred_head(Head) -->			% Infix operators
 	{ Head =.. [Functor,Left,Right],
-	  current_op(_,Type,Functor),
-	  op_type(Type, infix), !
-	},
+	  is_op_type(Functor, infix)
+	}, !,
 	html([ var(class=arglist, \pred_arg(Left, 1)),
-	       b(class=pred, Functor),
+	       ' ', b(class=pred, Functor), ' ',
 	       var(class=arglist, \pred_arg(Right, 2))
 	     ]).
 pred_head(Head) -->			% Prefix operators
 	{ Head =.. [Functor,Arg],
-	  current_op(_,Type,Functor),
-	  op_type(Type, prefix), !
-	},
-	html([ b(class=pred, Functor),
+	  is_op_type(Functor, prefix)
+	}, !,
+	html([ b(class=pred, Functor), ' ',
 	       var(class=arglist, \pred_arg(Arg, 1))
 	     ]).
 pred_head(Head) -->			% Postfix operators
 	{ Head =.. [Functor,Arg],
-	  current_op(_,Type,Functor),
-	  op_type(Type, postfix), !
-	},
+	  is_op_type(Functor, postfix)
+	}, !,
 	html([ var(class=arglist, \pred_arg(Arg, 1)),
-	       b(class=pred, Functor)
+	       ' ', b(class=pred, Functor)
 	     ]).
 pred_head(Head) -->			% Plain terms
 	{ Head =.. [Functor|Args] },
@@ -782,6 +892,15 @@ pred_head(Head) -->			% Plain terms
 		   [ '(', \pred_args(Args, 1), ')' ])
 	     ]).
 
+%%	is_op_type(+Atom, ?Type)
+%
+%	True if Atom is an operator of   Type.  Type is one of =prefix=,
+%	=infix= or =postfix=.
+
+is_op_type(Functor, Type) :-
+	current_op(_Pri, F, Functor),
+	op_type(F, Type).
+
 op_type(fx,  prefix).
 op_type(fy,  prefix).
 op_type(xf,  postfix).
@@ -842,7 +961,7 @@ pred_det(Det) -->
 %%	term(+Term, +Bindings)// is det.
 %
 %	Process the \term element as produced by doc_wiki.pl.
-%	
+%
 %	@tbd	Properly merge with pred_head//1
 
 term(Atom, []) -->
@@ -854,9 +973,9 @@ term(Term, Bindings) -->
 	}, !,
 	pred_head(Term).
 term(Term, Bindings) -->
-	{ bind_vars(Bindings) },
+	{ bind_vars(Term, Bindings) },
 	argtype(Term).
-	
+
 
 		 /*******************************
 		 *	       PREDREF		*
@@ -868,7 +987,7 @@ term(Term, Bindings) -->
 %	Create a reference to a predicate. The reference consists of the
 %	relative path to the  file  using   the  predicate  indicator as
 %	anchor.
-%	
+%
 %	Current file must  be  available   through  the  global variable
 %	=pldoc_file=. If this variable not  set   it  creates  a link to
 %	/doc/<file>#anchor.  Such links only work in the online browser.
@@ -923,21 +1042,22 @@ manref(Name/Arity, HREF, Options) :-
 	format(string(FragmentId), '~w/~d', [Name, Arity]),
 	www_form_encode(FragmentId, EncId),
 	(   option(files(_Map), Options)
-	->  option(man_server(Root), Options,
-		   'http://gollem.science.uva.nl/SWI-Prolog/pldoc/')
-	;   doc_server_root(Root)
+	->  option(man_server(ManHandler), Options,
+		   'http://gollem.science.uva.nl/SWI-Prolog/pldoc/man')
+	;   http_location_by_id(pldoc_man, ManHandler)
 	),
-	format(string(HREF), '~wman?predicate=~w', [Root, EncId]).
-	
+	http_location_by_id(pldoc_man, ManHandler),
+	format(string(HREF), '~w?predicate=~w', [ManHandler, EncId]).
+
 
 %%	pred_href(+NameArity, +Module, -HREF) is semidet.
 %
 %	Create reference.  Prefer:
-%	
+%
 %		1. Local definition
 %		2. If from package and documented: package documentation
 %		3. From any file
-%	
+%
 %	@bug	Should analyse import list to find where the predicate
 %		comes from.
 
@@ -945,7 +1065,6 @@ pred_href(Name/Arity, Module, HREF) :-
 	format(string(FragmentId), '~w/~d', [Name, Arity]),
 	www_form_encode(FragmentId, EncId),
 	functor(Head, Name, Arity),
-	doc_server_root(Root),
 	(   catch(relative_file(Module:Head, File), _, fail)
 	->  format(string(HREF), '~w#~w', [File, EncId])
 	;   in_file(Module:Head, File)
@@ -954,8 +1073,10 @@ pred_href(Name/Arity, Module, HREF) :-
 		prolog:doc_object_summary(Name/Arity, packages, _, _)
 	    ->	format(string(FragmentId), '~w/~d', [Name, Arity]),
 		www_form_encode(FragmentId, EncId),
-		format(string(HREF), '~wman?predicate=~w', [Root, EncId])
-	    ;	format(string(HREF), '~wdoc~w#~w', [Root, File, EncId])
+		http_location_by_id(pldoc_man, ManHandler),
+		format(string(HREF), '~w?predicate=~w', [ManHandler, EncId])
+	    ;	http_location_by_id(pldoc_doc, DocHandler),
+	        format(string(HREF), '~w~w#~w', [DocHandler, File, EncId])
 	    )
 	).
 
@@ -966,7 +1087,23 @@ relative_file(Head, RelFile) :-
 	b_getval(pldoc_file, CurrentFile), CurrentFile \== [],
 	in_file(Head, DefFile),
 	relative_file_name(DefFile, CurrentFile, RelFile).
-	
+
+%%	pred_source_href(+Pred:predicate_indicator, +Module, -HREF) is semidet.
+%
+%	HREF is a URL to show the predicate source in its file.
+
+pred_source_href(Name/Arity, Module, HREF) :-
+	format(string(FragmentId), '~w/~d', [Name, Arity]),
+	www_form_encode(FragmentId, EncId),
+	functor(Head, Name, Arity),
+	(   catch(relative_file(Module:Head, File), _, fail)
+	->  format(string(HREF), '~w?source=true#~w', [File, EncId])
+	;   in_file(Module:Head, File),
+	    http_location_by_id(pldoc_doc, DocHandler),
+	    format(string(HREF), '~w~w?source=true#~w', [DocHandler, File, EncId])
+	).
+
+
 %%	object_ref(+Object, +Options)// is det.
 %
 %	Create a hyperlink to Object. Points to the /doc_for URL. Object
@@ -986,7 +1123,7 @@ object_ref(Obj, Options) -->
 	{ object_href(Obj, HREF, Options)
 	},
 	html(a(href(HREF), \object_link(Obj, Options))).
-	
+
 %%	object_href(+Object, -HREF) is det.
 %%	object_href(+Object, -HREF, +Options) is det.
 %
@@ -997,24 +1134,35 @@ object_href(Obj, HREF) :-
 
 object_href(M:PI0, HREF, Options) :-
 	option(files(Map), Options),
-	current_module(M, File),
+	module_property(M, file(File)),
 	memberchk(file(File, DocFile), Map), !,
 	expand_pi(PI0, PI),
 	term_to_string(PI, PIS),
 	www_form_encode(PIS, PIEnc),
 	file_base_name(DocFile, LocalFile),	% TBD: proper directory index
 	format(string(HREF), '~w#~w', [LocalFile, PIEnc]).
-object_href(Obj, HREF, _Options) :-
+object_href(Obj0, HREF, _Options) :-
+	localise_object(Obj0, Obj),
 	term_to_string(Obj, String),
 	www_form_encode(String, Enc),
-	doc_server_root(Root),
-	format(string(HREF), '~wdoc_for?object=~w', [Root, Enc]).
+	http_location_by_id(pldoc_object, ObjHandler),
+	format(string(HREF), '~w?object=~w', [ObjHandler, Enc]).
 
 expand_pi(Name//Arity0, Name/Arity) :- !,
 	Arity is Arity0+2.
 expand_pi(PI, PI).
 
 
+%%	localise_object(+ObjIn, -ObjOut) is det.
+%
+%	Abstract  path-details  to  make  references  more  stable  over
+%	versions.
+
+localise_object(Obj0, Obj) :-
+	prolog:doc_canonical_object(Obj0, Obj), !.
+localise_object(Obj, Obj).
+
+
 %%	term_to_string(+Term, -String) is det.
 %
 %	Convert Term, possibly  holding  variables,   into  a  canonical
@@ -1040,16 +1188,17 @@ term_to_string(Term, String) :-
 
 object_link(Obj, Options) -->
 	prolog:doc_object_link(Obj, Options), !.
-object_link(_M:PI, _) --> !,
-	pi(PI).
 object_link(PI, _) -->
-	pi(PI), !.
+	{ is_pi(PI) }, !,
+	pi(PI).
 object_link(Module:module(_Title), _) -->
-	{ current_module(Module, File),
+	{ module_property(Module, file(File)),
 	  file_base_name(File, Base)
 	}, !,
 	html(Base).
 
+pi(_M:PI) --> !,
+	pi(PI).
 pi(Name/Arity) --> !,
 	html([Name, /, Arity]).
 pi(Name//Arity) -->
@@ -1086,15 +1235,15 @@ in_file(Module, Head, File) :-
 %%	relative_file_name(+Path:atom, +RelTo:atom, -RelPath:atom) is det.
 %
 %	Create a relative path from an absolute one.
-%	
+%
 %	@tbd	move to library?
 
 relative_file_name(Path, RelTo, RelPath) :-
-        concat_atom(PL, /, Path),
-        concat_atom(RL, /, RelTo),
+        atomic_list_concat(PL, /, Path),
+        atomic_list_concat(RL, /, RelTo),
         delete_common_prefix(PL, RL, PL1, PL2),
         to_dot_dot(PL2, DotDot, PL1),
-        concat_atom(DotDot, /, RelPath).
+        atomic_list_concat(DotDot, /, RelPath).
 
 delete_common_prefix([H|T01], [H|T02], T1, T2) :- !,
         delete_common_prefix(T01, T02, T1, T2).
@@ -1107,25 +1256,110 @@ to_dot_dot([_|T0], ['..'|T], Tail) :-
 
 
 %%     file(+FileName)// is det.
+%%     file(+FileName, +Options)// is det.
 %
-%      Create a link to another filename if the file exists.  Called by
-%      \file(File) terms in the DOM term generated by wiki.pl.
+%      Create a link to another filename if   the file exists. Called by
+%      \file(File) terms in the DOM term generated by wiki.pl. Supported
+%      options are:
+%
+%          * label(+Label)
+%          Label to use for the link to the file.
+%
+%          * absolute_path(+Path)
+%          Absolute location of the referenced file.
+%
+%          * href(+HREF)
+%          Explicitely provided link; overrule link computation.
+%
+%          * map_extension(+Pairs)
+%          Map the final extension if OldExt-NewExt is in Pairs.
+%
+%          * files(+Map)
+%          List of file(Name, Link) that specifies that we must
+%	   user Link for the given physical file Name.
+%
+%	@tbd	Translation of files to HREFS is a mess.  How to relate
+%		these elegantly?
 
 file(File) -->
-	{ catch(nb_getval(pldoc_options, Options), _, Options = []) },
-	file(File, Options).
-	  
+	file(File, []).
+
 file(File, Options) -->
-	{ existing_linked_file(File, Path) }, !,
-	{ (   option(files(Map), Options),
-	      memberchk(file(Path, DocFile), Map)
-	  ->  file_base_name(DocFile, HREF)	% TBD: proper location
-	  ;   HREF=File
-	  )
+	{ catch(nb_getval(pldoc_options, GenOptions), _, GenOptions = []),
+	  merge_options(Options, GenOptions, FinalOptions)
 	},
-	html(a([class(file), href(HREF)], File)).
+	link_file(File, FinalOptions), !.
 file(File, _) -->
-	html(code(class(file), File)).
+	html(code(class(nofile), File)).
+
+link_file(File, Options) -->
+	{ file_href(File, HREF, Options),
+	  option(label(Label), Options, File),
+	  option(class(Class), Options, file)
+	},
+	html(a([class(Class), href(HREF)], Label)).
+
+%%	file_href(+FilePath, -HREF, +Options) is det.
+%
+%	Find URL for refering to FilePath based on Options.
+
+file_href(_, HREF, Options) :-
+	option(href(HREF), Options), !.
+file_href(File, HREF, Options) :-
+	file_href_real(File, HREF0, Options),
+	map_extension(HREF0, HREF, Options).
+
+%%	map_extension(+HREFIn, -HREFOut, Options) is det.
+%
+%	Replace extension using the option
+%
+%	    * map_extension(+Pairs)
+
+map_extension(HREF0, HREF, Options) :-
+	option(map_extension(Map), Options),
+	file_name_extension(Base, Old, HREF0),
+	memberchk(Old-New, Map), !,
+	file_name_extension(Base, New, HREF).
+map_extension(HREF, HREF, _).
+
+
+file_href_real(File, HREF, Options) :-
+	(   option(absolute_path(Path), Options)
+	;   existing_linked_file(File, Path)
+	), !,
+	(   option(files(Map), Options),
+	    memberchk(file(Path, LinkFile), Map)
+	->  true
+	;   LinkFile = Path
+	),
+	file_href(LinkFile, HREF).
+file_href_real(File, HREF, _) :-
+	http:location(Alias, _, _),
+	Term =.. [Alias,File],
+	absolute_file_name(Term, _,
+			   [ access(read),
+			     file_errors(fail)
+			   ]), !,
+	http_absolute_location(Term, HREF, []).
+
+%%	file_href(+FilePath, -HREF) is det.
+%
+%	Create a relative URL from  the   current  location to the given
+%	absolute file name. It resolves  the   filename  relative to the
+%	file being processed  that  is   available  through  the  global
+%	variable =pldoc_file=.
+
+file_href(Path, HREF) :-
+	nb_current(pldoc_file, CFile),
+	CFile \== [], !,
+	relative_file_name(Path, CFile, HREF).
+file_href(Path, Path).
+
+
+%%	existing_linked_file(+File, -Path) is semidet.
+%
+%	True if File is a path to an existing file relative to the
+%	current file.  Path is the absolute location of File.
 
 existing_linked_file(File, Path) :-
 	catch(b_getval(pldoc_file, CurrentFile), _, fail),
@@ -1137,17 +1371,79 @@ existing_linked_file(File, Path) :-
 			   ]).
 
 
-%%	image(+FileName)// is det.
+%%	include(+FileName, +Type, +Options)// is det.
 %
-%	Create an inline image from a link   to another file if the file
-%	exists. Called by \image(File) terms in   the DOM term generated
-%	by wiki.pl.
+%	Inline FileName. If this is an image file, show an inline image.
+%	Else we create a link  like   file//1.  Called by \include(File,
+%	Type)  terms  in  the  DOM  term  generated  by  wiki.pl  if  it
+%	encounters [[file.ext]].
 
-image(File) -->
-	{ existing_linked_file(File, _) }, !,
-	html(img([src(File), alt(File)])).
-image(File) -->
-	html(code(class(file), ['[[',File,']]'])).
+include(PI, predicate, _) --> !,
+	(   html_tokens_for_predicates(PI, [])
+	->  []
+	;   html(['[[', \predref(PI), ']]'])
+	).
+include(File, image, Options) -->
+	{ file_href(File, HREF, Options),
+	  include(image_attribute, Options, Attrs0),
+	  merge_options(Attrs0,
+			[ alt(File),
+			  border(0),
+			  src(HREF)
+			], Attrs)
+	},
+	html(img(Attrs)).
+include(File, _Type, Options) -->
+	link_file(File, Options), !.
+include(File, _, _) -->
+	html(code(class(nofile), ['[[',File,']]'])).
+
+image_attribute(src(_)).
+image_attribute(alt(_)).
+image_attribute(align(_)).
+image_attribute(width(_)).
+image_attribute(height(_)).
+image_attribute(border(_)).
+
+
+%%	html_tokens_for_predicates(+PI, +Options)// is semidet.
+%
+%	Inline description for a predicate as produced by the text below
+%	from wiki processing.
+%
+%	==
+%		* [[member/2]]
+%		* [[append/3]]
+%	==
+
+html_tokens_for_predicates([], _Options) -->
+	[].
+html_tokens_for_predicates([H|T], Options) --> !,
+	html_tokens_for_predicates(H, Options),
+	html_tokens_for_predicates(T, Options).
+html_tokens_for_predicates(PI, Options) -->
+	{ PI = _:_/_, !,
+	  (   doc_comment(PI, Pos, _Summary, Comment)
+	  ->  true
+	  ;   Comment = ''
+	  )
+	},
+	object(PI, Pos, Comment, [dl], _, Options).
+html_tokens_for_predicates(Spec, Options) -->
+	{ findall(PI, documented_pi(Spec, PI), List),
+	  List \== [], !
+	},
+	html_tokens_for_predicates(List, Options).
+html_tokens_for_predicates(Spec, Options) -->
+	man_page(Spec, [links(false)|Options]).
+
+
+documented_pi(Spec, PI) :-
+	generalise_spec(Spec, PI),
+	doc_comment(PI, _Pos, _Summary, _Comment).
+
+generalise_spec(Name/Arity, _M:Name/Arity).
+generalise_spec(Name//Arity, _M:Name//Arity).
 
 
 		 /*******************************
@@ -1155,36 +1451,45 @@ image(File) -->
 		 *******************************/
 
 
-%%	doc_for_wiki_file(+File, +Out:stream, +Options) is det.
+%%	doc_for_wiki_file(+File, +Options) is det.
 %
 %	Write HTML for the File containing wiki data.
 
-doc_for_wiki_file(FileSpec, Out, _Options) :-
+doc_for_wiki_file(FileSpec, _Options) :-
 	absolute_file_name(FileSpec, File,
 			   [ access(read)
 			   ]),
 	read_file_to_codes(File, String, []),
 	b_setval(pldoc_file, File),
-	call_cleanup((wiki_string_to_dom(String, [], DOM),
-		      phrase(html(DOM), Tokens),
-		      print_html_head(Out),
-		      print_html(Out, Tokens)
-		     ),
+	call_cleanup(reply_wiki_page(File, String),
 		     nb_delete(pldoc_file)).
 
+reply_wiki_page(File, String) :-
+	wiki_codes_to_dom(String, [], DOM),
+	title(DOM, File, Title),
+	reply_html_page(title(Title),
+			[ \html_requires(pldoc)
+			| DOM
+			]).
+
+title(DOM, _, Title) :-
+	sub_term(h1(_,Title), DOM), !.
+title(_, File, Title) :-
+	file_base_name(File, Title).
+
 
 		 /*******************************
 		 *	      ANCHORS		*
 		 *******************************/
 
-%%	anchor_name(+Head, -PI:atom/integer, -Anchor:string) is det.
+%%	pred_anchor_name(+Head, -PI:atom/integer, -Anchor:string) is det.
 %
 %	Create an HTML anchor name from Head.
 
-anchor_name(//(Head), Name/Arity, Anchor) :- !,
+pred_anchor_name(//(Head), Name/Arity, Anchor) :- !,
 	functor(Head, Name, DCGArity),
 	Arity is DCGArity+2,
 	format(string(Anchor), '~w/~d', [Name, Arity]).
-anchor_name(Head, Name/Arity, Anchor) :-
+pred_anchor_name(Head, Name/Arity, Anchor) :-
 	functor(Head, Name, Arity),
 	format(string(Anchor), '~w/~d', [Name, Arity]).
