@@ -1,4 +1,4 @@
-/*  $Id: doc_index.pl,v 1.27 2007/06/21 09:19:47 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -30,7 +30,7 @@
 */
 
 :- module(pldoc_index,
-	  [ doc_for_dir/3,		% +Dir, +Out, +Options
+	  [ doc_for_dir/2,		% +Dir, +Options
 	    dir_index/4,		% +Dir, +Options, //
 	    object_summaries/5,		% +Objs, +Section, +Options, //
 	    file_index_header/4,	% +File, +Options, //
@@ -42,7 +42,10 @@
 :- use_module(doc_html).
 :- use_module(doc_wiki).
 :- use_module(doc_search).
-:- use_module(library('http/html_write')).
+:- use_module(doc_util).
+:- use_module(library(http/http_dispatch)).
+:- use_module(library(http/html_write)).
+:- use_module(library(http/html_head)).
 :- use_module(library(readutil)).
 :- use_module(library(url)).
 :- use_module(library(option)).
@@ -52,14 +55,14 @@
 /** <module> Create indexes
 */
 
-%%	doc_for_dir(+Dir, +Out, +Options) is det.
+%%	doc_for_dir(+Dir, +Options) is det.
 %
 %	Write summary index for all files  in   Dir  to  Out. The result
 %	consists of the =README= file  (if   any),  a  table holding with
 %	links to objects and summary  sentences   and  finaly the =TODO=
 %	file (if any).
 
-doc_for_dir(DirSpec, Out, Options) :-
+doc_for_dir(DirSpec, Options) :-
 	absolute_file_name(DirSpec,
 			   [ file_type(directory),
 			     access(read)
@@ -67,10 +70,8 @@ doc_for_dir(DirSpec, Out, Options) :-
 			   Dir),
 	file_base_name(Dir, Base),
 	Title = Base,
-	doc_page_dom(Title, \dir_index(Dir, Options), DOM),
-	phrase(html(DOM), Tokens),
-	print_html_head(Out),
-	print_html(Out, Tokens).
+	reply_html_page(title(Title),
+			\dir_index(Dir, Options)).
 
 
 %%	dir_index(+Dir, +Options)//
@@ -133,10 +134,10 @@ dir_footer(_, _) -->
 
 wiki_file(Dir, Type) -->
 	{ wiki_file_type(Type, Base),
-	  concat_atom([Dir, /, Base], File),
+	  atomic_list_concat([Dir, /, Base], File),
 	  access_file(File, read), !,
 	  read_file_to_codes(File, String, []),
-	  wiki_string_to_dom(String, [], DOM)
+	  wiki_codes_to_dom(String, [], DOM)
 	},
 	pldoc_html:html(DOM).
 
@@ -157,7 +158,7 @@ file_indices([], _) -->
 file_indices([H|T], Options) -->
 	file_index(H, Options),
 	file_indices(T, Options).
-	
+
 %%	file_index(+File, +Options)// is det.
 %
 %	Create an index for File.
@@ -216,15 +217,16 @@ doc_file_href(File, HREF, _) :-
 %
 %	Create a /doc HREF from Path.  There   are  some nasty things we
 %	should take care of.
-%	
-%		* Windows paths may start with Drive:
+%
+%		* Windows paths may start with =|L:|= (mapped to =|/L:|=)
 %		* Paths may contain spaces and other weird stuff
 
-doc_file_href(File, HREF) :-
+doc_file_href(File0, HREF) :-
+	insert_alias(File0, File),
 	ensure_slash_start(File, SlashFile),
 	http_location([path(SlashFile)], Escaped),
-	doc_server_root(Root),
-	format(string(HREF), '~wdoc~w', [Root, Escaped]).
+	http_location_by_id(pldoc_doc, DocRoot),
+	format(string(HREF), '~w~w', [DocRoot, Escaped]).
 
 %%	ensure_slash_start(+File0, -File) is det.
 %
@@ -246,7 +248,7 @@ object_summaries(Objects, Section, Options) -->
 	  keysort(Tagged, Ordered)
 	},
 	obj_summaries(Ordered, Section, Options).
-	
+
 obj_summaries([], _, _) -->
 	[].
 obj_summaries([_Tag-H|T], Section, Options) -->
@@ -266,11 +268,12 @@ tag_pub_priv([H|T0], [Tag-H|T], Options) :-
 %
 %	Create a summary for Object.  Summary consists of a link to
 %	the Object and a summary text as a table-row.
-%	
+%
 %	@tbd	Hacky interface.  Do we demand Summary to be in Wiki?
 
 object_summary(doc(Obj, _Pos, Summary), _Section, Options) --> !,
-	(   { wiki_string_to_dom(Summary, [], DOM0),
+	(   { string_to_list(Summary, Codes),
+	      wiki_codes_to_dom(Codes, [], DOM0),
 	      strip_leading_par(DOM0, DOM),
 	      (	  private(Obj, Options)
 	      ->  Class = private		% private definition
@@ -280,8 +283,11 @@ object_summary(doc(Obj, _Pos, Summary), _Section, Options) --> !,
 	->  html(tr(class(Class),
 		    [ td(\object_ref(Obj, Options)),
 		      td(class(summary), DOM),
-		      td(align(right),
-			 \object_edit_button(Obj, Options))
+		      td([align(right), width('30')],
+			 span(style('float:right'),
+			      [ \object_source_button(Obj, Options),
+				\object_edit_button(Obj, Options)
+			      ]))
 		    ]))
 	;   []
 	).
@@ -296,7 +302,7 @@ object_summary(_, _, _) -->
 		 /*******************************
 		 *	    NAVIGATION		*
 		 *******************************/
-	       
+
 %%	doc_links(+Directory, +Options)// is det.
 %
 %	Provide overview links and search facilities.
@@ -309,14 +315,16 @@ doc_links(Directory, Options) -->
 	->  working_directory(Dir, Dir)
 	;   Dir = Directory
 	},
-	html(div(class(navhdr),
-		 [ span(style('float:left'),
-			div([ \source_dir_menu(Dir),
-			      \version
-			    ])),
-		   span(style('float:right'), \search_form(Options)),
-		   br(clear(both))
-		 ])).
+	html([ \html_requires(pldoc),
+	       div(class(navhdr),
+		   [ div(class(jump),
+			  div([ \source_dir_menu(Dir),
+				\version
+			      ])),
+		     div(class(search), \search_form(Options)),
+		     br(clear(right))
+		   ])
+	     ]).
 
 
 %%	version// is det.
@@ -324,10 +332,7 @@ doc_links(Directory, Options) -->
 %	Prolog version
 
 version -->
-	{ current_prolog_flag(version, Version),
-	  Major is Version // 10000,
-	  Minor is (Version // 100) mod 100,
-	  Patch is Version mod 100
+	{ current_prolog_flag(version_data, swi(Major, Minor, Patch, _))
 	},
 	html(a([ class(prolog_version),
 		 href('http://www.swi-prolog.org')
@@ -341,16 +346,15 @@ version -->
 
 source_dir_menu(Dir) -->
 	{ findall(D, source_directory(D), List),
-	  sort(List, Dirs),
-	  doc_server_root(Root)
+	  sort(List, Dirs)
 	},
-	html(form([ action(Root+'directory')
+	html(form([ action(location_by_id(pldoc_dir))
 		  ],
 		  [ input([type(submit), value('Go')]),
 		    select(name(dir),
 			   \source_dirs(Dirs, Dir))
 		  ])).
-	     
+
 source_dirs([], _) -->
 	[].
 source_dirs([H|T], WD) -->
@@ -363,14 +367,18 @@ source_dirs([H|T], WD) -->
 	  format(string(Call), 'document.location=\'~w\';', [HREF])
 	},
 	html(option([onClick(Call)|Attrs], H)),
-	source_dirs(T, WD).	
+	source_dirs(T, WD).
 
-%%	source_directory(?Dir) is nondet.
+%%	source_directory(+Dir) is semidet.
+%%	source_directory(-Dir) is det.
 %
 %	True if Dir is a directory  from   which  we  have loaded Prolog
 %	sources.
 
 source_directory(Dir) :-
-	source_file(File),
-	once(source_file(_, File)),
-	file_directory_name(File, Dir).
+	(   ground(Dir)
+	->  '$time_source_file'(File, _Time1, _System1),
+	    file_directory_name(File, Dir), !
+	;   '$time_source_file'(File, _Time2, _System2),
+	    file_directory_name(File, Dir)
+	).
