@@ -166,17 +166,17 @@ ns(serql,'http://www.openrdf.org/schema/serql#').
 %%	rdf_register_ns(+Alias, +URI, +Options) is det.
 %
 %	Register Alias as an abbreviateion for URI. Options:
-%	
+%
 %		* force(Boolean)
 %		If =true=, Replace existing namespace alias. Please note
 %		that replacing a namespace is dangerous as namespaces
 %		affect preprocessing. Make sure all code that depends on
 %		a namespace is compiled after changing the registration.
-%		
+%
 %		* keep(Boolean)
 %		If =true= and Alias is already defined, keep the
 %		original message and succeed silently.
-%		
+%
 %	Without options, an attempt  to  redefine   an  alias  raises  a
 %	permission error.
 
@@ -201,13 +201,13 @@ rdf_register_ns(Alias, URI, _) :-
 	assert(ns(Alias, URI)).
 
 %%	register_file_ns(+Map:list(NS=URL)) is det.
-%	
+%
 %	Register a namespace as encounted in   the  namespace list of an
 %	RDF document. We only register if  both the abbreviation and URL
 %	are not already known. Is there a   better  way? This code could
 %	also do checks on the consistency   of  RDF and other well-known
 %	namespaces.
-%	
+%
 %	@tbd	Better error handling
 
 register_file_ns([]) :- !.
@@ -238,11 +238,11 @@ rdf_global_id(Global, Global).
 
 %%	rdf_global_object(+Object, -GlobalObject) is semidet.
 %%	rdf_global_object(-Object, +GlobalObject) is semidet.
-%	
+%
 %	Same as rdf_global_id/2,  but  intended   for  dealing  with the
 %	object part of a  triple,  in   particular  the  type  for typed
 %	literals.
-%	
+%
 %	@error	existence_error(rdf_namespace, NS)
 
 rdf_global_object(NS:Local, Global) :-
@@ -251,7 +251,7 @@ rdf_global_object(literal(type(NS:Local, Value)),
 		  literal(type(Global, Value))) :-
 	global(NS, Local, Global), !.
 rdf_global_object(Global, Global).
-	    
+
 global(NS, Local, Global) :-
 	(   atom(Global)
 	->  ns(NS, Full),
@@ -265,7 +265,7 @@ global(NS, Local, Global) :-
 
 
 %%	rdf_global_term(+TermIn, -GlobalTerm) is det.
-%	
+%
 %	Does rdf_global_id/2 on all terms NS:Local by recursively analysing
 %	the term.
 
@@ -330,7 +330,7 @@ valid_arg(A) :-
 	throw(error(type_error(rdf_meta_argument, A), _)).
 
 %%	rdf_meta(+Heads)
-%	
+%
 %	This   directive   is   expanded   using   term-expansion.   The
 %	implementation just throws an error in   case  it is called with
 %	the wrong context.
@@ -368,7 +368,7 @@ rdf_expand_arg(:, A, E) :- !,
 rdf_expand_arg(_, A, A).
 
 %%	mk_global(+Src, -Resource)
-%	
+%
 %	Realised rdf_global_id(+, -), but adds compiletime checking,
 %	notably to see whether a namespace is not yet defined.
 
@@ -411,14 +411,14 @@ must_be_atom(X) :-
 	rdf_transaction(:, +).
 
 %%	rdf_equal(?Resource1, ?Resource2)
-%	
+%
 %	Simple equality test to exploit goal-expansion
 
 rdf_equal(Resource, Resource).
 
 
 %%	rdf_has(?Subject, +Predicate, ?Object)
-%	
+%
 %	Succeeds if the triple rdf(Subject, Predicate, Object) is true
 %	exploiting the rdfs:subPropertyOf predicate.
 
@@ -431,7 +431,7 @@ rdf_has(Subject, Predicate, Object) :-
 		 *******************************/
 
 %%	rdf_member_property(?Prop, ?Index)
-%	
+%
 %	Deal with the rdf:_1, ... properties.
 
 term_expansion(member_prefix(x),
@@ -457,7 +457,7 @@ rdf_member_property(P, N) :-
 %%	rdf_node(-Id)
 %
 %	Generate a unique blank node identifier for a subject.
-%	
+%
 %	@deprecated	New code should use rdf_bnode/1.
 
 rdf_node(Resource) :-
@@ -476,10 +476,10 @@ rdf_bnode(Value) :-
 
 
 %%	rdf_is_bnode(+Id)
-%	
+%
 %	Tests if a resource is a blank node (i.e. is an anonymous
 %	resource).
-%	
+%
 %	@see rdf_bnode/1.
 
 rdf_is_bnode(Id) :-
@@ -492,7 +492,7 @@ rdf_is_bnode(Id) :-
 		 *******************************/
 
 %%	rdf_source_location(+Subject, -File:Line)
-%	
+%
 %	Return the source-locations for triples for this subject.
 
 rdf_source_location(Subject, Source) :-
@@ -506,9 +506,9 @@ rdf_source_location(Subject, Source) :-
 		 *******************************/
 
 %%	rdf_statistics(?KeyValue) is nondet.
-%	
+%
 %	Obtain statistics on the RDF database.
-%	
+%
 %	@param KeyValue	Term of the form Key(Value).
 
 rdf_statistics(sources(Count)) :-
@@ -560,7 +560,7 @@ index(rdf(+,+,+), 7).
 		 *******************************/
 
 %%	rdf_current_predicate(?Predicate)
-%	
+%
 %	True if Predicate is a currently defined predicate.
 
 rdf_current_predicate(P) :-
@@ -579,7 +579,7 @@ rdf_current_predicate(P, DB) :-
 	once(rdf(_,P,_,DB:_)).
 
 %%	rdf_predicate_property(?Predicate, ?Property)
-%	
+%
 %	Enumerate predicates and their properties
 
 
@@ -598,7 +598,7 @@ rdf_predicate_property(P, Prop) :-
 
 %%	rdf_transaction(:Goal) is semidet.
 %%	rdf_transaction(:Goal, +Id) is semidet.
-%	
+%
 %	Backward compatibility
 
 rdf_transaction(Goal) :-
@@ -647,7 +647,7 @@ update_mask(X, Mask0, Mask) :-
 	Mask is Mask0 \/ M.
 
 %%	monitor_mask(Name, Mask)
-%	
+%
 %	Mask bit for the monitor events.  Note that this must be kept
 %	consistent with the enum broadcast_id defined in rdf_db.c
 
@@ -679,7 +679,7 @@ monitor_mask(all,	   0xffff).
 
 %%	rdf_save_db(+File) is det.
 %%	rdf_save_db(+File, +DB) is det.
-%	
+%
 %	Save triples into File in a   quick-to-load binary format. If DB
 %	is supplied only triples flagged to originate from that database
 %	are  added.  Files  created  this  way    can  be  loaded  using
@@ -714,7 +714,7 @@ rdf_load_db_no_admin(File, Id, Graphs) :-
 %
 %	Verify the loaded cache file and optionally fix the modification
 %	time (new versions save this along with the snapshot).
-%	
+%
 %	@tbd	What to do if there is a cache mismatch? Delete the loaded
 %		graphs and fail?
 
@@ -763,36 +763,36 @@ rdf_load_db(File) :-
 %
 %	Load RDF file.  Options provides additional processing options.
 %	Currently defined options are:
-%	
+%
 %	    * result(-Action, -Triples, -MD5)
 %	    Return action taken (load, reload, none) and number
 %	    of triples loaded from the file as well as the MD5
 %	    digest.
-%	    
+%
 %	    * blank_nodes(+ShareMode)
 %	    How to handle equivalent blank nodes.  If =share= (default),
 %	    equivalent blank nodes are shared in the same resource.
-%	    
+%
 %	    * base_uri(+URI)
 %	    URI that is used for rdf:about="" and other RDF constructs
 %	    that are relative to the base uri.  Default is the source
 %	    URL.
-%	    
+%
 %	    * db(+DB)
 %	    Named graph in which to load the data.  It is *not* allowed
 %	    to load two sources into the same named graph.
-%	    
+%
 %	    * if(Condition)
 %	    When to load the file. One of =true=, =changed= (default) or
 %	    =not_loaded=.
-%	    
+%
 %	    * cache(Bool)
 %	    If =false=, do not use or create a cache file.
-%	    
+%
 %	    * register_namespaces(Bool)
 %	    If =true= (default =false=), register xmlns= namespace
 %	    declarations as ns/2 namespaces if there is no conflict.
-%	    
+%
 %	Other options are forwarded to process_rdf/3.
 
 rdf_load(Spec) :-
@@ -840,7 +840,6 @@ rdf_load(Spec, Options0) :-
 	    ;   rdf_input_open(Input, Stream, Format),
 		must_be(ground, Format),
 		rdf_format(Format, RDFFormat),
-		rdf_set_graph_source(DB, SourceURL, Modified),
 		call_cleanup(rdf_load_stream(RDFFormat, Stream,
 					     [ base_uri(BaseURI),
 					       blank_nodes(ShareMode),
@@ -848,6 +847,7 @@ rdf_load(Spec, Options0) :-
 					     | RDFOptions
 					     ]),
 			     close_input(Input, Stream)), !,
+		rdf_set_graph_source(DB, SourceURL, Modified),
 		register_file_ns(NSList),
 		(   Cache == true,
 		    rdf_cache_file(SourceURL, write, CacheFile)
@@ -899,11 +899,11 @@ close_input(_, Stream) :-
 %
 %	Resolve input description Term.  Unify   Input  with a canonical
 %	description of the input, which is one of:
-%	
+%
 %	  * stream(Stream)
 %	  * file(AbsolutePath)
 %	  * url(Protocol, URL)
-%	
+%
 %	BaseURI is unified with a default base   URI. Note that this may
 %	be overruled from the options of rdf_load/2.
 
@@ -957,7 +957,7 @@ clean_base_uri(BaseURI0, BaseURI) :-
 	rdf_storage_encoding(Ext, _), Ext \== '',
 	file_name_extension(BaseURI, Ext, BaseURI0), !.
 clean_base_uri(BaseURI, BaseURI).
-	
+
 %%	is_url(+Term, -Protocol) is semidet.
 %
 %	True if Term is an atom denoting a URL of the given Protocol.
@@ -1037,7 +1037,7 @@ rdf_storage_encoding('', plain).
 %%	rdf_load_stream(+Format, +Stream, +Options)
 %
 %	Load RDF data from Stream.
-%	
+%
 %	@tbd	Handle mime-types?
 
 rdf_load_stream(xml, Stream, Options) :- !,
@@ -1094,7 +1094,7 @@ report_loaded(Action, Source, DB, Triples, T0, Options) :-
 
 
 %%	rdf_unload(+Spec)
-%	
+%
 %	Remove the triples loaded from the specified source and remove
 %	the source from the database.
 
@@ -1140,14 +1140,14 @@ rdf_source(DB, SourceURL) :-
 	rdf_graph_source_(DB, SourceURL, _Modified).
 
 %%	rdf_source(?Source)
-%	
+%
 %	True if Source is a loaded source.
-%	
+%
 %	@deprecated	Use rdf_graph/1 or rdf_source/2.
 
 rdf_source(DB) :-
 	rdf_source(DB,_,_,_,_).		% loaded files
-rdf_source(DB) :- 
+rdf_source(DB) :-
 	rdf_graphs_(Sources),		% other sources
 	member(DB, Sources),
 	\+ rdf_source(DB,_,_,_,_),
@@ -1155,7 +1155,7 @@ rdf_source(DB) :-
 	Triples > 0.
 
 %%	rdf_make
-%	
+%
 %	Reload all loaded files that have been modified since the last
 %	time they were loaded.
 
@@ -1192,7 +1192,7 @@ assert_triples([H|_], _) :-
 		 *******************************/
 
 %%	rdf_reset_db
-%	
+%
 %	Remove all triples from the RDF database and reset all its
 %	statistics.
 
@@ -1205,24 +1205,27 @@ rdf_reset_db :-
 		 *	     SAVE RDF		*
 		 *******************************/
 
-%%	rdf_save(File) is det.
-%%	rdf_save(File, +Options) is det.
+%%	rdf_save(Out) is det.
+%%	rdf_save(Out, +Options) is det.
+%
+%	Write RDF data as RDF/XML. Options is a list of one or more of
+%	the following options:
+%
+%		* graph(+Graph)
+%		Save only triples associated to the given named Graph.
 %
-%	Save RDF data to file.  Options is a list of one or more of the
-%	following options:
-%	
 %		* db(+DB)
-%		Save only triples associated to the given DB
-%		
+%		Deprecated synonym for graph(DB).
+%
 %		* anon(Bool)
 %		If false (default true) do not save blank nodes that do
 %		not appear (indirectly) as object of a named resource.
-%		
+%
 %		* base_uri(URI)
 %		BaseURI used. If present, all URIs that can be
 %		represented relative to this base are written using
 %		their shorthand.  See also =write_xml_base= option
-%		
+%
 %		* write_xml_base(Bool)
 %		If =false=, do _not_ include the =|xml:base|=
 %		declaration that is written normally when using the
@@ -1232,19 +1235,20 @@ rdf_reset_db :-
 %		Call Convertor(-Type, -Content, +RDFObject), providing
 %		the opposite for the convert_typed_literal option of
 %		the RDF parser.
-%		
+%
 %		* encoding(Encoding)
 %		Encoding for the output.  Either utf8 or iso_latin_1
-%		
+%
 %		* document_language(+Lang)
 %		Initial xml:lang saved with rdf:RDF element
-%		
+%
 %		* sorted(+Boolean)
 %		If =true= (default =false=), emit subjects sorted on
 %		the full URI.  Useful to make file comparison easier.
-%	
-%	@param File	Location to save the data.  This can also be a 
-%			file-url (=|file://path|=).
+%
+%	@param Out	Location to save the data.  This can also be a
+%			file-url (=|file://path|=) or a stream wrapped
+%			in a term stream(Out).
 
 :- module_transparent
 	rdf_transaction/1,
@@ -1267,16 +1271,17 @@ rdf_save(Spec, Options0) :-
 rdf_save(Spec, DB) :-
 	atom(DB), !,			% backward compatibility
 	to_file(Spec, File),
-	rdf_save2(File, [db(DB)]).
+	rdf_save2(File, [graph(DB)]).
 
 to_file(URL, File) :-
+	atom(URL),
 	file_name_to_url(File, URL), !.
 to_file(File, File).
 
 rdf_save2(File, Options) :-
 	option(encoding(Encoding), Options, utf8),
 	valid_encoding(Encoding),
-	open(File, write, Out, [encoding(Encoding)]),
+	open_output(File, Encoding, Out, Close),
 	flag(rdf_db_saved_subjects, OSavedSubjects, 0),
 	flag(rdf_db_saved_triples, OSavedTriples, 0),
 	call_cleanup(rdf_do_save(Out, Options),
@@ -1285,8 +1290,15 @@ rdf_save2(File, Options) :-
 				  File,
 				  OSavedSubjects,
 				  OSavedTriples,
-				  Out)).
+				  Close)).
 
+open_output(stream(Out), Encoding, Out,
+	    set_stream(Out, encoding(Old))) :-
+	stream_property(Out, encoding(Old)),
+	set_stream(Out, encoding(Encoding)).
+open_output(File, Encoding, Out,
+	    close(Out)) :-
+	open(File, write, Out, [encoding(Encoding)]).
 
 valid_encoding(Enc) :-
 	(   xml_encoding_name(Enc, _)
@@ -1299,8 +1311,8 @@ cleanup_save(Reason,
 	     File,
 	     OSavedSubjects,
 	     OSavedTriples,
-	     Out) :-
-	close(Out),
+	     Close) :-
+	call(Close),
 	flag(rdf_db_saved_subjects, SavedSubjects, OSavedSubjects),
 	flag(rdf_db_saved_triples, SavedTriples, OSavedTriples),
 	retractall(named_anon(_, _)),
@@ -1323,25 +1335,27 @@ rdf_do_save(Out, Options) :-
 					% cleanup handlers isn't called!?
 
 rdf_subject(Subject, Options) :-
-	db(Options, DB),
+	graph(Options, DB),
 	var(DB), !,
 	rdf_subject(Subject).
 rdf_subject(Subject, Options) :-
-	db(Options, DB),
+	graph(Options, DB),
 	rdf_subject(Subject),
 	(   rdf(Subject, _, _, DB:_)
 	->  true
 	).
 
-db(Options, DB) :-
-	(   memberchk(db(DB0), Options)
+graph(Options, DB) :-
+	(   memberchk(graph(DB0), Options)
+	->  DB = DB0
+	;   memberchk(db(DB0), Options)
 	->  DB = DB0
 	;   true			% leave unbound
 	).
 
 
 %%	meta_options(+OptionsIn, -OptionsOut)
-%	
+%
 %	Do module qualification for options that are module sensitive.
 
 meta_options([], []).
@@ -1373,7 +1387,8 @@ rdf_save_header(Out, Options) :-
 	(   member(Id, NSList),
 	    ns(Id, NS),
 	    rdf_quote_uri(NS, QNS),
-	    xml_quote_attribute(QNS, NSText, Enc),
+	    xml_quote_attribute(QNS, NSText0, Enc),
+	    xml_escape_parameter_entity(NSText0, NSText),
 	    format(Out, '~N    <!ENTITY ~w \'~w\'>', [Id, NSText]),
 	    fail
 	;   true
@@ -1399,8 +1414,8 @@ rdf_save_header(Out, Options) :-
 	format(Out, '>~n', []).
 rdf_save_header(Out, FileRef) :-	% compatibility
 	atom(FileRef),
-	rdf_save_header(Out, [db(FileRef)]).
-	
+	rdf_save_header(Out, [graph(FileRef)]).
+
 xml_encoding(Enc, Encoding) :-
 	(   xml_encoding_name(Enc, Encoding)
 	->  true
@@ -1412,17 +1427,37 @@ xml_encoding_name(iso_latin_1, 'ISO-8859-1').
 xml_encoding_name(utf8,        'UTF-8').
 
 
+%%	xml_escape_parameter_entity(+In, -Out) is det.
+%
+%	Escape % as &#37; for entity declarations.
+
+xml_escape_parameter_entity(In, Out) :-
+	sub_atom(In, _, _, _, '%'), !,
+	atom_codes(In, Codes),
+	phrase(escape_parent(Codes), OutCodes),
+	atom_codes(Out, OutCodes).
+xml_escape_parameter_entity(In, In).
+
+escape_parent([]) --> [].
+escape_parent([H|T]) -->
+	(   { H == 37 }
+	->  "&#37;"
+	;   [H]
+	),
+	escape_parent(T).
+
+
 %%	header_namespaces(Options, -List)
-%	
+%
 %	Get namespaces we will define as entities
 
 header_namespaces(Options, List) :-
 	memberchk(namespaces(NSL0), Options), !,
 	sort([rdf,rdfs|NSL0], List).
 header_namespaces(Options, List) :-
-	db(Options, DB),
+	graph(Options, DB),
 	used_namespace_entities(List, DB).
-	
+
 %%	used_namespace_entities(-List, ?DB)
 %
 %	Return the list of namespaces used in an RDF database.
@@ -1451,7 +1486,7 @@ ns_abbreviations([H0|T0], [H|T]) :-
 	ns_abbreviations(T0, T).
 ns_abbreviations([_|T0], T) :-
 	ns_abbreviations(T0, T).
-	
+
 
 add_ns(S, Set) :-
 	rdf_url_namespace(S, Full),
@@ -1491,15 +1526,18 @@ decl_used_predicate_ns(DB) :-
 decl_predicate_ns(Pred) :-
 	predicate_ns(Pred, _), !.
 decl_predicate_ns(Pred) :-
-	rdf_global_id(NS:_Local, Pred),
-	assert(predicate_ns(Pred, NS)), !.
+	rdf_global_id(NS:Local, Pred),
+	xml_name(Local), !,
+	assert(predicate_ns(Pred, NS)).
 decl_predicate_ns(Pred) :-
 	atom_codes(Pred, Codes),
 	append(NSCodes, LocalCodes, Codes),
 	xml_codes(LocalCodes), !,
 	(   NSCodes \== []
 	->  atom_codes(NS, NSCodes),
-	    (   between(1, infinite, N),
+	    (   ns(Id, NS)
+	    ->	assert(predicate_ns(Pred, Id))
+	    ;	between(1, infinite, N),
 		atom_concat(ns, N, Id),
 		\+ ns(Id, _)
 	    ->  rdf_register_ns(Id, NS),
@@ -1529,14 +1567,14 @@ rdf_save_footer(Out) :-
 	format(Out, '</rdf:RDF>~n', []).
 
 %%	rdf_save_non_anon_subject(+Out, +Subject, +Options)
-%	
+%
 %	Save an object.  Anonymous objects not saved if anon(false)
 %	is present in the Options list.
 
 rdf_save_non_anon_subject(_Out, Subject, Options) :-
 	rdf_is_bnode(Subject),
 	(   memberchk(anon(false), Options)
-	;   db(Options, DB),
+	;   graph(Options, DB),
 	    rdf_db(_, _, Subject, DB)
 	), !.
 rdf_save_non_anon_subject(Out, Subject, Options) :-
@@ -1558,19 +1596,19 @@ rdf_save_subject(Out, Subject, Options) :-
 rdf_save_subject(Out, Subject, DB) :-
 	(   var(DB)
 	->  rdf_save_subject(Out, Subject, [])
-	;   rdf_save_subject(Out, Subject, [db(DB)])
+	;   rdf_save_subject(Out, Subject, [graph(DB)])
 	).
-		  
+
 
 %%	rdf_save_subject(+Out:stream, +Subject:resource, +BaseURI,
 %%			 +Indent:int, +Options) is det.
 %
 %	Save properties of Subject.
-%	
+%
 %	@param Indent	Current indentation
 
 rdf_save_subject(Out, Subject, BaseURI, Indent, Options) :-
-	db(Options, DB),
+	graph(Options, DB),
 	findall(Pred=Object, rdf_db(Subject, Pred, Object, DB), Atts0),
 	sort(Atts0, Atts),		% remove duplicates
 	length(Atts, L),
@@ -1598,6 +1636,7 @@ rdf_db(Subject, Pred, Object, DB) :-
 rdf_save_subject(Out, Subject, BaseURI, Atts, Indent, Options) :-
 	rdf_equal(rdf:type, RdfType),
 	select(RdfType=Type, Atts, Atts1),
+	\+ rdf_is_bnode(Type),
 	rdf_id(Type, BaseURI, TypeId),
 	xml_is_name(TypeId), !,
 	format(Out, '~*|<', [Indent]),
@@ -1650,7 +1689,7 @@ save_attributes(Atts, BaseURI, Out, Element, Indent, Options) :-
 	).
 
 %%	split_attributes(+Attributes, -HeadAttrs, -BodyAttr)
-%	
+%
 %	Split attribute (Name=Value) list into attributes for the head
 %	and body. Attributes can only be in the head if they are literal
 %	and appear only one time in the attribute list.
@@ -1661,7 +1700,7 @@ split_attributes(Atts, HeadAttr, BodyAttr) :-
 	append(Dupls, Rest, BodyAttr).
 
 %%	duplicate_attributes(+Attrs, -Duplicates, -Singles)
-%	
+%
 %	Extract attributes that appear more than onces as we cannot
 %	dublicate an attribute in the head according to the XML rules.
 
@@ -1822,7 +1861,7 @@ rdf_save_list(Out, List, BaseURI, Indent, Options) :-
 
 
 %%	rdf_id(+Resource, +BaseURI, -NSLocal)
-%	
+%
 %	Generate a NS:Local  name  for   Resource  given  the  indicated
 %	default namespace. This call is used for elements.
 
@@ -1851,7 +1890,7 @@ rdf_write_id(Out, Atom) :-
 	write(Out, Atom).
 
 %%	rdf_value(+Resource, +BaseURI, -Text, +Encoding)
-%	
+%
 %	According  to  "6.4  RDF  URI  References"  of  the  RDF  Syntax
 %	specification, a URI reference is  UNICODE string not containing
 %	control sequences, represented as  UTF-8   and  then  as escaped
@@ -1868,7 +1907,7 @@ rdf_value(V, _, Text, Encoding) :-
 	atom_concat(Full, Local, V), !,
 	rdf_quote_uri(Local, QLocal0),
 	xml_quote_attribute(QLocal0, QLocal, Encoding),
-	concat_atom(['&', NS, (';'), QLocal], Text).
+	atomic_list_concat(['&', NS, (';'), QLocal], Text).
 rdf_value(V, _, Q, Encoding) :-
 	rdf_quote_uri(V, Q0),
 	xml_quote_attribute(Q0, Q, Encoding).
@@ -1928,7 +1967,7 @@ in_time(Triples, ParseTime) -->
 		 *******************************/
 
 %	prolog:meta_goal(+Goal, -MetaArgs)
-%	
+%
 %	Used by the XPCE/Prolog cross-referencer as well as PceEmacs for
 %	colouring code.
 
