@@ -1,4 +1,4 @@
-/*  $Id: rdf_db_test.pl,v 1.33 2007/07/09 13:56:38 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -105,7 +105,7 @@ save_reload(Encoding) :-
 	delete_file(File).
 
 %	convert_typed(+Type, +Content, -Object)
-%	
+%
 %	Convert to type(Type, PrologValue), providing the inverse of
 %	the default RDF as produced by rdf_db.pl
 
@@ -219,7 +219,7 @@ lang(1) :-
 		lang(en, 'John'),
 		lang(en, ''),
 		'Johannes'
-	      ]. 
+	      ].
 lang(2) :-
 	lang_data,
 	findall(X, rdf(x, a, literal(lang(nl, X))), Xs),
@@ -242,10 +242,10 @@ lang(save_db) :-
 	save_reload_db,
 	X = lang(_,_),
 	findall(X, rdf(x, a, literal(X)), Xs),
-	(   Xs =@= [ lang(nl, 'Jan'), 
+	(   Xs =@= [ lang(nl, 'Jan'),
 		     lang(en, 'John'),
 		     lang(en, ''),
-		     lang(_, 'Johannes') 
+		     lang(_, 'Johannes')
 		   ]
 	->  true
 	;   format(user_error, 'Xs = ~w~n', [Xs]),
@@ -259,7 +259,7 @@ lang(save) :-
 		     [ lang(nl, 'Jan'),
 		       lang(en, 'John'),
 		       lang(en, ''),
-		       'Johannes' 
+		       'Johannes'
 		     ])
 	->  true
 	;   format(user_error, 'Xs = ~q~n', [Xs]),
@@ -267,6 +267,38 @@ lang(save) :-
 	).
 
 
+		 /*******************************
+		 *	    NAMESPACES		*
+		 *******************************/
+
+term_expansion(ns_data(S0,P0,O0),
+	       ns_data(S,P,O)) :-
+	rdf_global_id(S0, S),
+	rdf_global_id(P0, P),
+	rdf_global_id(O0, O).
+
+:- rdf_register_ns(dynamic, 'http://www.dynamic.org/').
+
+ns_data(x, rdf:type, rdf:is).
+ns_data(y, rdf:type, rdf:(dynamic)).
+ns_data(z, rdf:type, (dynamic):rdf).
+ns_data(z, (dynamic):attr1, literal(dynamic)).
+ns_data(z, (dynamic):attr2, (dynamic):rdf).
+
+namespace(save) :-
+	findall(rdf(S,P,O), ns_data(S,P,O), Triples),
+	forall(member(rdf(S,P,O), Triples), rdf_assert(S,P,O)),
+	save_reload,
+	findall(rdf(S,P,O), rdf(S,P,O), NewTriples),
+	(   same_set(Triples, NewTriples)
+	->  true
+	;   format(user_error, 'NewTriples = ~q~n', [NewTriples]),
+	    fail
+	).
+
+
+
+
 		 /*******************************
 		 *	 LITERAL SHARING	*
 		 *******************************/
@@ -553,7 +585,7 @@ prefix(7) :- tprefix(p2, bbbb).
 prefix(8) :- tprefix(p2, bbbbb).
 prefix(9) :- tprefix(p2, 'Bbbbb').
 prefix(10) :- tprefix(p2, 'BBBBB').
-				
+
 prefix(like-1) :-
 	mkprefix_db(_),
 	findall(L, rdf(_,_,literal(like('a*b'), L)), Ls),
@@ -709,7 +741,8 @@ duplicates(1) :-
 
 source(1) :-
 	rdf_assert(a,b,c,test),
-	rdf_db:rdf_set_graph_source(test, 'test.rdf'),
+	get_time(Now),
+	rdf_db:rdf_set_graph_source(test, 'test.rdf', Now),
 	rdf_source(test, X),
 	X == 'test.rdf'.
 
@@ -748,7 +781,7 @@ run_test_script(Script) :-
 
 run_test_scripts(Directory) :-
 	(   script_dir(ScriptDir),
-	    concat_atom([ScriptDir, /, Directory], Dir),
+	    atomic_list_concat([ScriptDir, /, Directory], Dir),
 	    exists_directory(Dir)
 	->  true
 	;   Dir = Directory
@@ -794,6 +827,7 @@ testset(same).
 testset(typed).
 testset(lang).
 testset(wide).
+testset(namespace).
 testset(update).
 testset(transaction).
 testset(label).
@@ -808,7 +842,7 @@ testset(duplicates).
 testset(source).
 
 %	testdir(Dir)
-%	
+%
 %	Enumerate directories holding tests.
 
 testdir('Tests').
@@ -873,7 +907,7 @@ runtest(Name) :-
 	fail.
 runtest(_) :-
 	format(' done.~n').
-	
+
 test_failed(R, Except) :-
 	clause(Head, _, R),
 	functor(Head, Name, 1),
