@@ -1,11 +1,11 @@
-/*  $Id: rdf_http_plugin.pl,v 1.5 2007/03/06 20:15:42 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -43,14 +43,25 @@ from an HTTP server.
 
 :- multifile
 	rdf_db:rdf_open_hook/3,
-	rdf_db:rdf_input_info/3,
+	rdf_db:rdf_input_info_hook/3,
 	rdf_db:url_protocol/1.
 
 rdf_db:rdf_open_hook(url(http, URL), Stream, Format) :-
+	atom(Format), !,
 	http_open(URL, Stream,
-		  [ header(content_type, Type)
+		  [ header(content_type, Type),
+		    request_header('Accept' = 'text/rdf+xml; q=1, \
+					       application/rdf+xml; q=1, \
+					       application/x-turtle; q=0.9, \
+					       text/xml; q=0.5, \
+					       text/html; q=0.5, \
+					       application/xhtml+xml; q=0.5, \
+					       */*; q=0.1')
 		  ]),
-	guess_format(Type, URL, Format).
+	(   ground(Format)
+	->  true
+	;   guess_format(Type, URL, Format)
+	).
 
 
 %%	guess_format(+ContentType, +URL, -Format)
@@ -58,7 +69,7 @@ rdf_db:rdf_open_hook(url(http, URL), Stream, Format) :-
 %	Guess the file format. We first try the official mime-types, but
 %	as it is  quite  likely  many   web-servers  do  not  have these
 %	registered, we use the filename extension as a backup.
-%	
+%
 %	@bug	The turtle parser only parses a subset of n3.
 
 guess_format('text/rdf+xml',	      _, xml).
@@ -68,12 +79,18 @@ guess_format('application/turtle',    _, turtle).
 guess_format('text/rdf+n3',	      _, turtle). % Bit dubious
 guess_format('text/html',	      _, xhtml).
 guess_format('application/xhtml+xml', _, xhtml).
-guess_format(_, URL, Format) :-
-	file_name_extension(_, Ext, URL),
-	rdf_db:rdf_file_type(Ext, Format).
+guess_format(_Mime, URL, Format) :-
+	zip_extension(URL, PlainUrl, Format, PlainFormat),
+	file_name_extension(_Base, Ext, PlainUrl),
+	rdf_db:rdf_file_type(Ext, PlainFormat).
 
 
-rdf_db:rdf_input_info(url(http, URL), Modified, Format) :-
+zip_extension(URL0, URL, gzip(Format), Format) :-
+	file_name_extension(URL, gz, URL0), !.
+zip_extension(URL, URL, Format, Format).
+
+
+rdf_db:rdf_input_info_hook(url(http, URL), Modified, Format) :-
 	http_open(URL, Stream,
 		  [ header(content_type, Type),
 		    header(last_modified, Date),
@@ -85,3 +102,9 @@ rdf_db:rdf_input_info(url(http, URL), Modified, Format) :-
 	parse_time(Date, Modified).
 
 rdf_db:url_protocol(http).
+
+rdf_db:exists_url(url(http, URL)) :-
+	catch(http_open(URL, Stream,
+			[ method(head)
+			]), _, fail),
+	close(Stream).
