@@ -1,11 +1,11 @@
-/*  $Id: rdf_turtle.pl,v 1.5 2007/07/09 13:56:38 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2004-2006, University of Amsterdam
+    Copyright (C): 2004-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -31,13 +31,14 @@
 
 :- module(rdf_turtle,
 	  [ rdf_load_turtle/3,		% +Input, -Triples, +Options
-	    rdf_process_turtle/3	% +Input, :ObObject, +Options
+	    rdf_process_turtle/3	% +Input, :OnObject, +Options
 	  ]).
 :- use_module(library(assoc)).
 :- use_module(library(option)).
 :- use_module(library('semweb/rdf_db')).
 :- use_module(library(debug)).
 :- use_module(library(url)).
+:- use_module(library(http/http_open)).
 
 :- meta_predicate
 	rdf_process_turtle(+,:,+).
@@ -48,7 +49,8 @@ This module implements the Turtle  language   for  representing  the RDF
 triple model as defined by Dave Beckett  from the Institute for Learning
 and Research Technology University of Bristol in the document:
 
-	* http://www.ilrt.bris.ac.uk/discovery/2004/01/turtle/
+	* http://www.w3.org/TeamSubmission/turtle/
+	* http://www.ilrt.bris.ac.uk/discovery/2004/01/turtle/ (original)
 
 The current parser handles all positive   and negative examples provided
 by the above document at october 17, 2004.
@@ -58,28 +60,28 @@ by the above document at october 17, 2004.
 */
 
 %%	rdf_load_turtle(+Input, -Triples, +Options)
-%	
+%
 %	Read a stream or file into a set of triples of the format
-%	
+%
 %		rdf(Subject, Predicate, Object)
-%		
+%
 %	The representation is consistent with the SWI-Prolog RDF/XML
 %	and ntriples parsers.  Provided options are:
-%	
+%
 %		* base_uri(+BaseURI)
 %		Initial base URI.  Defaults to file://<file> for loading
 %		files.
-%		
+%
 %		* anon_prefix(+Prefix)
 %		Blank nodes are generated as <Prefix>1, <Prefix>2, etc.
 %		If Prefix is not an atom blank nodes are generated as
 %		node(1), node(2), ...
 
 rdf_load_turtle(In, Triples, Options) :-
-	open_input(In, Stream),
+	open_input(In, Stream, Close),
 	init_state(In, Stream, Options, State),
 	call_cleanup(phrase(turtle_file(State, Stream), Triples),
-		     close(Stream)).
+		     Close).
 
 
 %%	rdf_process_turtle(+Input, :OnObject, +Options) is det.
@@ -88,10 +90,10 @@ rdf_load_turtle(In, Triples, Options) :-
 %	of triples.  Options is the same as for rdf_load_turtle/3.
 
 rdf_process_turtle(In, OnObject, Options) :-
-	strip_module(OnObject, M, G),
-	open_input(In, Stream),
+	open_input(In, Stream, Close),
 	init_state(In, Stream, Options, State),
-	process_stream(State, Stream, M:G).
+	call_cleanup(process_stream(State, Stream, OnObject),
+		     Close).
 
 
 process_stream(State, In, OnObject) :-
@@ -114,10 +116,29 @@ process_stream(State, In, OnObject) :-
 	).
 
 
-open_input(stream(Stream), Stream) :- !.
-open_input(Stream, Stream) :-
-	is_stream(Stream), !.
-open_input(File, Stream) :-
+%%	open_input(+Input, -Stream, -Close) is det.
+%
+%	Open given input.
+%
+%	@param  Close goal to undo the open action
+%	@tbd	Synchronize with input handling of rdf_db.pl.
+%	@error	existence_error, permission_error
+
+open_input(stream(Stream), Stream, true) :- !,
+	stream_property(Stream, encoding(Old)),
+	(   Old == utf8
+	->  Close = true
+	;   set_stream(Stream, encoding(utf8)),
+	    Close = set_stream(Stream, encoding(Old))
+	).
+open_input(Stream, Stream, Close) :-
+	is_stream(Stream), !,
+	open_input(stream(Stream), Stream, Close).
+open_input(URL, Stream, close(Stream)) :-
+	sub_atom(URL, 0, _, _, 'http://'), !,
+	http_open(URL, Stream, []),
+	set_stream(Stream, encoding(utf8)).
+open_input(File, Stream, close(Stream)) :-
 	absolute_file_name(File, Path,
 			   [ access(read),
 			     extensions([ttl, ''])
@@ -146,6 +167,8 @@ init_state(In, Stream, Options, State) :-
 	->  true
 	;   In = stream(_)
 	->  BaseURI = []
+	;   is_absolute_url(In)
+	->  BaseURI = In
 	;   absolute_file_name(In, File),
 	    atom_concat('file://', File, BaseURI)
 	),
@@ -190,6 +213,11 @@ triples(State, []) -->
 	uri(State, URI),
 	{ setarg(1, State, URI)
 	}.
+triples(State, []) -->
+      ['@',name(base)],!,
+      uri(State,URI),
+      { setarg(1,State,URI)
+      }.
 triples(State, Triples) -->
 	subject(State, Subject, Triples, T),
 	(   predicate_object_list(State, Subject, T, [])
@@ -197,7 +225,7 @@ triples(State, Triples) -->
 	;   { Triples \== T }		% [ p o ; ... ] .
 	->  { T = [] }
 	).
-	
+
 subject(State, Subject, T, T) -->
 	resource(State, Subject), !.
 subject(State, Subject, T0, T) -->
@@ -240,7 +268,7 @@ object(State, Object, T, T) -->
 object(_, literal(type(Type, N)), T, T) -->
 	[ numeric(Tp, Codes) ], !,
 	{ numeric_url(Tp, Type),
-	  atom_codes(N, Codes)
+	  normalise_number(Tp, Codes, N)
 	}.
 object(State, Object, T, T) -->
 	resource(State, Object), !.
@@ -249,6 +277,17 @@ object(State, Object, T0, T) -->
 object(State, _, _, _) -->
 	syntax_error(State, expected_object).
 
+%%	normalise_number(+Type, +Codes:list, -Literal:atom) is det.
+%
+%	Turtle normalisation of numbers. Currently  only implemented for
+%	integers. This ensures that 0001 is parsed as "1"^^xsd:integer.
+
+normalise_number(integer, Codes, N) :- !,
+	number_codes(I, Codes),
+	atom_number(N, I).
+normalise_number(_, Codes, N) :-
+	atom_codes(N, Codes).
+
 term_expansion(numeric_url(I, Local),
 	       numeric_url(I, URI)) :-
 	rdf_global_id(Local, URI).
@@ -281,7 +320,7 @@ resource(State, BaseURI) -->
 	[ : ], !,
 	{ arg(1, State, BaseURI)
 	}.
-	
+
 
 uri(State, URI) -->
 	[ relative_uri(Rel)
@@ -382,7 +421,7 @@ syntax_error(State, Error) -->
 		 *******************************/
 
 %%	turtle_tokens(+In, -List)
-%	
+%
 %	Read a statement from a turtle file, returning the contents as a
 %	list of tokens.
 
@@ -470,7 +509,7 @@ turtle_token(Punct, In, C, P) :-
 	get_code(In, C).
 
 %%	turtle_number(+Char0, +In, -CharNext, -Value)
-%	
+%
 %	Value is Type:CodeList
 
 turtle_number(0'-, In, CN, numeric(T, [0'-|Codes])) :- !,
@@ -483,7 +522,7 @@ turtle_number(C0, In, CN, Value) :-
 	turtle_number_nn(C0, In, CN, Value).
 
 turtle_number_nn(C, In, CN, numeric(Type, Codes)) :-
-	turtle_integer_codes(C, In, CN0, Codes, T0), 	% [0-9]+ 
+	turtle_integer_codes(C, In, CN0, Codes, T0), 	% [0-9]+
 	(   CN0 == 0'.
 	->  T0 = [CN0|T1],
 	    get_code(In, C1),
@@ -522,20 +561,52 @@ e(0'e).
 e(0'E).
 
 sign(0'-).
-sign(0'+).
+sign(0'+).				%'
 
-					% string
-turtle_string(-1, In, _, []) :- !,
+%%	turtle_string(+Char, +In, -NextChar, -String) is det.
+%
+%	Extract a turtle string. We have  seen   a  single  ". There are
+%	-like Python- two types  of  turtle   strings,  one  written  as
+%	"hello" and the other as """hello""".
+
+turtle_string(0'", In, Next, String) :-
+	get_code(In, C),
+	(   C == 0'"
+	->  get_code(In, C2),
+	    turtle_q_string(C2, In, Next, q3, String)
+	;   Next = C,
+	    String = []
+	).
+turtle_string(C, In, Next, String) :-
+	turtle_q_string(C, In, Next, q1, String).
+
+
+turtle_q_string(-1, In, _, _, []) :- !,
 	syntax_error(In, -1, unexpected_end_of_input).
-turtle_string(0'", In, C, []) :- !,
-	get_code(In, C).
-turtle_string(0'\\, In, C, [H|T]) :- !,
+turtle_q_string(0'", In, C, Q, String) :- !,
+	get_code(In, C2),
+	(   Q == q1
+	->  C = C2,
+	    String = []
+	;   /* Q == q3 */
+	    C2 == 0'"
+        ->  get_code(In, C3),
+	    (	C3 == 0'"
+	    ->  get_code(In, C),
+	        String = []
+	    ;   String = [0'", 0'"|Rest],
+	        turtle_q_string(C3, In, C, q3, Rest)
+	    )
+	;   String = [0'"|Rest],
+	    turtle_q_string(C2, In, C, q3, Rest)
+	).
+turtle_q_string(0'\\, In, C, Q, [H|T]) :- !,
 	get_code(In, C1),
 	string_escape(C1, In, C2, H),
-	turtle_string(C2, In, C, T).
-turtle_string(C0, In, C, [C0|T]) :-
+	turtle_q_string(C2, In, C, Q, T).
+turtle_q_string(C0, In, C, Q, [C0|T]) :-
 	get_code(In, C1),
-	turtle_string(C1, In, C, T).
+	turtle_q_string(C1, In, C, Q, T).
 
 
 string_escape(0'n, In, C, 0'\n) :- !,
@@ -569,7 +640,10 @@ language(C0, In, C, [C0|Codes]) :-
 	code_type(C0, lower),
 	get_code(In, C1),
 	lwr_word(C1, In, C2, Codes, Tail),
-	sub_langs(C2, In, C, Tail, []).
+	sub_langs(C2, In, C, Tail, []), !.
+language(_, In, _, _) :-
+	line_count(In, LineNo),
+	syntax_error(In, LineNo, language_specifier).
 
 lwr_word(C0, In, C, [C0|T0], T) :-
 	code_type(C0, lower), !,
@@ -616,12 +690,16 @@ resource_token(C0, In, C, Prefix:Name) :-
 	name(C2, In, C, Name).
 
 
+%%	uri_chars(+Char, +In:stream, -NextChar, -UriChars) is semidet.
+
 uri_chars(0'>, In, C, []) :- !,
 	get_code(In, C).
 uri_chars(0'\\, In, C, [H|T]) :- !,
 	get_code(In, C1),
 	string_escape(C1, In, C2, H),
 	uri_chars(C2, In, C, T).
+uri_chars(-1, _, _, _) :- !,
+	fail.
 uri_chars(C0, In, C, [C0|T]) :-
 	get_code(In, C1),
 	uri_chars(C1, In, C, T).
@@ -684,13 +762,17 @@ turtle_ws(0xA).
 turtle_ws(0xD).
 turtle_ws(0x20).
 
-syntax_error(Stream, _StartLine, Which) :-
+syntax_error(Stream, -1, Which) :- !,
 	stream_property(Stream, file_name(File)),
 	line_count(Stream, LineNo),
 	line_position(Stream, LinePos),
 	character_count(Stream, CharIndex),
 	throw(error(syntax_error(Which),
 		    file(File, LineNo, LinePos, CharIndex))).
+syntax_error(Stream, LineNo, Which) :-
+	stream_property(Stream, file_name(File)),
+	throw(error(syntax_error(Which),
+		    file(File, LineNo, -1, -1))).
 
 		 /*******************************
 		 *	    RDF-DB HOOK		*
@@ -711,3 +793,15 @@ assert_triples([rdf(S,P,O)|T], Location) :-
 	assert_triples(T, Location).
 
 rdf_db:rdf_file_type(ttl, turtle).
+rdf_db:rdf_file_type(n3,  turtle).	% not really, but good enough
+rdf_db:rdf_file_type(nt,  turtle).	% not really, but good enough
+
+
+		 /*******************************
+		 *	   XREF SUPPORT		*
+		 *******************************/
+
+:- multifile
+	prolog:meta_goal/2.
+
+prolog:meta_goal(rdf_process_turtle(_,G,_), [G+2]).
