@@ -1,11 +1,11 @@
-/*  $Id: rdf.pl,v 1.34 2006/08/31 09:44:01 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2002-2006, University of Amsterdam
+    Copyright (C): 2002-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -51,16 +51,20 @@
 %	Parse an XML file holding an RDF term into a list of RDF triples.
 %	see rdf_triple.pl for a definition of the output format. Options:
 %
-%	# base_uri(URI)
-%	URI to use as base
+%		* base_uri(+URI)
+%		URI to use as base
 %
-%	# expand_foreach(Bool)
-%	Apply each(Container, Pred, Object) on the members of Container
-%		
-%	# namespaces([NS=URL, ...])
-%	Return list of namespaces declared using xmlns:NS=URL in
-%	the document.  This can be used to update the namespace
-%	list with rdf_register_ns/2.
+%		* expand_foreach(+Bool)
+%		Apply each(Container, Pred, Object) on the members of
+%		Container
+%
+%		* namespaces(-Namespaces:list(NS=URL))
+%		Return list of namespaces declared using xmlns:NS=URL in
+%		the document.  This can be used to update the namespace
+%		list with rdf_register_ns/2.
+%
+%	@see	Use process_rdf/3 for processing large documents in
+%		_|call-back|_ style.
 
 load_rdf(File, Triples) :-
 	load_rdf(File, Triples, []).
@@ -82,7 +86,7 @@ load_rdf(File, Triples, Options0) :-
 		     rdf_end_file(Cleanup)),
 	exit_ns_collect(NSList),
 	post_process(Options, Triples0, Triples).
-	
+
 entity_options([], [], []).
 entity_options([H|T0], Entities, Rest) :-
 	(   H = entity(_,_)
@@ -145,40 +149,44 @@ member_attribute(A) :-
 		 *******************************/
 
 %%	process_rdf(+Input, :OnObject, +Options)
-%	
+%
 %	Process RDF from Input. Input is either an atom or a term of the
 %	format stream(Handle). For each   encountered  description, call
 %	OnObject(+Triples) to handle the  triples   resulting  from  the
 %	description. Defined Options are:
-%	
+%
 %		* base_uri(+URI)
 %		Determines the reference URI.
-%		
+%
 %		* db(DB)
 %		When loading from a stream, the source is taken from
 %		this option or -if non-existent- from base_uri.
-%		
+%
 %		* lang(LanguageID)
 %		Set initial language (as xml:lang)
-%		
+%
 %		* convert_typed_literal(:Convertor)
 %		Call Convertor(+Type, +Content, -RDFObject) to create
 %		a triple rdf(S, P, RDFObject) instead of rdf(S, P,
 %		literal(type(Type, Content)).
-%		
-%		* namespaces([NS=URL, ...])
+%
+%		*  namespaces(-Namespaces:list(NS=URL))
 %		Return list of namespaces declared using xmlns:NS=URL in
 %		the document.  This can be used to update the namespace
 %		list with rdf_register_ns/2.
-%		
+%
 %		* entity(Name, Value)
 %		Overrule entity values found in the file
-
+%
+%		* embedded(Boolean)
+%		If =true=, do not give warnings if rdf:RDF is embedded
+%		in other XML data.
 
 process_rdf(File, OnObject, Options0) :-
 	is_list(Options0), !,
 	entity_options(Options0, EntOptions, Options1),
-	meta_options(Options1, Options),
+	meta_options(Options1, Options2),
+	process_options(Options2, ProcessOptions, Options),
 	option(base_uri(BaseURI), Options, []),
 	rdf_start_file(Options, Cleanup),
 	strip_module(OnObject, Module, Pred),
@@ -200,7 +208,7 @@ process_rdf(File, OnObject, Options0) :-
 	set_sgml_parser(Parser, file(Source)),
 	set_sgml_parser(Parser, dialect(xmlns)),
 	set_sgml_parser(Parser, space(sgml)),
-	do_process_rdf(Parser, In, NSList, Close, Cleanup).
+	do_process_rdf(Parser, In, NSList, Close, Cleanup, ProcessOptions).
 process_rdf(File, BaseURI, OnObject) :-
 	process_rdf(File, OnObject, [base_uri(BaseURI)]).
 
@@ -218,11 +226,12 @@ def_entity(DTD, Name, Value) :-
 	close(Stream).
 
 
-do_process_rdf(Parser, In, NSList, Close, Cleanup) :-
+do_process_rdf(Parser, In, NSList, Close, Cleanup, Options) :-
 	call_cleanup((   sgml_parse(Parser,
 				    [ source(In),
 				      call(begin, rdf:on_begin),
 				      call(xmlns, rdf:on_xmlns)
+				    | Options
 				    ]),
 			 exit_ns_collect(NSList)
 		     ),
@@ -247,20 +256,26 @@ on_begin(NS:'RDF', Attr, _) :-
 	nb_setval(rdf_state, Options).
 on_begin(Tag, Attr, Parser) :-
 	nb_getval(rdf_state, Options),
-	Options \== (-), !,
-	get_sgml_parser(Parser, line(Start)),
-	get_sgml_parser(Parser, file(File)),
-	sgml_parse(Parser,
-		   [ document(Content),
-		     parse(content)
-		   ]),
-	nb_getval(rdf_object_handler, OnTriples),
-	element_to_plrdf(element(Tag, Attr, Content), Objects, Options),
-	rdf_triples(Objects, Triples),
-	call(OnTriples, Triples, File:Start).
+	(   Options == (-)
+	->  nb_getval(rdf_options, RdfOptions),
+	    (	memberchk(embedded(true), RdfOptions)
+	    ->	true
+	    ;	print_message(warning, rdf(unexpected(Tag, Parser)))
+	    )
+	;   get_sgml_parser(Parser, line(Start)),
+	    get_sgml_parser(Parser, file(File)),
+	    sgml_parse(Parser,
+		       [ document(Content),
+			 parse(content)
+		       ]),
+	    nb_getval(rdf_object_handler, OnTriples),
+	    element_to_plrdf(element(Tag, Attr, Content), Objects, Options),
+	    rdf_triples(Objects, Triples),
+	    call(OnTriples, Triples, File:Start)
+	).
 
 %%	on_xmlns(+NS, +URL, +Parser)
-%	
+%
 %	Build up the list of   encountered xmlns:NS=URL declarations. We
 %	use  destructive  assignment  here   as    an   alternative   to
 %	assert/retract, ensuring thread-safety and better performance.
@@ -299,7 +314,7 @@ modify_state1(xml:lang = Lang, Options0, Options) :- !,
 modify_state1(_, Options, Options).
 
 %%	remove_fragment(+URI, -WithoutFragment)
-%	
+%
 %	When handling xml:base, we must delete the possible fragment.
 
 remove_fragment(URI, Plain) :-
@@ -315,7 +330,7 @@ set_option(Opt, Options0, [Opt|Options]) :-
 
 
 %%	meta_options(+OptionsIn, -OptionsOut)
-%	
+%
 %	Do module qualification for options that are module sensitive.
 
 :- module_transparent
@@ -335,6 +350,12 @@ meta_options([H0|T0], [H|T]) :-
 	meta_options(T0, T).
 
 
+process_options(Options, Process, RestOptions) :-
+	select_option(content_length(Len), Options, RestOptions), !,
+	Process = [content_length(Len)].
+process_options(Options, [], Options).
+
+
 		 /*******************************
 		 *	      MESSAGES		*
 		 *******************************/
@@ -354,6 +375,11 @@ prolog:message(rdf(not_a_name(Name))) -->
 	[ 'RDF: argument to rdf:ID is not an XML name: ~p'-[Name] ].
 prolog:message(rdf(redefined_id(Id))) -->
 	[ 'RDF: rdf:ID ~p: multiple definitions'-[Id] ].
+prolog:message(rdf(unexpected(Tag, Parser))) -->
+	{ get_sgml_parser(Parser, file(File)),
+	  get_sgml_parser(Parser, line(Line))
+	},
+	[ 'RDF: ~w:~d: Unexpected element ~w'-[File, Line, Tag] ].
 
 
 		 /*******************************
@@ -379,7 +405,7 @@ unparse_xml(element(Name, Attr, Content)) -->
 	;   ">",
 	    unparse_xml(Content)
 	).
-	
+
 attributes([]) -->
 	[].
 attributes([H|T]) -->
