@@ -1,4 +1,4 @@
-/*  $Id: rdf_parser.pl,v 1.43 2007/07/09 13:44:50 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -142,7 +142,7 @@ description(description, About, BagID, Properties, Options) ::=
 	{ !, append(PropAttrs, PropElts, Properties)
 	}.
 description(Type, About, BagID, Properties, Options) ::=
-	element(Type,
+	element(\uri(Type, Options),
 		\attrs([ \?idAboutAttr(About, Options),
 			 \?bagIdAttr(BagID, Options)
 		       | \propAttrs(PropAttrs, Options)
@@ -150,7 +150,7 @@ description(Type, About, BagID, Properties, Options) ::=
 		\propertyElts(PropElts, Options)),
 	{ append(PropAttrs, PropElts, Properties)
 	}.
-		
+
 propAttrs([], _) ::=
 	[], !.
 propAttrs([H|T], Options) ::=
@@ -195,6 +195,8 @@ propertyElt(Id, Name, Value, Options0) ::=
 	{ modify_state(E0, Options0, E, Options), !,
 	  rewrite(\propertyElt(Id, Name, Value, Options), E)
 	}.
+propertyElt(Id, Name, Value, Options) ::=
+	\literalPropertyElt(Id, Name, Value, Options), !.
 					% 5.14 emptyPropertyElt
 propertyElt(Id, Name, Value, Options) ::=
 	element(Name, A, \all_ws),
@@ -223,14 +225,6 @@ propertyElt(Id, Name, collection(Elements), Options) ::=
 			 \?idAttr(Id, Options)
 		       ]),
 		\nodeElementList(Elements, Options)).
-propertyElt(Id, Name, Literal, Options) ::=
-	element(Name,
-		\attrs([ \typeAttr(Type, Options),
-			 \?idAttr(Id, Options)
-		       ]),
-		Content),
-	{ typed_literal(Type, Content, Literal, Options)
-	}.
 propertyElt(Id, Name, Literal, Options) ::=
 	element(Name,
 		\attrs([ \?idAttr(Id, Options)
@@ -250,6 +244,15 @@ propertyElt(Id, Name, unparsed(Value), Options) ::=
 		       ]),
 		Value).
 
+literalPropertyElt(Id, Name, Literal, Options) ::=
+	element(Name,
+		\attrs([ \typeAttr(Type, Options),
+			 \?idAttr(Id, Options)
+		       ]),
+		Content),
+	{ typed_literal(Type, Content, Literal, Options)
+	}.
+
 emptyPropertyElt(Id, Literal, Options) ::=
 	\attrs([ \?idAttr(Id, Options),
 		 \?parseLiteral
@@ -274,10 +277,10 @@ aboutResourceEmptyElt(node(URI), _Options) ::=
 	\nodeIDAttr(URI).
 
 %%	literal_value(+In, -Value, +Options)
-%	
+%
 %	Create the literal value for rdf:parseType="Literal" attributes.
 %	The content is the Prolog XML DOM tree for the literal.
-%	
+%
 %	@tbd	Note that the specs demand a canonical textual representation
 %		of the XML data as a Unicode string.  For now the user can
 %		achieve this using the convert_typed_literal hook.
@@ -285,7 +288,7 @@ aboutResourceEmptyElt(node(URI), _Options) ::=
 literal_value(Value, literal(type(rdf:'XMLLiteral', Value)), _).
 
 %%	mkliteral(+Atom, -Object, +Options)
-%	
+%
 %	Translate attribute value Atom into an RDF object using the
 %	lang(Lang) option from Options.
 
@@ -298,7 +301,7 @@ mkliteral(Text, literal(Val), Options) :-
 	).
 
 %%	typed_literal(+Type, +Content, -Literal, +Options)
-%	
+%
 %	Handle a literal attribute with rdf:datatype=Type qualifier. NB:
 %	possibly  it  is  faster  to  use  a  global  variable  for  the
 %	conversion hook.
@@ -312,12 +315,10 @@ typed_literal(Type, Content, literal(Object), Options) :-
 	    )
 	;   Object = error(cannot_convert(Type, Content), _)
 	).
+typed_literal(Type, [], literal(type(Type, '')), _Options) :- !.
 typed_literal(Type, [Text], literal(type(Type, Text)), _Options) :- !.
 typed_literal(Type, Content, literal(type(Type, Content)), _Options).
-	
 
-idTermAttr(id(Id), Options) ::=
-	\idAttr(Id, Options).
 
 idAboutAttr(id(Id), Options) ::=
 	\idAttr(Id, Options), !.
@@ -328,11 +329,6 @@ idAboutAttr(node(About), _Options) ::=
 idAboutAttr(AboutEach, Options) ::=
 	\aboutEachAttr(AboutEach, Options).
 
-idRefAttr(Id, Options) ::=
-	\idAttr(Id, Options), !.
-idRefAttr(about(URI), Options) ::=
-	\resourceAttr(URI, Options).
-
 %%	an_rdf_object(-Object, +OptionsURI)
 %
 %	Deals with an object, but there may be spaces around.  I'm still
@@ -402,15 +398,21 @@ typeAttr(Type, Options) ::=
 	\rdf_or_unqualified(datatype) = \uri(Type, Options).
 
 uri(URI, Options) ::=
-	A,
-	{   memberchk(base_uri(Base), Options),
-	    Base \== []
-	->  canonical_uri(A, Base, URI)
-	;   sub_atom(A, 0, _, _, #)
-	->  sub_atom(A, 1, _, 0, URI)
-	;   decode_uri(A, URI)
+	T,
+	{   make_atom(T, A),
+	    (	memberchk(base_uri(Base), Options),
+		Base \== []
+	    ->  canonical_uri(A, Base, URI)
+	    ;   sub_atom(A, 0, _, _, #)
+	    ->  sub_atom(A, 1, _, 0, URI)
+	    ;   url_iri(A, URI)
+	    )
 	}.
 
+make_atom(NS:L, A) :- !,
+	atom_concat(NS, L, A).
+make_atom(A, A).
+
 globalid(Id, Options) ::=
 	A,
 	{   make_globalid(A, Options, Id)
@@ -432,53 +434,27 @@ make_globalid(In, Options, Id) :-
 	(   memberchk(base_uri(Base), Options),
 	    Base \== []
 	->  (   is_absolute_url(In)
-	    ->	decode_uri(In, Id)
-	    ;	concat_atom([Base, In], #, Id0),
-		decode_uri(Id0, Id)
+	    ->	url_iri(In, Id)
+	    ;	atomic_list_concat([Base, In], #, Id0),
+		url_iri(Id0, Id)
 	    )
 	;   sub_atom(In, 0, _, _, #)
 	->  sub_atom(In, 1, _, 0, Id)
-	;   decode_uri(In, Id)
+	;   url_iri(In, Id)
 	).
 
 
 %%	canonical_uri(+In, +Base, -Absolute)
-%	
+%
 %	Make the URI absolute and decode special sequences. For the last
 %	clause, which is the correct order?
 
 canonical_uri('', Base, Base) :- !.	% '' expands to xml:base
 canonical_uri(URI0, [], URI) :- !,	% do not use one
-	decode_uri(URI0, URI).
+	url_iri(URI0, URI).
 canonical_uri(URI, Base, Global) :-	% use our generic library
 	global_url(URI, Base, Global0),
-	decode_uri(Global0, Global).
-
-
-%%	decode_uri(+Encoded, -Decoded)
-%	
-%	RDF URIs are encoded  Unicode  strings.   First  the  unicode is
-%	translated using UTF-8 and then  the   result  is represented as
-%	US-ASCII using %XX sequences. Here we have to do the reverse.
-%	
-%	NOTE: this encoding is known under the name IRI.
-
-decode_uri(Encoded, Decoded) :-
-	sub_atom(Encoded, _, _, _, '%'), !,
-	atom_codes(Encoded, Codes),
-	unescape_precent(Codes, UTF8),
-	phrase(utf8_codes(Unicodes), UTF8),
-	atom_codes(Decoded, Unicodes).
-decode_uri(Encoded, Encoded).
-
-unescape_precent([], []).
-unescape_precent([0'%,C1,C2|T0], [H|T]) :- !,
-	code_type(C1, xdigit(D1)),
-	code_type(C2, xdigit(D2)),
-	H is D1*16 + D2,
-	unescape_precent(T0, T).
-unescape_precent([H|T0], [H|T]) :-
-	unescape_precent(T0, T).
+	url_iri(Global0, Global).
 
 
 		 /*******************************
@@ -625,7 +601,7 @@ do_attrs(C, L) :-
 	rewrite(C, L).
 
 %	\noMoreAttrs
-%	
+%
 %	Check attribute-list is empty.  Reserved xml: attributes are
 %	excluded from this test.
 
@@ -637,13 +613,13 @@ noMoreAttrs ::=
 	].
 
 %%	modify_state(+Element0, +Options0, -Element, -Options)
-%	
+%
 %	If Element0 contains xml:base = Base, strip it from the
 %	attributes list and update base_uri(_) in the Options
-%	
+%
 %	It Element0 contains xml:lang = Lang, strip it from the
 %	attributes list and update lang(_) in the Options
-%	
+%
 %	Remove all xmlns=_, xmlns:_=_ and xml:_=_.  Only succeed
 %	if something changed.
 
@@ -699,14 +675,14 @@ xml_attr(xml:_).
 
 
 %%	remove_fragment(+URI, -WithoutFragment)
-%	
+%
 %	When handling xml:base, we must delete the possible fragment.
 
 remove_fragment(URI, Plain) :-
 	sub_atom(URI, B, _, _, #), !,
 	sub_atom(URI, 0, B, _, Plain).
 remove_fragment(URI, URI).
-	
+
 
 		 /*******************************
 		 *     HELP PCE-EMACS A BIT	*
@@ -745,7 +721,14 @@ emacs_prolog_colours:goal_classification(\_, expanded).
 :- dynamic
 	prolog:meta_goal/2.
 :- multifile
-	prolog:meta_goal/2.
+	prolog:meta_goal/2,
+	prolog:called_by/2.
 
 prolog:meta_goal(rewrite(A, _), [A]).
 prolog:meta_goal(\A,		[A+1]).
+
+prolog:called_by(attrs(Attrs, _Term), Called) :-
+	findall(G+1, sub_term(\?G, Attrs), Called, Tail),
+	findall(G+1, sub_term(\G, Attrs), Tail).
+
+
