@@ -1,11 +1,11 @@
-/*  $Id: sgml.pl,v 1.30 2004/09/21 14:22:24 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2005, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -37,6 +37,7 @@
 	    load_structure/3,		% +File, -Term, +Options
 
 	    load_dtd/2,			% +DTD, +File
+	    load_dtd/3,			% +DTD, +File, +Options
 	    dtd/2,			% +Type, -DTD
 	    dtd_property/2,		% +DTD, ?Property
 
@@ -52,10 +53,15 @@
 
 	    sgml_register_catalog_file/2, % +File, +StartOrEnd
 
+	    xml_quote_attribute/3,	% +In, -Quoted, +Encoding
+	    xml_quote_cdata/3,		% +In, -Quoted, +Encoding
 	    xml_quote_attribute/2,	% +In, -Quoted
 	    xml_quote_cdata/2,		% +In, -Quoted
-	    xml_name/1			% +In
+	    xml_name/1,			% +In
+	    xml_is_dom/1		% +Term
 	  ]).
+:- use_module(library(lists)).
+:- use_module(library(option)).
 
 :- multifile user:file_search_path/2.
 :- dynamic   user:file_search_path/2.
@@ -87,7 +93,7 @@ init :-
 
 :- initialization
 	init.
-	
+
 
 		 /*******************************
 		 *	   DTD HANDLING		*
@@ -104,6 +110,10 @@ diagnosed to mess with the entity resolution by Fabien Todescato.
 	current_dtd/2.
 :- volatile
 	current_dtd/2.
+:- thread_local
+	registered_cleanup/0.
+:- volatile
+	registered_cleanup/0.
 
 :- multifile
 	dtd_alias/2.
@@ -123,17 +133,41 @@ dtd(Type, DTD) :-
 			     access(read)
 			   ], DtdFile),
 	load_dtd(DTD, DtdFile),
+	register_cleanup,
 	asserta(current_dtd(Type, DTD)).
 
+%%	load_dtd(+DTD, +DtdFile, +Options)
+%
+%	Load file into a DTD.  Defined options are:
+%
+%		* dialect(+Dialect)
+%		Dialect to use (xml, xmlns, sgml)
+%
+%		* encoding(+Encoding)
+%		Encoding of DTD file
+
 load_dtd(DTD, DtdFile) :-
-	open_dtd(DTD, [], DtdOut),
-	open(DtdFile, read, DtdIn),
+	load_dtd(DTD, DtdFile, []).
+load_dtd(DTD, DtdFile, Options) :-
+	split_dtd_options(Options, DTDOptions, FileOptions),
+	open_dtd(DTD, DTDOptions, DtdOut),
+	open(DtdFile, read, DtdIn, FileOptions),
 	copy_stream_data(DtdIn, DtdOut),
 	close(DtdIn),
 	close(DtdOut).
 
-%	destroy_dtds
-%	
+split_dtd_options([], [], []).
+split_dtd_options([H|T], [H|TD], S) :-
+	dtd_option(H), !,
+	split_dtd_options(T, TD, S).
+split_dtd_options([H|T], TD, [H|S]) :-
+	split_dtd_options(T, TD, S).
+
+dtd_option(dialect(_)).
+
+
+%%	destroy_dtds
+%
 %	Destroy  DTDs  cached  by  this  thread   as  they  will  become
 %	unreachable anyway.
 
@@ -144,7 +178,15 @@ destroy_dtds :-
 	;   true
 	).
 
-:- thread_at_exit(destroy_dtds).
+%%	register_cleanup
+%
+%	Register cleanup of DTDs created for this thread.
+
+register_cleanup :-
+	registered_cleanup, !.
+register_cleanup :-
+	catch(thread_at_exit(destroy_dtds), _, true),
+	assert(registered_cleanup).
 
 
 		 /*******************************
@@ -232,7 +274,8 @@ load_structure(stream(In), Term, Options) :- !,
 	new_sgml_parser(Parser,
 			[ dtd(DTD)
 			]),
-	call_cleanup(parse(Parser, Options2, TermRead, In),
+	def_entities(Options2, DTD, Options3),
+	call_cleanup(parse(Parser, Options3, TermRead, In),
 		     free_sgml_parser(Parser)),
 	(   ExplicitDTD == true
 	->  (   DTD = dtd(_, DocType),
@@ -244,7 +287,7 @@ load_structure(stream(In), Term, Options) :- !,
 	),
 	Term = TermRead.
 load_structure(Stream, Term, Options) :-
-	Stream = '$stream'(_), !,
+	is_stream(Stream), !,
 	load_structure(stream(Stream), Term, Options).
 load_structure(File, Term, Options) :-
 	open(File, read, In, [type(binary)]),
@@ -259,6 +302,19 @@ parse(Parser, Options, Document, In) :-
 		   | Options1
 		   ]).
 
+def_entities([], _, []).
+def_entities([entity(Name, Value)|T], DTD, Opts) :- !,
+	def_entity(DTD, Name, Value),
+	def_entities(T, DTD, Opts).
+def_entities([H|T0], DTD, [H|T]) :-
+	def_entities(T0, DTD, T).
+
+def_entity(DTD, Name, Value) :-
+	open_dtd(DTD, [], Stream),
+	xml_quote_attribute(Value, QValue),
+	format(Stream, '<!ENTITY ~w "~w">~n', [Name, QValue]),
+	close(Stream).
+
 
 		 /*******************************
 		 *	     UTILITIES		*
@@ -280,24 +336,66 @@ load_html_file(File, Term) :-
 
 
 		 /*******************************
-		 *	      UTIL		*
+		 *	      ENCODING		*
 		 *******************************/
 
-%	option(Option(?Value), OptionList, Default)
+%	xml_quote_attribute(+In, -Quoted)
+%	xml_quote_cdata(+In, -Quoted)
+%
+%	Backward  compatibility  for  versions  that  allow  to  specify
+%	encoding. All characters that cannot fit the encoding are mapped
+%	to XML character entities (&#dd;).  Using   ASCII  is the safest
+%	value.
 
-option(Opt, Options) :-
-	memberchk(Opt, Options), !.
-option(Opt, Options) :-
-	functor(Opt, OptName, 1),
-	arg(1, Opt, OptVal),
-	memberchk(OptName=OptVal, Options), !.
+xml_quote_attribute(In, Quoted) :-
+	xml_quote_attribute(In, Quoted, ascii).
+
+xml_quote_cdata(In, Quoted) :-
+	xml_quote_cdata(In, Quoted, ascii).
+
+xml_name(In) :-
+	xml_name(In, ascii).
+
+
+		 /*******************************
+		 *	   TYPE CHECKING	*
+		 *******************************/
+
+%	xml_is_dome(@Term)
+%
+%	True  if  term  statisfies   the    structure   as  returned  by
+%	load_structure/3 and friends.
+
+xml_is_dom(0) :- !, fail.		% catch variables
+xml_is_dom([]) :- !.
+xml_is_dom([H|T]) :- !,
+	xml_is_dom(H),
+	xml_is_dom(T).
+xml_is_dom(element(Name, Attributes, Content)) :- !,
+	dom_name(Name),
+	dom_attributes(Attributes),
+	xml_is_dom(Content).
+xml_is_dom(pi(Pi)) :- !,
+	atom(Pi).
+xml_is_dom(CDATA) :-
+	atom(CDATA).
+
+dom_name(NS:Local) :-
+	atom(NS),
+	atom(Local), !.
+dom_name(Local) :-
+	atom(Local).
+
+dom_attributes(0) :- !, fail.
+dom_attributes([]).
+dom_attributes([H|T]) :-
+	dom_attribute(H),
+	dom_attributes(T).
+
+dom_attribute(Name=Value) :-
+	dom_name(Name),
+	atomic(Value).
 
-select_option(Opt, Options, Rest) :-
-	select(Opt, Options, Rest), !.
-select_option(Opt, Options, Rest) :-
-	functor(Opt, OptName, 1),
-	arg(1, Opt, OptVal),
-	select(OptName=OptVal, Options, Rest), !.
 
 		 /*******************************
 		 *	      MESSAGES		*
@@ -311,3 +409,19 @@ prolog:message(sgml(Parser, File, Line, Message)) -->
 	{ get_sgml_parser(Parser, dialect(Dialect))
 	},
 	[ 'SGML2PL(~w): ~w:~w: ~w'-[Dialect, File, Line, Message] ].
+
+
+		 /*******************************
+		 *	   XREF SUPPORT		*
+		 *******************************/
+
+:- multifile
+	prolog:called_by/2.
+
+prolog:called_by(sgml_parse(_, Options), Called) :-
+	is_list(Options),
+	findall(G+3,
+		(   member(call(_, G), Options),
+		    callable(G)
+		),
+		Called).
