@@ -1,4 +1,4 @@
-/*  $Id: pce_compile.pl,v 1.10 1995/06/09 08:12:44 jan Exp $
+/*  $Id$
 
     Part of XPCE
     Designed and implemented by Anjo Anjewierden and Jan Wielemaker
@@ -44,7 +44,7 @@ pce_ifhostproperty(prolog(quintus),
 :- require([ between/3
 	   , call/3
 	   , concat/3
-	   , concat_atom/2
+	   , atomic_list_concat/2
 	   , forall/2
 	   , genarg/3
 	   ]).
@@ -84,7 +84,7 @@ pce_begin_class(TermDef, Super, Doc) :-
 	strip_module(TermDef, Module, TD),
 	asserta(pce_compile:load_module(Module)),
 	pce_begin_class_(TD, Super, Doc).
-	
+
 
 pce_begin_class_(TermDef, Super, Doc) :-
 	TermDef =.. [ClassName|TermArgs],
@@ -189,7 +189,7 @@ pce_end_class :-
 	).
 
 %	push_compile_operators.
-%	Push the current 
+%	Push the current
 
 push_compile_operators :-
 	push_operators(
@@ -204,7 +204,7 @@ push_compile_operators :-
 
 pop_compile_operators :-
 	pop_operators.
-	
+
 
 %	current_class(-Class)
 %	Class for which we are currently compiling.
@@ -263,7 +263,7 @@ pce_term_expansion(Term, _) :-
 	),
 	pce_error(context_error(Term, nomethod, clause)),
 	fail.
-	
+
 is_string([]).
 is_string([H|T]) :-
 	between(0, 255, H),
@@ -377,7 +377,7 @@ pce_ifhostproperty(need_extern_declaration,
 	feedback(expand_send(ClassName, Selector)))).
 
 
-pce_ifhostproperty(need_extern_declaration, 
+pce_ifhostproperty(need_extern_declaration,
 (do_expand((Head :<- DocBody),			% Prolog get
 	[ (PlHead :- Body)
 	, pce_compile:lazy_method(Selector, ClassName, get,
@@ -425,7 +425,7 @@ pce_ifhostproperty(need_extern_declaration, [
 	add_plus_term(NN, M, T))]).
 
 
-pce_ifhostproperty(string, 
+pce_ifhostproperty(string,
 (extract_documentation((DocText::Body), Str, Body) :- !,
 	(   string(DocText)
 	->  Str = DocText
@@ -465,7 +465,7 @@ create_type_vector(List,    new(VectorTerm)) :-
 predicate_name(SendGet, Selector, Name) :-
 	current_class(Class),
 	class_name(Class, ClassName),
-	concat_atom([SendGet, '_', Selector, '_', ClassName], Name).
+	atomic_list_concat([SendGet, '_', Selector, '_', ClassName], Name).
 
 
 pl_head_args(send, [], [], []) :- !.
@@ -633,7 +633,7 @@ pce_get_method(Class, Selector, RType, Head, Types, Doc, Loc, Group) :-
        message(@prolog, call, bind_lazy, @arg1, @arg2, @arg3)).
 
 pce_ifhostproperty(prolog(swi),		% Hide from the tracer
-		   (:- '$hide'(bind_lazy, 3))).
+		   (:- '$hide'(bind_lazy/3))).
 
 bind_lazy(Kind, ClassName, @default) :- !,
 	retract(lazy_method(Selector, ClassName, Kind, Arguments)),
@@ -649,7 +649,7 @@ lazy_send_method(Head, Types, Doc, Loc, Group,
 	get(@pce, convert, ClassName, class, Class),
 	send(Class, send_method,
 	     send_method(Selector, Types, Message, Doc, Loc, Group)).
-	
+
 lazy_get_method(RType, Head, Types, Doc, Loc, Group,
 		ClassName, Selector) :-
 	pce_get_method_message(Head, Message),
@@ -667,7 +667,7 @@ lazy_get_method(RType, Head, Types, Doc, Loc, Group,
 %
 %	Get the default value for an argument.
 
-default(@default, resource(Obj, Name), Value) :- !, 
+default(@default, resource(Obj, Name), Value) :- !,
 	(   get(Obj, resource_value, Name, Value)
 	->  true
 	;   pce_error(get_resource_failed(Name, Obj)),
