@@ -1,4 +1,4 @@
-/*  $Id: application.pl,v 1.20 2006/10/25 11:34:16 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -63,7 +63,7 @@ initialise(Emacs, Buffers:dict) :->
 	new(@emacs_mark_list, emacs_bookmark_editor),
 	ignore(send(Emacs, server_start)),
 	ignore(send(Emacs, load_user_init_file)).
-	
+
 unlink(Emacs) :->
 	(   get(Emacs, exit_message, Msg),
 	    send(@pce?exit_messages, delete, Msg)
@@ -86,7 +86,7 @@ show_buffer_menu(Emacs) :->
 	->  send(Menu, expose)
 	;   send(emacs_buffer_menu(Emacs), open)
 	).
-	    
+
 
 selection(Emacs, B:emacs_buffer*) :->
 	"Select emacs buffer"::
@@ -163,8 +163,8 @@ open_object(_Emacs, Object:prolog, _NewWindow:new_window=[bool]) :->
 	edit(Object).
 
 show_bookmarks(_) :->
-	"Show PceEmacs bookmarks"::
-	send(@emacs_mark_list, open).
+	"Show PceEmacs bookmarks window"::
+	send(@emacs_mark_list, expose).
 
 
 		 /*******************************
@@ -187,7 +187,7 @@ save_some_buffers(BM, Confirm:[bool]) :->
 	;   send(@pce, report, status, 'No buffers need saving')
 	).
 
-	
+
 check_saved_at_exit(BM) :->
 	"Check for unsaved buffers when called from exit"::
 	send(BM, save_some_buffers, @on),
@@ -253,9 +253,6 @@ modes(_Emacs, ModeNames:chain) :<-
 
 :- pce_group(help).
 
-:- dynamic
-	help_file/1.
-
 :- pce_help_file(emacs, pce_help('emacs.hlp')).
 :- pce_help_file(emacs_customise, pce_help('customise.hlp')).
 
@@ -281,7 +278,8 @@ server_start(Emacs, Force:[bool]) :->
 	    ;	get(@emacs_server, status, listen)
 	    )
 	->  (   current_prolog_flag(windows, true)
-	    ->	catch(start_emacs_dde_server(false), _, true)
+	    ->	Goal = start_emacs_dde_server(false), % fool xref
+	        catch(Goal, _, true)
 	    ;	true
 	    )
 	;   (	send(@emacs_server_address, exists, @off)
@@ -298,7 +296,7 @@ server_start(Emacs, Force:[bool]) :->
 	    ),
 	    ignore(send(@emacs_server, listen))
 	).
-	
+
 
 :- pce_group(customise).
 
@@ -336,4 +334,4 @@ load_user_init_file(_Emacs) :->
 	).
 
 :- pce_end_class(emacs).
-	  
+
