@@ -1,4 +1,4 @@
-/*  $Id: language_mode.pl,v 1.32 2007/07/04 08:22:03 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -32,6 +32,7 @@
 :- module(emacs_language_mode, []).
 :- use_module(library(pce)).
 :- use_module(library(hyper)).
+:- use_module(library(socket), [gethostname/1]).
 :- require([ auto_call/1
 	   , chain_list/2
 	   , default/3
@@ -135,7 +136,7 @@ comment_lines(TB, S0, End, Comment) :-
 	->  comment_lines(TB, S2, End1, Comment)
 	;   true
 	).
-	
+
 
 fill_comment_paragraph(M, Justify:justify=[bool|int], From:[int]) :->
 	"Fill paragraph in (line) comment"::
@@ -191,7 +192,7 @@ fill_comment(M,
 	send(M, caret, CaretF?start),
 	free(EndF),
 	free(CaretF).
-	
+
 uncomment(_M, _Re, _LeadCol, Here, EndF) :-
 	get(EndF, start, End),
 	Here >= End, !.
@@ -230,7 +231,7 @@ insert_comment_block(E) :->
 
 insert_section_header(E) :->
 	"Insert Prolog/C section header"::
-	send(E, insert, 
+	send(E, insert,
 '		 /*******************************
 		 *               C		*
 		 *******************************/
@@ -299,6 +300,13 @@ file_header_parameter('%U', _, UserName) :-
 	->  true
 	;   UserName = User
 	).
+file_header_parameter('%E', _, Email) :-
+	(   getenv('EMAIL', Email)
+	->  true
+	;   get(@pce, user, User),
+	    gethostname(Host),
+	    atomic_list_concat([User, Host], @, Email)
+	).
 file_header_parameter('%D', _, Date) :-
 	new(D, date),
 	get(D, year, Year),
@@ -414,7 +422,7 @@ indent_expression_line(E, Brackets:[name], Base:[int]) :->
 	    ),
 	    send(E, align_line, Col), !.
 
-		
+
 %	->indent_comment_line
 %
 
