@@ -3,9 +3,9 @@
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2008, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -34,9 +34,11 @@
 :- use_module(library(pce)).
 :- use_module(library(emacs_extend)).
 :- use_module(library(pce_prolog_xref)).
+:- use_module(library(prolog_source)).
 :- use_module(library(lists)).
 :- use_module(library(operators)).
 :- use_module(library(debug)).
+:- use_module(library(edit)).
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 User extension hooks.
@@ -201,7 +203,7 @@ restore_settings(state(Fd, Style, Esc)) :-
 	close(Fd).
 
 %	emacs_push_op(+Prec, +Type, :Name)
-%	
+%
 %	Define operators into the default source module and register
 %	them to be undone by pop_operators/0.
 
@@ -215,7 +217,7 @@ emacs_push_op(P, T, N0) :-
 	debug(emacs, ':- ~w.', [op(P,T,N)]).
 
 %	syntax_error(+Error)
-%	
+%
 %	Print syntax errors if the debugging topic emacs is active.
 
 syntax_error(E) :-
@@ -226,7 +228,7 @@ syntax_error(E) :-
 	fail.
 
 %	fix_operators(+Term, +Src)
-%	
+%
 %	Fix flags that affect the  syntax,   such  as operators and some
 %	style checking options. Src is the  canonical source as required
 %	by the cross-referencer.
@@ -255,7 +257,7 @@ process_directive(Directive, _) :-
 	erase(Ref).
 
 %	process_use_module(+Imports, +Src)
-%	
+%
 %	Get the exported operators from the referenced files.
 
 process_use_module([], _).
@@ -284,7 +286,7 @@ colourise(TB, Fd) :-
 	->  colourise_term(Term, TB, TermPos)
 	;   show_syntax_error(TB, Error)
 	).
-	
+
 show_syntax_error(TB, Pos:_Message) :-
 	get(TB, scan, Pos, line, 0, start, BOL),
 	get(TB, scan, Pos, line, 0, end, EOL),
@@ -334,18 +336,29 @@ colourise_term(Fact, TB, Pos) :- !,
 	colour_item(clause, TB,	Pos),
 	colourise_clause_head(Fact, TB, Pos).
 
+%%	colourise_extended_head(+Head, +ExtraArgs, +TB, +Pos) is det.
+%
+%	Colourise a clause-head that  is   extended  by  term_expansion,
+%	getting ExtraArgs more  arguments  (e.g.,   DCGs  add  two  more
+%	arguments.
+
 colourise_extended_head(Head, N, TB, Pos) :-
-	functor_position(Pos, FPos, _),
+	extend(Head, N, TheHead),
+	colourise_clause_head(TheHead, TB, Pos).
+
+extend(M:Head, N, M:ExtHead) :-
+	nonvar(Head), !,
+	extend(Head, N, ExtHead).
+extend(Head, N, ExtHead) :-
+	callable(Head), !,
 	Head =.. List,
 	length(Extra, N),
 	append(List, Extra, List1),
-	TheHead =.. List1,
-	classify_head(TB, TheHead, Class),
-	colour_item(head(Class), TB, FPos),
-	colourise_term_args(Head, TB, Pos).
+	ExtHead =.. List1.
+extend(Head, _, Head).
+
 
 colourise_clause_head(Head, TB, Pos) :-
-	nonvar(Head),
 	head_colours(Head, ClassSpec-ArgSpecs), !,
 	functor_position(Pos, FPos, ArgPos),
 	(   ClassSpec == classify
@@ -361,7 +374,7 @@ colourise_clause_head(Head, TB, Pos) :-
 	colourise_term_args(Head, TB, Pos).
 
 %	colourise_extern_head(+Head, +Module, +TB, +Pos)
-%	
+%
 %	Colourise the head specified as Module:Head. Normally used for
 %	adding clauses to multifile predicates in other modules.
 
@@ -378,7 +391,7 @@ colour_method_head(SGHead, TB, Pos) :-
 	colourise_term_args(Head, TB, Pos).
 
 %	functor_position(+Term, -FunctorPos, -ArgPosList)
-%	
+%
 %	Get the position of a functor   and  its argument. Unfortunately
 %	this goes wrong for lists, who have two `functor-positions'.
 
@@ -389,7 +402,7 @@ functor_position(Pos, Pos, []).
 
 
 %	colourise_directive(+Body, +TB, +Pos)
-%	
+%
 %	Colourise the body of a directive.
 
 colourise_directive(Body, TB, Pos) :-
@@ -397,7 +410,7 @@ colourise_directive(Body, TB, Pos) :-
 
 
 %	colourise_body(+Body, +TB, +Pos)
-%	
+%
 %	Breaks down to colourise_goal/3.
 
 colourise_body(Body, TB, Pos) :-
@@ -408,7 +421,7 @@ colourise_body(Body, Origin, TB, Pos) :-
 	colourise_goals(Body, Origin, TB, Pos).
 
 %	colourise_method_body(+MethodBody, +TB, +Pos)
-%	
+%
 %	Colourise the optional "comment":: as pce(comment) and proceed
 %	with the body.
 
@@ -447,7 +460,7 @@ colourise_subgoals([Pos|T], N, Body, Origin, TB) :-
 	colourise_subgoals(T, NN, Body, Origin, TB).
 
 %	colourise_dcg(+Body, +Head, +TB, +Pos)
-%	
+%
 %	Breaks down to colourise_dcg_goal/3.
 
 colourise_dcg(Body, Head, TB, Pos) :-
@@ -503,7 +516,7 @@ colourise_dcg_goal(Goal, _, TB, Pos) :-
 
 
 %	colourise_goal(+Goal, +Origin, +TB, +Pos)
-%	
+%
 %	Colourise access to a single goal.
 
 					% Deal with list as goal (consult)
@@ -541,7 +554,7 @@ colourise_goal(Goal, Origin, TB, Pos) :-
 	colourise_goal_args(Goal, TB, Pos).
 
 %	colourise_goal_args(+Goal, +TB, +Pos)
-%	
+%
 %	Colourise the arguments to a goal. This predicate deals with
 %	meta- and database-access predicates.
 
@@ -563,13 +576,13 @@ colourise_meta_args(N, Goal, MetaArgs, TB, [P0|PT]) :-
 	colourise_meta_args(NN, Goal, MetaArgs, TB, PT).
 
 %	meta_args(+Goal, -ArgSpec)
-%	
+%
 %	Return a copy of Goal, where each meta-argument is an integer
 %	representing the number of extra arguments. The non-meta
 %	arguments are unbound variables.
-%	
+%
 %	E.g. meta_args(maplist(foo,x,y), X) --> X = maplist(2,_,_)
-%	
+%
 %	NOTE: this could be cached if performance becomes an issue.
 
 meta_args(Goal, VarGoal) :-
@@ -589,7 +602,7 @@ instantiate_meta([H|T]) :-
 	instantiate_meta(T).
 
 %	expand_meta(+MetaSpec, +Goal, -Expanded)
-%	
+%
 %	Add extra arguments to the goal if the meta-specifier is an
 %	integer (see above).
 
@@ -607,7 +620,7 @@ expand_meta(MetaSpec, Goal, Expanded) :-
 	Expanded =.. List.
 
 %	colourise_db(+Arg, +TB, +Pos)
-%	
+%
 %	Colourise database modification calls (assert/1, retract/1 and
 %	friends.
 
@@ -666,7 +679,7 @@ colourise_directory(Spec, TB, Pos) :-
 
 %	colourise_class(ClassName, TB, Pos)
 %
-%	Colourise an XPCE class.  
+%	Colourise an XPCE class.
 
 colourise_class(ClassName, TB, Pos) :-
 	classify_class(TB, ClassName, Classification),
@@ -694,7 +707,7 @@ colourise_term_arg(Var, TB, Pos) :-			% variable
 colourise_term_arg(Atom, TB, Pos) :-			% single quoted atom
 	atom(Atom),
 	arg(1, Pos, From),
-	get(TB, character, From, 39), !, 	
+	get(TB, character, From, 39), !,
 	colour_item(quoted_atom, TB, Pos).
 colourise_term_arg(List, TB, list_position(_, _, Elms, Tail)) :- !,
 	colourise_list_args(Elms, Tail, List, TB, classify).	% list
@@ -705,7 +718,7 @@ colourise_term_arg(_, TB, string_position(F, T)) :- !,	% string
 	colour_item(string, TB, F-T).
 colourise_term_arg(_Arg, _TB, _Pos) :-
 	true.
-	
+
 colourise_list_args([HP|TP], Tail, [H|T], TB, How) :-
 	specified_item(How, H, TB, HP),
 	colourise_list_args(TP, Tail, T, TB, How).
@@ -715,7 +728,7 @@ colourise_list_args([], TP, T, TB, How) :-
 
 
 %	colourise_exports(+List, +TB, +Pos)
-%	
+%
 %	Colourise the module export-list (or any other list holding
 %	terms of the form Name/Arity referring to predicates).
 
@@ -736,7 +749,7 @@ colourise_exports2(_, _, _).
 
 
 %	colourise_imports(+List, +File, +TB, +Pos)
-%	
+%
 %	Colourise import list from use_module/2, importing from File.
 
 colourise_imports(List, File, TB, Pos) :-
@@ -758,7 +771,7 @@ colourise_imports(except(Except), File, Public, TB,
 	colour_item(keyword(except), TB, FF-FT),
 	colourise_imports(Except, File, Public, TB, LP).
 colourise_imports(_, _, _, TB, Pos) :-
-	colour_item(type_error(list), TB, Pos).	
+	colour_item(type_error(list), TB, Pos).
 
 colourise_imports2([G0|GT], File, Public, TB, [P0|PT]) :- !,
 	colourise_import(G0, File, TB, P0),
@@ -779,7 +792,7 @@ colourise_import(PI, _, TB, Pos) :-
 
 
 %	colourise_declarations(+Term, +TB, +Pos)
-%	
+%
 %	Colourise the Predicate indicator lists of dynamic, multifile, etc
 %	declarations.
 
@@ -823,12 +836,12 @@ colour_item(Class, TB, Pos) :-
 	L is T - F,
 	make_fragment(Class, TB, F, L, Name).
 colour_item(_, _, _).
-	
+
 colour_item(Class, TB, F, T) :-
 	colour_item(Class, TB, F-T).
 
 %	make_fragment(+Class, +TB, +From, +Len, +StyleName)
-%	
+%
 %	Actually create the fragment.
 
 make_fragment(goal(Class, Goal), TB, F, L, Style) :-
@@ -877,7 +890,7 @@ body_compiled((_;_)).
 body_compiled(\+_).
 
 %	goal_classification(+TB, +Goal, +Origin, -Class)
-%	
+%
 %	Classify Goal appearing in TB and called from a clause with head
 %	Origin.  For directives Origin is [].
 
@@ -894,7 +907,7 @@ goal_classification(_TB, Goal, _, Class) :-
 goal_classification(_TB, _Goal, _, undefined).
 
 %	goal_classification(+Goal, -Class)
-%	
+%
 %	Multifile hookable classification for non-local goals.
 
 goal_classification(Goal, built_in) :-
@@ -908,7 +921,7 @@ goal_classification(SS, expanded) :-	% XPCE (TBD)
 	functor(SS, send_super, A),
 	A >= 2, !.
 goal_classification(SS, expanded) :-	% XPCE (TBD)
-	functor(SS, get_super, A), 
+	functor(SS, get_super, A),
 	A >= 3, !.
 
 classify_head(TB, Goal, exported) :-
@@ -945,6 +958,7 @@ goal_colours(reexport(_),	     built_in-[file]).
 goal_colours(reexport(File,_),       built_in-[file,imports(File)]).
 goal_colours(dynamic(_),	     built_in-[predicates]).
 goal_colours(thread_local(_),	     built_in-[predicates]).
+goal_colours(module_transparent(_),  built_in-[predicates]).
 goal_colours(multifile(_),	     built_in-[predicates]).
 goal_colours(volatile(_),	     built_in-[predicates]).
 goal_colours(consult(_),	     built_in-[file]).
@@ -991,6 +1005,9 @@ head_colours(file_search_path(_,_), hook-[identifier,classify]).
 head_colours(library_directory(_),  hook-[file]).
 head_colours(resource(_,_,_),	    hook-[identifier,classify,file]).
 
+head_colours(Var, _) :-
+	var(Var), !,
+	fail.
 head_colours(M:H, Colours) :-
 	atom(M), callable(H),
 	xref_hook(M:H), !,
@@ -1000,7 +1017,7 @@ head_colours(M:H, Colours) :-
 	head_colours(H, HC),
 	HC = hook - _, !,
 	Colours = hook - [ hook, HC ].
-head_colours(M:_,		    meta-[module(M),extern(M)]).
+head_colours(M:_, meta-[module(M),extern(M)]).
 
 
 		 /*******************************
@@ -1386,7 +1403,7 @@ popup(_GF, Popup:popup) :<-
 	      make_prolog_mode_goal_popup).
 
 %	make_prolog_mode_goal_popup(-Popup)
-%	
+%
 %	Create the popup and define actions for handling the right-menu
 %	on goals.
 
@@ -1461,7 +1478,7 @@ has_source(F) :->
 	), !.
 
 %	->edit
-%	
+%
 %	Find the predicate and invoke ->find_definition on the
 %	@emacs_mode, which is the mode object of the current editor.
 
@@ -1472,7 +1489,7 @@ edit(F, NewWindow:[bool]) :->
 
 
 %	->listing
-%	
+%
 %	List the predicate in an XPCE buffer
 
 listing(F) :->
@@ -1524,17 +1541,26 @@ identify(F) :->
 	send(TB, report, status, Report).
 
 %	identify_pred(+XrefClass, +Fragment, -Summary)
-%	
+%
 %	Generate an identifying description for the predicate.
 
 identify_pred(Term, _, Summary) :-
 	identify(Term, String),
 	new(Summary, string(String)).
 identify_pred(Class, F, Summary) :-		% SWI-Prolog documented built-in
-	get(F, predicate, Pred),	% & PlDoc summaries
+	get(F, predicate, Pred),		% & PlDoc summaries
 	get(Pred, summary, Summary0), !,
 	functor(Class, ClassName, _),
-	new(Summary, string('%N: [%s] %s', Pred, ClassName, Summary0)).
+	(   (   Class == autoload,
+		autoload_source(F, From)
+	    ->	true
+	    ;	Class = imported(From)
+	    ),
+	    file_name_on_path(From, Alias)
+	->  term_to_atom(Alias, Atom),
+	    new(Summary, string('%N: [%s from %s] %s', Pred, ClassName, Atom, Summary0))
+	;   new(Summary, string('%N: [%s] %s', Pred, ClassName, Summary0))
+	).
 identify_pred(built_in, F, Summary) :-
 	get(F, head, Head),
 	predicate_property(system:Head, foreign), !,
@@ -1544,16 +1570,10 @@ identify_pred(built_in, F, Summary) :-
 	sub_atom(Name, 0, _, _, $), !,
 	new(Summary, string('%N: SWI-Prolog private built-in', F)).
 identify_pred(autoload, F, Summary) :-	% Autoloaded predicates
-	get(F, name, Name),
-	get(F, arity, Arity),
-	'$find_library'(_Module, Name, Arity, _LoadModule, Library), !,
-	absolute_file_name(Library,
-			   [ file_type(prolog),
-			     access(read),
-			     file_errors(fail)
-			   ],
-			   Source),
-	new(Summary, string('%N: autoload from %s', F, Source)).
+	autoload_source(F, Source),
+	file_name_on_path(Source, Alias),
+	term_to_atom(Alias, Atom),
+	new(Summary, string('%N: autoload from %s', F, Atom)).
 identify_pred(local(Line), F, Summary) :-	% Local predicates
 	new(Summary, string('%N: locally defined at line %d', F, Line)).
 identify_pred(foreign(Line), F, Summary) :-	% Foreign predicates
@@ -1561,7 +1581,9 @@ identify_pred(foreign(Line), F, Summary) :-	% Foreign predicates
 identify_pred(constraint(Line), F, Summary) :-	% Local constraint
 	new(Summary, string('%N: constraint defined at line %d', F, Line)).
 identify_pred(imported(From), F, Summary) :-
-	new(Summary, string('%N: imported from %s', F, From)).
+	file_name_on_path(From, Alias),
+	term_to_atom(Alias, Atom),
+	new(Summary, string('%N: imported from %s', F, Atom)).
 identify_pred(recursion, _, 'Recursive reference') :- !.
 identify_pred(dynamic(_Line), F, Summary) :-
 	get(F, loaded_specifier, Spec),
@@ -1575,6 +1597,17 @@ identify_pred(global, _, 'Predicate defined in global module user') :- !.
 identify_pred(Class, _, ClassName) :-
 	term_to_atom(Class, ClassName).
 
+autoload_source(F, Source) :-
+	get(F, name, Name),
+	get(F, arity, Arity),
+	'$find_library'(_Module, Name, Arity, _LoadModule, Library), !,
+	absolute_file_name(Library,
+			   [ file_type(prolog),
+			     access(read),
+			     file_errors(fail)
+			   ],
+			   Source).
+
 :- pce_end_class(emacs_goal_fragment).
 
 
@@ -1731,7 +1764,7 @@ classify_class(_, Name, user(File)) :-
 classify_class(_, Name, user) :-
 	get(@classes, member, Name, _), !.
 classify_class(_, _, undefined).
-	
+
 :- pce_end_class(emacs_class_fragment).
 
 
@@ -1773,7 +1806,7 @@ file(F, File:name) :<-
 	(   get(F, classification, file)
 	->  File = Context
 	;   get(F, classification, module)
-	->  current_module(Context, File)
+	->  module_property(Context, file(File))
 	).
 
 
@@ -1798,7 +1831,7 @@ identify_fragment(directory(Path), _, Summary) :-
 identify_fragment(type_error(Type), _, Summary) :-
 	new(Summary, string('Type error: argument must be a %s', Type)).
 identify_fragment(module(Module), _, Summary) :-
-	current_module(Module, Path),
+	module_property(Module, file(Path)),
 	new(Summary, string('Module %s loaded from %s', Module, Path)).
 identify_fragment(method(send), _, 'XPCE send method').
 identify_fragment(method(get), _, 'XPCE get method').
@@ -1813,4 +1846,4 @@ identify_fragment(Class, _, Summary) :-
 	term_to_atom(Class, Summary).
 
 :- pce_end_class(emacs_prolog_fragment).
-		   
+
