@@ -1,11 +1,11 @@
-/*  $Id: trace.pl,v 1.24 2004/09/21 14:22:25 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2006, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -30,18 +30,21 @@
 */
 
 :- module(pce_prolog_tracer,
-	  [ prolog_show_frame/2,	% +Frame, +Options
-	    clause_info/4,		% +Ref, -File, -TermPos, -VarOffset
-	    clear_clause_info_cache/0
+	  [ prolog_show_frame/2		% +Frame, +Options
 	  ]).
+:- use_module(library(prolog_clause)).
+:- use_module(library(lists)).
+:- use_module(library(debug)).
 :- consult([ clause,
 	     util,
 	     source,
 	     break,
-	     gui
+	     gui,
+	     stack
 	   ]).
 
-:- visible(+cut_call).
+:- initialization
+   visible(+cut_call).
 
 
 		 /*******************************
@@ -50,15 +53,27 @@
 
 :- thread_local
 	last_action/1,
-	show_unify_as/2,
-	showing/1.
+	show_unify_as/2.
 
 user:prolog_trace_interception(Port, Frame, CHP, Action) :-
 	current_prolog_flag(gui_tracer, true),
-	notrace(intercept(Port, Frame, CHP, GuiAction)),
-	map_action(GuiAction, Frame, Action).
+	(   notrace(intercept(Port, Frame, CHP, GuiAction)),
+	    map_action(GuiAction, Frame, Action)
+	->  true
+	;   print_message(warning,
+			  guitracer(intercept_failed(Port, Frame,
+						     CHP, Action))),
+	    Action = continue
+	).
 
 
+%%	map_action(+GuiAction, +Frame, -Action) is det.
+%
+%	Map the abstract action of the gui-tracer into actions for the
+%	low-level tracer.  Runs in the debugged thread.
+%
+%	@tbd	The argument frame is not used.  Delete?
+
 map_action(creep, _, continue) :-
 	traceall.
 map_action(skip, _, continue) :-
@@ -90,16 +105,21 @@ map_action(finish, _, continue) :-
 	trace,
 	prolog_skip_level(_, Level).
 
+%%	traceall is det.
+%
+%	Go into non-skipping trace mode.
+
 traceall :-
 	prolog_skip_level(_, very_deep),
 	trace.
 
-intercept(_, _, _, _) :-
-	setting(active, false), !,
-	fail.
+%%	intercept(+Port, +Frame, +Choice, -Action) is semidet.
+%
+%	Toplevel of the tracer interception.  Runs in debugged thread.
+
 intercept(Port, Frame, CHP, Action) :-
 	send_tracer(current_break(@nil)),
-	debug('*** do_intercept(~w, ~w, ~w, _) ...~n', [Port, Frame, CHP]), 
+	debug('*** do_intercept(~w, ~w, ~w, _) ...~n', [Port, Frame, CHP]),
 	do_intercept(Port, Frame, CHP, Action0),
 	fix_action(Port, Action0, Action),
 	debug('*** ---> Action = ~w~n', [Action]),
@@ -111,6 +131,11 @@ fix_action(fail, skip,   creep) :- !.
 fix_action(exit, skip,   creep) :- !.
 fix_action(_,    Action, Action).
 
+%%	do_intercept(+Port, +Frame, +Choice, -Action) is det.
+%
+%	Actual core of the tracer intercepting code. Runs in the
+%	debugged thread.
+
 do_intercept(call, Frame, CHP, Action) :-
 	(   (   last_action(retry)
 	    ;	prolog_frame_attribute(Frame, top, true),
@@ -231,15 +256,23 @@ do_intercept(cut_exit(PC), Frame, CHP, Action) :-
 	action(Action).
 
 
+%%	show(+StartFrame, +Choice, +Up, +Port) is det.
+%
+%	Show current location from StartFrame.  Must be called in the
+%	context of the debugged thread.
+%
+%	@param Up	Skip bottom Up frames.  Use to show call port
+%			in the parent frame.
+
 show(StartFrame, CHP, Up, exception(Except)) :- !,
 	show(StartFrame, CHP, Up, exception, exception),
 	message_to_string(Except, Message),
 	send_tracer(report(warning, 'Exception: %s', Message)).
 show(StartFrame, CHP, Up, Port) :-
+	show(StartFrame, CHP, Up, Port, Port),
 	prolog_frame_attribute(StartFrame, goal, Goal),
 	predicate_name(user:Goal, Pred),
-	send_tracer(report(status, '%s: %s', Port?label_name, Pred)),
-	show(StartFrame, CHP, Up, Port, Port).
+	send_tracer(report(status, '%s: %s', Port?label_name, Pred)).
 
 show(StartFrame, CHP, Up, Port, Style) :-
 	send_tracer(current_frame(StartFrame)),
@@ -249,8 +282,6 @@ show(StartFrame, CHP, Up, Port, Style) :-
 			    stack
 			  ]),
 	find_frame(Up, StartFrame, Port, PC, Frame),
-	retractall(showing(_)),
-	assert(showing(Frame)),
 	prolog_show_frame(Frame,
 			  [ pc(PC),
 			    port(Port),
@@ -260,65 +291,16 @@ show(StartFrame, CHP, Up, Port, Style) :-
 			  ]).
 
 
-		 /*******************************
-		 *         SHOW LOCATION	*
-		 *******************************/
-
-attribute(Attributes, Att) :-
-	memberchk(Att, Attributes), !.
-
-attribute(Attributes, Att, _) :-
-	memberchk(Att, Attributes), !.
-attribute(_, Att, Def) :-
-	arg(1, Att, Def).
-
-
-prolog_show_frame(Frame, Attributes) :-
-	show_stack(Frame, Attributes),
-	show_bindings(Frame, Attributes),
-	show_source(Frame, Attributes),
-	(   setting(auto_raise, true)
-	->  send_tracer(expose)
-	;   true
-	).
-
-
-show_source(Frame, Attributes) :-
-	attribute(Attributes, source), !,
-	debug('source for #~w: ', [Frame]),
-	(   attribute(Attributes, pc(PC)),
-	    attribute(Attributes, port(Port), call),
-	    attribute(Attributes, style(Style), Port),
-	    debug('Show source, PC = ~w, Port = ~w~n', [PC, Port]),
-	    (   (PC == call ; PC == fail ; PC == exception )
-	    ->  prolog_frame_attribute(Frame, goal, Goal),
-		find_source(Goal, File, Line),
-		debug('At ~w:~d~n', [File, Line]),
-		send_tracer(show_line(File, Line, Style))
-	    ;   (   prolog_frame_attribute(Frame, clause, ClauseRef),
-		    debug('ClauseRef = ~w, PC = ~w~n', [ClauseRef, PC]),
-		    ClauseRef \== 0
-		->  subgoal_position(ClauseRef, PC, File, CharA, CharZ),
-		    debug('~p.~n', [show_range(File, CharA, CharZ, Style)]),
-		    send_tracer(show_range(File, CharA, CharZ, Style)),
-		    (	clause_property(ClauseRef, erased)
-		    ->	send_tracer(report(warning,
-					   'Running erased clause; source location may be incorrect'))
-		    ;	true
-		    )
-		;   prolog_frame_attribute(Frame, goal, Goal),
-		    find_source(Goal, File, Line),
-		    send_tracer(show_line(File, Line, Style))
-		)
-	    )
-	->  true
-	;   send_tracer(file(@nil))
-	).
-show_source(_, _).
-
+%%	find_frame(+Up, +StartFrame, +Port, -PC, -Frame)
+%
+%	Find the parent frame Up levels above StartFrame. Must be called
+%	in the context of the debugged thread.
+%
+%	@param PC	PC in parent frame
+%	@param Frame	Parent frame
 
 find_frame(N, Start, _, PC, Frame) :-
-	N > 0, 
+	N > 0,
 	debug('Frame = ~w; ', [Start]),
 	prolog_frame_attribute(Start, pc, PC0),
 	prolog_frame_attribute(Start, parent, Frame0), !,
@@ -334,17 +316,128 @@ find_frame2(N, F0, _, F, PC) :-
 	NN is N - 1,
 	find_frame2(NN, F1, PC1, F, PC).
 
+
+		 /*******************************
+		 *         SHOW LOCATION	*
+		 *******************************/
+
+%%	attribute(+Attributes, ?Att) is semidet.
+%%	attribute(+Attributes, ?Att, +Default) is semidet.
+%
+%	Attribute parsing
+%
+%	@bug	Merge with option library.
+
+attribute(Attributes, Att) :-
+	memberchk(Att, Attributes), !.
+
+attribute(Attributes, Att, _) :-
+	memberchk(Att, Attributes), !.
+attribute(_, Att, Def) :-
+	arg(1, Att, Def).
+
+%%	tracer_gui(+Attributes, -GUI) is det.
+%
+%	Find the tracer GUI object.
+
+tracer_gui(Attributes, GUI) :-
+	attribute(Attributes, gui(GUI)), !,
+	debug('GUI = ~p (given)~n', [GUI]).
+tracer_gui(_, GUI) :-
+	thread_self(Thread),
+	prolog_tracer(Thread, GUI),
+	debug('GUI = ~p (from thread ~p)~n', [GUI, Thread]).
+
+%%	prolog_show_frame(+Frame, +Attributes) is det.
+%
+%	Show given Prolog Frame in GUI-tracer, updating information as
+%	provided by Attributes.  Defined attributes:
+%
+%		* pc(PC)
+%		* choice(CHP)
+%		* port(Port)
+%		* style(Style)
+%		Style to use for editor fragment indicating location
+%		* source
+%		Update source window
+%		* bindings
+%		Update variable bindings window
+%		* stack
+%		Update stack window
+%		* gui(Object)
+%		Gui to address
+
+prolog_show_frame(Frame, Attributes) :-
+	debug('prolog_show_frame(~p, ~p)~n', [Frame, Attributes]),
+	show_stack(Frame, Attributes),
+	show_bindings(Frame, Attributes),
+	show_source(Frame, Attributes),
+	(   setting(auto_raise, true)
+	->  tracer_gui(Attributes, GUI),
+	    send_tracer(GUI, expose)
+	;   true
+	).
+
+
+%%	show_source(+Frame, +Attributes) is det.
+%
+%	Update the current location in the source window. If called from
+%	the GUI, the attribute gui(GUI) must be   given to relate to the
+%	proper thread.
+
+show_source(Frame, Attributes) :-
+	attribute(Attributes, source), !,
+	tracer_gui(Attributes, GUI),
+	debug('source for #~w~n: ', [Frame]),
+	(   attribute(Attributes, pc(PC)),
+	    attribute(Attributes, port(Port), call),
+	    attribute(Attributes, style(Style), Port),
+	    debug('Show source, PC = ~w, Port = ~w~n', [PC, Port]),
+	    (   atom(PC), PC \== exit		% PC is a port-name
+	    ->  prolog_frame_attribute(GUI, Frame, goal, Goal),
+		find_source(Goal, File, Line),
+		debug('At ~w:~d~n', [File, Line]),
+		send_tracer(GUI, show_line(File, Line, Style))
+	    ;   (   prolog_frame_attribute(GUI, Frame, clause, ClauseRef),
+		    debug('ClauseRef = ~w, PC = ~w~n', [ClauseRef, PC]),
+		    ClauseRef \== 0
+		->  subgoal_position(ClauseRef, PC, File, CharA, CharZ),
+		    debug('~p.~n', [show_range(File, CharA, CharZ, Style)]),
+		    send_tracer(GUI, show_range(File, CharA, CharZ, Style)),
+		    (	clause_property(ClauseRef, erased)
+		    ->	send_tracer(GUI,
+				    report(warning,
+					   'Running erased clause; \
+					   source location may be incorrect'))
+		    ;	true
+		    )
+		;   prolog_frame_attribute(GUI, Frame, goal, Goal),
+		    find_source(Goal, File, Line),
+		    send_tracer(GUI, show_line(File, Line, Style))
+		)
+	    )
+	->  true
+	;   send_tracer(GUI, file(@nil))
+	).
+show_source(_, _).
+
+
+%%	subgoal_position(+Clause, +PortOrPC, -File, -CharA, -CharZ) is det.
+%
+%	Character  range  CharA..CharZ  in  File   is  the  location  to
+%	highlight for the given clause at the given location.
+
 subgoal_position(ClauseRef, unify, File, CharA, CharZ) :- !,
-	clause_info(ClauseRef, File, TPos, _),
+	pce_clause_info(ClauseRef, File, TPos, _),
 	head_pos(ClauseRef, TPos, PosTerm),
 	arg(1, PosTerm, CharA),
 	arg(2, PosTerm, CharZ).
 subgoal_position(ClauseRef, choice, File, CharA, CharZ) :- !,
-	clause_info(ClauseRef, File, TPos, _),
+	pce_clause_info(ClauseRef, File, TPos, _),
 	arg(2, TPos, CharA),
 	CharZ is CharA + 1.		% i.e. select the dot.
 subgoal_position(ClauseRef, exit, File, CharA, CharZ) :- !,
-	clause_info(ClauseRef, File, TPos, _),
+	pce_clause_info(ClauseRef, File, TPos, _),
 	arg(2, TPos, CharA),
 	CharZ is CharA + 1.		% i.e. select the dot.
 subgoal_position(ClauseRef, fail, File, CharA, CharZ) :- !,
@@ -352,7 +445,7 @@ subgoal_position(ClauseRef, fail, File, CharA, CharZ) :- !,
 subgoal_position(ClauseRef, exception, File, CharA, CharZ) :- !,
 	subgoal_position(ClauseRef, exit, File, CharA, CharZ).
 subgoal_position(ClauseRef, PC, File, CharA, CharZ) :-
-	clause_info(ClauseRef, File, TPos, _),
+	pce_clause_info(ClauseRef, File, TPos, _),
 	(   '$clause_term_position'(ClauseRef, PC, List)
 	->  debug('Term-position: for ref=~w at PC=~w: ~w~n',
 		  [ClauseRef, PC, List]),
@@ -392,10 +485,71 @@ find_subgoal([1|T], brace_term_position(_,_,Pos), SPos) :-
 		 *             ACTION		*
 		 *******************************/
 
+%%	action(-Action) is det.
+%
+%	Wait for the user to perform some   action. We are called in the
+%	context of the debugged thread. If we are in the main thread, we
+%	use classical XPCE <-confirm. Otherwise  we   hang  waiting on a
+%	message queue. While waiting, we must  be prepared to call goals
+%	on behalf of in_debug_thread/2 started by   the  debugger gui to
+%	get additional information  on  the   state  of  our (debugging)
+%	thread.
+%
+%	@tbd	Synchronise with send_pce/1 and in_debug_thread/2.
+
 action(Action) :-
+	thread_self(main), !,
 	get_tracer(action, Action0),
 	debug('Got action ~w~n', [Action0]),
 	action(Action0, Action).
+action(Action) :-
+	send_tracer(prepare_action),
+	repeat,
+	debug(' ---> action: wait~n', []),
+	(   repeat,
+	    catch(thread_get_message('$trace'(Result)), E, wait_error(E))
+	->  true
+	;   debug('thread_get_message() failed; retrying ...~n'),
+	    fail
+	),
+	debug(' ---> action: result = ~p~n', [Result]),
+	(   Result = call(Goal, GVars, Caller)
+	->  run_in_debug_thread(Goal, GVars, Caller),
+	    fail
+	;   Result = action(Action)
+	->  !
+	;   assertion(fail)
+	).
+
+%%	wait_error(+ErrorTerm)
+%
+%	thread_get_message/1 can only fail due   to  signals throwing an
+%	exception. For example,  if  the  traced   goal  is  guarded  by
+%	call_with_time_limit/2. Here we  print  the   message  and  keep
+%	waiting. Note that this causes the  system   to  be lost for the
+%	application.
+%
+%	@tbd	Allow passing the error to the application
+%	@tbd	Deal with similar signals in other part of the tracing
+%		code.
+
+wait_error(E) :-
+	message_to_string(E, Message),
+	format(user_error, 'Error while waiting for for user: ~w~n\
+			   Retrying~n', [Message]),
+	fail.
+
+
+run_in_debug_thread(Goal, GVars, Caller) :-
+	(   catch(Goal, Error, true)
+	->  (   var(Error)
+	    ->	Result = true(GVars)
+	    ;	Result = error(Error)
+	    )
+	;   Result = false
+	),
+	debug(' ---> run_in_debug_thread: send ~p~n', [Result]),
+	thread_send_message(Caller, '$trace'(Result)).
 
 action(break, Action) :- !,
 	break,
@@ -408,24 +562,40 @@ action(Action, Action).
 		 *	      STACK		*
 		 *******************************/
 
+%%	show_stack(+Frame, +Attributes) is det.
+%
+%	Show call- and choicepoint stack. Run in the context of the GUI.
+
 show_stack(Frame, Attributes) :-
 	attribute(Attributes, stack), !,
+	tracer_gui(Attributes, GUI),
 	debug('stack ...', []),
+	in_debug_thread(GUI,
+			stack_info(Frame,
+				   CallFrames, ChoiceFrames,
+				   Attributes)),
+	send_tracer(GUI, show_stack(CallFrames, ChoiceFrames)).
+show_stack(_, _).
+
+%%	stack_info(+Frame, -CallFrames, -ChoiceFrames, +Attributes) is det.
+%
+%	Find the callstack and choicepoints that must be made visible in
+%	the stack window. Must  run  in   the  context  of  the debugged
+%	thread.
+
+stack_info(Frame, CallFrames, ChoiceFrames, Attributes) :-
 	attribute(Attributes, port(Port), call),
 	attribute(Attributes, pc(PC), Port),
 	attribute(Attributes, choice(CHP), Frame),
 	setting(stack_depth, Depth),
 	setting(choice_depth, MaxChoice),
-	get_tracer(member(stack), StackBrowser),
-	send(StackBrowser, clear),
 	stack_frames(Depth, Frame, PC, CallFrames),
 	debug('Stack frames: ~w~n', [CallFrames]),
 	level_range(CallFrames, Range),
 	debug('Levels ~w, CHP = ~w~n', [Range, CHP]),
 	choice_frames(MaxChoice, CHP, Range, [], ChoiceFrames),
-	debug('Choicepoints: ~p~n', [ChoiceFrames]),
-	display_stack(StackBrowser, CallFrames, ChoiceFrames).
-show_stack(_, _).
+	debug('Choicepoints: ~p~n', [ChoiceFrames]).
+
 
 stack_frames(0, _, _, []) :- !.
 stack_frames(Depth, F, PC, Frames) :-
@@ -463,7 +633,7 @@ choice_frames(Max, CHP, Range, Seen, [frame(Frame, choice)|Frames]) :-
 choice_frames(_, _, _, _, []).
 
 %	earlier_choice(+Here, -Visible)
-%	
+%
 %	Return earliers choices on backgtracking.
 
 earlier_choice(CHP, CHP).
@@ -472,7 +642,7 @@ earlier_choice(CHP, Next) :-
 	earlier_choice(Parent, Next).
 
 %	visible_choice(+CHP)
-%	
+%
 %	A visible choice is a choice-point that realises a real choice
 %	and is created by a visible frame.
 
@@ -503,8 +673,12 @@ in_range(Level, Low-_High) :-
 	Level >= Low.
 %	between(Low, High, Level).
 
-show_stack_location(Frame, PC) :-
-	get_tracer(member(stack), StackBrowser),
+%%	show_stack_location(+GUI, +Frame, +PC)
+%
+%	Highlight Frame in the stack-view.
+
+show_stack_location(GUI, Frame, PC) :-
+	get_tracer(GUI, member(stack), StackBrowser),
 	send(StackBrowser, selection, Frame, PC).
 
 
@@ -512,45 +686,55 @@ show_stack_location(Frame, PC) :-
 		 *	       BINDINGS		*
 		 *******************************/
 
-show_args_pc(call).
-show_args_pc(fail).
-show_args_pc(exception).
-show_args_pc(foreign).
+%%	show_bindings(+Frame, +Attributes) is det.
+%
+%	Show argument bindings.
 
 show_bindings(Frame, Attributes) :-
 	attribute(Attributes, bindings), !,
+	tracer_gui(Attributes, GUI),
 	debug('bindings ... ', []),
-	get_tracer(member(bindings), Browser),
+	get_tracer(GUI, member(bindings), Browser),
 	(   attribute(Attributes, pc(PC))
 	->  true
 	;   PC = @default
 	),
-	show_stack_location(Frame, PC),
+	show_stack_location(GUI, Frame, PC),
 	send(Browser, clear),
 	send(Browser, prolog_frame, Frame),
 	(   show_args_pc(PC)
-	->  show_arguments(Frame, Attributes)
+	->  send(Browser, label, 'Arguments'),
+	    show_arguments(GUI, Frame, Attributes)
 	;   send(Browser, label, 'Bindings'),
-	    prolog_frame_attribute(Frame, clause, ClauseRef),
+	    prolog_frame_attribute(GUI, Frame, clause, ClauseRef),
 	    debug('(clause ~w) ', [ClauseRef]),
-	    catch(clause_info(ClauseRef, _, _, VarNames), E,
+	    catch(pce_clause_info(ClauseRef, _, _, VarNames), E,
 		  (print_message(error, E), fail)),
-	    frame_bindings(Frame, VarNames, Bindings),
+	    in_debug_thread(GUI, frame_bindings(Frame, VarNames, Bindings)),
+	    debug('(bindings ~p) ', [Bindings]),
 	    send(Browser, bindings, Bindings),
 	    debug('(ok) ', [])
 	).
 show_bindings(_, _).
 
-show_arguments(Frame, _Attributes) :-
-	get_tracer(member(bindings), Browser),
-	send(Browser, label, 'Arguments'),
-	frame_arguments(Frame, Args),
+%%	show_args_pc(+Port) is semidet.
+%
+%	If we are at Port, we must simple show the arguments.
+
+show_args_pc(call).
+show_args_pc(fail).
+show_args_pc(exception).
+show_args_pc(foreign).
+
+show_arguments(GUI, Frame, _Attributes) :-
+	get_tracer(GUI, member(bindings), Browser),
+	in_debug_thread(GUI, frame_arguments(Frame, Args)),
 	send(Browser, bindings, Args).
 
-%	frame_arguments(+Frame, -Args)
+%%	frame_arguments(+Frame, -Args)
 %
 %	Return arguments of the frame as [I:I=Value, ...], compatible with
-%	the normal binding list.
+%	the normal binding list. Must run in context of debugged thread.
 
 frame_arguments(Frame, Args) :-
 	prolog_frame_attribute(Frame, goal, Goal),
@@ -568,6 +752,11 @@ frame_arguments(I, Arity, Frame, [I:I=Value|T]) :-
 frame_arguments(_, _, _, []).
 
 
+%%	frame_bindings(+Frame, +VarNames, -Bindings) is det.
+%
+%	Get the variable bindings for Frame. Must run the the context of
+%	the debugged thread.
+
 frame_bindings(Frame, VarNames, Bindings) :-
 	functor(VarNames, _, Arity),
 	frame_bindings(0, Arity, Frame, VarNames, B0),
