@@ -1,11 +1,11 @@
-/*  $Id: pl-atom.c,v 1.81 2007/04/18 09:12:53 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -39,7 +39,7 @@ the address of the structure.
 
 Next, there is a hash-table, which is a normal `open' hash-table mapping
 char * to the atom structure. This   thing is dynamically rehashed. This
-table is used by lookupAtom() below. 
+table is used by lookupAtom() below.
 
 Atom garbage collection
 -----------------------
@@ -230,6 +230,7 @@ PL_unregister_blob_type(PL_blob_t *type)
 { unsigned int i;
   PL_blob_t **t;
   int discarded = 0;
+  Atom *ap;
 
   PL_LOCK(L_MISC);
   for(t = &GD->atoms.types; *t; t = &(*t)->next)
@@ -243,13 +244,15 @@ PL_unregister_blob_type(PL_blob_t *type)
   PL_register_blob_type(&unregistered_blob_atom);
 
   LOCK();
-  for(i=0; i < entriesBuffer(&atom_array, Atom); i++ )
+
+  ap = GD->atoms.array;
+  for(i=0; i < GD->atoms.count; i++, ap++ )
   { Atom atom;
 
-    if ( (atom = baseBuffer(&atom_array, Atom)[i]) )
+    if ( (atom = *ap) )
     { if ( atom->type == type )
       { atom->type = &unregistered_blob_atom;
-	
+
 	atom->name = "<discarded blob>";
 	atom->length = strlen(atom->name);
 
@@ -284,9 +287,9 @@ another atom.
 
 static void
 registerAtom(Atom a)
-{ size_t n = entriesBuffer(&atom_array, Atom);
+{ size_t n = GD->atoms.count;
 #ifdef O_ATOMGC				/* try to find a hole! */
-  Atom *ap = baseBuffer(&atom_array, Atom);
+  Atom *ap = GD->atoms.array;
   Atom *ep = ap+n;
   Atom *p;
 
@@ -308,8 +311,19 @@ registerAtom(Atom a)
 #endif /*O_ATOMGC*/
 
   a->atom = (n<<LMASK_BITS)|TAG_ATOM;
+  if ( n >= GD->atoms.array_allocated )
+  { size_t newcount = GD->atoms.array_allocated * 2;
+    size_t newsize  = newcount*sizeof(Atom);
+    Atom *np = PL_malloc(newsize);
 
-  addBuffer(&atom_array, a, Atom);
+    memcpy(np, ap, newsize/2);
+    GD->atoms.array = np;
+    GD->atoms.array_allocated = newcount;
+    PL_free(ap);
+    ap = np;
+  }
+  ap[n++] = a;
+  GD->atoms.count = n;
 }
 
 
@@ -332,15 +346,17 @@ treat the signal as bogus if agc has already been performed.
 
 word
 lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
-{ int v0, v;
+{ GET_LD
+  unsigned int v0, v;
   uintptr_t oldheap;
   Atom a;
 
   if ( !type->registered )		/* avoid deadlock */
     PL_register_blob_type(type);
 
+  startCritical;
   LOCK();
-  v0 = unboundStringHashValue(s, length);
+  v0 = MurmurHashAligned2(s, length, MURMUR_SEED);
   v  = v0 & (atom_buckets-1);
   DEBUG(0, lookups++);
 
@@ -351,11 +367,13 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
 	if ( length == a->length &&
 	     type == a->type &&
 	     memcmp(s, a->name, length) == 0 )
-	{ 
+	{
 #ifdef O_ATOMGC
-	  a->references++;
+	  if ( indexAtom(a->atom) >= GD->atoms.builtin )
+	    a->references++;
 #endif
           UNLOCK();
+	  endCritical;
 	  *new = FALSE;
 	  return a->atom;
 	}
@@ -367,11 +385,12 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
 	if ( length == a->length &&
 	     type == a->type &&
 	     s == a->name )
-	{ 
+	{
 #ifdef O_ATOMGC
 	  a->references++;
 #endif
           UNLOCK();
+	  endCritical;
 	  *new = FALSE;
 	  return a->atom;
 	}
@@ -379,31 +398,29 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
     }
   }
 
-  { GET_LD
-    oldheap = GD->statistics.heap;
-    a = allocHeap(sizeof(struct atom));
-    a->length = length;
-    a->type = type;
-    if ( false(type, PL_BLOB_NOCOPY) )
-    { a->name = allocHeap(length+1);
-      memcpy(a->name, s, length);
-      a->name[length] = EOS;
-    } else
-    { a->name = (char *)s;
-    }
-#ifdef O_HASHTERM
-    a->hash_value = v0;
+  oldheap = GD->statistics.heap;
+  a = allocHeap(sizeof(struct atom));
+  a->length = length;
+  a->type = type;
+  if ( false(type, PL_BLOB_NOCOPY) )
+  { a->name = allocHeap(length+1);
+    memcpy(a->name, s, length);
+    a->name[length] = EOS;
+  } else
+  { a->name = (char *)s;
+  }
+#ifdef O_TERMHASH
+  a->hash_value = v0;
 #endif
 #ifdef O_ATOMGC
-    a->references = 1;
+  a->references = 1;
 #endif
-    registerAtom(a);
-    if ( true(type, PL_BLOB_UNIQUE) )
-    { a->next       = atomTable[v];
-      atomTable[v]  = a;
-    }
-    GD->statistics.atoms++;
+  registerAtom(a);
+  if ( true(type, PL_BLOB_UNIQUE) )
+  { a->next       = atomTable[v];
+    atomTable[v]  = a;
   }
+  GD->statistics.atoms++;
 
 #ifdef O_ATOMGC
   if ( GD->atoms.margin != 0 &&
@@ -414,17 +431,18 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
       PL_raise(SIG_ATOM_GC);
   }
 #endif
-    
+
   if ( atom_buckets * 2 < GD->statistics.atoms )
     rehashAtoms();
 
   GD->statistics.atomspace += (GD->statistics.heap - oldheap);
 
   UNLOCK();
-  
+
   *new = TRUE;
   if ( type->acquire )
     (*type->acquire)(a->atom);
+  endCritical;
 
   return a->atom;
 }
@@ -537,7 +555,7 @@ atoms each pass.
 
 void
 lockAtoms()
-{ GD->atoms.builtin     = entriesBuffer(&atom_array, Atom);
+{ GD->atoms.builtin     = GD->atoms.count;
   GD->atoms.non_garbage = GD->atoms.builtin;
 }
 
@@ -550,12 +568,12 @@ markAtom(atom_t a)
 { size_t i = indexAtom(a);
   Atom ap;
 
-  if ( i >= entriesBuffer(&atom_array, Atom) )
+  if ( i >= GD->atoms.count )
     return;				/* not an atom */
   if ( i < GD->atoms.builtin )
     return;				/* locked range */
 
-  ap = fetchBuffer(&atom_array, i, Atom);
+  ap = GD->atoms.array[i];
 
   if ( ap )
   {
@@ -602,13 +620,13 @@ destroyAtom(Atom *ap, uintptr_t mask ARG_LD)
 
   for( ; ; ap2 = &(*ap2)->next )
   { assert(*ap2);			/* MT: TBD: failed a few times!? */
-	
+
     if ( *ap2 == a )
     { *ap2 = a->next;
       break;
     }
   }
-      
+
   *ap = NULL;			/* delete from index array */
   GD->atoms.collected++;
   GD->statistics.atoms--;
@@ -622,9 +640,9 @@ destroyAtom(Atom *ap, uintptr_t mask ARG_LD)
 static void
 collectAtoms(void)
 { GET_LD
-  Atom *ap0 = baseBuffer(&atom_array, Atom);
+  Atom *ap0 = GD->atoms.array;
   Atom *ap  = ap0 + GD->atoms.builtin;
-  Atom *ep  = ap0 + entriesBuffer(&atom_array, Atom);
+  Atom *ep  = ap0 + GD->atoms.count;
   int hole_seen = FALSE;
   uintptr_t mask = atom_buckets-1;
 
@@ -662,12 +680,13 @@ GC. These issues are described with enterGC() in pl-gc.c.
 word
 pl_garbage_collect_atoms()
 { GET_LD
-  intptr_t oldcollected, oldheap, freed;
+  intptr_t oldheap, freed;
+  int64_t oldcollected;
   int verbose;
   double t;
   sigset_t set;
 
-  PL_LOCK(L_GC);			
+  PL_LOCK(L_GC);
   if ( gc_status.blocked )		/* Tricky things; avoid problems. */
   { PL_UNLOCK(L_GC);
     succeed;
@@ -684,7 +703,7 @@ pl_garbage_collect_atoms()
 
   gc_status.blocked++;			/* avoid recursion */
 
-  if ( (verbose = trueFeature(TRACE_GC_FEATURE)) )
+  if ( (verbose = truePrologFlag(PLFLAG_TRACE_GC)) )
   {
 #ifdef O_DEBUG_ATOMGC
 /*
@@ -700,11 +719,13 @@ pl_garbage_collect_atoms()
   }
 
   PL_LOCK(L_THREAD);
+  PL_LOCK(L_AGC);
   LOCK();
   blockSignals(&set);
   t = CpuTime(CPU_USER);
   markAtomsOnStacks(LD);
 #ifdef O_PLMT
+  markAtomsThreads();
   forThreadLocalData(markAtomsOnStacks, 0);
 #endif
   oldcollected = GD->atoms.collected;
@@ -719,15 +740,16 @@ pl_garbage_collect_atoms()
   GD->statistics.atomspace -= freed;
   unblockSignals(&set);
   UNLOCK();
+  PL_UNLOCK(L_AGC);
   PL_UNLOCK(L_THREAD);
   gc_status.blocked--;
   PL_UNLOCK(L_GC);
-    
+
   if ( verbose )
     printMessage(ATOM_informational,
 		 PL_FUNCTOR_CHARS, "agc", 1,
 		   PL_FUNCTOR_CHARS, "done", 3,
-		     PL_LONG, GD->atoms.collected - oldcollected,
+		     PL_INT64, GD->atoms.collected - oldcollected,
 		     PL_INT, GD->statistics.atoms,
 		     PL_DOUBLE, (double)t);
 
@@ -751,13 +773,13 @@ PL_agc_hook(PL_agc_hook_t new)
 
 void
 resetAtoms()
-{ 
+{
 }
 
 
 void
 PL_register_atom(atom_t a)
-{ 
+{
 #ifdef O_ATOMGC
   size_t index = indexAtom(a);
 
@@ -765,7 +787,7 @@ PL_register_atom(atom_t a)
   { Atom p;
 
     LOCK();
-    p = fetchBuffer(&atom_array, index, Atom);
+    p = GD->atoms.array[index];
     p->references++;
     UNLOCK();
   }
@@ -774,7 +796,7 @@ PL_register_atom(atom_t a)
 
 void
 PL_unregister_atom(atom_t a)
-{ 
+{
 #ifdef O_ATOMGC
   size_t index = indexAtom(a);
 
@@ -782,10 +804,10 @@ PL_unregister_atom(atom_t a)
   { Atom p;
 
     LOCK();
-    p = fetchBuffer(&atom_array, index, Atom);
+    p = GD->atoms.array[index];
     p->references--;
     if ( p->references == (unsigned)-1 )
-    { Sdprintf("OOPS: -1 references to %s\n", p->name);
+    { Sdprintf("OOPS: -1 references to '%s'\n", p->name);
       trap_gdb();
     }
     UNLOCK();
@@ -805,7 +827,7 @@ rehashAtoms()
 { GET_LD
   Atom *oldtab   = atomTable;
   int   oldbucks = atom_buckets;
-  intptr_t mx = entriesBuffer(&atom_array, Atom);
+  size_t mx = GD->atoms.count;
   uintptr_t mask;
   Atom *ap, *ep;
 
@@ -814,10 +836,10 @@ rehashAtoms()
   mask = atom_buckets-1;
   atomTable = allocHeap(atom_buckets * sizeof(Atom));
   memset(atomTable, 0, atom_buckets * sizeof(Atom));
-  
+
   DEBUG(0, Sdprintf("rehashing atoms (%d --> %d)\n", oldbucks, atom_buckets));
 
-  for(ap = baseBuffer(&atom_array, Atom), ep = ap+mx;
+  for(ap = GD->atoms.array, ep = ap+mx;
       ap < ep;
       ap++)
   { Atom a = *ap;
@@ -837,7 +859,7 @@ pl_atom_hashstat(term_t idx, term_t n)
 { GET_LD
   long i, m;
   Atom a;
-  
+
   if ( !PL_get_long(idx, &i) || i < 0 || i >= (long)atom_buckets )
     fail;
   for(m = 0, a = atomTable[i]; a; a = a->next)
@@ -858,8 +880,8 @@ registerBuiltinAtoms()
 
   for(s = atoms; *s; s++, a++)
   { size_t len = strlen(*s);
-    int v0 = unboundStringHashValue(*s, len);
-    int v = v0 & (atom_buckets-1);
+    unsigned int v0 = MurmurHashAligned2(*s, len, MURMUR_SEED);
+    unsigned int v = v0 & (atom_buckets-1);
 
     a->name       = (char *)*s;
     a->length     = len;
@@ -867,7 +889,7 @@ registerBuiltinAtoms()
 #ifdef O_ATOMGC
     a->references = 0;
 #endif
-#ifdef O_HASHTERM
+#ifdef O_TERMHASH
     a->hash_value = v0;
 #endif
     a->next       = atomTable[v];
@@ -895,10 +917,11 @@ initAtoms(void)
     atomTable = allocHeap(atom_buckets * sizeof(Atom));
 
     memset(atomTable, 0, atom_buckets * sizeof(Atom));
-    initBuffer(&atom_array);
+    GD->atoms.array_allocated = 4096;
+    GD->atoms.array = PL_malloc(GD->atoms.array_allocated*sizeof(Atom));
     registerBuiltinAtoms();
 #ifdef O_ATOMGC
-    GD->atoms.margin     = 10000;
+    GD->atoms.margin = 10000;
     lockAtoms();
 #endif
     PL_register_blob_type(&text_atom);
@@ -911,7 +934,7 @@ initAtoms(void)
 
 void
 cleanupAtoms(void)
-{ discardBuffer(&atom_array);
+{ PL_free(GD->atoms.array);
 }
 
 
@@ -928,7 +951,7 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t i ARG_LD)
 	  return PL_unify_atom(type, bt->atom_name);
 	else if ( false(bt, PL_BLOB_TEXT) )
 	  fail;
-	  
+
 	succeed;
       }
       if ( !PL_is_variable(a) )
@@ -950,10 +973,10 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t i ARG_LD)
       fail;
   }
 
-  for( ; i < (int)entriesBuffer(&atom_array, Atom); i++ )
+  for( ; i < (int)GD->atoms.count; i++ )
   { Atom atom;
 
-    if ( (atom = baseBuffer(&atom_array, Atom)[i]) )
+    if ( (atom = GD->atoms.array[i]) )
     { if ( type )
       { if ( type_name && type_name != atom->type->atom_name )
 	  continue;
@@ -961,7 +984,7 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t i ARG_LD)
 	PL_unify_atom(type, atom->type->atom_name);
       } else if ( false(atom->type, PL_BLOB_TEXT) )
 	continue;
-	
+
       PL_unify_atom(a, atom->atom);
       ForeignRedoInt(i+1);
     }
@@ -986,6 +1009,19 @@ PRED_IMPL("current_atom", 1, current_atom, PL_FA_NONDETERMINISTIC)
   return current_blob(A1, 0, CTX_CNTRL, CTX_INT PASS_LD);
 }
 
+static
+PRED_IMPL("$atom_references", 2, atom_references, 0)
+{ PRED_LD
+  atom_t atom;
+
+  if ( PL_get_atom_ex(A1, &atom) )
+  { Atom av = atomValue(atom);
+
+    return PL_unify_integer(A2, av->references);
+  }
+
+  fail;
+}
 
 		 /*******************************
 		 *	 ATOM COMPLETION	*
@@ -1013,17 +1049,16 @@ allAlpha(const char *s)
 
 static int
 extendAtom(char *prefix, bool *unique, char *common)
-{ intptr_t i, mx = entriesBuffer(&atom_array, Atom);
+{ intptr_t i, mx = GD->atoms.count;
   Atom a;
   bool first = TRUE;
   int lp = (int) strlen(prefix);
-  
+  Atom *ap = GD->atoms.array;
+
   *unique = TRUE;
 
-  for(i=0; i<mx; i++)
-  { a = baseBuffer(&atom_array, Atom)[i];
-
-    if ( a && strprefix(a->name, prefix) )
+  for(i=0; i<mx; i++, ap++)
+  { if ( (a=*ap) && strprefix(a->name, prefix) )
     { if ( strlen(a->name) >= LINESIZ )
 	continue;
       if ( first == TRUE )
@@ -1050,7 +1085,7 @@ pl_complete_atom(term_t prefix, term_t common, term_t unique)
   bool u;
   char buf[LINESIZ];
   char cmm[LINESIZ];
-    
+
   if ( !PL_get_chars_ex(prefix, &p, CVT_ALL) )
     fail;
   strcpy(buf, p);
@@ -1077,16 +1112,17 @@ compareMatch(const void *m1, const void *m2)
 
 static bool
 extend_alternatives(char *prefix, struct match *altv, int *altn)
-{ intptr_t i, mx = entriesBuffer(&atom_array, Atom);
-  Atom a;
+{ intptr_t i, mx = GD->atoms.count;
+  Atom a, *ap;
   char *as;
   int l;
 
   *altn = 0;
-  for(i=0; i<mx; i++)
-  { if ( !(a = baseBuffer(&atom_array, Atom)[i]) )
+  ap = GD->atoms.array;
+  for(i=0; i<mx; i++, ap++)
+  { if ( !(a=*ap) )
       continue;
-    
+
     as = a->name;
     if ( strprefix(as, prefix) &&
 	 allAlpha(as) &&
@@ -1098,7 +1134,7 @@ extend_alternatives(char *prefix, struct match *altv, int *altn)
 	break;
     }
   }
-  
+
   qsort(altv, *altn, sizeof(struct match), compareMatch);
 
   succeed;
@@ -1121,7 +1157,7 @@ pl_atom_completions(term_t prefix, term_t alternatives)
   strcpy(buf, p);
 
   extend_alternatives(buf, altv, &altn);
-  
+
   for(i=0; i<altn; i++)
   { if ( !PL_unify_list(alts, head, alts) ||
 	 !PL_unify_atom(head, altv[i].name->atom) )
@@ -1129,13 +1165,13 @@ pl_atom_completions(term_t prefix, term_t alternatives)
   }
 
   return PL_unify_nil(alts);
-} 
+}
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Completeness generation for the GNU readline library. This function uses
 a state variable to indicate  the   generator  should maintain/reset its
-state. Horrible! 
+state. Horrible!
 
 We must use thread-local data here.  Worse   is  we can't use the normal
 Prolog one as there might  not  be   a  Prolog  engine associated to the
@@ -1182,7 +1218,8 @@ alnum_text(PL_chars_t *txt)
 static int
 atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
 { GET_LD
-  intptr_t i, mx = entriesBuffer(&atom_array, Atom);
+  intptr_t i, mx = GD->atoms.count;
+  Atom *ap;
 
 #ifdef O_PLMT
   if ( !key )
@@ -1200,12 +1237,12 @@ atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
 #endif
   }
 
-  for(; i<mx; i++)
+  for(ap=&GD->atoms.array[i]; i<mx; i++, ap++)
   { Atom a;
 
-    if ( !(a = baseBuffer(&atom_array, Atom)[i]) )
+    if ( !(a = *ap) )
       continue;
-    
+
     if ( is_signalled() )		/* Notably allow windows version */
       PL_handle_signals();		/* to break out on ^C */
 
@@ -1216,10 +1253,10 @@ atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
     {
 #ifdef O_PLMT
       pthread_setspecific(key, (void *)(i+1));
-#else   
+#else
       LD->atoms.generator = i+1;
 #endif
-  
+
       return TRUE;
     }
   }
@@ -1267,7 +1304,7 @@ PL_atom_generator_w(const pl_wchar_t *prefix,
       { const unsigned char *s = (const unsigned char *)hit.text.t;
 	const unsigned char *e = &s[hit.length];
 	pl_wchar_t *o;
-	
+
 	for(o=buffer; s<e;)
 	  *o++ = *s++;
 	*o = EOS;
@@ -1288,4 +1325,5 @@ PL_atom_generator_w(const pl_wchar_t *prefix,
 BeginPredDefs(atom)
   PRED_DEF("current_blob",  2, current_blob, PL_FA_NONDETERMINISTIC)
   PRED_DEF("current_atom", 1, current_atom, PL_FA_NONDETERMINISTIC)
+  PRED_DEF("$atom_references", 2, atom_references, 0)
 EndPredDefs
