@@ -50,10 +50,20 @@ vName(Word adr)
     Ssprintf(name, "_G%ld", (Word)adr - (Word)gBase);
 
   return name;
-} 
+}
 #endif
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+(*) Although this is an assignment from var   to value, we use a trailed
+assignment  to  exploit  mergeTrailedAssignments()   in  GC,  discarding
+multiple  assignments  in  the  same  segment,  needed  to  ensure  that
+deterministic wakeup does not leak  space.   The  test  program is this,
+which must run in constant space.
+
+	loop :- freeze(X, true), X = a, loop.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static int
 registerWakeup(Word name, Word value ARG_LD)
 { Word wake;
@@ -80,12 +90,13 @@ registerWakeup(Word name, Word value ARG_LD)
     DEBUG(1, Sdprintf("appended to wakeup\n"));
   } else				/* empty list */
   { Word head = valTermRef(LD->attvar.head);
-    
+
     assert(isVar(*head));
+    TrailAssignment(head);		/* See (*) */
     *head = consPtr(wake, TAG_COMPOUND|STG_GLOBAL);
-    Trail(head);			/* not gc-ed!? */
+    TrailAssignment(tail);
     *tail = makeRef(wake+3);
-    Trail(tail);			/* not gc-ed!? */
+    LD->alerted |= ALERT_WAKEUP;
     DEBUG(1, Sdprintf("new wakeup\n"));
   }
 
@@ -99,7 +110,7 @@ registerWakeup(Word name, Word value ARG_LD)
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 assignAttVar(Word var, Word value)		(var := value)
-	
+
 Assign  value  to  the  given  attributed    variable,   adding  a  term
 wake(Attribute, Value, Tail) to the global variable resembling the goals
 that should be awoken.
@@ -214,7 +225,7 @@ find_attr(Word av, atom_t name, Word *vp ARG_LD)
 
   for(;;)
   { deRef(l);
-    
+
     if ( isNil(*l) )
     { *vp = l;
       fail;
@@ -283,7 +294,7 @@ get_attr(term_t list, atom_t name, term_t value)
 
   for(;;)
   { deRef(l);
-    
+
     if ( isTerm(*l) )
     { Functor f = valueTerm(*l);
 
@@ -371,24 +382,38 @@ so we will use save/restore in all places.
 The functions below provide a way to   realise the save/restore. It must
 be nicely nested in the  same  way   and  using  the same constraints as
 PL_open_foreign_frame/PL_close_foreign_frame.
+
+NOTE: Now we also save pending  exceptions,   for  which  the same rules
+apply. The environment has size 1 if there  is a pending exception, 2 if
+a wakeup was saved and 3 if both where saved.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 fid_t
 saveWakeup(ARG1_LD)
 { Word h;
 
-  if ( *(h=valTermRef(LD->attvar.head)) )
+  if ( *(h=valTermRef(LD->attvar.head)) || exception_term )
   { fid_t fid = PL_open_foreign_frame();
-    term_t s = PL_new_term_refs(2);
-    
-    DEBUG(1, pl_write(LD->attvar.head); pl_nl());
+    term_t s;
 
-    *valTermRef(s+0) = *h;
-    setVar(*h);
-    h = valTermRef(LD->attvar.tail);
-    *valTermRef(s+1) = *h;
-    setVar(*h);    
-    DEBUG(1, Sdprintf("Saved wakeup to %p\n", valTermRef(s)));
+    if ( exception_term )
+    { s = PL_new_term_ref();
+      *valTermRef(s) = *valTermRef(exception_term);
+      exception_term = 0;
+    }
+
+    if ( *h )
+    { s = PL_new_term_refs(2);
+
+      DEBUG(1, pl_write(LD->attvar.head); pl_nl());
+
+      *valTermRef(s+0) = *h;
+      setVar(*h);
+      h = valTermRef(LD->attvar.tail);
+      *valTermRef(s+1) = *h;
+      setVar(*h);
+      DEBUG(1, Sdprintf("Saved wakeup to %p\n", valTermRef(s)));
+    }
 
     return fid;
   }
@@ -397,18 +422,48 @@ saveWakeup(ARG1_LD)
 }
 
 
+static void
+restore_exception(Word p ARG_LD)
+{ DEBUG(1, Sdprintf("Restore exception from %p\n", p));
+
+  *valTermRef(exception_bin) = p[0];
+  exception_term = exception_bin;
+
+  DEBUG(1, pl_writeln(exception_term));
+}
+
+
+static void
+restore_wakeup(Word p ARG_LD)
+{
+
+  *valTermRef(LD->attvar.head) = p[0];
+  *valTermRef(LD->attvar.tail) = p[1];
+
+  DEBUG(1, pl_writeln(LD->attvar.head));
+}
+
+
 void
 restoreWakeup(fid_t fid ARG_LD)
 { if ( fid )
   { FliFrame fr = (FliFrame) valTermRef(fid);
     Word p = (Word)(fr+1);
 
-    DEBUG(1, Sdprintf("Restore wakeup from %p\n", p));
-
-    *valTermRef(LD->attvar.head) = p[0];
-    *valTermRef(LD->attvar.tail) = p[1];
-
-    DEBUG(1, pl_write(LD->attvar.head); pl_nl());
+    switch(fr->size)
+    { case 3:
+	restore_exception(p PASS_LD);
+        p++;
+	/*FALLTHROUGH*/
+      case 2:
+	restore_wakeup(p PASS_LD);
+        break;
+      case 1:
+	restore_exception(p PASS_LD);
+        break;
+      default:
+	assert(0);
+    }
 
     PL_discard_foreign_frame(fid);
   }
@@ -455,7 +510,7 @@ PRED_IMPL("get_attr", 3, get_attr3, 0) /* +Var, +Name, -Value */
     fail;
   if ( !PL_get_attr(A1, al) )
     fail;
-  
+
   return get_attr(al, name, A3);
 }
 
@@ -475,7 +530,7 @@ PRED_IMPL("put_attr", 3, put_attr3, 0)	/* +Var, +Name, +Value */
   if ( isVar(*vp) && !onStackArea(global, vp) )
   { Word p = allocGlobal(1);		/* attribute values should be on */
 					/* the global stack! */
-    
+
     setVar(*p);
     *vp = makeRefG(p);
     Trail(vp);
@@ -592,7 +647,7 @@ PRED_IMPL("$freeze", 2, freeze, PL_FA_TRANSPARENT)
       if ( find_attr(v, ATOM_freeze, &vp PASS_LD) )
       { Word gc = allocGlobal(3);		/* 3 cells global */
 
-	gc[0] = FUNCTOR_comma2;
+	gc[0] = FUNCTOR_dand2;
 	gc[1] = linkVal(vp);
 	gc[2] = goal;
 
@@ -628,7 +683,7 @@ PRED_IMPL("$freeze", 2, freeze, PL_FA_TRANSPARENT)
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 $new_choice_point(-Chp) is det.
 
-Unify Chp with a reference to a new choicepoint. 
+Unify Chp with a reference to a new choicepoint.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static
@@ -688,14 +743,14 @@ has_attributes_after(Word av, Choice ch ARG_LD)
 
   for(;;)
   { deRef(l);
-    
+
     if ( isNil(*l) )
     { fail;
     } else if ( isTerm(*l) )
     { Functor f = valueTerm(*l);
 
       DEBUG(1, Sdprintf("\tterm at %p\n", f));
-      
+
       if ( (Word)f >= ch->mark.globaltop )
 	succeed;
 
@@ -758,7 +813,7 @@ PRED_IMPL("$attvars_after_choicepoint", 2, attvars_after_choicepoint, 0)
 retry:
   if ( !PL_get_intptr_ex(A1, &off) )
     fail;
-  
+
   ch = (Choice)((Word)lBase+off);
   list = tailp = allocGlobalNoShift(1);
   if ( !list )
@@ -788,7 +843,7 @@ retry:
       }
     }
   }
-  
+
   scan_trail(FALSE);
   endCritical;
 
