@@ -1,4 +1,4 @@
-/*  $Id: pl-error.c,v 1.38 2007/01/12 19:52:55 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -46,6 +46,31 @@ put_name_arity(term_t t, functor_t f)
 }
 
 
+static void
+rewrite_callable(atom_t *expected, term_t actual)
+{ term_t a = 0;
+  int loops = 0;
+
+  while ( PL_is_functor(actual, FUNCTOR_colon2) )
+  { if ( !a )
+     a = PL_new_term_ref();
+
+    PL_get_arg(1, actual, a);
+    if ( !PL_is_atom(a) )
+    { *expected = ATOM_atom;
+      PL_put_term(actual, a);
+      return;
+    } else
+    { PL_get_arg(2, actual, a);
+      PL_put_term(actual, a);
+    }
+
+    if ( ++loops > 100 && !PL_is_acyclic(actual) )
+      break;
+  }
+}
+
+
 int
 PL_error(const char *pred, int arity, const char *msg, int id, ...)
 { Definition caller;
@@ -60,7 +85,8 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
   else
     caller = NULL;
 
-  if ( id == ERR_FILE_OPERATION && !fileerrors )
+  if ( id == ERR_FILE_OPERATION &&
+       !truePrologFlag(PLFLAG_FILEERRORS) )
     fail;
 
   if ( msg == MSG_ERRNO )
@@ -81,10 +107,28 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       err_instantiation:
       PL_unify_atom(formal, ATOM_instantiation_error);
       break;
+    case ERR_MUST_BE_VAR:
+    { int argn = va_arg(args, int);
+      char buf[50];
+      /*term_t bound =*/ va_arg(args, term_t);
+
+      if ( !msg )
+      { Ssprintf(buf, "%d-%s argument",
+		 argn, argn == 1 ? "st" : argn == 2 ? "nd" : "th");
+	msg = buf;
+      }
+
+      PL_unify_term(formal,
+		    PL_FUNCTOR, FUNCTOR_representation_error1,
+		      PL_ATOM, ATOM_variable);
+      break;
+    }
     case ERR_TYPE:			/* ERR_INSTANTIATION if var(actual) */
     { atom_t expected = va_arg(args, atom_t);
       term_t actual   = va_arg(args, term_t);
 
+      if ( expected == ATOM_callable )
+	rewrite_callable(&expected, actual);
       if ( PL_is_variable(actual) && expected != ATOM_variable )
 	goto err_instantiation;
 
@@ -94,6 +138,19 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 		      PL_TERM, actual);
       break;
     }
+    case ERR_CHARS_TYPE:		/* ERR_INSTANTIATION if var(actual) */
+    { const char *expected = va_arg(args, const char*);
+      term_t actual        = va_arg(args, term_t);
+
+      if ( PL_is_variable(actual) && !streq(expected, "variable") )
+	goto err_instantiation;
+
+      PL_unify_term(formal,
+		    PL_FUNCTOR, FUNCTOR_type_error2,
+		      PL_CHARS, expected,
+		      PL_TERM, actual);
+      break;
+    }
     case ERR_AR_TYPE:			/* arithmetic type error */
     { atom_t expected = va_arg(args, atom_t);
       Number num      = va_arg(args, Number);
@@ -175,15 +232,15 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     case ERR_PERMISSION_PROC:
     { atom_t op = va_arg(args, atom_t);
       atom_t type = va_arg(args, atom_t);
-      Definition def = va_arg(args, Definition);
-      term_t pred = PL_new_term_ref();
+      predicate_t pred = va_arg(args, predicate_t);
+      term_t pi = PL_new_term_ref();
 
-      unify_definition(pred, def, 0, GP_NAMEARITY|GP_HIDESYSTEM);
+      PL_unify_predicate(pi, pred, GP_NAMEARITY|GP_HIDESYSTEM);
       PL_unify_term(formal,
 		    PL_FUNCTOR, FUNCTOR_permission_error3,
 		    PL_ATOM, op,
 		    PL_ATOM, type,
-		    PL_TERM, pred);
+		    PL_TERM, pi);
       break;
     }
     case ERR_NOT_IMPLEMENTED_PROC:
@@ -220,7 +277,7 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       term_t actual = PL_new_term_ref();
 
       put_name_arity(actual, f);
-      
+
       PL_unify_term(formal,
 		    PL_FUNCTOR, FUNCTOR_type_error2,
 		      PL_ATOM, ATOM_evaluable,
@@ -246,6 +303,17 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 
       break;
     }
+    case ERR_OCCURS_CHECK:
+    { Word p1  = va_arg(args, Word);
+      Word p2  = va_arg(args, Word);
+
+      PL_unify_term(formal,
+			PL_FUNCTOR, FUNCTOR_occurs_check2,
+			  PL_TERM, wordToTermRef(p1),
+			  PL_TERM, wordToTermRef(p2));
+
+      break;
+    }
     case ERR_TIMEOUT:
     { atom_t op   = va_arg(args, atom_t);
       term_t obj  = va_arg(args, term_t);
@@ -284,7 +352,7 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 	case EMFILE:
 	case ENFILE:
 	  PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_representation_error1,
+			PL_FUNCTOR, FUNCTOR_resource_error1,
 			  PL_ATOM, ATOM_max_files);
 	  break;
 #ifdef EPIPE
@@ -306,7 +374,7 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     case ERR_STREAM_OP:
     { atom_t action = va_arg(args, atom_t);
       term_t stream = va_arg(args, term_t);
-      
+
       PL_unify_term(formal,
 		    PL_FUNCTOR, FUNCTOR_io_error2,
 		      PL_ATOM, action,
@@ -333,7 +401,7 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 		      PL_CHARS, err);
       break;
     }
-    case ERR_NOT_IMPLEMENTED_FEATURE:		/* non-ISO */
+    case ERR_NOT_IMPLEMENTED:		/* non-ISO */
     { const char *what = va_arg(args, const char *);
 
       PL_unify_term(formal,
@@ -502,7 +570,7 @@ tostr(char *buf, const char *fmt, ...)
   va_start(args, fmt);
   Svsprintf(buf, fmt, args);
   va_end(args);
-  
+
   return buf;
 }
 
@@ -511,15 +579,32 @@ tostr(char *buf, const char *fmt, ...)
 		 *	PRINTING MESSAGES	*
 		 *******************************/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+printMessage(atom_t severity, ...)
+
+Calls print_message(severity, term), where  ...   are  arguments  as for
+PL_unify_term(). This predicate saves possible   pending  exceptions and
+restores them to make the call from B_THROW possible.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 void
 printMessage(atom_t severity, ...)
-{ fid_t fid = PL_open_foreign_frame();
-  term_t av = PL_new_term_refs(2);
+{ fid_t fid;
+  term_t ex, av;
   predicate_t pred = PROCEDURE_print_message2;
   va_list args;
 
-  gc_status.blocked++;			/* sometimes called from dangerous */
+  blockGC(PASS_LD1);			/* sometimes called from dangerous */
 					/* places */
+  fid = PL_open_foreign_frame();
+  av = PL_new_term_refs(2);
+
+  if ( exception_term )
+  { ex = PL_copy_term_ref(exception_term);
+    exception_term = 0;
+  } else
+    ex = 0;
+
   va_start(args, severity);
   PL_put_atom(av+0, severity);
   PL_unify_termv(av+1, args);
@@ -532,9 +617,14 @@ printMessage(atom_t severity, ...)
     PL_write_term(Serror, av+1, 1200, 0);
     Sfprintf(Serror, "\n");
   }
-  gc_status.blocked--;
+
+  if ( ex )
+  { PL_put_term(exception_bin, ex);
+    exception_term = exception_bin;
+  }
 
   PL_discard_foreign_frame(fid);
+  unblockGC(PASS_LD1);
 }
 
 
@@ -554,13 +644,6 @@ PL_get_chars_ex(term_t t, char **s, unsigned int flags)
 }
 
 
-int
-PL_get_text_ex(term_t t, PL_chars_t *txt, unsigned int flags)
-{ return PL_get_text(t, txt, flags);
-}
-
-
-
 int
 PL_get_atom_ex(term_t t, atom_t *a)
 { if ( PL_get_atom(t, a) )
@@ -575,6 +658,9 @@ PL_get_integer_ex(term_t t, int *i)
 { if ( PL_get_integer(t, i) )
     succeed;
 
+  if ( PL_is_integer(t) )
+    return PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_int);
+
   return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_integer, t);
 }
 
@@ -584,6 +670,9 @@ PL_get_long_ex(term_t t, long *i)
 { if ( PL_get_long(t, i) )
     succeed;
 
+  if ( PL_is_integer(t) )
+    return PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_long);
+
   return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_integer, t);
 }
 
@@ -593,6 +682,9 @@ PL_get_int64_ex(term_t t, int64_t *i)
 { if ( PL_get_int64(t, i) )
     succeed;
 
+  if ( PL_is_integer(t) )
+    return PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_int64_t);
+
   return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_integer, t);
 }
 
@@ -642,7 +734,7 @@ PL_unify_list_ex(term_t l, term_t h, term_t t)
 
   if ( PL_get_nil(l) )
     fail;
-  
+
   return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, l);
 }
 
@@ -666,7 +758,7 @@ PL_get_list_ex(term_t l, term_t h, term_t t)
 
   if ( PL_get_nil(l) )
     fail;
-  
+
   return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, l);
 }
 
@@ -709,3 +801,12 @@ PL_get_arg_ex(int n, term_t term, term_t arg)
     return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_natural, a);
   }
 }
+
+
+int
+PL_get_module_ex(term_t name, Module *m)
+{ if ( !PL_get_module(name, m) )
+    return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, name);
+
+  succeed;
+}
