@@ -1,4 +1,4 @@
-/*  $Id: pl-fmt.c,v 1.76 2007/06/19 09:05:31 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -76,7 +76,7 @@ update_column(int col, int c)
     case '\b':	return (col <= 0 ? 0 : col - 1);
     default:	return col + 1;
   }
-}   
+}
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -216,7 +216,7 @@ pl_format_predicate(term_t chr, term_t descr)
 
   if ( !format_predicates )
     format_predicates = newHTable(8);
-  
+
   if ( (s = lookupHTable(format_predicates, (void *)(intptr_t)c)) )
     s->value = proc;
   else
@@ -254,7 +254,7 @@ pl_current_format_predicate(term_t chr, term_t descr, control_t h)
 	 unify_definition(descr, ((Procedure)s->value)->definition, 0, 0) )
     { ForeignRedoPtr(e);
     }
-    
+
     PL_rewind_foreign_frame(fid);
   }
 
@@ -274,7 +274,7 @@ format_impl(IOSTREAM *out, term_t format, term_t Args)
   if ( !PL_get_text(format, &fmt, CVT_ALL|BUF_RING) )
     return PL_error("format", 3, NULL, ERR_TYPE, ATOM_text, format);
 
-  if ( (argc = lengthList(args, FALSE)) >= 0 )
+  if ( (argc = (int)lengthList(args, FALSE)) >= 0 )
   { term_t head = PL_new_term_ref();
     int n = 0;
 
@@ -287,7 +287,8 @@ format_impl(IOSTREAM *out, term_t format, term_t Args)
 
     PL_put_term(argv, args);
   }
-  
+
+  startCritical;
   switch(fmt.storage)			/* format can to call-back! */
   { case PL_CHARS_RING:
     case PL_CHARS_STACK:
@@ -299,6 +300,7 @@ format_impl(IOSTREAM *out, term_t format, term_t Args)
 
   rval = do_format(out, &fmt, argc, argv);
   PL_free_text(&fmt);
+  endCritical;
 
   return rval;
 }
@@ -320,7 +322,7 @@ pl_format3(term_t out, term_t format, term_t args)
 		    /*cleanup*/,
 		    DEBUG(1, Sdprintf("Cleanup after throw()\n"));
 		    discardOutputRedirect(&ctx);
-		    PL_rethrow(););
+		    rc = PL_rethrow(););
 
   return rc;
 }
@@ -406,7 +408,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 
 					/* Check for user defined format */
 	  if ( format_predicates &&
-	       (s = lookupHTable(format_predicates, (Void)((intptr_t)c))) )
+	       (s = lookupHTable(format_predicates, (void*)((intptr_t)c))) )
 	  { Procedure proc = (Procedure) s->value;
 	    FunctorDef fdef = proc->definition->functor;
 	    term_t av = PL_new_term_refs(fdef->arity);
@@ -478,7 +480,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  NEED_ARG;
 		  if ( !valueExpression(argv, &n PASS_LD) )
 		  { char f[2];
-		    
+
 		    f[0] = c;
 		    f[1] = EOS;
 		    FMT_ARG(f, argv);
@@ -504,7 +506,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  if ( !valueExpression(argv, &i PASS_LD) ||
 		       !toIntegerNumber(&i, 0) )
 		  { char f[2];
-		    
+
 		    f[0] = c;
 		    f[1] = EOS;
 		    FMT_ARG(f, argv);
@@ -518,7 +520,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  } else
 		  { if ( arg < 1 || arg > 36 )
 		    { term_t r = PL_new_term_ref();
-		      
+
 		      PL_put_integer(r, arg);
 		      Sunlock(fd);
 		      return PL_error(NULL, 0, NULL, ERR_DOMAIN,
@@ -527,7 +529,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		    formatNumber(FALSE, 0, arg, c == 'r', &i, (Buffer)&b);
 		  }
 		  clearNumber(&i);
-		  outstring0(&state, baseBuffer(&b, char));			
+		  outstring0(&state, baseBuffer(&b, char));
 		  discardBuffer(&b);
 		  here++;
 		  break;
@@ -554,7 +556,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  char *str;
 
 	      case 'k':			/* write_canonical */
-		  f = pl_write_canonical; 
+		  f = pl_write_canonical;
 	          goto pl_common;
 	      case 'p':			/* print */
 		  f = pl_print;
@@ -672,9 +674,13 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  if ( str != buf )
 		    free(str);
 
-		  if ( !rval && ex )
+		  if ( !rval )
 		  { Sunlock(fd);
-		    return PL_raise_exception(ex);
+
+		    if ( ex )
+		      return PL_raise_exception(ex);
+		    else
+		      fail;
 		  }
 
 		  SHIFT;
@@ -798,7 +804,7 @@ emit_rubber(format_state *state)
   for(j = 0; s <= e; j++)
   { int chr;
 
-    if ( r->where == j && rn )
+    if ( rn && r->where == j )
     { size_t n;
 
       for(n=0; n<r->size; n++)
@@ -840,14 +846,21 @@ formatNumber(bool split, int div, int radix, bool small, Number i,
 { switch(i->type)
   { case V_INTEGER:
     { int64_t n = i->value.i;
-
-      char tmp[100]; 
-      char *end = &tmp[sizeof(tmp)];
-      char *s = end;			/* i.e. start at the end */
+      char buf[100];
+      char *tmp, *end, *s;
       int before = (div == 0);
       int digits = 0;
       bool negative = FALSE;
 
+      if ( div+3 > (int)sizeof(buf) )	/* 0.000NNNN with div digits after 0. */
+      { tmp = PL_malloc(div+3);
+	end = tmp+div+3;
+      } else
+      { tmp = buf;
+	end = tmp+sizeof(buf);
+      }
+
+      s = end;				/* i.e. start at the end */
       *--s = EOS;
       if ( n < 0 )
       { n = -n;
@@ -867,10 +880,12 @@ formatNumber(bool split, int div, int radix, bool small, Number i,
 	  n /= radix;
 	}
 	if ( negative )
-	  *--s = '-';  
+	  *--s = '-';
       }
 
       addMultipleBuffer(out, s, end-s, char);
+      if ( tmp != buf )
+	PL_free(tmp);
 
       return baseBuffer(out, char);
     }
@@ -942,7 +957,7 @@ formatFloat(int how, int arg, Number f, Buffer out)
     arg = 6;
 
   switch(f->type)
-  { 
+  {
 #ifdef O_GMP
     mpf_t mpf;
     case V_MPZ:
@@ -954,7 +969,7 @@ formatFloat(int how, int arg, Number f, Buffer out)
       int size;
       int written;
       int fbits;
-      
+
       switch(how)
       { case 'f':
 	case 'g':
@@ -970,10 +985,10 @@ formatFloat(int how, int arg, Number f, Buffer out)
 	default:
 	  fbits = 4*arg;
       }
-      
+
       mpf_init2(mpf, fbits);
       mpf_set_q(mpf, f->value.mpq);
-      
+
     print:
       Ssprintf(tmp, "%%.%dF%c", arg, how);
       size = 0;
@@ -991,9 +1006,9 @@ formatFloat(int how, int arg, Number f, Buffer out)
     }
 #endif
     case V_INTEGER:
-      promoteToRealNumber(f);
+      promoteToFloatNumber(f);
       /*FALLTHROUGH*/
-    case V_REAL:
+    case V_FLOAT:
     { char tmp[12];
 
       if ( arg > 256 )
