@@ -1,4 +1,4 @@
-/*  $Id: pl-gmp.c,v 1.22 2007/05/08 15:40:48 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -38,31 +38,154 @@ static mpz_t MPZ_MIN_LONG;		/* Prolog int64_t integers */
 static mpz_t MPZ_MAX_LONG;
 #endif
 
-#if 0
+
 		 /*******************************
 		 *	MEMMORY MANAGEMENT	*
 		 *******************************/
 
-/*See also clearNumber()*/
+#if O_MY_GMP_ALLOC
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+GMP doesn't (yet) allow for handling  memory overflows. You can redefine
+the allocation handles, but you are not  allowed to return NULL or abort
+the execution using longjmp(). As our  normal   GMP  numbers live on the
+global stack, we however can  cleanup   the  temporary  numbers that are
+created during the Prolog function evaluation  and use longjmp() through
+STACK_OVERFLOW_THROW.   Patrick Pelissier acknowledged this should work.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static int
+gmp_too_big()
+{ GET_LD
+
+  DEBUG(1, Sdprintf("Signalling GMP overflow\n"));
+
+  return (int)outOfStack((Stack)&LD->stacks.global, STACK_OVERFLOW_THROW);
+}
+
+#define TOO_BIG_GMP(n) ((n) > 1000 && (n) > (size_t)limitStack(global))
 
 static void *
 mp_alloc(size_t bytes)
-{
+{ GET_LD
+  mp_mem_header *mem;
+
+  if ( LD->gmp.persistent )
+    return malloc(bytes);
+
+  if ( TOO_BIG_GMP(bytes) ||
+       !(mem = malloc(sizeof(mp_mem_header)+bytes)) )
+  { gmp_too_big();
+    pl_abort(ABORT_FATAL);
+    return NULL;			/* make compiler happy */
+  }
+
+  GMP_LEAK_CHECK(LD->gmp.allocated += bytes);
+
+  mem->next = NULL;
+  mem->context = LD->gmp.context;
+  if ( LD->gmp.tail )
+  { mem->prev = LD->gmp.tail;
+    LD->gmp.tail->next = mem;
+    LD->gmp.tail = mem;
+  } else
+  { mem->prev = NULL;
+    LD->gmp.head = LD->gmp.tail = mem;
+  }
+  DEBUG(9, Sdprintf("GMP: alloc %ld@%p\n", bytes, &mem[1]));
+
+  return &mem[1];
 }
 
 
 static void *
 mp_realloc(void *ptr, size_t oldsize, size_t newsize)
-{
+{ GET_LD
+  mp_mem_header *oldmem, *newmem;
+
+  if ( LD->gmp.persistent )
+    return realloc(ptr, newsize);
+
+  oldmem = ((mp_mem_header*)ptr)-1;
+  if ( TOO_BIG_GMP(newsize) ||
+       !(newmem = realloc(oldmem, sizeof(mp_mem_header)+newsize)) )
+  { gmp_too_big();
+    pl_abort(ABORT_FATAL);
+    return NULL;			/* make compiler happy */
+  }
+
+  if ( oldmem != newmem )		/* re-link if moved */
+  { if ( newmem->prev )
+      newmem->prev->next = newmem;
+    else
+      LD->gmp.head = newmem;
+
+    if ( newmem->next )
+      newmem->next->prev = newmem;
+    else
+      LD->gmp.tail = newmem;
+  }
+
+  GMP_LEAK_CHECK(LD->gmp.allocated -= oldsize;
+		 LD->gmp.allocated += newsize);
+  DEBUG(9, Sdprintf("GMP: realloc %ld@%p --> %ld@%p\n", oldsize, ptr, newsize, &newmem[1]));
+
+  return &newmem[1];
 }
 
 
 static void
-free(void *ptr, size_t size)
-{
+mp_free(void *ptr, size_t size)
+{ GET_LD
+  mp_mem_header *mem;
+
+  if ( LD->gmp.persistent )
+  { free(ptr);
+    return;
+  }
+
+  mem = ((mp_mem_header*)ptr)-1;
+
+  if ( mem == LD->gmp.head )
+  { LD->gmp.head = LD->gmp.head->next;
+    if ( LD->gmp.head )
+      LD->gmp.head->prev = NULL;
+    else
+      LD->gmp.tail = NULL;
+  } else if ( mem == LD->gmp.tail )
+  { LD->gmp.tail = LD->gmp.tail->prev;
+    LD->gmp.tail->next = NULL;
+  } else
+  { mem->prev->next = mem->next;
+    mem->next->prev = mem->prev;
+  }
+
+  free(mem);
+  DEBUG(9, Sdprintf("GMP: free: %ld@%p\n", size, ptr));
+  GMP_LEAK_CHECK(LD->gmp.allocated -= size);
+}
+
+
+void
+mp_cleanup(ar_context *ctx)
+{ GET_LD
+  mp_mem_header *mem, *next;
+
+  if ( LD->gmp.context )
+  { for(mem=LD->gmp.head; mem; mem=next)
+    { next = mem->next;
+      if ( mem->context == LD->gmp.context )
+      { DEBUG(9, Sdprintf("GMP: cleanup of %p\n", &mem[1]));
+	mp_free(&mem[1], 0);
+      }
+    }
+  }
+
+  LD->gmp.context = ctx->parent;
 }
 #endif
 
+
 #ifdef __WINDOWS__
 #undef isascii
 int
@@ -80,30 +203,31 @@ globalMPZ() pushes an mpz type GMP  integer   onto  the local stack. The
 saved version is the _mp_size field, followed by the limps.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-int
-wordSizeofMPZ(mpz_t mpz)
-{ int size = sizeof(mp_limb_t)*abs(mpz->_mp_size);
-  
-  return (size+sizeof(word)-1)/sizeof(word);
-}
-
-
 static word
 globalMPZ(mpz_t mpz)
 { GET_LD
+  Word p;
+  word r;
+
+  size_t size = sizeof(mp_limb_t)*abs(mpz->_mp_size);
+  size_t wsz  = (size+sizeof(word)-1)/sizeof(word);
+  word m      = mkIndHdr(wsz+1, TAG_INTEGER);
+
+  if ( wsizeofInd(m) != wsz+1 )
+  { PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_integer);
+    return 0;
+  }
+
+  p = allocGlobal(wsz+3);
+  r = consPtr(p, TAG_INTEGER|STG_GLOBAL);
 
-  int size = sizeof(mp_limb_t)*abs(mpz->_mp_size);
-  int wsz  = (size+sizeof(word)-1)/sizeof(word);
-  Word p   = allocGlobal(wsz+3);
-  word r   = consPtr(p, TAG_INTEGER|STG_GLOBAL);
-  word m   = mkIndHdr(wsz+1, TAG_INTEGER);
 
   *p++     = m;
   p[wsz]   = 0L;			/* pad out */
   p[wsz+1] = m;
   *p++     = (word)mpz->_mp_size;
   memcpy(p, mpz->_mp_d, size);
-  
+
   return r;
 }
 
@@ -210,7 +334,7 @@ loadMPZFromCharp(const char *data, Word r, Word *store)
     size = -size;
   } else
     neg = FALSE;
-  
+
   limpsize = (size+sizeof(mp_limb_t)-1)/sizeof(mp_limb_t);
   wsize = (limpsize*sizeof(mp_limb_t)+sizeof(word)-1)/sizeof(word);
   p = *store;
@@ -264,7 +388,11 @@ skipMPZOnCharp(const char *data)
 
 static void
 mpz_init_set_si64(mpz_t mpz, int64_t i)
-{ DEBUG(2, Sdprintf("Converting " INT64_FORMAT " to MPZ\n", i));
+{
+#if SIZEOF_LONG == 8
+  mpz_init_set_si(mpz, (long)i);
+#else
+  DEBUG(2, Sdprintf("Converting " INT64_FORMAT " to MPZ\n", i));
 
   if ( i >= LONG_MIN && i <= LONG_MAX )
   { mpz_init_set_si(mpz, (long)i);
@@ -278,8 +406,8 @@ mpz_init_set_si64(mpz_t mpz, int64_t i)
       mpz_neg(mpz, mpz);
     }
   }
-
   DEBUG(2, gmp_printf("\t--> %Zd\n", mpz));
+#endif
 }
 
 
@@ -304,7 +432,7 @@ promoteToMPZNumber(number *n)
       n->value.mpz[0] = mpz[0];
       break;
     }
-    case V_REAL:
+    case V_FLOAT:
       mpz_init_set_d(n->value.mpz, n->value.f);
       n->type = V_MPZ;
       break;
@@ -326,7 +454,7 @@ promoteToMPQNumber(number *n)
     }
     case V_MPQ:
       break;
-    case V_REAL:
+    case V_FLOAT:
     { double v = n->value.f;
 
       n->type = V_MPQ;
@@ -338,6 +466,43 @@ promoteToMPQNumber(number *n)
 }
 
 
+		 /*******************************
+		 *		RW		*
+		 *******************************/
+
+void
+ensureWritableNumber(Number n)
+{ switch(n->type)
+  { case V_MPZ:
+      if ( !n->value.mpz->_mp_alloc )
+      { mpz_t tmp;
+
+	tmp[0] = n->value.mpz[0];
+	mpz_init_set(n->value.mpz, tmp);
+	break;
+      }
+    case V_MPQ:
+    { if ( !mpq_numref(n->value.mpq)->_mp_alloc )
+      { mpz_t tmp;
+
+	tmp[0] = mpq_numref(n->value.mpq)[0];
+	mpz_init_set(mpq_numref(n->value.mpq), tmp);
+      }
+      if ( !mpq_denref(n->value.mpq)->_mp_alloc )
+      { mpz_t tmp;
+
+	tmp[0] = mpq_denref(n->value.mpq)[0];
+	mpz_init_set(mpq_denref(n->value.mpq), tmp);
+      }
+      break;
+    }
+    default:
+      break;
+  }
+}
+
+
+
 		 /*******************************
 		 *	       CLEAR		*
 		 *******************************/
@@ -351,7 +516,7 @@ numbers that are created by GMP only.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
-clearNumber(Number n)
+clearGMPNumber(Number n)
 { switch(n->type)
   { case V_MPZ:
       if ( n->value.mpz->_mp_alloc )
@@ -375,14 +540,22 @@ clearNumber(Number n)
 
 void
 initGMP()
-{ mpz_init_set_si64(MPZ_MIN_TAGGED, PLMINTAGGEDINT);
-  mpz_init_set_si64(MPZ_MAX_TAGGED, PLMAXTAGGEDINT);
-  mpz_init_set_si64(MPZ_MIN_PLINT, PLMININT);
-  mpz_init_set_si64(MPZ_MAX_PLINT, PLMAXINT);
+{ if ( !GD->gmp.initialised )
+  { GD->gmp.initialised = TRUE;
+
+    mpz_init_set_si64(MPZ_MIN_TAGGED, PLMINTAGGEDINT);
+    mpz_init_set_si64(MPZ_MAX_TAGGED, PLMAXTAGGEDINT);
+    mpz_init_set_si64(MPZ_MIN_PLINT, PLMININT);
+    mpz_init_set_si64(MPZ_MAX_PLINT, PLMAXINT);
 #if SIZEOF_LONG < SIZEOF_VOIDP
-  mpz_init_set_si64(MPZ_MIN_LONG, LONG_MIN);
-  mpz_init_set_si64(MPZ_MAX_LONG, LONG_MAX);
+    mpz_init_set_si64(MPZ_MIN_LONG, LONG_MIN);
+    mpz_init_set_si64(MPZ_MAX_LONG, LONG_MAX);
 #endif
+#ifdef O_MY_GMP_ALLOC
+    if ( !GD->gmp.keep_alloc_functions )
+      mp_set_memory_functions(mp_alloc, mp_realloc, mp_free);
+#endif
+  }
 }
 
 		 /*******************************
@@ -442,7 +615,7 @@ put_mpz(mpz_t mpz)
     return consInt(v);
   } else if ( mpz_to_int64(mpz, &v) )
   { GET_LD
-    
+
 #if SIZEOF_LONG < SIZEOF_VOIDP
     return makeNum(v);
 #else
@@ -456,10 +629,8 @@ put_mpz(mpz_t mpz)
 #endif /*O_GMP*/
 
 word
-put_number(Number n)
-{ GET_LD
-
-  switch(n->type)
+put_number__LD(Number n ARG_LD)
+{ switch(n->type)
   { case V_INTEGER:
     { word w = consInt(n->value.i);
 
@@ -475,18 +646,25 @@ put_number(Number n)
     { if ( mpz_cmp_ui(mpq_denref(n->value.mpq), 1L) == 0 )
       { return put_mpz(mpq_numref(n->value.mpq));
       } else
-      { Word p = allocGlobal(3);
-	
+      { word num, den;
+	Word p;
+
+	if ( !(num=put_mpz(mpq_numref(n->value.mpq))) ||
+	     !(den=put_mpz(mpq_denref(n->value.mpq))) )
+	  fail;
+
+	p = allocGlobal(3);
+
 	p[0] = FUNCTOR_rdiv2;
-	p[1] = put_mpz(mpq_numref(n->value.mpq));
-	p[2] = put_mpz(mpq_denref(n->value.mpq));
+	p[1] = num;
+	p[2] = den;
 
 	return consPtr(p, TAG_COMPOUND|STG_GLOBAL);
       }
     }
 #endif
-    case V_REAL:
-      return globalReal(n->value.f);
+    case V_FLOAT:
+      return globalFloat(n->value.f);
   }
 
   assert(0);
@@ -502,12 +680,13 @@ PL_unify_number(term_t t, Number n)
   deRef(p);
 
   if ( canBind(*p) )
-  { word w = put_number(n);
+  { word w;
+
+    if ( !(w = put_number(n)) )
+      fail;
 
-#ifdef O_SHIFT_STACKS
     p = valTermRef(t);			/* put_number can shift the stacks */
     deRef(p);
-#endif
 
     bindConst(p, w);
     succeed;
@@ -523,7 +702,8 @@ PL_unify_number(term_t t, Number n)
 	int rc;
 
 	get_integer(*p, &n2);
-	rc = ar_compare(n, &n2, EQ);
+	same_type_numbers(n, &n2);
+	rc = ar_compare_eq(n, &n2);
 	clearNumber(&n2);
 
 	return rc;
@@ -531,15 +711,17 @@ PL_unify_number(term_t t, Number n)
       break;
 #ifdef O_GMP
     case V_MPQ:
-    { term_t q = PL_new_term_ref();
-      
-      *valTermRef(q) = put_number(n);
-      return PL_unify(t, q);
+    { word w;
+
+      if ( !(w=put_number(n)) )
+	fail;
+
+      return _PL_unify_atomic(t, w);
     }
 #endif
-    case V_REAL:
-      if ( isReal(*p) )
-	return n->value.f == valReal(*p);
+    case V_FLOAT:
+      if ( isFloat(*p) )
+	return n->value.f == valFloat(*p);
       break;
   }
 
@@ -552,8 +734,8 @@ get_number(word w, Number n ARG_LD)
 { if ( isInteger(w) )
   { get_integer(w, n);
   } else
-  { n->type = V_REAL;
-    n->value.f = valReal(w);
+  { n->type = V_FLOAT;
+    n->value.f = valFloat(w);
   }
 }
 
@@ -568,9 +750,9 @@ PL_get_number(term_t t, Number n)
   { get_integer(*p, n);
     succeed;
   }
-  if ( isReal(*p) )
-  { n->value.f = valReal(*p);
-    n->type = V_REAL;
+  if ( isFloat(*p) )
+  { n->value.f = valFloat(*p);
+    n->type = V_FLOAT;
     succeed;
   }
 
@@ -583,11 +765,11 @@ PL_get_number(term_t t, Number n)
 		 *******************************/
 
 void
-promoteToRealNumber(Number n)
+promoteToFloatNumber(Number n)
 { switch(n->type)
   { case V_INTEGER:
       n->value.f = (double)n->value.i;
-      n->type = V_REAL;
+      n->type = V_FLOAT;
       break;
 #ifdef O_GMP
     case V_MPZ:
@@ -595,7 +777,7 @@ promoteToRealNumber(Number n)
 
       clearNumber(n);
       n->value.f = val;
-      n->type = V_REAL;
+      n->type = V_FLOAT;
       break;
     }
     case V_MPQ:
@@ -603,11 +785,11 @@ promoteToRealNumber(Number n)
 
       clearNumber(n);
       n->value.f = val;
-      n->type = V_REAL;
+      n->type = V_FLOAT;
       break;
     }
 #endif
-    case V_REAL:
+    case V_FLOAT:
       break;
   }
 }
@@ -626,8 +808,8 @@ promoteNumber(Number n, numtype t)
       promoteToMPQNumber(n);
       break;
 #endif
-    case V_REAL:
-      promoteToRealNumber(n);
+    case V_FLOAT:
+      promoteToFloatNumber(n);
       break;
   }
 }
@@ -640,18 +822,12 @@ same_type_numbers(n1, n2)
     total ordering between the number types.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#ifndef max
-#define max(a,b) ((a) >= (b) ? (a) : (b))
-#endif
-
 void
-same_type_numbers(Number n1, Number n2)
-{ if ( n1->type != n2->type )
-  { if ( (int)n1->type > (int)n2->type )
-      promoteNumber(n2, n1->type);
-    else
-      promoteNumber(n1, n2->type);
-  }
+make_same_type_numbers(Number n1, Number n2)
+{ if ( (int)n1->type > (int)n2->type )
+    promoteNumber(n2, n1->type);
+  else
+    promoteNumber(n1, n2->type);
 }
 
 
@@ -684,7 +860,7 @@ cmpNumbers(Number n1, Number n2)
       return rc < 0 ? LESS : rc == 0 ? EQUAL : GREATER;
     }
 #endif
-    case V_REAL:
+    case V_FLOAT:
       return n1->value.f  < n2->value.f ? LESS :
 	     n1->value.f == n2->value.f ? EQUAL : GREATER;
   }
@@ -712,7 +888,7 @@ cpNumber(Number to, Number from)
       mpq_set(to->value.mpq, from->value.mpq);
       break;
 #endif
-    case V_REAL:
+    case V_FLOAT:
       to->value.f = from->value.f;
   }
 }
@@ -728,15 +904,24 @@ int
 PL_get_mpz(term_t t, mpz_t mpz)
 { GET_LD
   Word p = valTermRef(t);
-  
+
   deRef(p);
   if ( isInteger(*p) )
   { number n;
 
     get_integer(*p, &n);
-    promoteToMPZNumber(&n);
-    mpz_set(mpz, n.value.mpz);
-    clearNumber(&n);
+    switch(n.type)
+    { case V_INTEGER:
+	promoteToMPZNumber(&n);
+        mpz_set(mpz, n.value.mpz);
+	clearNumber(&n);
+	break;
+      case V_MPZ:
+	mpz_set(mpz, n.value.mpz);
+        break;
+      default:
+	assert(0);
+    }
 
     return TRUE;
   }
@@ -750,7 +935,7 @@ PL_get_mpq(term_t t, mpq_t mpq)
 { if ( PL_is_rational(t) )
   { GET_LD
     number n;
-    
+
     if ( valueExpression(t, &n PASS_LD) )
     { switch(n.type)
       { case V_INTEGER:
@@ -772,7 +957,7 @@ PL_get_mpq(term_t t, mpq_t mpq)
 	  ;
       }
       clearNumber(&n);
-    }	 
+    }
   }
 
   return FALSE;
