@@ -5,7 +5,7 @@
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -51,22 +51,8 @@
 #include <parms.h>			/* pick from the working dir */
 #endif
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-First, include config.h or, if MD is  specified, this file.  This allows
-for -DMD="config/win64.h"
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-#ifndef CONFTEST
-#ifdef MD
-#include MD
-#else
-#include <config.h>
-#endif
-#endif
-
-#if defined(O_PLMT) && !defined(_REENTRANT)
-#define _REENTRANT 1
-#endif
+#define PL_KERNEL		1
+#include "pl-builtin.h"
 
 #ifdef HAVE_DMALLOC_H
 #include <dmalloc.h>			/* Use www.dmalloc.com debugger */
@@ -93,6 +79,8 @@ handy for it someone wants to add a data type to the system.
       has been taken out.
   O_COMPILE_ARITH
       Include arithmetic compiler (compiles is/2, >/2, etc. into WAM).
+  O_COMPILE_IS
+      Compile Var = Value in the body.
   O_PROLOG_FUNCTIONS
       Include evaluable Prolog functions into the arithmetic module.
   O_BLOCK
@@ -129,19 +117,18 @@ handy for it someone wants to add a data type to the system.
       Use GNU gmp library for infinite precision arthmetic
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define PL_KERNEL		1
 #define O_COMPILE_OR		1
 #define O_SOFTCUT		1
 #define O_COMPILE_ARITH		1
+#define O_COMPILE_IS		1
 #define O_STRING		1
 #define O_PROLOG_FUNCTIONS	1
 #define O_BLOCK			1
 #define O_CATCHTHROW		1
-#define O_INLINE_FOREIGNS	1
 #define O_DEBUGGER		1
 #define O_INTERRUPT		1
 #define O_DESTRUCTIVE_ASSIGNMENT 1
-#define O_HASHTERM		1
+#define O_TERMHASH		1
 #define O_LIMIT_DEPTH		1
 #define O_SAFE_SIGNALS		1
 #define O_LOGICAL_UPDATE	1
@@ -255,19 +242,6 @@ void *alloca ();
 #undef O_LARGEFILES
 #endif
 
-#if HAVE_XOS_H
-#include <xos.h>
-#endif
-#ifdef HAVE_UXNT_H
-#include <uxnt.h>
-#endif
-
-#ifdef __WINDOWS__
-#include <winsock2.h>
-#endif
-#include "pl-mutex.h"
-#include "pl-stream.h"
-
 #include <sys/types.h>
 #include <setjmp.h>
 #ifdef ASSERT_H_REQUIRES_STDIO_H
@@ -350,12 +324,6 @@ A common basis for C keywords.
 #define volatile
 #endif
 
-#ifdef HAVE_VISIBILITY_ATTRIBUTE
-#define SO_LOCAL __attribute__((visibility("hidden")))
-#else
-#define SO_LOCAL
-#endif
-
 #if defined(__GNUC__) && !defined(NORETURN)
 #define NORETURN __attribute__ ((noreturn))
 #else
@@ -373,14 +341,10 @@ A common basis for C keywords.
 #define forwards static		/* forwards function declarations */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Booleans,  addresses,  strings  and other   goodies.   Note that  ANSI
-compilers have  `Void'.   This should  be  made part  of  the  general
-platform as well.
+Booleans,  addresses,  strings  and other   goodies.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 typedef int			bool;
-typedef double			real;
-typedef void *			Void;
 
 #if __GNUC__ && !__STRICT_ANSI__
 #define LocalArray(t, n, s)	t n[s]
@@ -407,10 +371,6 @@ typedef void *			caddress;
 #define ESC			((char) 27)
 #define streq(s, q)		((strcmp((s), (q)) == 0))
 
-#define CHAR_MODE 0		/* See PL_unify_char() */
-#define CODE_MODE 1
-#define BYTE_MODE 2
-
 #ifndef abs
 #define abs(x)			((x) < 0 ? -(x) : (x))
 #endif
@@ -436,7 +396,7 @@ them.  Descriptions:
 	use malloc() for allocating the stack as the local and global
 	stack need to be apart by this amount.  Also, an interrupt
 	skips this amount of stack.
-	
+
 	* MAXVARIABLES
 	Maximum number of variables in a clause.  May be increased
 	further, but the global local stacks need to be separated
@@ -490,223 +450,13 @@ sizes  of  the  hash  tables are defined.  Note that these should all be
 #define FLAGHASHSIZE		16	/* global flag/3 table */
 #define ARITHHASHSIZE		64	/* arithmetic function table */
 
-#define TABLE_UNLOCKED		0x10000000L /* do not create mutex for table */
-#define TABLE_MASK		0xf0000000UL
-
-#define pointerHashValue(p, size) ((((intptr_t)(p) >> LMASK_BITS) ^ \
-				    ((intptr_t)(p) >> (LMASK_BITS+5)) ^ \
-				    ((intptr_t)(p))) & \
-				   ((size)-1))
-
-#define TABLE_REF_MASK		0x1UL
-#define isTableRef(p)		((uintptr_t)(p) & TABLE_REF_MASK)
-#define makeTableRef(p)		((void*)((uintptr_t)(p) | TABLE_REF_MASK))
-#define unTableRef(s, p)	(*((s*)((uintptr_t)(p) & ~TABLE_REF_MASK)))
-
-#define return_next_table(t, v, clean) \
-	{ for((v) = (v)->next; isTableRef(v) && (v); (v) = unTableRef(t, v)) \
-	  if ( (v) == (t)NULL ) \
-	  { clean; \
-	    succeed; \
-	  } \
-	  ForeignRedoPtr(v); \
-	}
-
-#define for_table(ht, s, code) \
-	{ int _k; \
-	  PL_LOCK(L_TABLE); \
-	  for(_k = 0; _k < (ht)->buckets; _k++) \
-	  { Symbol _n, s; \
-	    for(s=(ht)->entries[_k]; s; s = _n) \
-	    { _n = s->next; \
-	      code; \
-	    } \
-	  } \
-          PL_UNLOCK(L_TABLE); \
-	}
-#define for_unlocked_table(ht, s, code) \
-	{ int _k; \
-	  for(_k = 0; _k < (ht)->buckets; _k++) \
-	  { Symbol _n, s; \
-	    for(s=(ht)->entries[_k]; s; s = _n) \
-	    { _n = s->next; \
-	      code; \
-	    } \
-	  } \
-	}
+#include "pl-table.h"
 
 /* Definition->indexPattern is set to NEED_REINDEX if the definition's index
    pattern needs to be recomputed */
-#define NEED_REINDEX (1UL << (LONGBITSIZE-1))
+#define NEED_REINDEX (1U << (INTBITSIZE-1))
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Foreign language interface definitions.  Note that these macros MUST  be
-consistent  with  the  definitions  in  pl-itf.h, which is included with
-users foreign language code.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-typedef enum
-{ FRG_FIRST_CALL = 0,		/* Initial call */
-  FRG_CUTTED     = 1,		/* Context was cutted */
-  FRG_REDO	 = 2		/* Normal redo */
-} frg_code;
-
-typedef struct foreign_context *control_t;
-
-struct foreign_context
-{ uintptr_t		context;	/* context value */
-  frg_code		control;	/* FRG_* action */
-  struct PL_local_data *engine;		/* invoking engine */
-};
-
-#define FRG_REDO_MASK	0x00000003L
-#define FRG_REDO_BITS	2
-#define REDO_INT	0x02		/* Returned an integer */
-#define REDO_PTR	0x03		/* returned a pointer */
-
-#define ForeignRedoIntVal(v)	(((uintptr_t)(v)<<FRG_REDO_BITS)|REDO_INT)
-#define ForeignRedoPtrVal(v)	(((uintptr_t)(v))|REDO_PTR)
-
-#define ForeignRedoInt(v)	return ForeignRedoIntVal(v)
-#define ForeignRedoPtr(v)	return ForeignRedoPtrVal(v)
-
-#define ForeignControl(h)	((h)->control)
-#define ForeignContextInt(h)	((intptr_t)(h)->context)
-#define ForeignContextPtr(h)	((void *)(h)->context)
-#define ForeignEngine(h)	((h)->engine)
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Virtual machine instruction declarations.  Prefixes:
-
-  I_	General instructions
-  B_	Body specific version
-  H_	Head specific versin
-  A_	Arithmetic compilation specific
-  C_	Control (compilation of ;/2, etc.)
-
-Numbering these things is arbitrary,  but  for  fast  operation  of  the
-switch  in  pl-wam.c,  numbering  should start at 0 and be without gaps.
-I_HIGHEST must be made equal to the highest  value  of  the  instruction
-codes.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-#define I_NOP		((code) 0)		/* nop */
-#define I_ENTER		((code) 1)		/* enter body */
-#define I_CALL		((code) 2)		/* call procedure */
-#define I_DEPART	((code) 3)		/* last call of procedure */
-#define I_EXIT		((code) 4)		/* exit procedure */
-#define B_FUNCTOR	((code) 5)		/* start functor */
-#define B_RFUNCTOR	((code) 6)		/* start functor */
-#define H_FUNCTOR	((code) 7)
-#define H_RFUNCTOR	((code) 8)
-#define I_POPF		((code) 9)		/* end functor */
-#define B_VAR		((code)10)		/* variable */
-#define H_VAR		((code)11)
-#define B_CONST		((code)12)		/* constant (atomic) */
-#define H_CONST		((code)13)
-#define B_STRING	((code)14)		/* String in body */
-#define H_STRING	((code)15)		/* String in the head */
-#define B_MPZ		((code)16)		/* String in body */
-#define H_MPZ		((code)17)		/* String in the head */
-#define B_INTEGER	((code)18)		/* bignum in the head */
-#define H_INTEGER	((code)19)		/* bignum in the body */
-#define B_INT64		((code)20)		/* bignum in the head */
-#define H_INT64		((code)21)		/* bignum in the body */
-#define B_FLOAT		((code)22)		/* double in the head */
-#define H_FLOAT		((code)23)		/* double in the body */
-
-#define B_FIRSTVAR	((code)24)		/* first occurrence of var */
-#define H_FIRSTVAR	((code)25)
-#define B_VOID		((code)26)		/* anonimous variables */
-#define H_VOID		((code)27)
-#define B_ARGFIRSTVAR	((code)28)		/* body vars nested in funct */
-#define B_ARGVAR	((code)29)
-
-#define H_NIL		((code)30)		/* [] in the head */
-#define B_NIL		((code)31)		/* [] in the body */
-#define H_LIST		((code)32)		/* ./2 in the head */
-#define H_RLIST		((code)33)		/* right-recursive list */
-#define B_LIST		((code)34)		/* ./2 in the body */
-#define B_RLIST		((code)35)		/* right-recursive list */
-
-#define B_VAR0		((code)36)		/* B_VAR 0 */
-#define B_VAR1		((code)37)		/* B_VAR 1 */
-#define B_VAR2		((code)38)		/* B_VAR 2 */
-
-#define I_USERCALL0	((code)39)		/* variable in body (call/1) */
-#define I_USERCALLN	((code)40)		/* call/[2...] */
-#define I_CUT		((code)41)		/* ! */
-#define I_APPLY		((code)42)		/* apply/2 */
-
-#if O_COMPILE_ARITH
-#define A_ENTER		((code)43)		/* start arithmetic sequence */
-#define A_INTEGER	((code)44)		/* 32-bit signed int */
-#define A_INT64		((code)45)		/* 64-bit signed int */
-#define A_MPZ		((code)46)		/* GMP mpz_t */
-#define A_DOUBLE	((code)47)		/* 64-bit double */
-#define A_VAR0		((code)48)		/* variable-0 */
-#define A_VAR1		((code)49)		/* variable-1 */
-#define A_VAR2		((code)50)		/* variable-2 */
-#define A_VAR		((code)51)		/* variable-n */
-#define A_FUNC0		((code)52)		/* nullary arithmic function */
-#define A_FUNC1		((code)53)		/* unary arithmic function */
-#define A_FUNC2		((code)54)		/* binary arithmic function */
-#define A_FUNC		((code)55)		/* n-ary arithmic function */
-#define A_LT		((code)56)		/* < */
-#define A_GT		((code)57)		/* > */
-#define A_LE		((code)58)		/* =< */
-#define A_GE		((code)59)		/* >= */
-#define A_EQ		((code)60)		/* =:= */
-#define A_NE		((code)61)		/* =\= */
-#define A_IS		((code)62)		/* is */
-#endif /* O_COMPILE_ARITH */
-
-#if O_COMPILE_OR
-#define C_OR		((code)63)		/* In-clause backtract point */
-#define C_JMP		((code)64)		/* Jump over code */
-#define C_MARK		((code)65)		/* Sub-clause cut mark */
-#define C_CUT		((code)66)		/* cut to corresponding mark */
-#define C_IFTHENELSE	((code)67)		/* if-then-else start */
-#define C_VAR		((code)68)		/* make a variable */
-#define C_END		((code)69)		/* dummy to help decompiler */
-#define C_NOT		((code)70)		/* same as C_IFTHENELSE */
-#define C_FAIL		((code)71)		/* fail */
-#endif /* O_COMPILE_OR */
-
-#if O_BLOCK
-#define I_CUT_BLOCK	((code)72)		/* !(block) */
-#define B_EXIT		((code)73)		/* exit(block, rval) */
-#endif /*O_BLOCK*/
-
-#if O_INLINE_FOREIGNS
-#define I_CALL_FV0	((code)74)		/* call foreign, no args */
-#define I_CALL_FV1	((code)75)		/* call foreign, 1 var arg */
-#define I_CALL_FV2	((code)76)		/* call foreign, 2 var args */
-#endif /*O_INLINE_FOREIGNS*/
-
-#define I_FAIL		((code)77)		/* fail */
-#define I_TRUE		((code)78)		/* true */
-
-#ifdef O_SOFTCUT
-#define C_SOFTIF	((code)79)		/* Start A *-> B ; C */
-#define C_SOFTCUT	((code)80)		/* `Cut' of A *-> B ; C */
-#endif /*O_SOFTCUT*/
-
-#define I_EXITFACT	((code)81)		/* exit from a fact */
-#define D_BREAK		((code)82)		/* Debugger break-point */
-
-#if O_CATCHTHROW
-#define I_CATCH		((code)83)		/* $catch (catch/3) */
-#define I_EXITCATCH	((code)84)		/* $exit_catch (catch/3) */
-#define B_THROW		((code)85)		/* throw(Exception) */
-#endif
-
-#define I_CONTEXT	((code)86)		/* Push context module */
-#define C_LCUT		((code)87)		/* ! local in \+ and -> */
-#define I_CALLCLEANUP	((code)88)		/* $call_cleanup */
-#define I_EXITCLEANUP	((code)89)		/* $exit_cleanup */
-
-#define I_HIGHEST	((code)89)		/* largest WAM code !!! */
+#include "pl-vmi.h"
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Arithmetic comparison
@@ -773,10 +523,13 @@ stored.  We call the latter type of data `indirect'.
 Below is a description of the  representation  used  for  each  type  of
 Prolog data:
 
+***TBD*** This is totally out of date.  The datatypes are accessed using
+macros defined in pl-data.h.
+
 INTEGER
     Integers are stored in the  27  remaining  bits  of  a  word.   This
     implies they are limited to +- 2^26.
-REAL
+FLOAT
     For a real, the 27 bits are a pointer to a 8 byte unit on the global
     stack.  For both words of the 8 byte unit, the top 3  and  bottom  2
     bits  are  reserved  for identification and garbage collection.  The
@@ -845,6 +598,7 @@ typedef uintptr_t		term_t;		/* external term-reference */
 typedef uintptr_t		word;		/* Anonymous 4 byte object */
 typedef word *			Word;		/* a pointer to anything */
 typedef word			atom_t;		/* encoded atom */
+typedef word			functor_t;	/* encoded functor */
 typedef uintptr_t		code WORD_ALIGNED; /* bytes codes */
 typedef code *			Code;		/* pointer to byte codes */
 typedef int			Char;		/* char that can pass EOF */
@@ -867,9 +621,6 @@ typedef struct recordList *	RecordList;	/* list of these */
 typedef struct module *		Module;		/* predicate modules */
 typedef struct sourceFile *	SourceFile;	/* file adminitration */
 typedef struct list_cell *	ListCell;	/* Anonymous list */
-typedef struct table *		Table;		/* (numeric) hash table */
-typedef struct symbol *		Symbol;		/* symbol of hash table */
-typedef struct table_enum *	TableEnum; 	/* Enumerate table entries */
 typedef struct localFrame *	LocalFrame;	/* environment frame */
 typedef struct local_definitions *LocalDefinitions; /* thread-local preds */
 typedef struct choice *		Choice;		/* Choice-point */
@@ -905,16 +656,16 @@ typedef uintptr_t PL_fid_t;		/* external foreign context-id */
 
 typedef enum
 { V_INTEGER,				/* integer (64-bit) value */
-#ifdef O_GMP    
+#ifdef O_GMP
   V_MPZ,				/* mpz_t */
   V_MPQ,				/* mpq_t */
 #endif
-  V_REAL				/* Floating point number (double) */
+  V_FLOAT				/* Floating point number (double) */
 } numtype;
 
 typedef struct
 { numtype type;				/* type of number */
-  union { double f;			/* value as real */
+  union { double f;			/* value as a floating point number */
 	  int64_t i;			/* value as integer */
 	  word  w[WORDS_PER_DOUBLE];	/* for packing/unpacking the double */
 #ifdef O_GMP
@@ -924,15 +675,19 @@ typedef struct
 	} value;
 } number, *Number;
 
+#define same_type_numbers(n1, n2) \
+	if ( (n1)->type != (n2)->type ) \
+	  make_same_type_numbers(n1, n2)
+
 #define TOINT_CONVERT_FLOAT	0x1	/* toIntegerNumber() */
 #define TOINT_TRUNCATE		0x2
 
 #ifdef O_GMP
 #define intNumber(n)	((n)->type <=  V_MPZ)
 #else
-#define intNumber(n)	((n)->type <  V_REAL)
+#define intNumber(n)	((n)->type <  V_FLOAT)
 #endif
-#define floatNumber(n)	((n)->type >= V_REAL)
+#define floatNumber(n)	((n)->type >= V_FLOAT)
 
 		 /*******************************
 		 *	   GET-PROCEDURE	*
@@ -956,6 +711,23 @@ typedef struct
 #define GF_PROCEDURE	2		/* check for max arity */
 
 
+		 /*******************************
+		 *	       ALERT		*
+		 *******************************/
+
+/* See updateAlerted()
+*/
+
+#define	ALERT_SIGNAL	 0x01
+#define	ALERT_GCREQ	 0x02
+#define	ALERT_OUTOFSTACK 0x04
+#define	ALERT_PROFILE	 0x08
+#define	ALERT_EXITREQ	 0x10
+#define	ALERT_DEPTHLIMIT 0x20
+#define	ALERT_WAKEUP	 0x40
+#define ALERT_DEBUG	 0x80
+
+
 		 /*******************************
 		 *	     CLEANUP		*
 		 *******************************/
@@ -999,7 +771,7 @@ with one operation, it turns out to be faster as well.
 #define SPY_ME			(0x00000080L) /* predicate */
 #define SYSTEM			(0x00000100L) /* predicate, module */
 #define TRACE_ME		(0x00000200L) /* predicate */
-#define METAPRED		(0x00000400L) /* predicate */
+#define P_TRANSPARENT		(0x00000400L) /* predicate */
 #define GC_SAFE			(0x00000800L) /* predicate */
 #define TRACE_CALL		(0x00001000L) /* predicate */
 #define TRACE_REDO		(0x00002000L) /* predicate */
@@ -1020,6 +792,7 @@ with one operation, it turns out to be faster as well.
 #define P_FOREIGN_CREF		(0x02000000L) /* predicate */
 #define P_DIRTYREG		(0x04000000L) /* predicate */
 #define P_ISO			(0x08000000L) /* predicate */
+#define P_META			(0x10000000L) /* predicate */
 
 #define ERASED			(0x0001) /* clause, record */
 					 /* clause flags */
@@ -1037,7 +810,6 @@ with one operation, it turns out to be faster as well.
 #define UNKNOWN_ERROR		(0x0040) /* module */
 #define UNKNOWN_WARNING		(0x0080) /* module */
 
-#define INLINE_F		(0x0001) /* functor (inline foreign) */
 #define CONTROL_F		(0x0002) /* functor (compiled controlstruct) */
 #define ARITH_F			(0x0004) /* functor (arithmetic operator) */
 
@@ -1050,16 +822,17 @@ with one operation, it turns out to be faster as well.
 Handling environment (or local stack) frames.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define FR_BITS			7	/* mask-bits */
-#define FR_NODEBUG		(0x01L)	/* Invisible frame */
+#define FR_BITS			8	/* mask-bits */
+#define FR_HIDE_CHILDS		(0x01L)	/* flag of pred after I_DEPART */
 #define FR_SKIPPED		(0x02L)	/* We have skipped on this frame */
 #define FR_MARKED		(0x04L)	/* GC */
 #define FR_MARKED_PRED		(0x08L)	/* GC predicates/clauses */
 #define FR_WATCHED		(0x10L)	/* GUI debugger */
 #define FR_CATCHED		(0x20L)	/* Frame caught an exception */
 #define FR_INBOX		(0x40L) /* Inside box (for REDO in built-in) */
+#define FR_CONTEXT		(0x80L)	/* fr->context is set */
 
-/* FR_LEVEL now handles levels upto 32M.  This is a bit low, but as it is
+/* FR_LEVEL now handles levels upto 16M.  This is a bit low, but as it is
    only used for the debugger (skip, etc) it is most likely acceptable.
    We must consider using a seperate slot in the localFrame
 */
@@ -1073,7 +846,7 @@ Handling environment (or local stack) frames.
 				  (fr)->flags |= ((l) << FR_BITS); \
 				}
 #define levelFrame(fr)		(fr->flags >> FR_BITS)
-#define incLevel(fr)		(fr->flags += (1<<FR_BITS))
+#define FR_LEVEL_STEP		((1<<FR_BITS))
 #define argFrameP(f, n)		((Word)((f)+1) + (n))
 #define argFrame(f, n)		(*argFrameP((f), (n)) )
 #define varFrameP(f, n)		((Word)(f) + (n))
@@ -1084,13 +857,16 @@ Handling environment (or local stack) frames.
 #define slotsFrame(f)		(true((f)->predicate, FOREIGN) ? \
 				      (f)->predicate->functor->arity : \
 				      (f)->clause->clause->prolog_vars)
-#define contextModule(f)	((f)->context)
 #ifdef O_LOGICAL_UPDATE
 #define generationFrame(f)	((f)->generation)
 #else
 #define generationFrame(f)	(0)
 #endif
 
+#define setNextFrameFlags(next, fr) \
+        (next)->flags = ((fr)->flags + FR_LEVEL_STEP) & \
+                        (~(FR_CONTEXT|FR_SKIPPED|FR_WATCHED|FR_CATCHED|FR_HIDE_CHILDS))
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Predicate reference counting. The aim  of   this  mechanism  is to avoid
 modifying the predicate structure while  it   has  choicepoints  or (MT)
@@ -1164,7 +940,7 @@ assume
   sizeof(*to) * n == sizeof(*double)
   	with n == 1 or n == 2.
 
-We assume the compiler will optimise this properly. 
+We assume the compiler will optimise this properly.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #define cpDoubleData(to, from) \
@@ -1202,7 +978,7 @@ Structure declarations that must be shared across multiple files.
 struct atom
 { Atom		next;		/* next in chain */
   word		atom;		/* as appearing on the global stack */
-#ifdef O_HASHTERM
+#ifdef O_TERMHASH
   unsigned int  hash_value;	/* hash-key value */
 #endif
 #ifdef O_ATOMGC
@@ -1240,7 +1016,8 @@ struct functorDef
   word		name;		/* Name of functor */
   unsigned	arity;		/* arity of functor */
   unsigned      flags;		/* Flag field holding: */
-  /* INLINE_F	   Inlined foreign (system) predicate */
+		  /* CONTROL_F	   Compiled control-structure */
+		  /* ARITH_F	   Arithmetic function */
 };
 
 
@@ -1274,11 +1051,10 @@ struct clause
     uintptr_t erased;		/* Generation I was erased */
   } generation;
 #endif /*O_LOGICAL_UPDATE*/
-  size_t		code_size;	/* size of ->codes */
   unsigned short	variables;	/* # of variables for frame */
   unsigned short	prolog_vars;	/* # real Prolog variables */
 #ifdef O_SHIFT_STACKS
-  unsigned short	marks;		/* C_MARK reserved */
+  unsigned short	marks;		/* C_IFTHEN reserved */
   unsigned short	line_no;	/* Source line-number */
 #else
   unsigned int		line_no; 	/* Source line-number */
@@ -1290,6 +1066,7 @@ struct clause
 		/* HAS_BREAKPOINTS Break-instructions in the clause */
 		/* GOAL_CLAUSE	   Temporary 'islocal' clause (no head) */
 		/* COMMIT_CLAUSE   Clause will commit (execute !) */
+  code		code_size;		/* size of ->codes */
   code		codes[1];		/* VM codes of clause */
 };
 
@@ -1298,22 +1075,32 @@ struct clause_ref
   ClauseRef	next;
 };
 
+#define VM_DYNARGC    255	/* compute argcount dynamically */
+
 #define CA1_PROC	1	/* code arg 1 is procedure */
 #define CA1_FUNC	2	/* code arg 1 is functor */
-#define CA1_DATA	3	/* code arg 2 is prolog data */
+#define CA1_DATA	3	/* code arg 2 is prolog data (H_CONST) */
 #define CA1_INTEGER	4	/* intptr_t value */
 #define CA1_INT64	5	/* int64 value */
 #define CA1_FLOAT	6	/* next WORDS_PER_DOUBLE are double */
 #define CA1_STRING	7	/* inlined string */
 #define CA1_MODULE	8	/* a module */
 #define CA1_VAR		9	/* a variable(-offset) */
-#define CA1_MPZ	       10	/* GNU mpz number */
+#define CA1_CHP	       10	/* ChoicePoint (also variable(-offset)) */
+#define CA1_MPZ	       11	/* GNU mpz number */
+#define CA1_FOREIGN    12	/* Foreign function pointer */
+#define CA1_CLAUSEREF  13	/* Clause reference */
+#define CA1_JUMP       14	/* Instructions to skip */
+#define CA1_AFUNC      15	/* Number of arithmetic function */
+
+#define VIF_BREAK      0x01	/* Can be a breakpoint */
 
 typedef struct
 { char		*name;		/* name of the code */
-  code		code;		/* number of the code */
-  char		arguments;	/* # arguments code takes */
-  char		argtype;	/* # `external' arguments code takes */
+  vmi		code;		/* number of the code */
+  unsigned char flags;		/* Addional flags (VIF_*) */
+  unsigned char	arguments;	/* #args code takes (or VM_DYNARGC) */
+  char		argtype[4];	/* Argument type(s) code takes */
 } code_info;
 
 struct mark
@@ -1357,21 +1144,23 @@ struct local_definitions
 
 struct definition
 { FunctorDef	functor;		/* Name/Arity of procedure */
+  Module	module;			/* module of the predicate */
+  Code		codes;			/* Executable code */
   union
   { ClauseRef	clauses;		/* clause list of procedure */
     Func	function;		/* function pointer of procedure */
     LocalDefinitions local;		/* P_THREAD_LOCAL predicates */
   } definition;
   ClauseRef	lastClause;		/* last clause of list */
-  Module	module;			/* module of the predicate */
   int		references;		/* reference count */
   unsigned int  erased_clauses;		/* #erased but not reclaimed clauses */
 #ifdef O_PLMT
   counting_mutex  *mutex;		/* serialize access to dynamic pred */
 #endif
   ClauseIndex 	hash_info;		/* clause hash-tables */
-  unsigned long indexPattern;		/* indexed argument pattern */
-  unsigned long flags;			/* booleans: */
+  unsigned int  indexPattern;		/* indexed argument pattern */
+  unsigned int  meta_info;		/* meta-predicate info */
+  unsigned int  flags;			/* booleans: */
 		/*	FOREIGN		   foreign predicate? */
 		/*	PROFILE_TICKED	   has been ticked this time? */
 		/*	TRACE_ME	   is my call visible? */
@@ -1380,7 +1169,7 @@ struct definition
 		/*	DYNAMIC		   dynamic predicate? */
 		/*	MULTIFILE	   defined over more files? */
 		/*	SYSTEM		   system predicate */
-		/*	METAPRED	   procedure transparent to modules */
+		/*	P_TRANSPARENT	   procedure transparent to modules */
 		/*	DISCONTIGUOUS	   procedure might be discontiguous */
 		/*	NONDETERMINISTIC   deterministic foreign (not used) */
 		/*	GC_SAFE		   Save to perform GC while active */
@@ -1396,6 +1185,10 @@ struct definition
   		/*	P_SHARED	   Multiple procs are using me */
   unsigned	indexCardinality : 8;	/* cardinality of index pattern */
   unsigned	number_of_clauses : 24;	/* number of associated clauses */
+#ifdef O_PROF_PENTIUM
+  int		prof_index;		/* index in profiling */
+  char	       *prof_name;		/* name in profiling */
+#endif
 };
 
 
@@ -1419,18 +1212,17 @@ struct localFrame
 #endif
   unsigned long	flags;			/* packed long holding: */
 		/*	LEVEL	   recursion level (28 bits) */
-		/*	FR_NODEBUG don't debug this frame ? */
+		/*	FR_HIDE_CHILDS don't debug this frame ? */
 		/*	FR_SKIPPED skipped in the tracer */
 		/*	FR_MARKED  Marked by GC */
 		/*	FR_WATCHED Watched by the debugger */
 		/*	FR_CATCHED Catched exception here */
-};  
+};
 
 
 typedef enum
 { CHP_JUMP = 0,				/* A jump due to ; */
   CHP_CLAUSE,				/* Next clause of predicate */
-  CHP_FOREIGN,				/* Foreign code choicepoint */
   CHP_TOP,				/* First (toplevel) choice */
   CHP_CATCH,				/* $catch initiated choice */
   CHP_DEBUG				/* Enable redo */
@@ -1440,7 +1232,7 @@ typedef enum
 { ABORT_NORMAL,				/* normal abort */
   ABORT_FATAL				/* abort on fatal error */
 } abort_type;
-   
+
 typedef enum
 { DBG_OFF = 0,				/* no debugging */
   DBG_ON,				/* switch on in current environment */
@@ -1524,12 +1316,10 @@ struct queryFrame
   LocalFrame	saved_environment;	/* Parent local-frame */
 					/* Do not put anything between */
 					/* or check parentFrame() */
-  struct localFrame frame;		/* The local frame */
+  struct localFrame top_frame;		/* The (dummy) top local frame */
+  struct localFrame frame;		/* The initial frame */
 };
 
-#define QueryOfTopFrame(fr) \
-	(QueryFrame)((char *)(fr) - offset(queryFrame, frame))
-
 
 #define FLI_MAGIC 		82649821
 #define FLI_MAGIC_CLOSED	42424242
@@ -1572,11 +1362,12 @@ struct recordRef
 struct sourceFile
 { atom_t	name;		/* name of source file */
   int		count;		/* number of times loaded */
-  intptr_t		time;		/* load time of file */
+  time_t	time;		/* load time of file */
   ListCell	procedures;	/* List of associated procedures */
   Procedure	current_procedure;	/* currently loading one */
   int		index;		/* index number (1,2,...) */
-  bool		system;		/* system sourcefile: do not reload */
+  unsigned	system : 1;	/* system sourcefile: do not reload */
+  unsigned	module_count:8;	/* # modules in the file */
 };
 
 
@@ -1618,30 +1409,20 @@ struct gc_trail_entry
 { word		address;	/* address of the variable */
 };
 
-struct table
-{ int		buckets;	/* size of hash table */
-  int		size;		/* # symbols in the table */
-  TableEnum	enumerators;	/* Handles for enumeration */
-#ifdef O_PLMT
-  simpleMutex  *mutex;		/* Mutex to guard table */
-#endif
-  void 		(*copy_symbol)(Symbol s);
-  void 		(*free_symbol)(Symbol s);
-  Symbol	*entries;	/* array of hash symbols */
-};
+		 /*******************************
+		 *	   META PREDICATE	*
+		 *******************************/
 
-struct symbol
-{ Symbol	next;		/* next in chain */
-  void *	name;		/* name entry of symbol */
-  void *	value;		/* associated value with name */
-};
+/*0..9*/				/* 0..9: `Extra meta arguments' */
+#define MA_META		10		/* : */
+#define MA_VAR		11		/* - */
+#define MA_ANY		12		/* ? */
+#define MA_NONVAR	13		/* + */
+
+#define MA_INFO(def, n)		(((def)->meta_info >> ((n)*4)) & 0xf)
+#define MA_SETINFO(def, n, i)	((def)->meta_info &= ~(0xf << (n)*4), \
+				 (def)->meta_info |= (i << (n)*4))
 
-struct table_enum
-{ Table		table;		/* Table we are working on */
-  int		key;		/* Index of current symbol-chain */
-  Symbol	current;	/* The current symbol */
-  TableEnum	next;		/* More choice points */
-};
 
 		 /*******************************
 		 *	 MEMORY ALLOCATION	*
@@ -1698,6 +1479,8 @@ struct alloc_pool
 			    } while(0)
 #endif /*O_DESTRUCTIVE_ASSIGNMENT*/
 
+#define NO_MARK_BAR	(Word)(~(uintptr_t)0)
+
 #define Mark(b)		do { (b).trailtop  = tTop; \
 			     LD->mark_bar = (b).globaltop = gTop; \
 			   } while(0)
@@ -1727,6 +1510,8 @@ be a good idea to remove the latter.
 
 Mark() sets LD->mark_bar, indicating  that   any  assignment  above this
 value need not be trailed.
+
+Note that the local stack is always _above_ the global stack.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #define Trail(p) \
@@ -1749,6 +1534,14 @@ value need not be trailed.
   }
 
 
+		 /*******************************
+		 *	    SUPERVISORS		*
+		 *******************************/
+
+#define SUPERVISOR(name)	(&PL_code_data.supervisors.name[1])
+
+
+
 		 /*******************************
 		 *	   FLI INTERNALS	*
 		 *******************************/
@@ -1760,49 +1553,15 @@ value need not be trailed.
 #define SaveLocalPtr(s, ptr)	term_t s = consTermRef(ptr)
 #define RestoreLocalPtr(s, ptr) (ptr) = (void *) valTermRef(s)
 #else
-#define SaveLocalPtr(s, ptr)	
-#define RestoreLocalPtr(s, ptr) 
+#define SaveLocalPtr(s, ptr)
+#define RestoreLocalPtr(s, ptr)
 #endif
 
 #define QueryFromQid(qid)	((QueryFrame) valTermRef(qid))
 #define QidFromQuery(f)		(consTermRef(f))
 #define QID_EXPORT_WAM_TABLE	(qid_t)(-1)
 
-#include "pl-itf.h"
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Defining built-in predicates using the new interface 
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-#define PRED_IMPL(name, arity, fname, flags) \
-        foreign_t \
-        pl_ ## fname ## _va(term_t PL__t0, int PL__ac, control_t PL__ctx)
-
-#define Arg(N)  (PL__t0+((n)-1))
-#define A1      (PL__t0)
-#define A2      (PL__t0+1)
-#define A3      (PL__t0+2)
-#define A3      (PL__t0+2)
-#define A4      (PL__t0+3)
-#define A5      (PL__t0+4)
-#define A6      (PL__t0+5)
-#define A7      (PL__t0+6)
-#define A8      (PL__t0+7)
-#define A9      (PL__t0+8)
-#define A10     (PL__t0+9)
-
-#define CTX_CNTRL ForeignControl(PL__ctx)
-#define CTX_PTR   ForeignContextPtr(PL__ctx)
-#define CTX_INT   ForeignContextInt(PL__ctx)
-#define CTX_ARITY PL__ac
-
-#define BeginPredDefs(id) \
-        PL_extension PL_predicates_from_ ## id[] = {
-#define PRED_DEF(name, arity, fname, flags) \
-        { name, arity, pl_ ## fname ## _va, (flags)|PL_FA_VARARGS },
-#define EndPredDefs \
-        { NULL, 0, NULL, 0 } \
-        };
+#include "SWI-Prolog.h"
 
 
 		 /*******************************
@@ -1834,27 +1593,6 @@ typedef struct
 #endif
 
 
-		 /*******************************
-		 *	   OPTION LISTS		*
-		 *******************************/
-
-#define OPT_BOOL	(0)		/* types */
-#define OPT_INT		(1)
-#define OPT_STRING	(2)
-#define OPT_ATOM	(3)
-#define OPT_TERM	(4)		/* arbitrary term */
-#define OPT_LONG	(5)
-#define OPT_NATLONG	(6)		/* > 0 */
-#define OPT_TYPE_MASK	0xff
-#define OPT_INF		0x100		/* allow 'inf' */
-
-#define OPT_ALL		0x1		/* flags */
-
-typedef struct
-{ word		name;			/* Name of option */
-  int		type;			/* Type of option */
-} opt_spec, *OptSpec;
-
 		 /*******************************
 		 *	      EVENTS		*
 		 *******************************/
@@ -1962,13 +1700,15 @@ typedef enum
   STACK_OVERFLOW_FATAL
 } stack_overflow_action;
 
+#define ensureRoomStack(s, n) ensure_room_stack((Stack)&LD->stacks.s, (n))
+
 #if O_DYNAMIC_STACKS
 #ifdef O_SEGV_HANDLING
 #define requireStack(s, n)
 #else
 #define requireStack(s, n) \
 	{ if ( roomStack(s) < (intptr_t)(n) ) \
- 	    ensureRoomStack((void*)&LD->stacks.s, n); \
+ 	    ensureRoomStack(s, n); \
 	}
 #endif /*O_SEGV_HANDLING*/
 #else
@@ -2000,8 +1740,11 @@ typedef struct
 		 *	    STREAM I/O		*
 		 *******************************/
 
+#define REDIR_MAGIC 0x23a9bef3
+
 typedef struct redir_context
-{ IOSTREAM     *stream;			/* temporary output */
+{ int		magic;			/* REDIR_MAGIC */
+  IOSTREAM     *stream;			/* temporary output */
   int		is_stream;		/* redirect to stream */
   int		redirected;		/* output is redirected */
   term_t	term;			/* redirect target */
@@ -2047,13 +1790,13 @@ typedef struct
   int64_t	trail_gained;		/* trail stack bytes collected */
   int64_t	global_left;		/* global stack bytes left after GC */
   int64_t	trail_left;		/* trail stack bytes left after GC */
-  real		time;			/* time spent in collections */
+  double	time;			/* time spent in collections */
 } pl_gc_status_t;
 
 
 typedef struct
 { int		blocked;		/* No shifts allowed */
-  real		time;			/* time spent in stack shifts */
+  double	time;			/* time spent in stack shifts */
   int		local_shifts;		/* Shifts of the local stack */
   int		global_shifts;		/* Shifts of the global stack */
   int		trail_shifts;		/* Shifts of the trail stack */
@@ -2068,14 +1811,12 @@ typedef struct
 #define MODULE_system	(GD->modules.system)
 #define MODULE_parse	(ReadingSource ? LD->modules.source \
 				       : MODULE_user)
-#define MODULE_context	(environment_frame ? contextModule(environment_frame) \
-					   : MODULE_user)
+
 
 		/********************************
 		*         PREDICATES            *
 		*********************************/
 
-#define PROCEDURE_garbage_collect0	(GD->procedures.garbage_collect0)
 #define PROCEDURE_block3		(GD->procedures.block3)
 #define PROCEDURE_catch3		(GD->procedures.catch3)
 #define PROCEDURE_true0			(GD->procedures.true0)
@@ -2083,12 +1824,13 @@ typedef struct
 #define PROCEDURE_event_hook1		(GD->procedures.event_hook1)
 #define PROCEDURE_print_message2	(GD->procedures.print_message2)
 #define PROCEDURE_dcall1		(GD->procedures.dcall1)
-#define PROCEDURE_setup_and_call_cleanup4 \
-				(GD->procedures.setup_and_call_cleanup4)
+#define PROCEDURE_setup_call_catcher_cleanup4 \
+				(GD->procedures.setup_call_catcher_cleanup4)
 #define PROCEDURE_dwakeup1		(GD->procedures.dwakeup1)
 #define PROCEDURE_call_residue_vars2	(GD->procedures.call_residue_vars2)
 #define PROCEDURE_dthread_init0		(GD->procedures.dthread_init0)
 #define PROCEDURE_exception_hook4	(GD->procedures.exception_hook4)
+#define PROCEDURE_dc_call_prolog	(GD->procedures.dc_call_prolog0)
 
 extern const code_info codeTable[]; /* Instruction info (read-only) */
 
@@ -2148,38 +1890,36 @@ typedef struct debuginfo
 #define FT_INT64	4		/* passed as int64_t */
 #define FT_MASK		0x0f		/* mask to get type */
 
-#define FF_READONLY	0x10		/* feature is read-only */
-#define FF_KEEP		0x20		/* keep value it already set */
-
-#define CHARESCAPE_FEATURE	  0x00001 /* handle \ in atoms */
-#define GC_FEATURE		  0x00002 /* do GC */
-#define TRACE_GC_FEATURE	  0x00004 /* verbose gc */
-#define TTY_CONTROL_FEATURE	  0x00008 /* allow for tty control */
-#define READLINE_FEATURE	  0x00010 /* readline is loaded */
-#define DEBUG_ON_ERROR_FEATURE	  0x00020 /* start tracer on error */
-#define REPORT_ERROR_FEATURE	  0x00040 /* print error message */
-#define FILE_CASE_FEATURE	  0x00080 /* file names are case sensitive */
-#define FILE_CASE_PRESERVING_FEATURE 0x0100 /* case preserving file names */
-#define DOS_FILE_NAMES_FEATURE    0x00200 /* dos (8+3) file names */
-#define ALLOW_VARNAME_FUNCTOR	  0x00400 /* Read Foo(x) as 'Foo'(x) */
-#define ISO_FEATURE		  0x00800 /* Strict ISO compliance */
-#define OPTIMISE_FEATURE	  0x01000 /* -O: optimised compilation */
-#define FILEVARS_FEATURE	  0x02000 /* Expand $var and ~ in filename */
-#define AUTOLOAD_FEATURE	  0x04000 /* do autoloading */
-#define CHARCONVERSION_FEATURE	  0x08000 /* do character-conversion */
-#define LASTCALL_FEATURE	  0x10000 /* Last call optimization enabled? */
-#define EX_ABORT_FEATURE	  0x20000 /* abort with exception */
-#define BACKQUOTED_STRING_FEATURE 0x40000 /* `a string` */
-#define SIGNALS_FEATURE		  0x80000 /* Handle signals */
-#define DEBUGINFO_FEATURE	  0x100000 /* generate debug info */
+#define PLFLAG_CHARESCAPE	    0x000001 /* handle \ in atoms */
+#define PLFLAG_GC		    0x000002 /* do GC */
+#define PLFLAG_TRACE_GC	  	    0x000004 /* verbose gc */
+#define PLFLAG_TTY_CONTROL	    0x000008 /* allow for tty control */
+#define PLFLAG_READLINE	  	    0x000010 /* readline is loaded */
+#define PLFLAG_DEBUG_ON_ERROR	    0x000020 /* start tracer on error */
+#define PLFLAG_REPORT_ERROR	    0x000040 /* print error message */
+#define PLFLAG_FILE_CASE	    0x000080 /* file names are case sensitive */
+#define PLFLAG_FILE_CASE_PRESERVING 0x000100 /* case preserving file names */
+#define PLFLAG_DOS_FILE_NAMES       0x000200 /* dos (8+3) file names */
+#define ALLOW_VARNAME_FUNCTOR	    0x000400 /* Read Foo(x) as 'Foo'(x) */
+#define PLFLAG_ISO		    0x000800 /* Strict ISO compliance */
+#define PLFLAG_OPTIMISE	  	    0x001000 /* -O: optimised compilation */
+#define PLFLAG_FILEVARS	  	    0x002000 /* Expand $var and ~ in filename */
+#define PLFLAG_AUTOLOAD	  	    0x004000 /* do autoloading */
+#define PLFLAG_CHARCONVERSION	    0x008000 /* do character-conversion */
+#define PLFLAG_LASTCALL	  	    0x010000 /* Last call optimization enabled? */
+#define PLFLAG_EX_ABORT	  	    0x020000 /* abort with exception */
+#define PLFLAG_BACKQUOTED_STRING    0x040000 /* `a string` */
+#define PLFLAG_SIGNALS		    0x080000 /* Handle signals */
+#define PLFLAG_DEBUGINFO	    0x100000 /* generate debug info */
+#define PLFLAG_FILEERRORS	    0x200000 /* Edinburgh file errors */
 
 typedef struct
-{ unsigned long flags;			/* the feature flags */
+{ unsigned long flags;			/* Fast access to some boolean Prolog flags */
 } pl_features_t;
 
-#define trueFeature(mask)	true(&features, mask)
-#define setFeatureMask(mask)	set(&features, mask)
-#define clearFeatureMask(mask)	clear(&features, mask)
+#define truePrologFlag(flag)	  true(&LD->prolog_flag.mask, flag)
+#define setPrologFlagMask(flag)	  set(&LD->prolog_flag.mask, flag)
+#define clearPrologFlagMask(flag) clear(&LD->prolog_flag.mask, flag)
 
 typedef enum
 { OCCURS_CHECK_FALSE = 0,
@@ -2189,7 +1929,7 @@ typedef enum
 
 
 #ifdef O_LIMIT_DEPTH
-#define DEPTH_NO_LIMIT	(~0x0L)		/* Highest value */
+#define DEPTH_NO_LIMIT	(~(uintptr_t)0x0) /* Highest value */
 #endif
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -2226,18 +1966,6 @@ decrease).
 
 #define REL(a)		((Word)(a) - (Word)(lBase))
 
-#if O_DEBUG
-#define DEBUG(n, g) { if (GD->debug_level >= (n)) { g; } }
-#else
-#define DEBUG(a, b) 
-#endif
-
-#if O_SECURE
-#define SECURE(g) {g;}
-#else
-#define SECURE(g)
-#endif
-
 #ifdef _DEBUG
 #define O_MAINTENANCE
 #endif
@@ -2248,7 +1976,10 @@ decrease).
 #include SYSLIB_H
 #endif
 
-#define COMMON(type) SO_LOCAL type
+#define NULL_ATOM ((atom_t)0)
+#define MK_ATOM(n)    		((atom_t)((n)<<7|TAG_ATOM|STG_STATIC))
+#include "pl-atom.ih"
+#include "pl-funct.ih"
 
 #include "pl-main.h"			/* Declarations needed by pl-main.c */
 #include "pl-error.h"			/* Exception generation */
@@ -2258,8 +1989,11 @@ decrease).
 #include "pl-gmp.h"			/* GNU-GMP support */
 #include "pl-global.h"			/* global data */
 #include "pl-funcs.h"			/* global functions */
+#include "pl-privitf.h"			/* private foreign interface */
 #include "pl-text.h"			/* text manipulation */
 #include "pl-hash.h"			/* Murmurhash function */
+#include "pl-option.h"			/* Option processing */
+#include "pl-files.h"			/* File management */
 
 #ifdef __DECC				/* Dec C-compiler: avoid conflicts */
 #undef leave
@@ -2267,9 +2001,4 @@ decrease).
 #undef try
 #endif
 
-#define NULL_ATOM ((atom_t)0)
-#define MK_ATOM(n)    		((atom_t)((n)<<7|TAG_ATOM|STG_STATIC))
-#include "pl-atom.ih"
-#include "pl-funct.ih"
-
 #endif /*_PL_INCLUDE_H*/
