@@ -35,6 +35,9 @@ option  parsing,  initialisation  and  handling  of errors and warnings.
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
 #endif
+#ifdef __WINDOWS__
+#include <process.h>			/* getpid() */
+#endif
 
 #define LOCK()   PL_LOCK(L_INIT)
 #define UNLOCK() PL_UNLOCK(L_INIT)
@@ -86,13 +89,38 @@ exec_var(const char *name)
 }
 
 
+static const char *
+longopt(const char *opt, int argc, const char **argv)
+{ size_t optlen = strlen(opt);
+
+  for(; argc > 0; argc--, argv++)
+  { const char *a = argv[0];
+
+    if ( *a++ == '-' && *a++ == '-' )
+    { if ( *a == EOS )		/* --: end of args */
+	return NULL;
+      if ( strncmp(a, opt, optlen) == 0 && a[optlen] == '=' )
+	return &a[optlen+1];
+    }
+  }
+
+  return NULL;
+}
+
+
 static char *
-findHome(char *symbols)
-{ char *home = NULL;
+findHome(const char *symbols, int argc, const char **argv)
+{ const char *home = NULL;
   char envbuf[MAXPATHLEN];
   char plp[MAXPATHLEN];
   const char *val;
-  
+
+  if ( (val=longopt("home", argc, argv)) )
+  { if ( (home=PrologPath(val, plp, sizeof(plp))) )
+      return store_string(home);
+    return NULL;
+  }
+
   if ( (val  = exec_var("homevar")) &&
        (home = Getenv(val, envbuf, sizeof(envbuf))) &&
        (home = PrologPath(home, plp, sizeof(plp))) )
@@ -170,7 +198,7 @@ basename of the running program, taking all the leading alnum characters.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static char *
-defaultSystemInitFile(char *a0)
+defaultSystemInitFile(const char *a0)
 { char plp[MAXPATHLEN];
   char *base = BaseName(PrologPath(a0, plp, sizeof(plp)));
   char buf[256];
@@ -229,41 +257,41 @@ setupGNUEmacsInferiorMode()
 
   if ( ((s = Getenv("EMACS", envbuf, sizeof(envbuf))) && s[0]) ||
        ((s = Getenv("INFERIOR", envbuf, sizeof(envbuf))) && streq(s, "yes")) )
-  { clearFeatureMask(TTY_CONTROL_FEATURE);
+  { clearPrologFlagMask(PLFLAG_TTY_CONTROL);
     val = TRUE;
   } else
   { val = FALSE;
-  } 
+  }
 
-  defFeature("emacs_inferior_process", FT_BOOL|FF_READONLY, val, 0);
+  PL_set_prolog_flag("emacs_inferior_process", PL_BOOL|FF_READONLY, val);
 }
 
 
 static void
-initPaths()
+initPaths(int argc, const char **argv)
 { char plp[MAXPATHLEN];
 
-  if ( GD->cmdline.argc > 0 )
+  if ( argc > 0 )
   { char plp1[MAXPATHLEN];
-    char *symbols = NULL;		/* The executable */
+    const char *symbols = NULL;		/* The executable */
 
-    if ( !(symbols = findExecutable(GD->cmdline.argv[0], plp1)) ||
+    if ( !(symbols = findExecutable(argv[0], plp1)) ||
 	 !(symbols = DeRefLink(symbols, plp)) )
-      symbols = GD->cmdline.argv[0];
-  
+      symbols = argv[0];
+
     DEBUG(2, Sdprintf("rc-module: %s\n", symbols));
-  
-    systemDefaults.home	       = findHome(symbols);
-  
+
+    systemDefaults.home	       = findHome(symbols, argc, argv);
+
 #ifdef __WINDOWS__			/* we want no module but the .EXE */
     GD->paths.module	       = store_string(symbols);
     symbols = findExecutable(NULL, plp);
     DEBUG(2, Sdprintf("Executable: %s\n", symbols));
 #endif
     GD->paths.executable       = store_string(symbols);
-    GD->options.systemInitFile = defaultSystemInitFile(GD->cmdline.argv[0]);
+    GD->options.systemInitFile = defaultSystemInitFile(argv[0]);
   } else
-  { systemDefaults.home	       = findHome(NULL);
+  { systemDefaults.home	       = findHome(NULL, argc, argv);
     GD->options.systemInitFile = store_string("none");
 #ifdef __WINDOWS__			/* we want no module but the .EXE */
     GD->paths.module	       = store_string("libpl.dll");
@@ -303,11 +331,11 @@ initDefaults()
   GD->bootsession	     = FALSE;
 
   if ( systemDefaults.notty )
-    clearFeatureMask(TTY_CONTROL_FEATURE);
+    clearPrologFlagMask(PLFLAG_TTY_CONTROL);
   else
-    setFeatureMask(TTY_CONTROL_FEATURE);
+    setPrologFlagMask(PLFLAG_TTY_CONTROL);
 
-  setFeatureMask(DEBUGINFO_FEATURE);
+  setPrologFlagMask(PLFLAG_DEBUGINFO);
 }
 
 
@@ -410,7 +438,7 @@ static int
 parseCommandLineOptions(int argc0, char **argv, int *compile)
 { int argc = argc0;
 
-  
+
 
   for( ; argc > 0 && (argv[0][0] == '-' || argv[0][0] == '+'); argc--, argv++ )
   { char *s = &argv[0][1];
@@ -420,16 +448,16 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 
     if ( streq(s, "tty") )	/* +/-tty */
     { if ( s[-1] == '+' )
-	setFeatureMask(TTY_CONTROL_FEATURE);
+	setPrologFlagMask(PLFLAG_TTY_CONTROL);
       else
-	clearFeatureMask(TTY_CONTROL_FEATURE);
+	clearPrologFlagMask(PLFLAG_TTY_CONTROL);
 
       continue;
     } else if ( isoption(s, "nosignals") )
-    { clearFeatureMask(SIGNALS_FEATURE);
+    { clearPrologFlagMask(PLFLAG_SIGNALS);
       continue;
     } else if ( isoption(s, "nodebug") )
-    { clearFeatureMask(DEBUGINFO_FEATURE);
+    { clearPrologFlagMask(PLFLAG_DEBUGINFO);
       continue;
     } else if ( streq(s, "-quiet") )
     { GD->options.silent = TRUE;
@@ -478,7 +506,7 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 	case 'A':
 	case 'H':
         { uintptr_t size = memarea_limit(&s[1]);
-	  
+
 	  if ( size == MEMAREA_INVALID_SIZE )
 	    return -1;
 
@@ -544,7 +572,7 @@ openResourceDB(int argc, char **argv)
   int n;
 
   if ( !GD->bootsession )
-  { 
+  {
 #ifdef __WINDOWS__
     if ( GD->paths.module &&
 	 !streq(GD->paths.module, GD->paths.executable) &&
@@ -554,7 +582,7 @@ openResourceDB(int argc, char **argv)
     if ( (rc = rc_open_archive(GD->paths.executable, flags)) )
       return rc;
   }
-  
+
   for(n=0; n<argc-1; n++)
   { if ( argv[n][0] == '-' && argv[n][2] == EOS ) /* -? */
     { if ( argv[n][1] == '-' )
@@ -562,12 +590,12 @@ openResourceDB(int argc, char **argv)
       if ( GD->bootsession )
       { if ( argv[n][1] == 'o' )
 	{ xfile = argv[n+1];
-	  break; 
+	  break;
 	}
       } else
       { if ( argv[n][1] == 'x' )
 	{ xfile = argv[n+1];
-	  break; 
+	  break;
 	}
       }
     }
@@ -812,7 +840,7 @@ PL_initialise(int argc, char **argv)
   GD->cmdline._c_argc = -1;
   DEBUG(1,
   { int i;
-    
+
     Sdprintf("argv =");
     for(i=0; i<argc; i++)
       Sdprintf(" %s", argv[i]);
@@ -835,11 +863,11 @@ properly on Linux. Don't bother with it.
 
   initOs();				/* Initialise OS bindings */
   initDefaults();			/* Initialise global defaults */
-  initPaths();				/* fetch some useful paths */
+  initPaths(argc, (const char**)argv);	/* fetch some useful paths */
 
   setupGNUEmacsInferiorMode();		/* Detect running under EMACS */
 #ifdef HAVE_SIGNAL
-  setFeatureMask(SIGNALS_FEATURE);	/* default: handle signals */
+  setPrologFlagMask(PLFLAG_SIGNALS);	/* default: handle signals */
 #endif
 
   if ( (GD->resourceDB = rc_open_archive(GD->paths.executable, RC_RDONLY)) )
@@ -889,7 +917,7 @@ properly on Linux. Don't bother with it.
   aliasThread(PL_thread_self(), ATOM_main);
   enableThreads(TRUE);
 #endif
-  defFeature("resource_database", FT_ATOM|FF_READONLY, rcpath);
+  PL_set_prolog_flag("resource_database", PL_ATOM|FF_READONLY, rcpath);
   initialiseForeign(GD->cmdline.argc, /* PL_initialise_hook() functions */
 		    GD->cmdline.argv);
   systemMode(TRUE);
@@ -902,7 +930,7 @@ properly on Linux. Don't bother with it.
     { PL_halt(1);
     }
     if ( Sclose(s) != 0 || !rc_save_archive(GD->resourceDB, NULL) )
-    { 
+    {
 #ifdef __WINDOWS__
       PlMessage("Failed to save system resources: %s", rc_strerror(rc_errno));
 #else
@@ -984,6 +1012,7 @@ usage()
     "    --nosignals      Do not modify any signal handling\n",
     "    --nodebug        Omit generation of debug info\n",
     "    --quiet          Quiet operation (also -q)\n",
+    "    --home=DIR       Use DIR as SWI-Prolog home\n",
     NULL
   };
   const cline *lp = lines;
@@ -1046,7 +1075,7 @@ runtime_vars(int format)
   char version[20];
 
   if ( systemDefaults.home )
-  { 
+  {
 #ifdef O_XOS
     if ( format == FMT_CMD )
     { _xos_os_filename(systemDefaults.home, base, MAXPATHLEN);
@@ -1118,17 +1147,17 @@ giveVersionInfo(const char *a)
 		 *	     CLEANUP		*
 		 *******************************/
 
-typedef void (*halt_function)(int, Void);
+typedef void (*halt_function)(int, void*);
 
 struct on_halt
 { halt_function	function;
-  Void		argument;
+  void*		argument;
   OnHalt	next;
 };
 
 
 void
-PL_on_halt(halt_function f, Void arg)
+PL_on_halt(halt_function f, void *arg)
 { if ( !GD->os.halting )
   { OnHalt h = allocHeap(sizeof(struct on_halt));
 
@@ -1234,6 +1263,7 @@ PL_cleanup(int rval)
   cleanupThreads();
 #endif
   cleanupForeign();
+  cleanupCodeToAtom();
   cleanupMemAlloc();
 
   UNLOCK();				/* requires GD->thread.enabled */
@@ -1315,8 +1345,14 @@ vsysError(const char *fm, va_list args)
     PL_halt(1);
 
 action:
-  Sfprintf(Serror, "\nAction? "); Sflush(Soutput);
+#ifdef HAVE_GETPID
+  Sfprintf(Serror, "\n[pid=%d] Action? ", getpid());
+#else
+  Sfprintf(Serror, "\nAction? ");
+#endif
+  Sflush(Soutput);
   ResetTty();
+
   switch(getSingleChar(Sinput, FALSE))
   { case 'a':
       pl_abort(ABORT_FATAL);
@@ -1348,7 +1384,7 @@ vfatalError(const char *fm, va_list args)
   Ssprintf(msg, "[FATAL ERROR:\n\t");
   Svsprintf(&msg[strlen(msg)], fm, args);
   Ssprintf(&msg[strlen(msg)], "]");
-  
+
   PlMessage(msg);
 #else
   Sfprintf(Serror, "[FATAL ERROR:\n\t");
@@ -1374,7 +1410,7 @@ bool
 vwarning(const char *fm, va_list args)
 { toldString();				/* play safe */
 
-  if ( trueFeature(REPORT_ERROR_FEATURE) )
+  if ( truePrologFlag(PLFLAG_REPORT_ERROR) )
   { if ( !GD->bootsession && GD->initialised &&
 	 !LD->outofstack && 		/* cannot call Prolog */
 	 !fm[0] == '$')			/* explicit: don't call Prolog */
@@ -1386,7 +1422,7 @@ vwarning(const char *fm, va_list args)
       term_t head = PL_new_term_ref();
 
       Svsprintf(message, fm, args);
-      
+
       for(;;)
       { char *eol = strchr(s, '\n');
 
@@ -1416,7 +1452,7 @@ vwarning(const char *fm, va_list args)
     }
   }
 
-  if ( !ReadingSource && trueFeature(DEBUG_ON_ERROR_FEATURE) )
+  if ( !ReadingSource && truePrologFlag(PLFLAG_DEBUG_ON_ERROR) )
     pl_trace();
 
   PL_fail;
