@@ -1,11 +1,11 @@
-/*  $Id: pl-modul.c,v 1.52 2007/03/07 08:17:17 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -96,7 +96,7 @@ _lookupModule(atom_t name)
   addHTable(GD->tables.modules, (void *)name, m);
   GD->statistics.modules++;
   PL_register_atom(name);
-  
+
   return m;
 }
 
@@ -108,7 +108,7 @@ lookupModule(atom_t name)
   LOCK();
   m = _lookupModule(name);
   UNLOCK();
-  
+
   return m;
 }
 
@@ -116,7 +116,7 @@ lookupModule(atom_t name)
 Module
 isCurrentModule(atom_t name)
 { Symbol s;
-  
+
   if ( (s = lookupHTable(GD->tables.modules, (void*)name)) )
     return (Module) s->value;
 
@@ -150,7 +150,7 @@ isSuperModule(Module s, Module m)	/* s is a super-module of m */
 next:
   if ( m == s )
     succeed;
-  
+
   for(c=m->supers; c; c=c->next)
   { if ( c->next )
     { if ( isSuperModule(s, c->value) )
@@ -356,19 +356,39 @@ PRED_IMPL("delete_import_module", 2, delete_import_module, 0)
 }
 
 
+static int
+get_existing_source_file(term_t file, SourceFile *sfp ARG_LD)
+{ SourceFile sf;
+  atom_t a;
+
+  if ( PL_get_atom(file, &a) )
+  { if ( (sf = lookupSourceFile(a, FALSE)) )
+    { *sfp = sf;
+      return TRUE;
+    }
+
+    return FALSE;
+  }
+
+  *sfp = NULL;
+  return TRUE;
+}
+
+
 word
 pl_current_module(term_t module, term_t file, control_t h)
 { GET_LD
   TableEnum e = NULL;
   Symbol symb;
   atom_t name;
+  SourceFile sf;
 
   if ( ForeignControl(h) == FRG_CUTTED )
   { e = ForeignContextPtr(h);
     freeTableEnum(e);
     succeed;
   }
-					/* deterministic cases */
+				/* deterministic case: module --> file */
   if ( PL_get_atom(module, &name) )
   { Module m;
 
@@ -378,28 +398,21 @@ pl_current_module(term_t module, term_t file, control_t h)
     }
 
     fail;
-  } else if ( PL_get_atom(file, &name) )
-  { int rval = FALSE;
-    for_table(GD->tables.modules, s,
-	      { Module m = s->value;
-
-		if ( m->file && m->file->name == name )
-		{ rval = PL_unify_atom(module, m->name);
-		  break;
-		}
-	      })
-    return rval;
   }
 
   switch(ForeignControl(h))
   { case FRG_FIRST_CALL:
+      if ( !get_existing_source_file(file, &sf PASS_LD) )
+	fail;				/* given, but non-existing file */
       e = newTableEnum(GD->tables.modules);
       break;
     case FRG_REDO:
       e = ForeignContextPtr(h);
+      get_existing_source_file(file, &sf PASS_LD);
       break;
-    default:
-      assert(0);
+    case FRG_CUTTED:
+      freeTableEnum(e);
+      succeed;
   }
 
   while( (symb = advanceTableEnum(e)) )
@@ -412,9 +425,16 @@ pl_current_module(term_t module, term_t file, control_t h)
     { fid_t cid = PL_open_foreign_frame();
       atom_t f = ( !m->file ? ATOM_nil : m->file->name);
 
-      if ( PL_unify_atom(module, m->name) &&
+      if ( (!sf || (m->file == sf)) &&
+	   PL_unify_atom(module, m->name) &&
 	   PL_unify_atom(file, f) )
-      { ForeignRedoPtr(e);
+      { PL_close_foreign_frame(cid);
+
+	if ( sf && sf->module_count == 1 )
+	{ freeTableEnum(e);
+	  succeed;
+	}
+	ForeignRedoPtr(e);
       }
 
       PL_discard_foreign_frame(cid);
@@ -438,7 +458,7 @@ PRED_IMPL("strip_module", 3, strip_module, PL_FA_TRANSPARENT)
     succeed;
 
   fail;
-}  
+}
 
 
 word
@@ -573,7 +593,7 @@ in it are abolished.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-declareModule(atom_t name, SourceFile sf, int line)
+declareModule(atom_t name, SourceFile sf, int line, int allow_newfile)
 { GET_LD
   Module module;
   term_t tmp = 0, rdef = 0, rtail = 0;
@@ -581,7 +601,7 @@ declareModule(atom_t name, SourceFile sf, int line)
   LOCK();
   module = _lookupModule(name);
 
-  if ( module->file && module->file != sf)
+  if ( !allow_newfile && module->file && module->file != sf)
   { term_t obj;
     char msg[256];
     UNLOCK();
@@ -593,8 +613,11 @@ declareModule(atom_t name, SourceFile sf, int line)
     return PL_error("module", 2, msg, ERR_PERMISSION,
 		    ATOM_redefine, ATOM_module, obj);
   }
-	    
-  module->file = sf;
+
+  if ( module->file != sf )
+  { module->file = sf;
+    sf->module_count++;		/* current determinism in $current_module/2 */
+  }
   module->line_no = line;
   LD->modules.source = module;
 
@@ -619,7 +642,7 @@ declareModule(atom_t name, SourceFile sf, int line)
 	    })
   clearHTable(module->public);
   UNLOCK();
-  
+
   if ( rdef )
   { PL_unify_nil(rtail);
 
@@ -634,19 +657,42 @@ declareModule(atom_t name, SourceFile sf, int line)
 }
 
 
-word
-pl_declare_module(term_t name, term_t file, term_t line)
+static
+PRED_IMPL("$declare_module", 4, declare_module, 0)
 { SourceFile sf;
   atom_t mname, fname;
-  int line_no;
+  int line_no, rdef;
 
-  if ( !PL_get_atom_ex(name, &mname) ||
-       !PL_get_atom_ex(file, &fname) ||
-       !PL_get_integer_ex(line, &line_no) )
+  if ( !PL_get_atom_ex(A1, &mname) ||
+       !PL_get_atom_ex(A2, &fname) ||
+       !PL_get_integer_ex(A3, &line_no) ||
+       !PL_get_bool_ex(A4, &rdef) )
     fail;
 
-  sf = lookupSourceFile(fname);
-  return declareModule(mname, sf, line_no);
+  sf = lookupSourceFile(fname, TRUE);
+  return declareModule(mname, sf, line_no, rdef);
+}
+
+
+static int
+unify_export_list(term_t public, Module module ARG_LD)
+{ term_t head = PL_new_term_ref();
+  term_t list = PL_copy_term_ref(public);
+  int rval = TRUE;
+
+  LOCKMODULE(module);
+  for_table(module->public, s,
+	    { if ( !PL_unify_list(list, head, list) ||
+		   !unify_functor(head, (functor_t)s->name, GP_NAMEARITY) )
+	      { rval = FALSE;
+		break;
+	      }
+	    })
+  UNLOCKMODULE(module);
+  if ( rval )
+    return PL_unify_nil(list);
+
+  fail;
 }
 
 
@@ -670,6 +716,8 @@ PRED_IMPL("$module_property", 2, module_property, 0)
       return PL_unify_atom(a, m->file->name);
     else
       fail;
+  } else if ( PL_is_functor(A2, FUNCTOR_exports1) )
+  { return unify_export_list(a, m PASS_LD);
   } else
     return PL_error(NULL, 0, NULL, ERR_DOMAIN,
 		    ATOM_module_property, A2);
@@ -677,98 +725,101 @@ PRED_IMPL("$module_property", 2, module_property, 0)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-export_list(+Module, -PublicPreds)
+export/1 exports a procedure specified by its name and arity or
+head from the context module.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-word
-pl_export_list(term_t modulename, term_t public)
-{ Module module;
-  atom_t mname;
+static int
+export_pi(term_t pi, Module module ARG_LD)
+{ functor_t fd;
+  Procedure proc;
 
-  if ( !PL_get_atom_ex(modulename, &mname) )
-    fail;
-  
-  if ( !(module = isCurrentModule(mname)) )
-    fail;
-  
-  { GET_LD
+  PL_strip_module(pi, &module, pi);
 
-    term_t head = PL_new_term_ref();
-    term_t list = PL_copy_term_ref(public);
-    int rval = TRUE;
+  if ( PL_is_functor(pi, FUNCTOR_comma2) )
+  { term_t a1 = PL_new_term_ref();
+    term_t a2 = PL_new_term_ref();
 
-    LOCKMODULE(module);
-    for_table(module->public, s,
-	      { if ( !PL_unify_list(list, head, list) ||
-		     !PL_unify_functor(head, (functor_t)s->name) )
-		{ rval = FALSE;
-		  break;
-		}
-	      })
-    UNLOCKMODULE(module);
-    if ( rval )
-      return PL_unify_nil(list);
+    _PL_get_arg(1, pi, a1);
+    _PL_get_arg(2, pi, a2);
 
-    fail;
+    TRY(export_pi(a1, module PASS_LD));
+    return export_pi(a2, module PASS_LD);
   }
-}
 
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-pl_export() exports a procedure specified by its name and arity from the
-context module.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+  if ( !get_functor(pi, &fd, &module, 0, GF_PROCEDURE) )
+    fail;
 
-word
-pl_export(term_t pred)
-{ GET_LD
-  Module module = NULL;
-  term_t head = PL_new_term_ref();
-  functor_t fd;
-
-  PL_strip_module(pred, &module, head);
-  if ( PL_get_functor(head, &fd) )
-  { Procedure proc;
-
-    if ( (proc = isStaticSystemProcedure(fd)) )
-      return PL_error(NULL, 0, NULL, ERR_PERMISSION_PROC,
-		      ATOM_export, ATOM_built_in_procedure, proc->definition);
-    proc = lookupProcedure(fd, module);
-
-    LOCKMODULE(module);
-    addHTable(module->public,
-	      (void *)proc->definition->functor->functor,
-	      proc);
-    UNLOCKMODULE(module);
+  if ( (proc = isStaticSystemProcedure(fd)) && true(proc->definition, P_ISO) )
     succeed;
-  }
+  proc = lookupProcedure(fd, module);
 
-  return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_callable, pred);
-}
-
-word
-pl_check_export()
-{ GET_LD
-  Module module = contextModule(environment_frame);
-
-  for_table(module->public, s,
-	    { Procedure proc = (Procedure) s->value;
-	      Definition def = proc->definition;
-
-	      if ( !isDefinedProcedure(proc) && /* not defined */
-		   proc->definition->module == module ) /* not imported */
-	      { printMessage(ATOM_error,
-			     PL_FUNCTOR_CHARS, "undefined_export", 2,
-			       PL_ATOM, module->name,
-			       PL_FUNCTOR, FUNCTOR_divide2,
-			         PL_ATOM, def->functor->name,
-			         PL_INT, (int)def->functor->arity);
-	      }
-	    })
+  LOCKMODULE(module);
+  addHTable(module->public,
+	    (void *)proc->definition->functor->functor,
+	    proc);
+  UNLOCKMODULE(module);
 
   succeed;
 }
 
+
+
+static
+PRED_IMPL("export", 1, export, PL_FA_TRANSPARENT)
+{ PRED_LD
+  Module module = NULL;
+
+  return export_pi(A1, module PASS_LD);
+}
+
+
+/** '$undefined_export'(+Module, -UndefExport:list(pi)) is det.
+
+Unify UndefExport with predicate indicators   of undefined predicates in
+Module.
+*/
+
+static
+PRED_IMPL("$undefined_export", 2, undefined_export, 0)
+{ PRED_LD
+  atom_t mname;
+  Module module;
+  TableEnum e;
+  Symbol symb;
+  term_t tail = PL_copy_term_ref(A2);
+  term_t head = PL_new_term_ref();
+
+  if ( !PL_get_atom_ex(A1, &mname) )
+    return FALSE;
+  if ( !(module = isCurrentModule(mname)) )
+    return PL_error(NULL, 0, NULL, ERR_EXISTENCE, ATOM_module, A1);
+
+  e = newTableEnum(module->public);
+
+  while( (symb = advanceTableEnum(e)) )
+  { Procedure proc = (Procedure) symb->value;
+    Definition def = proc->definition;
+    FunctorDef fd = def->functor;
+
+    if ( !isDefinedProcedure(proc) &&			/* not defined */
+	 def->module == module &&			/* not imported */
+	 !autoImport(fd->functor, module) )
+    { if ( !PL_unify_list(tail, head, tail) ||
+	   !unify_definition(head, proc->definition,
+			     0, GP_QUALIFY|GP_NAMEARITY) )
+      { freeTableEnum(e);
+	return FALSE;
+      }
+    }
+  }
+
+  freeTableEnum(e);
+  return PL_unify_nil(tail);
+}
+
+
 word
 pl_context_module(term_t module)
 { GET_LD
@@ -822,13 +873,12 @@ pl_import(term_t pred)
 { GET_LD
   Module source = NULL;
   Module destination = contextModule(environment_frame);
-  term_t head = PL_new_term_ref();
   functor_t fd;
   Procedure proc, old;
 
-  PL_strip_module(pred, &source, head);
-  if ( !PL_get_functor(head, &fd) )
-    return warning("import/1: instantiation fault");
+  if ( !get_functor(pred, &fd, &source, 0, GF_PROCEDURE) )
+    fail;
+
   proc = lookupProcedure(fd, source);
 
   if ( !isDefinedProcedure(proc) )
@@ -851,14 +901,14 @@ pl_import(term_t pred)
     }
 
     if ( old->definition->module == destination )
-      return warning("Cannot import %s into module %s: name clash", 
-		     procedureName(proc), 
+      return warning("Cannot import %s into module %s: name clash",
+		     procedureName(proc),
 		     stringAtom(destination->name) );
 
     if ( old->definition->module != source )
-    { warning("Cannot import %s into module %s: already imported from %s", 
-	      procedureName(proc), 
-	      stringAtom(destination->name), 
+    { warning("Cannot import %s into module %s: already imported from %s",
+	      procedureName(proc),
+	      stringAtom(destination->name),
 	      stringAtom(old->definition->module->name) );
       fail;
     }
@@ -870,16 +920,16 @@ pl_import(term_t pred)
   }
 
   if ( !isPublicModule(source, proc) )
-  { warning("import/1: %s is not declared public (still imported)", 
+  { warning("import/1: %s is not declared public (still imported)",
 	    procedureName(proc));
   }
-  
+
   { Procedure nproc = (Procedure)  allocHeap(sizeof(struct procedure));
-  
+
     nproc->type = PROCEDURE_TYPE;
     nproc->definition = proc->definition;
     set(proc->definition, P_SHARED);
-  
+
     LOCKMODULE(destination);
     addHTable(destination->procedures,
 	      (void *)proc->definition->functor->functor, nproc);
@@ -900,8 +950,11 @@ BeginPredDefs(module)
 	   PL_FA_NONDETERMINISTIC)
   PRED_DEF("import_module", 2, import_module,
 	   PL_FA_NONDETERMINISTIC)
+  PRED_DEF("$declare_module", 4, declare_module, 0)
   PRED_DEF("add_import_module", 3, add_import_module, 0)
   PRED_DEF("delete_import_module", 2, delete_import_module, 0)
   PRED_DEF("$module_property", 2, module_property, 0)
   PRED_DEF("strip_module", 3, strip_module, PL_FA_TRANSPARENT)
+  PRED_DEF("export", 1, export, PL_FA_TRANSPARENT)
+  PRED_DEF("$undefined_export", 2, undefined_export, 0)
 EndPredDefs
