@@ -1,4 +1,4 @@
-/*  $Id: pl-pro.c,v 1.85 2007/06/01 14:24:31 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -144,9 +144,9 @@ callProlog(Module module, term_t goal, int flags, term_t *ex)
 
     fail;
   }
-  
+
   proc = lookupProcedure(fd, module);
-  
+
   { int arity = arityFunctor(fd);
     term_t args = PL_new_term_refs(arity);
     qid_t qid;
@@ -169,11 +169,11 @@ callProlog(Module module, term_t goal, int flags, term_t *ex)
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Abort and toplevel. At the  moment,   prologToplevel()  sets a longjmp()
 context and pl_abort() jumps to this   context and resets the SWI-Prolog
-engine. 
+engine.
 
 Using the multi-threaded version, this is   not  acceptable. Each thread
 needs such a context, but worse  is   that  we cannot properly reset the
-reference count and ensure locks are all in a sane state. 
+reference count and ensure locks are all in a sane state.
 
 A cleaner solution is to  map  an   abort  onto  a Prolog exception. The
 exception-handling   code   should    ensure     proper    handling   of
@@ -200,11 +200,13 @@ pl_throw_abort()
   { fid_t fid = PL_open_foreign_frame();
     term_t ex = PL_new_term_ref();
 
+    clearSegStack(&LD->cycle.stack);	/* can do no harm */
+
     PL_put_atom(ex, ATOM_aborted);
     PL_throw(ex);			/* use longjmp() to ensure */
 
     PL_close_foreign_frame(fid);	/* should not be reached */
-    fail;				
+    fail;
   }
 }
 
@@ -224,7 +226,7 @@ static int can_abort;			/* embeded code can't abort */
 word
 pl_abort(abort_type type)
 { if ( !can_abort ||
-       (trueFeature(EX_ABORT_FEATURE) && type == ABORT_NORMAL) )
+       (truePrologFlag(PLFLAG_EX_ABORT) && type == ABORT_NORMAL) )
     return pl_throw_abort();
 
   if ( LD->critical > 0 )		/* abort in critical region: delay */
@@ -233,9 +235,10 @@ pl_abort(abort_type type)
     succeed;
   }
 
-  if ( !trueFeature(READLINE_FEATURE) )
+  if ( !truePrologFlag(PLFLAG_READLINE) )
     PopTty(Sinput, &ttytab);
   LD->outofstack = NULL;
+  clearSegStack(&LD->cycle.stack);
   closeFiles(FALSE);
   resetReferences();
 #ifdef O_PROFILE
@@ -246,7 +249,7 @@ pl_abort(abort_type type)
   resetSignals();
   resetForeign();
   resetAtoms();
-  resetBags();
+  updateAlerted(LD);
 
   longjmp(abort_context, 1);
   /*NOTREACHED*/
@@ -272,7 +275,6 @@ resetProlog()
 #endif
 
   gc_status.blocked    = 0;
-  gc_status.requested  = FALSE;
 #if O_SHIFT_STACKS
   LD->shift_status.blocked = 0;
 #endif
@@ -285,6 +287,8 @@ resetProlog()
 #ifndef O_ABORT_WITH_THROW
   can_abort = TRUE;
 #endif
+
+  updateAlerted(LD);
 }
 
 
@@ -297,7 +301,7 @@ prologToplevel(volatile atom_t goal)
   if ( setjmp(abort_context) != 0 )
   { if ( LD->current_signal )
       unblockSignal(LD->current_signal);
-    
+
     aborted = TRUE;
   } else
 #endif
@@ -311,10 +315,10 @@ prologToplevel(volatile atom_t goal)
     Procedure p;
     word gn;
 
-   resetProlog();
-   fid = PL_open_foreign_frame();
+    resetProlog();
+    fid = PL_open_foreign_frame();
 
-   if ( aborted )
+    if ( aborted )
     { aborted = FALSE;
       gn = PL_new_atom("$abort");
     } else
@@ -326,10 +330,10 @@ prologToplevel(volatile atom_t goal)
     rval = PL_next_solution(qid);
     if ( !rval && (except = PL_exception(qid)) )
     { atom_t a;
-      
+
       tracemode(FALSE, NULL);
       debugmode(DBG_OFF, NULL);
-      setFeatureMask(TAILRECURSION_FEATURE);
+      setPrologFlagMask(PLFLAG_LASTCALL);
       if ( PL_get_atom(except, &a) && a == ATOM_aborted )
       { aborted = TRUE;
       } else if ( !PL_is_functor(except, FUNCTOR_error2) )
@@ -424,6 +428,17 @@ last_arg:
 }
 
 
+static int				/* avoid false alarm in CHR */
+is_ht_capacity(int arity)
+{ int cap = 89;				/* chr_hashtable_store.pl */
+
+  while(cap < arity)
+    cap = cap*2+1;
+
+  return cap == arity;
+}
+
+
 static word
 check_data(Word p, int *recursive)
 { int arity; int n;
@@ -442,6 +457,8 @@ last_arg:
 #endif
          printk("Reference to higher address");
     }
+    if ( p2 == p )
+      printk("Reference to same address");
     if ( !onLocal(p2) && !onGlobal(p2) )
       printk("Illegal reference pointer at 0x%x --> 0x%x", p, p2);
 
@@ -490,22 +507,36 @@ last_arg:
       printk("Indirect data not on global");
     if ( isBignum(*p) )
       return key+(word) valBignum(*p);
-    if ( isReal(*p) )
-      return key+(word) valReal(*p);
+    if ( isFloat(*p) )
+      return key+(word) valFloat(*p);
     if ( isString(*p) )
-    { size_t sz, len;
-      char *s;
+    { if ( isBString(*p) )
+      { size_t sz, len;
+	char *s;
 
-      s = getCharsString(*p, &sz);
+	s = getCharsString(*p, &sz);
 
-      if ( sz != (len=strlen(s)) )
-      { if ( sz < len )
-	  printk("String has inconsistent length: 0x%x", *p);
-	else if ( s[sz] )
-	  printk("String not followed by NUL-char: 0x%x", *p);
+	if ( sz != (len=strlen(s)) )
+	{ if ( sz < len )
+	    printk("String has inconsistent length: 0x%x", *p);
+	  else if ( s[sz] )
+	    printk("String not followed by NUL-char: 0x%x", *p);
 /*	else
-	  printf("String contains NUL-chars: 0x%x", *p);
+	    printf("String contains NUL-chars: 0x%x", *p);
 */
+	}
+      } else
+      { size_t sz, len;
+	pl_wchar_t *s;
+
+	s = getCharsWString(*p, &sz);
+
+	if ( sz != (len=wcslen(s)) )
+	{ if ( sz < len )
+	    printk("String has inconsistent length: 0x%x", *p);
+	  else if ( s[sz] )
+	    printk("String not followed by NUL-char: 0x%x", *p);
+	}
       }
       return key + *addressIndirect(*p);
     }
@@ -519,7 +550,7 @@ last_arg:
 
   if ( isAtom(*p) )
   { uintptr_t idx;
-    uintptr_t mx = entriesBuffer(&atom_array, Atom);
+    uintptr_t mx = GD->atoms.count;
 
     assert(!is_marked(p));
     if ( storage(*p) != STG_STATIC )
@@ -545,16 +576,18 @@ last_arg:
 
     if ( !onGlobal(f) )
       printk("Term at %p not on global stack", f);
-      
+
     if ( tag(f->definition) != TAG_ATOM ||
          storage(f->definition) != STG_GLOBAL )
       printk("Illegal term: 0x%x", *p);
     arity = arityFunctor(f->definition);
-    if (arity <= 0 || arity > 256)
+    if ( arity < 0 )
+      printk("Illegal arity (%d)", arity);
+    else if ( arity > 256 && !is_ht_capacity(arity) )
       printk("Dubious arity (%d)", arity);
     for(n=0; n<arity-1; n++)
       key += check_data(&f->arguments[n], recursive);
-	
+
     p = &f->arguments[n];
     goto last_arg;
   }
