@@ -1,11 +1,11 @@
-/*  $Id: pl-prof.c,v 1.59 2007/06/28 15:05:06 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -53,7 +53,7 @@ static PL_prof_type_t *types[MAX_PROF_TYPES] = { &prof_default_type };
 #define PROFNODE_MAGIC 0x7ae38f24
 
 typedef struct call_node
-{ intptr_t 		    magic;		/* PROFNODE_MAGIC */
+{ intptr_t 	    magic;		/* PROFNODE_MAGIC */
   struct call_node *parent;
   void *            handle;		/* handle to procedure-id */
   PL_prof_type_t   *type;
@@ -80,6 +80,12 @@ activateProfiler(int active ARG_LD)
       (*types[i]->activate)(active);
   }
 
+  if ( active )
+  { /* consider using thread-time? */
+    LD->profile.time_at_start = CpuTime(CPU_USER);
+  }
+  updateAlerted(LD);
+
   sum_ok = FALSE;
 }
 
@@ -200,6 +206,13 @@ stopItimer(void)
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 POSIX version
+
+(*) Originally this code used 1,  which   appeared  to  mean anytime the
+system starts a new time slice for  the task. Hiroo Koshimoto discovered
+this caused very  fast  interrupts   on  MacOSX  `Leopart',  effectively
+causing the monitored program to stop  working.   We  now use 1000 us (1
+millisecond). That should be safe on   most modern hardware. If hardware
+becomes a lot faster we may wish to reduce this value.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #ifdef TIME_WITH_SYS_TIME
@@ -222,14 +235,13 @@ startProfiler(void)
   set_sighandler(SIGPROF, profile);
 
   value.it_interval.tv_sec  = 0;
-  value.it_interval.tv_usec = 1;
+  value.it_interval.tv_usec = 1000;	/* see (*) above */
   value.it_value.tv_sec  = 0;
-  value.it_value.tv_usec = 1;
-  
+  value.it_value.tv_usec = 1000;
+
   if (setitimer(ITIMER_PROF, &value, &ovalue) != 0)
     return PL_error(NULL, 0, MSG_ERRNO, ERR_SYSCALL, setitimer);
   activateProfiler(TRUE PASS_LD);
-  LD->profile.time_at_start = CpuTime(CPU_USER); /* consider using thread-time? */
 
   succeed;
 }
@@ -241,7 +253,7 @@ stopItimer(void)
   value.it_interval.tv_usec = 0;
   value.it_value.tv_sec  = 0;
   value.it_value.tv_usec = 0;
-  
+
   if ( !LD->profile.active )
     return;
   if (setitimer(ITIMER_PROF, &value, &ovalue) != 0)
@@ -270,7 +282,7 @@ stopProfiler(void)
 }
 
 
-static 
+static
 PRED_IMPL("profiler", 2, profiler, 0)
 { PRED_LD
   int val;
@@ -288,7 +300,7 @@ PRED_IMPL("profiler", 2, profiler, 0)
   else
     return stopProfiler();
 }
-	
+
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Prolog query API:
@@ -350,7 +362,7 @@ PRED_IMPL("$prof_sibling_of", 2, prof_sibling_of, PL_FA_NONDETERMINISTIC)
 	  sibling = roots;
 	else if ( get_node(A2, &parent PASS_LD) )
 	  sibling = parent->siblings;
-	else 
+	else
 	  fail;
       }
 
@@ -388,7 +400,7 @@ unify_node_id(term_t t, call_node *n)
   { return (*n->type->unify)(t, n->handle);
   } else
   { GET_LD
-  
+
     return PL_unify_pointer(t, n->handle);
   }
 }
@@ -683,7 +695,7 @@ PRED_IMPL("$prof_procedure_data", 7, prof_procedure_data, PL_FA_TRANSPARENT)
   memset(&sum, 0, sizeof(sum));
   for(n=roots; n; n=n->next)
     count += sumProfile(n, handle, &prof_default_type, &sum, 0 PASS_LD);
-  
+
   if ( count == 0 )
     fail;				/* nothing known about this one */
 
@@ -756,9 +768,11 @@ PRED_IMPL("$profile", 1, profile, PL_FA_TRANSPARENT)
   stopProfiler();
 
   DEBUG(0,
-	Sdprintf("Created %ld nodes (%ld bytes); %ld ticks (%ld overhead)\n",
-		 nodes, nodes*sizeof(call_node),
-		 LD->profile.ticks, LD->profile.accounting_ticks));
+	{ PRED_LD
+	    Sdprintf("Created %ld nodes (%ld bytes); %ld ticks (%ld overhead)\n",
+		     nodes, nodes*sizeof(call_node),
+		     LD->profile.ticks, LD->profile.accounting_ticks);
+	});
 
   return rc;
 }
@@ -783,13 +797,16 @@ clear the flags again.
 static void
 #ifdef __WINDOWS__
 profile(intptr_t count, PL_local_data_t *__PL_ld)
-{ 
+{
 #else /*__WINDOWS__*/
 profile(int sig)
 { GET_LD
 
 #define count 1
 
+  if ( !LD )
+    return;
+
 #if _AIX
   if ( !LD->profile.active )
     return;
@@ -801,7 +818,7 @@ profile(int sig)
 
 #endif /*__WINDOWS__*/
   LD->profile.ticks += count;
-  
+
   if ( accounting )
   { LD->profile.accounting_ticks += count;
   } else if ( current )
@@ -917,16 +934,23 @@ profCall(Definition def ARG_LD)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Exit, resuming execution in node
+
+Exit, resuming execution in node. Note that   we  ignore the node if the
+magic doesn't fit. That can  happen   using  tprofile/1 because the stop
+doesn't need to match the  start.   Actually,  we should clear prof_node
+references when clearing the data, but  this is rather complicated (must
+be synchornised with atom-gc) and  it   still  doesn't cope with foreign
+code.  Considering this is development only, we'll leave this for now.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
 profExit(struct call_node *node ARG_LD)
 { call_node *n;
 
-  accounting = TRUE;
-  assert(!node || node->magic == PROFNODE_MAGIC);
+  if ( node && node->magic != PROFNODE_MAGIC )
+    return;
 
+  accounting = TRUE;
   for(n=current; n && n != node; n=n->parent)
   { n->exits++;
   }
@@ -938,7 +962,8 @@ profExit(struct call_node *node ARG_LD)
 
 void
 profRedo(struct call_node *node ARG_LD)
-{ assert(!node || node->magic == PROFNODE_MAGIC);
+{ if ( node && node->magic != PROFNODE_MAGIC )
+    return;
 
   if ( node )
   { node->redos++;
@@ -947,6 +972,12 @@ profRedo(struct call_node *node ARG_LD)
 }
 
 
+void
+profSetHandle(struct call_node *node, void *handle)
+{ node->handle = handle;
+}
+
+
 		 /*******************************
 		 *	 FOREIGN ACCESS		*
 		 *******************************/
@@ -1010,10 +1041,10 @@ collectSiblingsNode(call_node *n)
 static void
 collectSiblingsTime()
 { GET_LD
-    
+
   if ( !sum_ok )
   { call_node *n;
-    
+
     for(n=roots; n; n=n->next)
       collectSiblingsNode(n);
 
@@ -1040,7 +1071,7 @@ static void
 freeProfileData()
 { GET_LD
   call_node *n;
-  
+
   for(n=roots; n; n=n->next)
   { freeProfileNode(n PASS_LD);
   }
@@ -1062,7 +1093,7 @@ stopItimer()
 {
 }
 
-static 
+static
 PRED_IMPL("profiler", 2, profiler, 0)
 { return notImplemented("profile", 2);
 }
@@ -1080,7 +1111,7 @@ PRED_IMPL("$prof_node", 7, prof_node, 0)
 static
 PRED_IMPL("$prof_sibling_of", 2, prof_sibling_of, PL_FA_NONDETERMINISTIC)
 { return notImplemented("profile_sibling_of", 2);
-} 
+}
 
 static
 PRED_IMPL("$profile", 1, profile, PL_FA_TRANSPARENT)
