@@ -1,4 +1,4 @@
-/*  $Id: pl-read.c,v 1.153 2007/05/16 10:04:41 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -44,7 +44,7 @@ static void	addUTF8Buffer(Buffer b, int c);
 		 *******************************/
 
 #define CharTypeW(c, t, w) \
-	((unsigned)(c) <= 0xff ? (_PL_char_types[(unsigned char)(c)] t) \
+	((unsigned)(c) <= 0xff ? (_PL_char_types[(unsigned)(c)] t) \
 			       : (uflagsW(c) & w))
 
 #define PlBlankW(c)	CharTypeW(c, <= SP, U_SEPARATOR)
@@ -56,6 +56,11 @@ static void	addUTF8Buffer(Buffer b, int c);
 #define PlPunctW(c)	CharTypeW(c, == PU, 0)
 #define PlSoloW(c)	CharTypeW(c, == SO, 0)
 
+int
+unicode_separator(pl_wchar_t c)
+{ return PlBlankW(c);
+}
+
 
 		 /*******************************
 		 *	   CHAR-CONVERSION	*
@@ -99,7 +104,7 @@ pl_current_char_conversion(term_t in, term_t out, control_t h)
 
       if ( !PL_is_variable(in) )
       { if ( PL_get_char(in, &cin, FALSE) )
-	  return PL_unify_char(out, char_conversion_table[cin], CHAR_MODE);
+	  return PL_unify_char(out, char_conversion_table[cin], PL_CHAR);
 	fail;
       }
       ctx = 0;
@@ -116,8 +121,8 @@ pl_current_char_conversion(term_t in, term_t out, control_t h)
 
   fid = PL_open_foreign_frame();
   for( ; ctx < 256; ctx++)
-  { if ( PL_unify_char(in, ctx, CHAR_MODE) &&
-	 PL_unify_char(out, char_conversion_table[ctx], CHAR_MODE) )
+  { if ( PL_unify_char(in, ctx, PL_CHAR) &&
+	 PL_unify_char(out, char_conversion_table[ctx], PL_CHAR) )
       ForeignRedoInt(ctx+1);
 
     PL_rewind_foreign_frame(fid);
@@ -218,6 +223,7 @@ multi-threading.
 typedef struct
 { unsigned char *here;			/* current character */
   unsigned char *base;			/* base of clause */
+  unsigned char *end;			/* end of the clause */
   unsigned char *token_start;		/* start of most recent read token */
   struct token token;			/* current token */
   bool _unget;				/* unget_token() */
@@ -241,12 +247,14 @@ typedef struct
   term_t	subtpos;		/* Report Subterm positions */
   term_t	comments;		/* Report comments */
 
+  atom_t	locked;			/* atom that must be unlocked */
   struct var_table vt;			/* Data about variables */
   struct read_buffer _rb;		/* keep read characters here */
 } read_data, *ReadData;
 
 #define	rdhere		  (_PL_rd->here)
 #define	rdbase		  (_PL_rd->base)
+#define	rdend		  (_PL_rd->end)
 #define	last_token_start  (_PL_rd->token_start)
 #define	cur_token	  (_PL_rd->token)
 #define	unget		  (_PL_rd->_unget)
@@ -256,7 +264,7 @@ typedef struct
 #define var_buffer	  (_PL_rd->vt._var_buffer)
 
 static void
-init_read_data(ReadData _PL_rd, IOSTREAM *in ARG_LD) 
+init_read_data(ReadData _PL_rd, IOSTREAM *in ARG_LD)
 { memset(_PL_rd, 0, sizeof(*_PL_rd));	/* optimise! */
 
   initBuffer(&var_name_buffer);
@@ -267,8 +275,8 @@ init_read_data(ReadData _PL_rd, IOSTREAM *in ARG_LD)
   _PL_rd->flags  = _PL_rd->module->flags; /* change for options! */
   _PL_rd->styleCheck = debugstatus.styleCheck;
   _PL_rd->on_error = ATOM_error;
-  _PL_rd->backquoted_string = trueFeature(BACKQUOTED_STRING_FEATURE);
-  if ( trueFeature(CHARCONVERSION_FEATURE) )
+  _PL_rd->backquoted_string = truePrologFlag(PLFLAG_BACKQUOTED_STRING);
+  if ( truePrologFlag(PLFLAG_CHARCONVERSION) )
     _PL_rd->char_conversion_table = char_conversion_table;
   else
     _PL_rd->char_conversion_table = NULL;
@@ -279,10 +287,32 @@ free_read_data(ReadData _PL_rd)
 { if ( rdbase && rdbase != rb.fast )
     PL_free(rdbase);
 
+  if ( _PL_rd->locked )
+    PL_unregister_atom(_PL_rd->locked);
+
   discardBuffer(&var_name_buffer);
   discardBuffer(&var_buffer);
 }
 
+
+#define NeedUnlock(a) need_unlock(a, _PL_rd)
+#define Unlock(a) unlock(a, _PL_rd)
+
+static void
+need_unlock(atom_t a, ReadData _PL_rd)
+{ _PL_rd->locked = a;
+}
+
+
+static void
+unlock(atom_t a, ReadData _PL_rd)
+{ if (  _PL_rd->locked == a )
+  { _PL_rd->locked = 0;
+    PL_unregister_atom(a);
+  }
+}
+
+
 #define DO_CHARESCAPE true(_PL_rd, CHARESCAPE)
 
 
@@ -443,7 +473,7 @@ reportReadError(ReadData rd)
     fail;
 
   printMessage(ATOM_error, PL_TERM, rd->exception);
-  
+
   if ( rd->on_error == ATOM_dec10 )
     succeed;
 
@@ -476,7 +506,7 @@ clearBuffer(ReadData _PL_rd)
 
   _PL_rd->posp = rdbase;
   _PL_rd->posi = 0;
-}      
+}
 
 
 static void
@@ -669,6 +699,17 @@ add_comment(Buffer b, IOPOS *pos, ReadData _PL_rd ARG_LD)
 }
 
 
+static void
+setErrorLocation(IOPOS *pos, ReadData _PL_rd)
+{ GET_LD
+
+  source_char_no = pos->charno;
+  source_line_pos = pos->linepos;
+  source_line_no = pos->lineno;
+  rb.here = rb.base+1;			/* see rawSyntaxError() */
+}
+
+
 static unsigned char *
 raw_read2(ReadData _PL_rd ARG_LD)
 { int c;
@@ -676,7 +717,7 @@ raw_read2(ReadData _PL_rd ARG_LD)
   bool dotseen = FALSE;
   IOPOS pbuf;					/* comment start */
   IOPOS *pos;
-  
+
   clearBuffer(_PL_rd);				/* clear input buffer */
   source_line_no = -1;
 
@@ -715,7 +756,7 @@ raw_read2(ReadData _PL_rd ARG_LD)
 		strcpy((char *)rb.base, "end_of_file. ");
 		rb.here = rb.base + 14;
 		return rb.base;
-      case '/': if ( _PL_rd->comments && rb.stream->position )
+      case '/': if ( rb.stream->position )
 		{ pbuf = *rb.stream->position;
 		  pbuf.charno--;
 		  pbuf.linepos--;
@@ -742,6 +783,7 @@ raw_read2(ReadData _PL_rd ARG_LD)
 		  if ((last = getchr()) == EOF)
 		  { if ( cbuf )
 		      discardBuffer(cbuf);
+		    setErrorLocation(pos, _PL_rd);
 		    rawSyntaxError("end_of_file_in_block_comment");
 		  }
 		  if ( cbuf )
@@ -763,6 +805,7 @@ raw_read2(ReadData _PL_rd ARG_LD)
 		    { case EOF:
 			if ( cbuf )
 			  discardBuffer(cbuf);
+		        setErrorLocation(pos, _PL_rd);
 			rawSyntaxError("end_of_file_in_block_comment");
 		      case '*':
 			if ( last == '/' )
@@ -787,8 +830,9 @@ raw_read2(ReadData _PL_rd ARG_LD)
 		} else
 		{ set_start_line;
 		  addToBuffer('/', _PL_rd);
-		  if ( isSymbol(c) )
-		  { while( c != EOF && isSymbol(c) )
+		  if ( isSymbolW(c) )
+		  { while( c != EOF && isSymbolW(c) &&
+			   !(c == '`' && _PL_rd->backquoted_string) )
 		    { addToBuffer(c, _PL_rd);
 		      c = getchr();
 		    }
@@ -821,14 +865,23 @@ raw_read2(ReadData _PL_rd ARG_LD)
 			addToBuffer(' ', _PL_rd);
 		    }
 		    if ( c == '\n' )
-		    { if ( something_read )
-			addToBuffer(c, _PL_rd);
+		    { IOPOS p, *pp;
+		      if ( (pp=rb.stream->position) )
+			p = *pp;
+
 		      c = getchr();
 		      if ( c == '%' )
-		      { addUTF8Buffer(cbuf, '\n');
+		      { if ( something_read )
+			{ addToBuffer('\n', _PL_rd);
+			}
+			addUTF8Buffer(cbuf, '\n');
 			addUTF8Buffer(cbuf, '%');
 			continue;
 		      }
+		      Sungetcode(c, rb.stream); /* unsafe: see Sungetcode() */
+		      if ( pp )
+			*pp = p;
+		      c = '\n';
 		    }
 		    break;
 		  }
@@ -841,7 +894,7 @@ raw_read2(ReadData _PL_rd ARG_LD)
 		      addToBuffer(' ', _PL_rd);
 		  }
 		}
-		goto handle_c;
+		goto handle_c;		/* is the newline */
      case '\'': if ( rb.here > rb.base && isDigit(rb.here[-1]) )
 		{ addToBuffer(c, _PL_rd); 		/* <n>' */
 		  if ( rb.here[-2] == '0' )		/* 0'<c> */
@@ -869,8 +922,9 @@ raw_read2(ReadData _PL_rd ARG_LD)
 		set_start_line;
 		dotseen++;
 		c = getchr();
-		if ( isSymbol(c) )
-		{ while( c != EOF && isSymbol(c) )
+		if ( isSymbolW(c) )
+		{ while( c != EOF && isSymbolW(c) &&
+			 !(c == '`' && _PL_rd->backquoted_string) )
 		  { addToBuffer(c, _PL_rd);
 		    c = getchr();
 		  }
@@ -960,19 +1014,24 @@ skipped.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static unsigned char *
-raw_read(ReadData _PL_rd ARG_LD)
-{ if ( (rb.stream->flags & SIO_ISATTY) && Sfileno(rb.stream) >= 0 )
-  { unsigned char *s;
-    ttybuf tab;
-    
+raw_read(ReadData _PL_rd, unsigned char **endp ARG_LD)
+{ unsigned char *s;
+
+  if ( (rb.stream->flags & SIO_ISATTY) && Sfileno(rb.stream) >= 0 )
+  { ttybuf tab;
+
     PushTty(rb.stream, &tab, TTY_SAVE);		/* make sure tty is sane */
     PopTty(rb.stream, &ttytab);
     s = raw_read2(_PL_rd PASS_LD);
     PopTty(rb.stream, &tab);
-
-    return s;
   } else
-    return raw_read2(_PL_rd PASS_LD);
+  { s = raw_read2(_PL_rd PASS_LD);
+  }
+
+  if ( endp )
+    *endp = _PL_rd->_rb.here;
+
+  return s;
 }
 
 
@@ -1025,7 +1084,7 @@ static Variable
 isVarAtom(word w, ReadData _PL_rd)
 { if ( tagex(w) == (TAG_ATOM|STG_GLOBAL) )
     return &baseBuffer(&var_buffer, variable)[w>>7];
-  
+
   return NULL;
 }
 
@@ -1043,7 +1102,7 @@ lookupVariable(const char *name, size_t len, ReadData _PL_rd)
 	       return v;
 	     })
   }
-       
+
   nv = entriesBuffer(&var_buffer, variable);
   next.name      = save_var_name(name, len, _PL_rd);
   next.namelen   = len;
@@ -1063,8 +1122,13 @@ warn_singleton(const char *name)	/* Name in UTF-8 */
     return TRUE;
   if ( name[1] == '_' )			/* __*: never warn */
     return FALSE;
-  if ( name[1] && !PlUpperW(name[1]) )	/* _a: warn */
-    return TRUE;
+  if ( name[1] )			/* _a: warn */
+  { int c;
+
+    utf8_get_char(&name[1], &c);
+    if ( !PlUpperW(c) )
+      return TRUE;
+  }
   return FALSE;
 }
 
@@ -1173,7 +1237,7 @@ skipSpaces(cucharp in)
 
     if ( !PlBlankW(chr) )
       return (ucharp)in;
-  }  
+  }
 
   return (ucharp)in;
 }
@@ -1189,7 +1253,25 @@ SkipIdCont(unsigned char *in)
 
     if ( !PlIdContW(chr) )
       return in;
-  }  
+  }
+
+  return in;
+}
+
+
+static unsigned char *
+SkipSymbol(unsigned char *in, ReadData _PL_rd)
+{ int chr;
+  unsigned char *s;
+
+  for( ; *in; in=s)
+  { s = (unsigned char*)utf8_get_char((char*)in, &chr);
+
+    if ( !isSymbolW(chr) )
+      return in;
+    if ( chr == '`' && _PL_rd->backquoted_string )
+      return in;
+  }
 
   return in;
 }
@@ -1198,17 +1280,17 @@ SkipIdCont(unsigned char *in)
 #define unget_token()	{ unget = TRUE; }
 
 #ifndef O_GMP
-static real
-uint64_to_real(uint64_t i)
+static double
+uint64_to_double(uint64_t i)
 {
 #ifdef __WINDOWS__
   int64_t s = (int64_t)i;
   if ( s >= 0 )
-    return (real)s;
+    return (double)s;
   else
-    return (real)s + 18446744073709551616.0;
+    return (double)s + 18446744073709551616.0;
 #else
-  return (real)i;
+  return (double)i;
 #endif
 }
 #endif
@@ -1223,12 +1305,12 @@ scan_decimal(cucharp *sp, Number n)
 
   for(c = *s; isDigit(c); c = *++s)
   { if ( t > maxi || t * 10 + c - '0' > PLMAXINT )
-    { 
+    {
 #ifdef O_GMP
       n->value.i = (int64_t)t;
       n->type = V_INTEGER;
       promoteToMPZNumber(n);
-      
+
       for(c = *s; isDigit(c); c = *++s)
       { mpz_mul_ui(n->value.mpz, n->value.mpz, 10);
 	mpz_add_ui(n->value.mpz, n->value.mpz, c - '0');
@@ -1238,7 +1320,7 @@ scan_decimal(cucharp *sp, Number n)
       succeed;
 #else
       double maxf = MAXREAL / (double) 10 - (double) 10;
-      double tf = uint64_to_real(t);
+      double tf = uint64_to_double(t);
 
       for(c = *s; isDigit(c); c = *++s)
       { if ( tf > maxf )
@@ -1247,12 +1329,12 @@ scan_decimal(cucharp *sp, Number n)
       }
       *sp = s;
       n->value.f = tf;
-      n->type = V_REAL;
+      n->type = V_FLOAT;
       succeed;
 #endif
     } else
       t = t * 10 + c - '0';
-  }  
+  }
 
   *sp = s;
 
@@ -1291,18 +1373,18 @@ scan_number(cucharp *s, int b, Number n)
 
       succeed;
 #else
-      real maxf = MAXREAL / (real) b - (real) b;
-      real tf = uint64_to_real(t);
+      double maxf = MAXREAL / (double) b - (double) b;
+      double tf = uint64_to_double(t);
 
-      tf = tf * (real)b + (real)d;
+      tf = tf * (double)b + (double)d;
       while((d = digitValue(b, *q)) >= 0)
       { q++;
         if ( tf > maxf )
 	  fail;				/* number too large */
-        tf = tf * (real)b + (real)d;
+        tf = tf * (double)b + (double)d;
       }
       n->value.f = tf;
-      n->type = V_REAL;
+      n->type = V_FLOAT;
       *s = q;
       succeed;
 #endif
@@ -1310,7 +1392,7 @@ scan_number(cucharp *s, int b, Number n)
     { q++;
       t = t * b + d;
     }
-  }  
+  }
 
   n->value.i = t;
   n->type = V_INTEGER;
@@ -1457,7 +1539,7 @@ addUTF8Buffer(Buffer b, int c)
 
 
 static int
-get_string(unsigned char *in, unsigned char **end, Buffer buf,
+get_string(unsigned char *in, unsigned char *ein, unsigned char **end, Buffer buf,
 	   ReadData _PL_rd)
 { int quote;
   int c;
@@ -1487,6 +1569,9 @@ get_string(unsigned char *in, unsigned char **end, Buffer buf,
       } while( c > 0x80 );
 
       goto next;
+    } else if ( in > ein )
+    { errorWarning("end_of_file_in_string", 0, _PL_rd);
+      return FALSE;
     }
 
     addBuffer(buf, c, char);
@@ -1496,7 +1581,7 @@ get_string(unsigned char *in, unsigned char **end, Buffer buf,
     *end = in;
 
   return TRUE;
-}  
+}
 
 
 static void
@@ -1509,7 +1594,7 @@ neg_number(Number n)
 	promoteToMPZNumber(n);
 	mpz_neg(n->value.mpz, n->value.mpz);
 #else
-	n->type = V_REAL;
+	n->type = V_FLOAT;
 	n->value.f = -(double)n->value.i;
 #endif
       } else
@@ -1523,7 +1608,7 @@ neg_number(Number n)
     case V_MPQ:
       assert(0);			/* are not read directly */
 #endif
-    case V_REAL:
+    case V_FLOAT:
       n->value.f = -n->value.f;
   }
 }
@@ -1546,7 +1631,7 @@ str_number(cucharp in, ucharp *end, Number value, int escape)
     switch(in[1])
     { case '\'':			/* 0'<char> */
       { int chr;
-	
+
 	if ( escape && in[2] == '\\' )	/* 0'\n, etc */
 	{ chr = escape_char(in+3, end, 0, NULL);
 	} else
@@ -1605,12 +1690,12 @@ str_number(cucharp in, ucharp *end, Number value, int escape)
     *end = (ucharp)in;
     succeed;
   }
-					/* Real numbers */
+					/* floating point numbers */
 					/* we should use strtod_l() here */
   if ( *in == '.' && isDigit(in[1]) )
   { double n;
 
-    promoteToRealNumber(value);
+    promoteToFloatNumber(value);
     n = 10.0, in++;
     while( isDigit(c = *in) )
     { in++;
@@ -1637,10 +1722,12 @@ str_number(cucharp in, ucharp *end, Number value, int escape)
         break;
     }
 
-    if ( !scan_decimal(&in, &exponent) || exponent.type != V_INTEGER )
+    if ( !isDigit(*in) ||
+	 !scan_decimal(&in, &exponent) ||
+	 exponent.type != V_INTEGER )
       fail;				/* too large exponent */
 
-    promoteToRealNumber(value);
+    promoteToFloatNumber(value);
 
     value->value.f *= pow((double)10.0,
 			  neg_exponent ? -(double)exponent.value.i
@@ -1698,7 +1785,6 @@ static Token
 get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 { int c;
   unsigned char *start;
-  int end;
 
   if ( unget )
   { unget = FALSE;
@@ -1720,14 +1806,15 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
   }
 
   switch(_PL_char_types[c])
-  { case LC:	
+  { case LC:
     lower:
 		{ PL_chars_t txt;
 
 		  rdhere = SkipIdCont(rdhere);
+		symbol:
 		  if ( _PL_rd->styleCheck & CHARSET_CHECK )
 		    checkASCII(start, rdhere-start, "atom");
-		  
+
 		functor:
 		  txt.text.t    = (char *)start;
 		  txt.length    = rdhere-start;
@@ -1735,6 +1822,8 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 		  txt.encoding  = ENC_UTF8;
 		  txt.canonical = FALSE;
 		  cur_token.value.atom = textToAtom(&txt);
+		  NeedUnlock(cur_token.value.atom);
+		  PL_free_text(&txt);
 
 		  cur_token.type = (*rdhere == '(' ? T_FUNCTOR : T_NAME);
 		  DEBUG(9, Sdprintf("%s: %s\n", c == '(' ? "FUNC" : "NAME",
@@ -1747,7 +1836,7 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 		{ rdhere = SkipIdCont(rdhere);
 		  if ( _PL_rd->styleCheck & CHARSET_CHECK )
 		    checkASCII(start, rdhere-start, "variable");
-		  if ( *rdhere == '(' && trueFeature(ALLOW_VARNAME_FUNCTOR) )
+		  if ( *rdhere == '(' && truePrologFlag(ALLOW_VARNAME_FUNCTOR) )
 		    goto functor;
 		  if ( start[0] == '_' &&
 		       rdhere == start + 1 &&
@@ -1780,7 +1869,7 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 		    syntaxError("illegal_number", _PL_rd);
 		}
     case_solo:
-    case SO:	{ cur_token.value.atom = codeToAtom(c);
+    case SO:	{ cur_token.value.atom = codeToAtom(c);		/* not registered */
 		  cur_token.type = (*rdhere == '(' ? T_FUNCTOR : T_NAME);
 		  DEBUG(9, Sdprintf("%s: %s\n",
 				  *rdhere == '(' ? "FUNC" : "NAME",
@@ -1791,34 +1880,20 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
     case SY:	if ( c == '`' && _PL_rd->backquoted_string )
 		  goto case_bq;
 
-                { start = rdhere - 1;
-		  while( isSymbol(*rdhere) &&
-			 !(*rdhere == '`' && _PL_rd->backquoted_string) )
-		    rdhere++;
-		  end = *rdhere;
-
-		  if ( rdhere == start+1 )
-		  { if ( (c == '+' || c == '-') &&	/* +- number */
-			 !must_be_op &&
-			 isDigit(*rdhere) )
-		    { goto case_digit;
-		    }
-		    if ( c == '.' && isBlank(*rdhere) )	/* .<blank> */
-		    { cur_token.type = T_FULLSTOP;
-		      break;
-		    }
+    	        rdhere = SkipSymbol(rdhere, _PL_rd);
+		if ( rdhere == start+1 )
+		{ if ( (c == '+' || c == '-') &&	/* +- number */
+		       !must_be_op &&
+		       isDigit(*rdhere) )
+		  { goto case_digit;
+		  }
+		  if ( c == '.' && isBlank(*rdhere) )	/* .<blank> */
+		  { cur_token.type = T_FULLSTOP;
+		    break;
 		  }
-
-		  if ( _PL_rd->styleCheck & CHARSET_CHECK )
-		    checkASCII(start, rdhere-start, "symbol");
-		  cur_token.value.atom = lookupAtom((char *)start, rdhere-start);
-		  cur_token.type = (end == '(' ? T_FUNCTOR : T_NAME);
-		  DEBUG(9, Sdprintf("%s: %s\n",
-				    end == '(' ? "FUNC" : "NAME",
-				    stringAtom(cur_token.value.atom)));
-
-		  break;
 		}
+
+		goto symbol;
     case PU:	{ switch(c)
 		  { case '{':
 		    case '[':
@@ -1828,7 +1903,6 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 		      { rdhere++;
 			cur_token.value.atom = (c == '[' ? ATOM_nil : ATOM_curl);
 			cur_token.type = rdhere[0] == '(' ? T_FUNCTOR : T_NAME;
-			PL_register_atom(cur_token.value.atom);
 			DEBUG(9, Sdprintf("NAME: %s\n",
 					  stringAtom(cur_token.value.atom)));
 			goto out;
@@ -1844,7 +1918,7 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 		  PL_chars_t txt;
 
 		  initBuffer(&b);
-		  if ( !get_string(rdhere-1, &rdhere, (Buffer)&b, _PL_rd) )
+		  if ( !get_string(rdhere-1, rdend, &rdhere, (Buffer)&b, _PL_rd) )
 		    fail;
 		  txt.text.t    = baseBuffer(&b, char);
 		  txt.length    = entriesBuffer(&b, char);
@@ -1852,6 +1926,7 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 		  txt.encoding  = ENC_UTF8;
 		  txt.canonical = FALSE;
 		  cur_token.value.atom = textToAtom(&txt);
+		  NeedUnlock(cur_token.value.atom);
 		  PL_free_text(&txt);
 		  cur_token.type = (rdhere[0] == '(' ? T_FUNCTOR : T_NAME);
 		  discardBuffer(&b);
@@ -1863,7 +1938,7 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 		  int type;
 
 		  initBuffer(&b);
-		  if ( !get_string(rdhere-1, &rdhere, (Buffer)&b, _PL_rd) )
+		  if ( !get_string(rdhere-1, rdend, &rdhere, (Buffer)&b, _PL_rd) )
 		    fail;
 		  txt.text.t    = baseBuffer(&b, char);
 		  txt.length    = entriesBuffer(&b, char);
@@ -1896,7 +1971,7 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 		  PL_chars_t txt;
 
 		  initBuffer(&b);
-		  if ( !get_string(rdhere-1, &rdhere, (Buffer)&b, _PL_rd) )
+		  if ( !get_string(rdhere-1, rdend, &rdhere, (Buffer)&b, _PL_rd) )
 		    fail;
 		  txt.text.t    = baseBuffer(&b, char);
 		  txt.length    = entriesBuffer(&b, char);
@@ -1911,7 +1986,7 @@ get_token__LD(bool must_be_op, ReadData _PL_rd ARG_LD)
 		  break;
 		}
 #endif
-    default:	
+    default:
     		{ sysError("read/1: tokeniser internal error");
     		  break;		/* make lint happy */
 		}
@@ -2094,7 +2169,7 @@ remainder of the values.
 	realloca(side, sizeof(*side), side_n+1); \
 	side[side_n++] = in_op; \
 	side_p = (side_n == 1 ? 0 : side_p+1);
-	
+
 #define Modify(cpri) \
 	if ( side_p >= 0 && cpri > side[side_p].right_pri ) \
 	{ term_t tmp; \
@@ -2158,8 +2233,8 @@ opPos(op_entry *op, out_entry *args ARG_LD)
 		    PL_LIST, 2, PL_TERM, args[0].tpos,
 		    		PL_TERM, args[1].tpos);
     } else
-    { int s, e;
-      
+    { long s, e;
+
       if ( op->kind == OP_PREFIX )
       { s = fs;
 	e = get_int_arg(args[0].tpos, 2 PASS_LD);
@@ -2176,7 +2251,7 @@ opPos(op_entry *op, out_entry *args ARG_LD)
 		    PL_LONG, fe,
 		    PL_LIST, 1, PL_TERM, args[0].tpos);
     }
-    
+
     return r;
   }
 
@@ -2290,7 +2365,7 @@ complex_term(const char *stop, term_t term, term_t positions,
       pin = PL_new_term_ref();
     else
       pin = 0;
-  
+
     if ( out_n != 0 || side_n != 0 )	/* Check for end of term */
     { if ( !(token = get_token(rmo == 1, _PL_rd)) )
 	fail;
@@ -2326,7 +2401,7 @@ complex_term(const char *stop, term_t term, term_t positions,
 
       if ( rmo == 0 && isOp(name, OP_PREFIX, &in_op, _PL_rd) )
       { DEBUG(9, Sdprintf("Prefix op: %s\n", stringAtom(name)));
-	
+
 	PushOp();
 
 	continue;
@@ -2349,8 +2424,8 @@ complex_term(const char *stop, term_t term, term_t positions,
 	Modify(in_op.left_pri);
 	if ( rmo == 1 )
 	{ Reduce(in_op.left_pri);
-	  PushOp();	
-	
+	  PushOp();
+
 	  continue;
 	}
       }
@@ -2427,7 +2502,7 @@ simple_term(bool must_be_op, term_t term, bool *name,
     case T_NAME:
       *name = TRUE;
       PL_put_atom(term, token->value.atom);
-      PL_unregister_atom(token->value.atom);
+      Unlock(token->value.atom);
       goto atomic_out;
     case T_NUMBER:
       _PL_put_number(term, &token->value.number);
@@ -2436,8 +2511,8 @@ simple_term(bool must_be_op, term_t term, bool *name,
       if ( positions )
       { PL_unify_term(positions,
 		      PL_FUNCTOR, FUNCTOR_minus2,
-		      PL_LONG, token->start,
-		      PL_LONG, token->end);
+		      PL_INTPTR, token->start,
+		      PL_INTPTR, token->end);
       }
       succeed;
     case T_STRING:
@@ -2445,15 +2520,15 @@ simple_term(bool must_be_op, term_t term, bool *name,
       if ( positions )
       { PL_unify_term(positions,
 		      PL_FUNCTOR, FUNCTOR_string_position2,
-		      PL_LONG, token->start,
-		      PL_LONG, token->end);
+		      PL_INTPTR, token->start,
+		      PL_INTPTR, token->end);
       }
       succeed;
     case T_FUNCTOR:
       { if ( must_be_op )
 	{ *name = TRUE;
 	  PL_put_atom(term, token->value.atom);
-	  PL_unregister_atom(token->value.atom);
+	  Unlock(token->value.atom);
 	  goto atomic_out;
 	} else
 	{ term_t av[16];
@@ -2464,6 +2539,7 @@ simple_term(bool must_be_op, term_t term, bool *name,
 	  term_t pa;			/* argument */
 	  term_t pe;			/* term-end */
 	  term_t ph;
+	  int unlock;
 
 	  if ( positions )
 	  { pa = PL_new_term_ref();
@@ -2471,15 +2547,17 @@ simple_term(bool must_be_op, term_t term, bool *name,
 	    ph = PL_new_term_ref();
 	    PL_unify_term(positions,
 			  PL_FUNCTOR, FUNCTOR_term_position5,
-			  PL_LONG, token->start,
+			  PL_INTPTR, token->start,
 			  PL_TERM, pe,
-			  PL_LONG, token->start,
-			  PL_LONG, token->end,
+			  PL_INTPTR, token->start,
+			  PL_INTPTR, token->end,
 			  PL_TERM, pa);
 	  } else
 	    pa = pe = ph = 0;
 
 	  functor = token->value.atom;
+	  unlock = (_PL_rd->locked == functor);
+	  _PL_rd->locked = 0;
 	  argc = 0, argv;
 	  get_token(must_be_op, _PL_rd); /* skip '(' */
 
@@ -2494,7 +2572,11 @@ simple_term(bool must_be_op, term_t term, bool *name,
 	    if ( positions )
 	    { PL_unify_list(pa, ph, pa);
 	    }
-	    TRY( complex_term(",)", argv[argc], ph, _PL_rd PASS_LD) );
+	    if ( !complex_term(",)", argv[argc], ph, _PL_rd PASS_LD) )
+	    { if ( unlock )
+		PL_unregister_atom(functor);
+	      fail;
+	    }
 	    argc++;
 	    token = get_token(must_be_op, _PL_rd); /* `,' or `)' */
 	  } while(token->value.character == ',');
@@ -2505,7 +2587,8 @@ simple_term(bool must_be_op, term_t term, bool *name,
 	  }
 
 	  build_term(term, functor, argc, argv, _PL_rd PASS_LD);
-	  PL_unregister_atom(functor);
+	  if ( unlock )
+	    PL_unregister_atom(functor);
 	}
 	succeed;
       }
@@ -2536,7 +2619,7 @@ simple_term(bool must_be_op, term_t term, bool *name,
 
 		PL_unify_term(positions,
 			      PL_FUNCTOR, FUNCTOR_brace_term_position3,
-			      PL_LONG, token->start,
+			      PL_INTPTR, token->start,
 			      PL_TERM, pe,
 			      PL_TERM, pa);
 	      } else
@@ -2563,7 +2646,7 @@ simple_term(bool must_be_op, term_t term, bool *name,
 
 		PL_unify_term(positions,
 			      PL_FUNCTOR, FUNCTOR_list_position4,
-			      PL_LONG, token->start,
+			      PL_INTPTR, token->start,
 			      PL_TERM, pe,
 			      PL_TERM, pa,
 			      PL_TERM, pt);
@@ -2591,7 +2674,7 @@ term is to be written.
 		readValHandle(tmp, argp++, _PL_rd PASS_LD);
 		setVar(*argp);
 		setHandle(tail, makeRef(argp));
-		
+
 		token = get_token(must_be_op, _PL_rd);
 
 		switch(token->value.character)
@@ -2666,7 +2749,7 @@ read_term(term_t term, ReadData rd ARG_LD)
   term_t result;
   Word p;
 
-  if ( !(rd->base = raw_read(rd PASS_LD)) )
+  if ( !(rd->base = raw_read(rd, &rd->end PASS_LD)) )
     fail;
 
   rd->here = rd->base;
@@ -2744,7 +2827,7 @@ backSkipBlanks(const unsigned char *start, const unsigned char *end)
 word
 pl_raw_read2(term_t from, term_t term)
 { GET_LD
-  unsigned char *s, *t2, *top;
+  unsigned char *s, *e, *t2, *top;
   read_data rd;
   word rval;
   IOSTREAM *in;
@@ -2755,19 +2838,22 @@ pl_raw_read2(term_t from, term_t term)
     fail;
 
   init_read_data(&rd, in PASS_LD);
-  if ( !(s = raw_read(&rd PASS_LD)) )
+  if ( !(s = raw_read(&rd, &e PASS_LD)) )
   { rval = PL_raise_exception(rd.exception);
     goto out;
   }
 
 					/* strip the input from blanks */
-  top = backSkipBlanks(s, rd._rb.here-1);
+  top = backSkipBlanks(s, e-1);
   t2 = backSkipUTF8(s, top, &chr);
   if ( chr == '.' )
     top = backSkipBlanks(s, t2);
+					/* watch for "0' ." */
+  if ( top < e && top-2 >= s && top[-1] == '\'' && top[-2] == '0' )
+    top++;
   *top = EOS;
   s = skipSpaces(s);
-  
+
   txt.text.t    = (char*)s;
   txt.length    = top-s;
   txt.storage   = PL_CHARS_HEAP;
@@ -2849,45 +2935,39 @@ retry:
 }
 
 
-static
-PRED_IMPL("read_clause", 2, read_clause, 0)
-{ PRED_LD
-  int rval;
+static int
+read_clause_pred(term_t from, term_t term ARG_LD)
+{ int rval;
+  IOSTREAM *s;
 
-  switch( CTX_ARITY )
-  { case 1:
-    { IOSTREAM *s;
+  if ( !getInputStream(from, &s) )
+    fail;
+  rval = read_clause(s, term PASS_LD);
+  if ( Sferror(s) )
+    return streamStatus(s);
+  else
+    PL_release_stream(s);
 
-      if ( !getInputStream(0, &s) )	/* Scurin */
-	fail;
-      rval = read_clause(s, A1 PASS_LD);
-      if ( Sferror(s) )
-	return streamStatus(s);
-      else
-	PL_release_stream(s);
-
-      return rval;
-    }
-    case 2:
-    { IOSTREAM *s;
-
-      if ( !getInputStream(A1, &s) )
-	fail;
-      rval = read_clause(s, A2 PASS_LD);
-      if ( Sferror(s) )
-	return streamStatus(s);
-      else
-	PL_release_stream(s);
-
-      return rval;
-    }
-  }
-  assert(0);
-  fail;
+  return rval;
 }
 
 
-static const opt_spec read_term_options[] = 
+static
+PRED_IMPL("read_clause", 1, read_clause, 0)
+{ PRED_LD
+
+  return read_clause_pred(0, A2 PASS_LD);
+}
+
+static
+PRED_IMPL("read_clause", 2, read_clause, 0)
+{ PRED_LD
+
+  return read_clause_pred(A1, A2 PASS_LD);
+}
+
+
+static const opt_spec read_term_options[] =
 { { ATOM_variable_names,    OPT_TERM },
   { ATOM_variables,         OPT_TERM },
   { ATOM_singletons,        OPT_TERM },
@@ -2940,7 +3020,7 @@ retry:
   if ( mname )
   { rd.module = lookupModule(mname);
     rd.flags  = rd.module->flags;
-  } 
+  }
 
   if ( charescapes != -1 )
   { if ( charescapes )
@@ -2998,15 +3078,15 @@ pl_read_term(term_t term, term_t options)
 		 *	   TERM <->ATOM		*
 		 *******************************/
 
-word
-pl_atom_to_term(term_t atom, term_t term, term_t bindings)
+static int
+atom_to_term(term_t atom, term_t term, term_t bindings)
 { GET_LD
   PL_chars_t txt;
 
-  if ( PL_is_variable(atom) )
+  if ( !bindings && PL_is_variable(atom) ) /* term_to_atom(+, -) */
   { char buf[1024];
     size_t bufsize = sizeof(buf);
-    word rval;
+    int rval;
     char *s = buf;
     IOSTREAM *stream;
     PL_chars_t txt;
@@ -3015,7 +3095,7 @@ pl_atom_to_term(term_t atom, term_t term, term_t bindings)
     stream->encoding = ENC_UTF8;
     PL_write_term(stream, term, 1200, PL_WRT_QUOTED);
     Sflush(stream);
-    
+
     txt.text.t = s;
     txt.length = bufsize;
     txt.storage = PL_CHARS_HEAP;
@@ -3030,18 +3110,20 @@ pl_atom_to_term(term_t atom, term_t term, term_t bindings)
     return rval;
   }
 
-  if ( PL_get_text_ex(atom, &txt, CVT_ALL) )
+  if ( PL_get_text(atom, &txt, CVT_ALL|CVT_EXCEPTION) )
   { GET_LD
     read_data rd;
-    word rval;
+    int rval;
     IOSTREAM *stream;
     source_location oldsrc = LD->read_source;
 
     stream = Sopen_text(&txt, "r");
 
     init_read_data(&rd, stream PASS_LD);
-    if ( PL_is_variable(bindings) || PL_is_list(bindings) )
+    if ( bindings && (PL_is_variable(bindings) || PL_is_list(bindings)) )
       rd.varnames = bindings;
+    else if ( bindings )
+      return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, bindings);
 
     if ( !(rval = read_term(term, &rd PASS_LD)) && rd.has_exception )
       rval = PL_raise_exception(rd.exception);
@@ -3056,6 +3138,18 @@ pl_atom_to_term(term_t atom, term_t term, term_t bindings)
 }
 
 
+static
+PRED_IMPL("atom_to_term", 3, atom_to_term, 0)
+{ return atom_to_term(A1, A2, A3);
+}
+
+
+static
+PRED_IMPL("term_to_atom", 2, term_to_atom, 0)
+{ return atom_to_term(A2, A1, 0);
+}
+
+
 int
 PL_chars_to_term(const char *s, term_t t)
 { GET_LD
@@ -3063,7 +3157,7 @@ PL_chars_to_term(const char *s, term_t t)
   int rval;
   IOSTREAM *stream = Sopen_string(NULL, (char *)s, -1, "r");
   source_location oldsrc = LD->read_source;
-    
+
   init_read_data(&rd, stream PASS_LD);
   PL_put_variable(t);
   if ( !(rval = read_term(t, &rd PASS_LD)) && rd.has_exception )
@@ -3082,4 +3176,6 @@ PL_chars_to_term(const char *s, term_t t)
 BeginPredDefs(read)
   PRED_DEF("read_clause", 1, read_clause, 0)
   PRED_DEF("read_clause", 2, read_clause, 0)
+  PRED_DEF("atom_to_term", 3, atom_to_term, 0)
+  PRED_DEF("term_to_atom", 2, term_to_atom, 0)
 EndPredDefs
