@@ -1,4 +1,4 @@
-/*  $Id: pl-rl.c,v 1.53 2007/03/20 20:40:37 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -25,10 +25,10 @@
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 This module binds the  SWI-Prolog  terminal   I/O  to  the  GNU readline
 library. Existence of this  this  library   is  detected  by  configure.
-Binding is achieved by rebinding the read function of the Sinput stream. 
+Binding is achieved by rebinding the read function of the Sinput stream.
 
 This  module  only  depends  on  the  public  interface  as  defined  by
-SWI-Prolog.h (pl-itf.h) and SWI-Stream.h (pl-stream.h).
+SWI-Prolog.h and SWI-Stream.h
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 
@@ -37,8 +37,8 @@ SWI-Prolog.h (pl-itf.h) and SWI-Stream.h (pl-stream.h).
 #endif
 #include <string.h>
 #include <stdlib.h>
-#include "pl-stream.h"
-#include "pl-itf.h"
+#include "SWI-Stream.h"
+#include "SWI-Prolog.h"
 
 #ifdef __WINDOWS__
 #ifdef WIN64
@@ -74,6 +74,7 @@ SWI-Prolog.h (pl-itf.h) and SWI-Stream.h (pl-stream.h).
 
 #undef ESC				/* will be redefined ... */
 #include <stdio.h>			/* readline needs it */
+#include <errno.h>
 #define savestring(x)			/* avoid definition there */
 #include <readline/readline.h>
 extern int rl_done;			/* should be in readline.h, but */
@@ -97,8 +98,8 @@ extern char *filename_completion_function(const char *, int);
 #define rl_completion_matches completion_matches
 #endif
 
-#ifndef HAVE_RL_READLINE_STATE		/* before version 4.2 */
-static int rl_readline_state = 0;
+#ifndef RL_STATE_INITIALIZED
+int rl_readline_state = 0;
 #define RL_STATE_INITIALIZED 0
 #endif
 #ifndef HAVE_RL_SET_PROMPT
@@ -111,6 +112,15 @@ static int rl_readline_state = 0;
 #define rl_cleanup_after_signal() (void)0
 #endif
 
+#if !defined(HAVE_RL_DONE) && defined(HAVE_DECL_RL_DONE) && !HAVE_DECL_RL_DONE
+/* surely not provided, so we provide a dummy.  We do this as
+   a global symbol, so if there is one in a dynamic library it
+   will work anyway.
+*/
+int rl_done;
+#endif
+
+
 static foreign_t
 pl_rl_read_init_file(term_t file)
 { char *f;
@@ -141,20 +151,20 @@ static foreign_t
 pl_rl_add_history(term_t text)
 { atom_t a;
   static atom_t last = 0;
-  PL_chars_t txt;
 
   if ( PL_get_atom_ex(text, &a) )
-  { if ( a != last )
+  { char *txt;
+
+    if ( a != last )
     { if ( last )
 	PL_unregister_atom(last);
       last = a;
       PL_register_atom(last);
 
-      PL_get_text(text, &txt, CVT_ATOM);
-      PL_text_recode(&txt, ENC_UTF8);
-
-      add_history(txt.text.t);
-      PL_free_text(&txt);
+      if ( PL_get_chars(text, &txt, CVT_ATOM|REP_MB|CVT_EXCEPTION) )
+	add_history(txt);
+      else
+	return FALSE;
     }
 
     return TRUE;
@@ -270,7 +280,7 @@ prepare_signals()
   for(s=signals; s->signo != -1; s++)
   { struct sigaction new;
 
-    memset(&new, 0, sizeof(new));	
+    memset(&new, 0, sizeof(new));
     new.sa_handler = rl_sighandler;
     sigaction(s->signo, &new, &s->old_state);
   }
@@ -299,6 +309,7 @@ rl_sighandler(int sig)
     rl_free_line_state ();
   rl_cleanup_after_signal ();
   restore_signals();
+  Sreset();
 
   for(s=signals; s->signo != -1; s++)
   { if ( s->signo == sig )
@@ -319,7 +330,6 @@ rl_sighandler(int sig)
   DEBUG(3, Sdprintf("Resetting after signal\n"));
   prepare_signals();
   rl_reset_after_signal ();
-  Sreset();
 }
 
 
@@ -344,7 +354,7 @@ reentrant access is tried.
 #ifdef HAVE_RL_EVENT_HOOK
 static int
 event_hook()
-{ if ( Sinput && Sinput->position )
+{ if ( Sinput->position )
   { intptr_t c0 = Sinput->position->charno;
 
     while( !input_on_fd(0) )
@@ -354,9 +364,7 @@ event_hook()
 	  rl_set_prompt(my_prompt);
 	rl_forced_update_display();
 	c0 = Sinput->position->charno;
-#ifdef HAVE_RL_DONE
 	rl_done = FALSE;
-#endif
       }
     }
   } else
@@ -391,12 +399,12 @@ Sread_readline(void *handle, char *buf, size_t size)
 #endif
 
   PL_write_prompt(ttymode == PL_NOTTY);
-  
+
   switch( ttymode )
   { case PL_RAWTTY:			/* get_single_char/1 */
 #ifdef O_RLC
     { int chr = getkey();
-      
+
       if ( chr == 04 || chr == 26 )
 	return 0;			/* EOF */
 
@@ -442,7 +450,7 @@ Sread_readline(void *handle, char *buf, size_t size)
       { char *oldp = my_prompt;
 
 	my_prompt = prompt ? store_string(prompt) : (char *)NULL;
-  
+
 	if ( sig_at_level == in_readline )
 	{ sig_at_level = -1;
 	  reset_readline();
@@ -450,7 +458,7 @@ Sread_readline(void *handle, char *buf, size_t size)
 
 	if ( in_readline++ )
 	{ int state = rl_readline_state;
-  
+
 	  rl_clear_pending_input();
 	  rl_discard_argument();
 	  rl_deprep_terminal();
@@ -458,9 +466,7 @@ Sread_readline(void *handle, char *buf, size_t size)
 	  line = pl_readline(prompt);
 	  rl_prep_terminal(FALSE);
 	  rl_readline_state = state;
-#ifdef HAVE_RL_DONE
 	  rl_done = 0;
-#endif
 	} else
 	  line = pl_readline(prompt);
 	in_readline--;
@@ -472,7 +478,7 @@ Sread_readline(void *handle, char *buf, size_t size)
 
       if ( line )
       { size_t l = strlen(line);
-	  
+
 	if ( l >= size )
 	{ PL_warning("Input line too long");	/* must be tested! */
 	  l = size-1;
@@ -507,7 +513,7 @@ prolog_complete(int ignore, int key)
       rl_delete_text(rl_point-1, rl_point);
       rl_point -= 1;
 #else
-      rl_delete(-1);
+      rl_delete(-1, key);
 #endif
     }
     rl_end_undo_group();
@@ -521,7 +527,7 @@ prolog_complete(int ignore, int key)
 static char *
 atom_generator(const char *prefix, int state)
 { char *s = PL_atom_generator(prefix, state);
-  
+
   if ( s )
     return strcpy(PL_malloc(1 + strlen(s)), s);
 
@@ -548,9 +554,9 @@ prolog_completion(const char *text, int start, int end)
 
 install_t
 PL_install_readline()
-{ 
+{
 #ifndef __WINDOWS__
-  if ( !trueFeature(TTY_CONTROL_FEATURE) || !isatty(0) )
+  if ( !truePrologFlag(PLFLAG_TTY_CONTROL) || !isatty(0) )
     return;
 #endif
 
@@ -578,8 +584,8 @@ PL_install_readline()
   PL_register_foreign("rl_add_history",    1, pl_rl_add_history, PL_FA_NOTRACE);
   PL_register_foreign("rl_write_history",  1, pl_rl_write_history, 0);
   PL_register_foreign("rl_read_history",   1, pl_rl_read_history, 0);
-  PL_set_feature("readline",    PL_BOOL, TRUE);
-  PL_set_feature("tty_control", PL_BOOL, TRUE);
+  PL_set_prolog_flag("readline",    PL_BOOL, TRUE);
+  PL_set_prolog_flag("tty_control", PL_BOOL, TRUE);
   PL_license("gpl", "GNU Readline library");
 }
 
