@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -52,7 +52,7 @@ forwards void initSignals(void);
 #undef I
 #define I TAGEX_INDIRECT
 
-const unsigned int tagtypeex[] = 
+const unsigned int tagtypeex[] =
 {
 	    /* var     attvar  int    float   atom   string    term     ref */
 /* static */	0,	0,	0,	0,	0,	0,	0,	0,
@@ -72,17 +72,21 @@ setupProlog(void)
   LD->pending_signals = 0;
 
   startCritical;
+  DEBUG(1, Sdprintf("wam_table ...\n"));
+  initWamTable();
+  DEBUG(1, Sdprintf("character types ...\n"));
   initCharTypes();
+  DEBUG(1, Sdprintf("foreign predicates ...\n"));
   initForeign();
 #if HAVE_SIGNAL
   DEBUG(1, Sdprintf("Prolog Signal Handling ...\n"));
-  if ( trueFeature(SIGNALS_FEATURE) )
+  if ( truePrologFlag(PLFLAG_SIGNALS) )
     initSignals();
 #endif
   DEBUG(1, Sdprintf("Stacks ...\n"));
-  if ( !initPrologStacks(GD->options.localSize, 
-			 GD->options.globalSize, 
-			 GD->options.trailSize, 
+  if ( !initPrologStacks(GD->options.localSize,
+			 GD->options.globalSize,
+			 GD->options.trailSize,
 			 GD->options.argumentSize) )
     fatalError("Not enough address space to allocate Prolog stacks");
   initPrologLocalData();
@@ -90,7 +94,7 @@ setupProlog(void)
   DEBUG(1, Sdprintf("Atoms ...\n"));
   initAtoms();
   DEBUG(1, Sdprintf("Features ...\n"));
-  initFeatures();
+  initPrologFlags();
   DEBUG(1, Sdprintf("Functors ...\n"));
   initFunctors();
   DEBUG(1, Sdprintf("Modules ...\n"));
@@ -115,9 +119,8 @@ setupProlog(void)
   DEBUG(1, Sdprintf("Tracer ...\n"));
   initTracer();
   debugstatus.styleCheck = SINGLETON_CHECK;
-  DEBUG(1, Sdprintf("wam_table ...\n"));
-  initWamTable();
   DEBUG(1, Sdprintf("IO ...\n"));
+  initFiles();
   initIO();
   initCharConversion();
   DEBUG(1, Sdprintf("Term ...\n"));
@@ -141,7 +144,8 @@ initPrologLocalData(void)
   environment_frame = (LocalFrame) NULL;
   LD->statistics.inferences = 0;
   LD->float_format = "%g";
-  LD->feature.write_attributes = PL_WRT_ATTVAR_IGNORE;
+  LD->prolog_flag.write_attributes = PL_WRT_ATTVAR_IGNORE;
+  updateAlerted(LD);
 }
 
 
@@ -181,7 +185,7 @@ static struct signame
 { int 	      sig;
   const char *name;
   int	      flags;
-} signames[] = 
+} signames[] =
 {
 #ifdef SIGHUP
   { SIGHUP,	"hup",    0},
@@ -190,13 +194,13 @@ static struct signame
 #ifdef SIGQUIT
   { SIGQUIT,	"quit",   0},
 #endif
-  { SIGILL,	"ill",    PLSIG_THROW},
+  { SIGILL,	"ill",    0},
   { SIGABRT,	"abrt",   0},
   { SIGFPE,	"fpe",    PLSIG_THROW},
 #ifdef SIGKILL
   { SIGKILL,	"kill",   0},
 #endif
-  { SIGSEGV,	"segv",   PLSIG_THROW},
+  { SIGSEGV,	"segv",   0},
 #ifdef SIGPIPE
   { SIGPIPE,	"pipe",   0},
 #endif
@@ -232,7 +236,7 @@ static struct signame
   { SIGTRAP,	"trap",   0},
 #endif
 #ifdef SIGBUS
-  { SIGBUS,	"bus",    PLSIG_THROW},
+  { SIGBUS,	"bus",    0},
 #endif
 #ifdef SIGSTKFLT
   { SIGSTKFLT,	"stkflt", 0},
@@ -312,7 +316,7 @@ PL_get_signum_ex(term_t sig, int *n)
   int i = -1;
 
   if ( PL_get_integer(sig, &i) )
-  { 
+  {
   } else if ( PL_get_chars(sig, &s, CVT_ATOM) )
   { i = signal_index(s);
   } else
@@ -454,7 +458,7 @@ dispatch_signal(int sig, int sync, Code PC)
     { predname = NULL;
       arity    = 0;
     }
-      
+
     PL_error(predname, arity, NULL, ERR_SIGNALLED, sig, signal_name(sig));
     if ( !sync )
       unblockGC(PASS_LD1);
@@ -513,7 +517,7 @@ set_sighandler(int sig, handler_t func)
 #else
   return signal(sig, func);
 #endif
-} 
+}
 
 #ifdef HAVE_SIGINFO_H
 #include <siginfo.h>
@@ -523,7 +527,7 @@ set_sighandler(int sig, handler_t func)
 #if defined(O_SEGV_HANDLING) && defined(O_DYNAMIC_STACKS)
 static handler_t
 set_stack_guard_handler(int sig, void *func)
-{ 
+{
 #ifdef HAVE_SIGACTION
   struct sigaction old;
   struct sigaction new;
@@ -537,7 +541,7 @@ set_stack_guard_handler(int sig, void *func)
   new.sa_flags     = SA_RESTART;
 #endif
 
-  if ( !trueFeature(SIGNALS_FEATURE) &&
+  if ( !truePrologFlag(PLFLAG_SIGNALS) &&
        !GD->options.silent )
   {			/* We need double \\ to get a single at Sdprintf() */
     Sdprintf("\\% Prolog still handles SIG_SEGV\n"
@@ -593,7 +597,7 @@ static void
 sig_exception_handler(int sig)
 { if ( LD && LD->pending_exception )
   { record_t ex = LD->pending_exception;
-	  
+
     LD->pending_exception = 0;
 
     PL_put_variable(exception_bin);
@@ -623,7 +627,7 @@ gc_handler(int sig)
 static void
 initSignals(void)
 { struct signame *sn = signames;
-  
+
 #ifdef SIGPIPE
   set_sighandler(SIGPIPE, SIG_IGN);
 #endif
@@ -653,7 +657,7 @@ initSignals(void)
 void
 cleanupSignals(void)
 { struct signame *sn = signames;
-  
+
   for( ; sn->name; sn++)
     unprepareSignal(sn->sig);
 }
@@ -691,7 +695,7 @@ listBlocked()
   int i;
 
   sigprocmask(SIG_BLOCK, NULL, &current);
-  
+
   Sdprintf("Blocked: ");
   for(i=1; i<32; i++)
   { if ( sigismember(&current, i) )
@@ -727,7 +731,7 @@ unblockSignals(sigset_t *old)
   { sigset_t set;
 
     allSignalMask(&set);
-  
+
     sigprocmask(SIG_UNBLOCK, &set, NULL);
     DEBUG(1, Sdprintf("UnBlocked all signals\n"));
   }
@@ -847,6 +851,8 @@ handleSignals(Code PC)
 out:
   if ( exception_term )
     return -1;
+  if ( done )
+    updateAlerted(ld);
 
   return done;
 }
@@ -860,19 +866,53 @@ endCritical__LD(ARG1_LD)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-signal(+Signal, :Old, :Handler)
-	Assign Handler to be called if signal arrises.  Example:
+on_signal(?SigNum, ?SigName, :OldHandler, :NewHandler)
 
-		signal(usr1, Old, handle_user_1/1).
+Assign NewHandler to be called if signal arrives.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-foreign_t
-pl_on_signal(term_t sig, term_t name, term_t old, term_t new)
+static int
+get_meta_arg(term_t arg, term_t m, term_t t)
+{ if ( PL_is_functor(arg, FUNCTOR_colon2) )
+  { _PL_get_arg(1, arg, m);
+    _PL_get_arg(2, arg, t);
+    return TRUE;
+  }
+
+  return PL_error(NULL, 0, NULL, ERR_TYPE,
+		  ATOM_meta_argument, arg);
+}
+
+
+static int
+get_module(term_t t, Module *m)
+{ atom_t a;
+
+  if ( !PL_get_atom_ex(t, &a) )
+    return FALSE;
+  *m = PL_new_module(a);
+
+  return TRUE;
+}
+
+
+static
+PRED_IMPL("$on_signal", 4, on_signal, 0)
 { int sign = -1;
   SigHandler sh;
   char *sn;
   atom_t a;
-  Module m = NULL;
+  term_t mold = PL_new_term_ref();
+  term_t mnew = PL_new_term_ref();
+
+  term_t sig  = A1;
+  term_t name = A2;
+  term_t old  = A3;
+  term_t new  = A4;
+
+  if ( !get_meta_arg(old, mold, old) ||
+       !get_meta_arg(new, mnew, new) )
+    return FALSE;
 
   if ( PL_get_integer(sig, &sign) && sign >= 1 && sign <= MAXSIGNAL )
   { TRY(PL_unify_atom_chars(name, signal_name(sign)));
@@ -893,25 +933,19 @@ pl_on_signal(term_t sig, term_t name, term_t old, term_t new)
   } else if ( sh->predicate )			/* call predicate */
   { Definition def = sh->predicate->definition;
 
-    if ( def->module == MODULE_user )
-    { TRY(PL_unify_atom(old, def->functor->name));
-    } else
-    { TRY(PL_unify_term(old,
-			PL_FUNCTOR, FUNCTOR_colon2,
-			   PL_ATOM, def->module->name, 
-			   PL_ATOM, def->functor->name));
-    }
+    if ( !PL_unify_atom(mold, def->module->name) ||
+	 !PL_unify_atom(old, def->functor->name) )
+      return FALSE;
   } else if ( sh->handler )
   { TRY(PL_unify_term(old,
 		      PL_FUNCTOR, FUNCTOR_foreign_function1,
 		      PL_POINTER, sh->handler));
-  }    
+  }
 
-  if ( PL_compare(old, new) == 0 )
+  if ( PL_compare(old, new) == 0 &&
+       PL_compare(mold, mnew) == 0 )
     succeed;					/* no change */
 
-  PL_strip_module(new, &m, new);
-
   if ( PL_get_atom(new, &a) )
   { if ( a == ATOM_default )
     { unprepareSignal(sign);
@@ -921,8 +955,13 @@ pl_on_signal(term_t sig, term_t name, term_t old, term_t new)
       sh->handler   = NULL;
       sh->predicate = NULL;
     } else
-    { predicate_t pred = lookupProcedure(PL_new_functor(a, 1), m);
-      
+    { Module m;
+      predicate_t pred;
+
+      if ( !get_module(mnew, &m) )
+	return FALSE;
+      pred = lookupProcedure(PL_new_functor(a, 1), m);
+
       sh = prepareSignal(sign);
       clear(sh, PLSIG_THROW);
       sh->handler = NULL;
@@ -1000,6 +1039,7 @@ initPrologStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argum
   base_addresses[STG_LOCAL]  = (uintptr_t)lBase;
   base_addresses[STG_GLOBAL] = (uintptr_t)gBase;
   base_addresses[STG_TRAIL]  = (uintptr_t)tBase;
+  *gBase++ = MARK_MASK;			/* see sweep_global_mark() */
   emptyStacks();
 
   DEBUG(1, Sdprintf("base_addresses[STG_LOCAL] = %p\n",
@@ -1041,6 +1081,7 @@ emptyStacks()
   exception_printed     = PL_new_term_ref();
   LD->exception.tmp     = PL_new_term_ref();
   LD->exception.pending = PL_new_term_ref();
+  LD->trim.dummy        = PL_new_term_ref();
 #ifdef O_ATTVAR
   LD->attvar.head	= PL_new_term_ref();
   LD->attvar.tail       = PL_new_term_ref();
@@ -1050,7 +1091,7 @@ emptyStacks()
   destroyGlobalVars();
 #endif
 
-  LD->mark_bar          = (Word)(~(uintptr_t)0);
+  LD->mark_bar          = NO_MARK_BAR;
 }
 
 #if O_DYNAMIC_STACKS
@@ -1239,7 +1280,7 @@ mapOrOutOf(Stack s)
 
 #ifndef O_SEGV_HANDLING
 void
-ensureRoomStack(Stack s, size_t bytes)
+ensure_room_stack(Stack s, size_t bytes)
 { while((char *)s->max <= &((char *)s->top)[bytes])
     mapOrOutOf(s);
 }
@@ -1273,15 +1314,38 @@ unmap(Stack s)
 
 /* mmap() version */
 
+#ifdef SECURE_GC
+#include <stdio.h>
+
+static void
+fixed_bases(void **gbase, void**tbase)
+{ const char *bases;
+
+  if ( PL_thread_self() == 1 && (bases=getenv("PL_STACK_BASES")) )
+  { unsigned long gb, tb;
+
+    if ( sscanf(bases, "G%lx+T%lx", &gb, &tb) == 2 )
+    { *gbase = (void*)gb;
+      *tbase = (void*)tb;
+    } else
+    { Sdprintf("Usage: export PL_STACK_BASES=\"G<gBase>+T<tBase>\"\n");
+      exit(1);
+    }
+  }
+}
+
+#endif /*SECURE_GC*/
+
 static int
 allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
-{ caddress lbase, gbase, tbase, abase;
+{ caddress lbase, gbase=NULL, tbase=NULL, abase=NULL;
+  int flags = MAP_FLAGS;
   intptr_t glsize;
   intptr_t minglobal   = 4*SIZEOF_VOIDP K;
   intptr_t minlocal    = 2*SIZEOF_VOIDP K;
   intptr_t mintrail    = 2*SIZEOF_VOIDP K;
   intptr_t minargument = 1*SIZEOF_VOIDP K;
-  
+
   size_alignment = getpagesize();
   while(size_alignment < 4*SIZEOF_VOIDP K)
     size_alignment *= 2;
@@ -1304,9 +1368,18 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   argument = (intptr_t) align_size(argument);
   glsize   = global+local;
 
-  tbase = mmap(NULL, trail,    MAPPROTECT, MAP_FLAGS, mapfd, 0L);
-  abase = mmap(NULL, argument, MAPPROTECT, MAP_FLAGS, mapfd, 0L);
-  gbase = mmap(NULL, glsize,   MAPPROTECT, MAP_FLAGS, mapfd, 0L);
+#ifdef SECURE_GC
+  fixed_bases(&gbase, &tbase);
+  if ( gbase || tbase )
+  { flags |= MAP_FIXED;
+    Sdprintf("Using fixed stack bases at gBase=0x%lx; tBase=0x%lx\n",
+	     gbase, tbase);
+  }
+#endif
+
+  tbase = mmap(tbase, trail,    MAPPROTECT, flags, mapfd, 0L);
+  abase = mmap(abase, argument, MAPPROTECT, MAP_FLAGS, mapfd, 0L);
+  gbase = mmap(gbase, glsize,   MAPPROTECT, flags, mapfd, 0L);
   lbase = addPointer(gbase, global);
 
   if ( tbase == MAP_FAILED || abase == MAP_FAILED || gbase == MAP_FAILED )
@@ -1320,6 +1393,12 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
     fail;
   }
 
+#ifdef SECURE_GC
+  if ( !(flags & MAP_FIXED) && PL_thread_self() == 1 )
+    Sdprintf("Realised export PL_STACK_BASES=\"G0x%lx+T0x%lx\"\n",
+	     (unsigned long)gbase, (unsigned long)tbase);
+#endif
+
 #define INIT_STACK(name, print, base, limit, minsize) \
   DEBUG(1, Sdprintf("%s stack at 0x%x; size = %ld\n", print, base, limit)); \
   init_stack((Stack) &LD->stacks.name, print, base, limit, minsize);
@@ -1333,7 +1412,7 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
 #ifdef O_SEGV_HANDLING
   set_stack_guard_handler(SIGSEGV, _PL_segv_handler);
 #endif
-  
+
   succeed;
 }
 
@@ -1347,6 +1426,7 @@ freeStacks(ARG1_LD)
 
 #undef LD
 #define LD LOCAL_LD
+  gBase--;				/* see initPrologStacks() */
   tlen  = limitStack(trail);
   alen  = limitStack(argument);
   gllen = limitStack(global) + limitStack(local);
@@ -1412,7 +1492,7 @@ mapOrOutOf(Stack s)
 
 #ifndef O_SEGV_HANDLING
 void
-ensureRoomStack(Stack s, size_t bytes)
+ensure_room_stack(Stack s, size_t bytes)
 { while((char *)s->max <= &((char *)s->top)[bytes])
     mapOrOutOf(s);
 }
@@ -1445,6 +1525,8 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
 
   GetSystemInfo(&info);
   size_alignment = info.dwPageSize;
+  while(size_alignment < 4*SIZEOF_VOIDP K)
+    size_alignment *= 2;
 
   local    = max(local,    minlocal + STACK_SIGNAL);
   global   = max(global,   minglobal + STACK_SIGNAL);
@@ -1501,6 +1583,7 @@ void
 freeStacks(PL_local_data_t *ld)
 { DEBUG(1, Sdprintf("[%d]: freeStacks()\n", PL_thread_self()));
 
+  ld->stacks.global.base--;		/* see initPrologStacks() */
   freeStack((Stack)&ld->stacks.global);	/* Region must be entirely committed */
   freeStack((Stack)&ld->stacks.local);	/* or decommitted */
   freeStack((Stack)&ld->stacks.trail);
@@ -1657,39 +1740,6 @@ why does IBM bring computers on the marked that are 10 years out-of-date
 at the moment of announcement?).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-word
-pl_stack_parameter(term_t name, term_t key, term_t old, term_t new)
-{ atom_t a, k;
-  Stack stack = NULL;
-  intptr_t *value = NULL;
-
-  if ( PL_get_atom(name, &a) )
-  { if ( a == ATOM_local )
-      stack = (Stack) &LD->stacks.local;
-    else if ( a == ATOM_global )
-      stack = (Stack) &LD->stacks.global;
-    else if ( a == ATOM_trail )
-      stack = (Stack) &LD->stacks.trail;
-    else if ( a == ATOM_argument )
-      stack = (Stack) &LD->stacks.argument;
-  }
-  if ( !stack )
-    return PL_error(NULL, 0, NULL, ERR_EXISTENCE, ATOM_stack, name);
-
-  if ( PL_get_atom(key, &k) )
-  { if ( k == ATOM_min_free )
-      value = &stack->minfree;
-  }
-  if ( !value )
-    return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_stack_parameter, key);
-
-  if ( !PL_unify_int64(old, *value) ||
-       !PL_get_intptr_ex(new, value) )
-    fail;
-
-  succeed;
-}
-
 
 static void
 init_stack(Stack s, char *name, intptr_t size, intptr_t limit, intptr_t minfree)
@@ -1761,6 +1811,7 @@ freeStacks(ARG1_LD)
 {
 #undef LD
 #define LD LOCAL_LD
+  gBase--;
   if ( gBase ) { free(gBase); gBase = NULL; lBase = NULL; }
   if ( tBase ) { free(tBase); tBase = NULL; }
   if ( aBase ) { free(aBase); aBase = NULL; }
@@ -1782,7 +1833,7 @@ trimStacks() reclaims all unused space on the stack. Note that the trail
 can have references to unused stack. We set the references to point to a
 dummy variable, so no harm  will  be   done.  Setting  it  to NULL would
 require a test in Undo(), which   is time-critical. trim_stacks normally
-isn't. This precaution is explicitly requires  for the trimStacks() that
+isn't. This precaution is explicitly required  for the trimStacks() that
 result from a stack-overflow.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
@@ -1792,7 +1843,6 @@ trimStacks(ARG1_LD)
 #undef LD
 #define LD LOCAL_LD
   TrailEntry te;
-  Word dummy = NULL;
 
   LD->trim_stack_requested = FALSE;
 
@@ -1808,19 +1858,24 @@ trimStacks(ARG1_LD)
 #endif /*O_DYNAMIC_STACKS*/
 #endif /*O_SHIFT_STACKS*/
 
+#ifdef SECURE_GC
+  { Word p;				/* clear the stacks */
+
+    for(p=gTop; p<gMax; p++)
+      *p = 0xbfbfbfbf;
+    for(p=(Word)lTop; p<(Word)lMax; p++)
+      *p = 0xbfbfbfbf;
+  }
+#endif
+
   for(te = tTop; --te >= tBase; )
   { Word p = te->address;
-    
+
     if ( isTrailVal(p) )
       continue;
 
     if ( !onStack(local, p) && !onStack(global, p) )
-    { if ( !dummy )
-      { dummy = allocGlobal(1);
-	setVar(*dummy);
-      }
-
-      te->address = dummy;
+    { te->address = valTermRef(LD->trim.dummy);
     }
   }
 #undef LD
@@ -1832,11 +1887,6 @@ word
 pl_trim_stacks()
 { trimStacks(PASS_LD1);
 
-#ifdef O_DYNAMIC_STACKS
-  gcPolicy((Stack) &LD->stacks.global, GC_FAST_POLICY);
-  gcPolicy((Stack) &LD->stacks.trail,  GC_FAST_POLICY);
-#endif
-
   succeed;
 }
 
@@ -1875,6 +1925,63 @@ freeLocalData(PL_local_data_t *ld)
   clearSegStack(&ld->cycle.stack);
 #endif
 
-  if ( ld->arith.stack.base )
-    PL_free(ld->arith.stack.base);
+  freeArithLocalData(ld);
 }
+
+
+
+		 /*******************************
+		 *	     PREDICATES		*
+		 *******************************/
+
+static
+PRED_IMPL("set_prolog_stack", 3, set_prolog_stack, 0)
+{ atom_t a, k;
+  Stack stack = NULL;
+
+  term_t name  = A1;
+  term_t prop  = A2;
+  term_t value = A3;
+
+  if ( PL_get_atom(name, &a) )
+  { if ( a == ATOM_local )
+      stack = (Stack) &LD->stacks.local;
+    else if ( a == ATOM_global )
+      stack = (Stack) &LD->stacks.global;
+    else if ( a == ATOM_trail )
+      stack = (Stack) &LD->stacks.trail;
+    else if ( a == ATOM_argument )
+      stack = (Stack) &LD->stacks.argument;
+  }
+  if ( !stack )
+    return PL_error(NULL, 0, NULL, ERR_EXISTENCE, ATOM_stack, name);
+
+  if ( PL_get_atom_ex(prop, &k) )
+  { if ( k == ATOM_low )
+      return PL_get_intptr_ex(value, &stack->small);
+    if ( k == ATOM_factor )
+      return PL_get_integer_ex(value, &stack->factor);
+    if ( k == ATOM_min_free )
+    {
+#ifdef O_SHIFT_STACKS
+      return PL_get_intptr_ex(value, &stack->minfree);
+#else
+      succeed;
+#endif
+    }
+
+    return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_stack_parameter, prop);
+  }
+
+  fail;
+}
+
+
+		 /*******************************
+		 *      PUBLISH PREDICATES	*
+		 *******************************/
+
+BeginPredDefs(setup)
+  PRED_DEF("set_prolog_stack", 3, set_prolog_stack, 0)
+  PRED_DEF("$on_signal", 4, on_signal, 0)
+EndPredDefs
