@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -43,7 +43,7 @@ MT:
 
 Multithreading is supported through  Slock()   and  Sunlock(). These are
 recursive locks. If a stream handle  might   be  known to another thread
-locking is required. 
+locking is required.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #ifdef MD
@@ -61,7 +61,7 @@ locking is required.
 #define PL_KERNEL 1
 #include <wchar.h>
 typedef wchar_t pl_wchar_t;
-#include "pl-stream.h"
+#include "SWI-Stream.h"
 #include "pl-utf8.h"
 #include <sys/types.h>
 #ifdef HAVE_SYS_TIME_H
@@ -213,6 +213,7 @@ S__setbuf(IOSTREAM *s, char *buffer, size_t size)
 
 	  free(newunbuf);
 	  errno = oldeno;
+	  S__seterror(s);
 	  return -1;
 	}
       }
@@ -237,8 +238,8 @@ S__setbuf(IOSTREAM *s, char *buffer, size_t size)
 
 void
 Ssetbuffer(IOSTREAM *s, char *buffer, size_t size)
-{ S__setbuf(s, buffer, size);
-  s->flags &= ~SIO_USERBUF;
+{ if ( S__setbuf(s, buffer, size) != (size_t)-1 )
+    s->flags &= ~SIO_USERBUF;
 }
 
 
@@ -281,16 +282,16 @@ print_trace(void)
   size_t size;
   char **strings;
   size_t i;
-     
+
   size = backtrace(array, sizeof(array)/sizeof(void *));
   strings = backtrace_symbols(array, size);
-     
+
   printf(" Stack:");
   for(i = 1; i < size; i++)
   { printf("\n\t[%ld] %s", (long)i, strings[i]);
   }
   printf("\n");
-       
+
   free(strings);
 }
 #endif /*DEBUG_IO_LOCKS*/
@@ -330,8 +331,8 @@ StryLock(IOSTREAM *s)
 }
 
 
-static int
-S__unlock(IOSTREAM *s)
+int
+Sunlock(IOSTREAM *s)
 { int rval = 0;
 
 #ifdef DEBUG_IO_LOCKS
@@ -350,15 +351,7 @@ S__unlock(IOSTREAM *s)
   { assert(0);
   }
 
-  return rval;
-}
-
-
-int
-Sunlock(IOSTREAM *s)
-{ int rval = S__unlock(s);
   SUNLOCK(s);
-
   return rval;
 }
 
@@ -368,7 +361,7 @@ Sunlock(IOSTREAM *s)
 		 *******************************/
 
 /* return values: -1: error, else #bytes written */
-			 
+
 static ssize_t
 S__flushbuf(IOSTREAM *s)
 { char *from, *to;
@@ -420,16 +413,15 @@ S__flushbufc(int c, IOSTREAM *s)
   } else
   { if ( s->flags & SIO_NBUF )
     { char chr = (char)c;
-    
+
       if ( (*s->functions->write)(s->handle, &chr, 1) != 1 )
-      { s->flags |= SIO_FERR;
+      { S__seterror(s);
 	c = -1;
       }
     } else
     { if ( S__setbuf(s, NULL, 0) == (size_t)-1 )
-      { s->flags |= SIO_FERR;
 	c = -1;
-      } else
+      else
 	*s->bufp++ = (char)c;
     }
   }
@@ -462,7 +454,7 @@ S__fillbuf(IOSTREAM *s)
     { fd_set wait;
       struct timeval time;
       int rc;
-      
+
       time.tv_sec  = s->timeout / 1000;
       time.tv_usec = (s->timeout % 1000) * 1000;
       FD_ZERO(&wait);
@@ -474,7 +466,7 @@ S__fillbuf(IOSTREAM *s)
 
       for(;;)
       { rc = select(fd+1, &wait, NULL, NULL, &time);
-	
+
 	if ( rc < 0 && errno == EINTR )
 	{ if ( PL_handle_signals() < 0 )
 	  { errno = EPLEXCEPTION;
@@ -548,6 +540,7 @@ S__fillbuf(IOSTREAM *s)
       } else if ( errno == EWOULDBLOCK )
       { s->bufp = s->buffer;
 	s->limitp = s->buffer;
+	S__seterror(s);
 	return -1;
 #endif
       } else
@@ -731,7 +724,7 @@ reperror(int c, IOSTREAM *s)
     { if ( put_byte(*q, s) < 0 )
 	return -1;
     }
-	
+
     return c;
   }
 
@@ -789,7 +782,7 @@ put_code(int c, IOSTREAM *s)
     case ENC_UTF8:
     { char buf[6];
       char *p, *end;
-      
+
       if ( c < 128 )
 	goto simple;
 
@@ -822,7 +815,7 @@ put_code(int c, IOSTREAM *s)
       { if ( put_byte(*q++, s) < 0 )
 	  return -1;
       }
-      
+
       break;
     }
     case ENC_UNKNOWN:
@@ -966,7 +959,7 @@ retry:
 	code = UTF8_FBV(c,extra);
 	for( ; extra > 0; extra-- )
 	{ int c2 = get_byte(s);
-	  
+
 	  if ( !ISUTF8_CB(c2) )
 	  { Sseterr(s, SIO_WARN, "Illegal UTF-8 continuation");
 	    c = UTF8_MALFORMED_REPLACEMENT;
@@ -1016,7 +1009,7 @@ retry:
 	    goto out;
 	  } else
 	  { Sseterr(s, SIO_WARN, "EOF in UCS character");
-	    c = UTF8_MALFORMED_REPLACEMENT; 
+	    c = UTF8_MALFORMED_REPLACEMENT;
 	    goto out;
 	  }
 	}
@@ -1219,16 +1212,16 @@ Sfread(void *data, size_t size, size_t elms, IOSTREAM *s)
 
       if ( (c = Sgetc(s)) == EOF )
 	break;
-      
+
       *buf++ = c & 0xff;
     }
-  } else 
+  } else
   { while(chars > 0)
     { int c;
 
       if ( s->bufp < s->limitp )
       { size_t avail = s->limitp - s->bufp;
-	
+
 	if ( chars <= avail )
 	{ memcpy(buf, s->bufp, chars);
 	  s->bufp += chars;
@@ -1240,7 +1233,7 @@ Sfread(void *data, size_t size, size_t elms, IOSTREAM *s)
 	  s->bufp += avail;
 	}
       }
-      
+
       if ( (c = S__fillbuf(s)) == EOF )
 	break;
 
@@ -1248,7 +1241,7 @@ Sfread(void *data, size_t size, size_t elms, IOSTREAM *s)
       chars--;
     }
   }
-  
+
   return (size*elms - chars)/size;
 }
 
@@ -1262,7 +1255,7 @@ Sfwrite(const void *data, size_t size, size_t elms, IOSTREAM *s)
   { if ( Sputc(*buf++, s) < 0 )
       break;
   }
-  
+
   return (size*elms - chars)/size;
 }
 
@@ -1397,19 +1390,22 @@ Sfeof(IOSTREAM *s)
   s->bufp--;
   return FALSE;
 }
-    
+
 
 static int
 S__seterror(IOSTREAM *s)
-{ if ( s->functions->control )
+{ s->io_errno = errno;
+
+  if ( !(s->flags&SIO_CLOSING) &&	/* s->handle is already invalid */
+       s->functions->control )
   { char *msg;
 
     if ( (*s->functions->control)(s->handle,
-				  SIO_LASTERROR, 
+				  SIO_LASTERROR,
 				  (void *)&msg) == 0 )
     { Sseterr(s, SIO_FERR, msg);
       return 0;
-    } 
+    }
   }
 
   s->flags |= SIO_FERR;
@@ -1421,7 +1417,7 @@ int
 Sferror(IOSTREAM *s)
 { return (s->flags & SIO_FERR) != 0;
 }
-    
+
 
 int
 Sfpasteof(IOSTREAM *s)
@@ -1432,6 +1428,7 @@ Sfpasteof(IOSTREAM *s)
 void
 Sclearerr(IOSTREAM *s)
 { s->flags &= ~(SIO_FEOF|SIO_WARN|SIO_FERR|SIO_FEOF2|SIO_TIMEOUT|SIO_CLEARERR);
+  s->io_errno = 0;
   Sseterr(s, 0, NULL);
 }
 
@@ -1476,7 +1473,7 @@ Ssetenc(IOSTREAM *s, IOENC enc, IOENC *old)
 
   if ( s->functions->control )
   { if ( (*s->functions->control)(s->handle,
-				  SIO_SETENCODING, 
+				  SIO_SETENCODING,
 				  (void *)&enc) != 0 )
       return -1;
   }
@@ -1554,6 +1551,7 @@ Ssize(IOSTREAM *s)
   }
 
   errno = ESPIPE;
+  S__seterror(s);
   return -1;
 }
 
@@ -1578,7 +1576,7 @@ Sseek64(IOSTREAM *s, int64_t pos, int whence)
     if ( now != -1 )
     { int64_t newpos;
       char *nbufp = (char *)-1;
-    
+
       if ( whence == SIO_SEEK_CUR )
       { nbufp = s->bufp + pos;
 	newpos = now + pos;
@@ -1599,11 +1597,12 @@ Sseek64(IOSTREAM *s, int64_t pos, int whence)
 
   if ( !s->functions->seek && !s->functions->seek64 )
   { errno = ESPIPE;
+    S__seterror(s);
     return -1;
   }
 
   Sflush(s);
-    
+
   s->bufp   = s->buffer;
   if ( (s->flags & SIO_INPUT) )
     s->limitp = s->buffer;
@@ -1612,18 +1611,19 @@ Sseek64(IOSTREAM *s, int64_t pos, int whence)
   { pos += Stell64(s);
     whence = SIO_SEEK_SET;
   }
-  
+
   if ( s->functions->seek64 )
     pos = (*s->functions->seek64)(s->handle, pos, whence);
   else if ( pos <= LONG_MAX )
     pos = (*s->functions->seek)(s->handle, (long)pos, whence);
   else
   { errno = EINVAL;
+    S__seterror(s);
     return -1;
   }
-  
+
   if ( pos < 0 )
-  { errno = EINVAL;
+  { S__seterror(s);
     return -1;
   }
 
@@ -1675,6 +1675,7 @@ Stell64(IOSTREAM *s)
     return pos;
   } else
   { errno = EINVAL;
+    S__seterror(s);
     return -1;
   }
 }
@@ -1684,10 +1685,13 @@ long
 Stell(IOSTREAM *s)
 { int64_t pos = Stell64(s);
 
+  if ( pos == -1 )
+    return -1;
   if ( pos <= LONG_MAX )
     return (long) pos;
 
   errno = EINVAL;
+  S__seterror(s);
   return -1;
 }
 
@@ -1708,7 +1712,7 @@ Sclose(IOSTREAM *s)
 { int rval = 0;
 
   if ( s->magic != SIO_MAGIC )		/* already closed!? */
-  { errno = EINVAL;
+  { s->io_errno = errno = EINVAL;
     return -1;
   }
 
@@ -1738,11 +1742,12 @@ Sclose(IOSTREAM *s)
   }
 #endif
   if ( s->functions->close && (*s->functions->close)(s->handle) < 0 )
-  { s->flags |= SIO_FERR;
+  { S__seterror(s);
     rval = -1;
   }
+
   while(s->locks > 0)			/* remove buffer-locks */
-  { int rc = S__unlock(s);
+  { int rc = Sunlock(s);
 
     if ( rval == 0 )
       rval = rc;
@@ -1750,7 +1755,6 @@ Sclose(IOSTREAM *s)
   if ( rval < 0 )
     reportStreamError(s);
   run_close_hooks(s);			/* deletes Prolog registration */
-
   SUNLOCK(s);
 
 #ifdef O_PLMT
@@ -1762,6 +1766,8 @@ Sclose(IOSTREAM *s)
 #endif
 
   s->magic = SIO_CMAGIC;
+  if ( s->message )
+    free(s->message);
   if ( !(s->flags & SIO_STATIC) )
     free(s);
 
@@ -1817,8 +1823,7 @@ Sgets(char *buf)
 
 int
 Sfputs(const char *q, IOSTREAM *s)
-{ 
-  for( ; *q; q++)
+{ for( ; *q; q++)
   { if ( Sputcode(*q&0xff, s) < 0 )
       return EOF;
   }
@@ -2159,7 +2164,7 @@ Svsprintf(char *buf, const char *fm, va_list args)
   s.buffer    = buf;
   s.flags     = SIO_FBUF|SIO_OUTPUT;
   s.encoding  = ENC_ISO_LATIN_1;
-  
+
   if ( (rval = Svfprintf(&s, fm, args)) >= 0 )
     *s.bufp = '\0';
 
@@ -2246,7 +2251,7 @@ Svfscanf(IOSTREAM *s, const char *fm, va_list args)
 	  continue;
 	}
       }
-	
+
       if ( *fm != '[' && *fm != c )
 	while(isblank(c))
 	  c = GET(s);
@@ -2399,13 +2404,13 @@ Svfscanf(IOSTREAM *s, const char *fm, va_list args)
 	      { float *fp = va_arg(args, float *);
 		*fp = v;
 		break;
-	      }	
+	      }
 	      case SZ_LONG:
 	      { double *fp = va_arg(args, double *);
 		*fp = v;
 		break;
 	      }
-	    }  
+	    }
 	    done++;
 	  }
 
@@ -2415,7 +2420,7 @@ Svfscanf(IOSTREAM *s, const char *fm, va_list args)
 	case 's':
 	  if ( !supress )
 	  { char *sp = va_arg(args, char *);
-	    
+
 	    while(!isblank(c) && field_width-- != 0)
 	    { *sp++ = c;
 	      c = GET(s);
@@ -2435,7 +2440,7 @@ Svfscanf(IOSTREAM *s, const char *fm, va_list args)
 	  continue;
 	case '[':
 	{ char set[256];
-	  
+
 	  memset(set, 0, sizeof(set));
 	  fm++;
 	  if ( *fm == ']' )
@@ -2446,7 +2451,7 @@ Svfscanf(IOSTREAM *s, const char *fm, va_list args)
 	  }
 	  while(*fm != ']')
 	  { if ( *fm == '-' )
-	      
+
 	  }
 	}
       }
@@ -2479,7 +2484,7 @@ Link two streams in a pipeline,  where   filter  filters data for stream
 `parent'. If parent is an output steam we have
 
 	application --> filter --> parent -->
-	  
+
 If parent is an input stream we have
 
 	--> parent --> filter --> application
@@ -2683,8 +2688,13 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
     recursiveMutexInit(s->mutex);
   }
 #endif
-  if ( (fd = Sfileno(s)) >= 0 && isatty(fd) )
-    s->flags |= SIO_ISATTY;
+  if ( (fd = Sfileno(s)) >= 0 )
+  { if ( isatty(fd) )
+      s->flags |= SIO_ISATTY;
+#if defined(F_SETFD) && defined(FD_CLOEXEC)
+    fcntl(fd, F_SETFD, FD_CLOEXEC);
+#endif
+  }
 
   return s;
 }
@@ -2773,7 +2783,7 @@ Sopen_file(const char *path, const char *how)
     return NULL;
 
   if ( lock )
-  { 
+  {
 #ifdef FCNTL_LOCKS
     struct flock buf;
 
@@ -3043,7 +3053,7 @@ Sread_memfile(void *handle, char *buf, size_t size)
     else
       size = mf->size - mf->here;
   }
-  
+
   memcpy(buf, &(*mf->buffer)[mf->here], size);
   mf->here += size;
 
@@ -3086,7 +3096,7 @@ Sclose_memfile(void *handle)
   { free(mf);
     return 0;
   }
-	 
+
   errno = EINVAL;			/* not opened */
   return -1;
 }
@@ -3387,7 +3397,7 @@ Sreset(void)
   if ( (s=Serror) && s->magic == SIO_MAGIC )
   { s->bufp = s->buffer;
   }
-} 
+}
 
 
 void
@@ -3407,6 +3417,14 @@ Scleanup(void)
 
     s->bufp = s->buffer;		/* avoid actual flush */
     S__removebuf(s);
+    if ( S__iob[i].mutex )
+    { recursiveMutex *m = S__iob[i].mutex;
+
+      S__iob[i].mutex = NULL;
+      recursiveMutexDelete(m);
+      free(m);
+    }
+
     *s = S__iob0[i];			/* re-initialise */
   }
 }
