@@ -1,4 +1,4 @@
-/*  $Id: pl-tai.c,v 1.27 2007/07/05 12:22:32 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -159,7 +159,7 @@ origin;
 static int
 unify_taia(term_t t, struct taia *taia)
 { double d = (double)((int64_t)taia->sec.x - TAI_UTC_OFFSET);
-  
+
   d += taia->nano / 1e9;
 
   return PL_unify_float(t, d);
@@ -334,15 +334,15 @@ PRED_IMPL("stamp_date_time", 3, stamp_date_time, 0)
       { time_t unixt;
 	int64_t ut64;
 	struct tm tm;
-  
+
 	utcoffset = tz_offset();
-  
+
 	ut64 = taia.sec.x - TAI_UTC_OFFSET;
 	unixt = (time_t) ut64;
-  
+
 	if ( (int64_t)unixt == ut64 )
 	{ double ip;
-  
+
 	  localtime_r(&unixt, &tm);
 	  sec = (double)tm.tm_sec + modf(argsec, &ip);
 	  ct.date.year  = tm.tm_year+1900;
@@ -508,7 +508,7 @@ foutstra(const char *str, IOSTREAM *fd)
 { wchar_t wbuf[256];
   size_t n;
 
-  if ( (n = mbstowcs(wbuf, str, sizeof(wbuf)/sizeof(wbuf[0])-1)) != (size_t)-1 ) 
+  if ( (n = mbstowcs(wbuf, str, sizeof(wbuf)/sizeof(wbuf[0])-1)) != (size_t)-1 )
   { wchar_t *p;
 
     for(p=wbuf; n-- > 0; p++)
@@ -521,7 +521,7 @@ static const char *abbred_weekday[] =
 { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
 static const char *weekday[] =
 { "Sunday", "Monday", "Tuesday", "Wednesday",
-  "Thursday", "Friday", "Satuerday" };
+  "Thursday", "Friday", "Saturday" };
 static const char *abbred_month[] =
 { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
@@ -578,7 +578,7 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm, int posix)
 	  { char fmt[3];
 	    char buf[256];
 	    size_t n;
-  
+
 	    fmt[0] = '%';
 	    fmt[1] = (char)c;
 	    fmt[2] = EOS;
@@ -618,7 +618,7 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm, int posix)
 	  case 'g':
 	  case 'V':
 	  { int year, days;
-	    
+
 	    cal_ftm(ftm, HAS_STAMP|HAS_WYDAY);
 	    year = ftm->tm.tm_year+1900;
 	    days = iso_week_days(ftm->tm.tm_yday, ftm->tm.tm_wday);
@@ -636,7 +636,7 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm, int posix)
 		days = d;
 	      }
 	    }
-	    
+
 	    switch(c)
 	    { case 'g':
 		OUT2DIGITS(fd, (year % 100 + 100) % 100);
@@ -709,7 +709,7 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm, int posix)
 	  }
 	  case 'U':			/* 00..53 weeknumber */
 	  { int wk;
-  
+
 	    cal_ftm(ftm, HAS_WYDAY);
 	    wk = (ftm->tm.tm_yday - (ftm->tm.tm_yday - ftm->tm.tm_wday + 7) % 7 + 7) / 7;
 	    OUT2DIGITS(fd, wk);
@@ -721,7 +721,7 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm, int posix)
 	    break;
 	  case 'W':			/* 00..53 monday-based week number */
 	  { int wk;
-  
+
 	    cal_ftm(ftm, HAS_WYDAY);
 	    wk = (ftm->tm.tm_yday - (ftm->tm.tm_yday - ftm->tm.tm_wday + 8) % 7 + 7) / 7;
 	    OUT2DIGITS(fd, wk);
