@@ -1,11 +1,11 @@
-/*  $Id: pl-term.c,v 1.35 2007/01/08 14:16:58 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2008, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -22,7 +22,9 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+#define bool pl_bool			/* avoid conflict with curses */
 #include "pl-incl.h"
+#undef bool
 
 #define LOCK()   PL_LOCK(L_TERM)
 #define UNLOCK() PL_UNLOCK(L_TERM)
@@ -42,12 +44,9 @@ windowing!
 #include <sys/ioctl.h>
 #endif
 
-extern int  tgetent();
-extern int  tgetnum();
-extern int  tgetflag();
-extern char *tgetstr();
-extern char *tgoto();
-extern int  tputs();
+#undef clear				/* conflicts */
+#include <curses.h>
+#include <term.h>
 
 #define MAX_TERMBUF	1024		/* Confirming manual */
 #define STAT_START	0
@@ -90,7 +89,7 @@ resetTerm()
 /* MT: Locked by calling lookupEntry()
 */
 
-static bool
+static int
 initTerm(void)
 { if ( term_initialised == STAT_START )
   { char term[100];
@@ -113,7 +112,7 @@ initTerm(void)
     switch( tgetent(buf_area, term) )
     { case -1:
       { term_t t = PL_new_term_ref();
-	
+
 	PL_put_atom_chars(t, "termcap");
 	PL_error(NULL, 0, "tgetent() cannot read database",
 		 ERR_PERMISSION,
@@ -148,7 +147,7 @@ lookupEntry(atom_t name, atom_t type)
 
   LOCK();
   if ( (s = lookupHTable(capabilities, (void*)name)) == NULL )
-  { if ( initTerm() == FALSE )
+  { if ( !initTerm() )
     { e = NULL;
       goto out;
     }
@@ -164,13 +163,13 @@ lookupEntry(atom_t name, atom_t type)
       if ( (n = tgetnum(stringAtom(name))) != -1 )
         e->value  = consInt(n);
     } else if ( type == ATOM_bool )
-    { bool b;
-    
+    { int b;
+
       if ( (b = tgetflag(stringAtom(name))) != -1 )
         e->value = (b ? ATOM_on : ATOM_off);
     } else if ( type == ATOM_string )
     { char *s;
-    
+
       if ( (s = tgetstr(stringAtom(name), &string_area)) != NULL )
         e->value  = PL_new_atom(s);	/* locked: ok */
     } else
@@ -183,12 +182,12 @@ lookupEntry(atom_t name, atom_t type)
     addHTable(capabilities, (void *)name, e);
   } else
     e = (Entry)s->value;
-    
+
 out:
   UNLOCK();
   return e;
 }
-      
+
 
 word
 pl_tty_get_capability(term_t name, term_t type, term_t value)
@@ -205,11 +204,11 @@ pl_tty_get_capability(term_t name, term_t type, term_t value)
 
   fail;
 }
-  
+
 
 static int
 tputc(int chr)
-{ return Sputc(chr, Suser_output); 
+{ return Sputc(chr, Suser_output);
 }
 
 
@@ -266,7 +265,7 @@ pl_tty_size(term_t r, term_t c)
 #ifdef TIOCGSIZE
   struct ttysize ws;
   iorval = ioctl(0, TIOCGSIZE, &ws);
-	
+
   rows = ws.ts_lines;
   cols = ws.ts_cols;
 #else
