@@ -76,7 +76,7 @@ term_expansion((:- if(G)), []) :-
 term_expansion((:- else), []) :-
 	(   retract(user:include_code(X))
 	->  (   X == true
-	    ->  X2 = false 
+	    ->  X2 = false
 	    ;   X2 = true
 	    ),
 	    asserta(user:include_code(X2))
@@ -88,7 +88,7 @@ term_expansion((:- endif), []) :-
 term_expansion(_, []) :-
 	user:include_code(X), !,
 	X == false.
-	    
+
 
 		 /*******************************
 		 *	      SYNTAX		*
@@ -150,7 +150,7 @@ syntax(latin-1) :-
 	atom_codes(A, [247]),
 	atom_to_term(A, T, []),
 	atom_codes(T, [247]).
-	
+
 
 		 /*******************************
 		 *	       WRITE		*
@@ -232,7 +232,7 @@ occurs_check(simple-2) :-
 	unify_with_occurs_check(_A, _B).
 occurs_check(attvar-1) :-		% test wakeup
 	freeze(X, X = Y),
-	unify_with_occurs_check(X, a), 
+	unify_with_occurs_check(X, a),
 	Y == a.
 occurs_check(attvar-2) :-		% test occurs-check
 	freeze(A, true),
@@ -370,7 +370,7 @@ arithmetic(truncate-1) :-
 	1 is truncate(1.9),
 	-1 is truncate(-1.1),
 	-1 is truncate(-1.9).
-:- if(current_prolog_flag(bounded, false)). 
+:- if(current_prolog_flag(bounded, false)).
 arithmetic(floor-2) :-
 	A is floor(9223372036854775808.000000),
 	A == 9223372036854775808.
@@ -519,7 +519,15 @@ idiv(Dd,Dr,Iq):-
         rational(Q,Qt,Qn),
         Iq is Qt//Qn.
 
+dec(X, Y) :-
+	Y is X - 1.
 
+gmp(add-promote1) :-
+	A is 1 + 9223372036854775807,
+	A =:= 9223372036854775808.
+gmp(add-promote2) :-
+	A is -9223372036854775808 + -1,
+	A =:= -9223372036854775809.
 gmp(neg-1) :-				% check conversion of PLMININT
 	A is -9223372036854775808,
 	-A =:= 9223372036854775808.
@@ -530,7 +538,7 @@ gmp(abs-1) :-
 	abs(A) =:= 9223372036854775808.
 gmp(sign-1) :-
 	-1 =:= sign(-5 rdiv 3),
-	0 =:= sign(0 rdiv 1), 
+	0 =:= sign(0 rdiv 1),
 	1 =:= sign(2 rdiv 7),
 	fac(60, X),
 	-1 =:= sign(-X),
@@ -590,7 +598,7 @@ gmp(arith-1) :-
 	Y is B * A * B * A * B * A,
 	integer(Y),
 	R is X / Y,
-	R == A. 
+	R == A.
 gmp(pow-1) :-
 	A is 10**50, integer(A).
 gmp(pow-2) :-
@@ -628,6 +636,9 @@ gmp(rational-1) :-				% IEEE can represent 0.25
 gmp(rational-2) :-
 	A is 2 rdiv 4,
 	rational(A, 1, 2).
+gmp(rational-3) :-
+	dec(6 rdiv 5, X),
+	X == 1 rdiv 5.
 gmp(rationalize-1) :-
 	A is rationalize(0.0), A == 0,
 	B is rationalize(0.1), B == 1 rdiv 10,
@@ -721,6 +732,10 @@ gmp(length) :-
 	N is 1<<66,
 	catch(length(_L, N), Error, true),
 	Error = error(resource_error(stack), global).
+gmp(ar_add_ui) :-			% check realloc of gmp number
+	A = 1000000000000000000000,
+	X is A+1,
+	X == 1000000000000000000001.
 
 :- endif.
 
@@ -806,7 +821,7 @@ meta(call-11) :-
 	flag(a, 3, Old).
 meta(call-12) :-
 	catch(call(1), E, true),
-	error(E, type_error(callable, 1)).
+	error(E, type_error(callable, _)).
 meta(apply-1) :-
 	apply(=, [a,a]).
 meta(apply-2) :-
@@ -914,7 +929,7 @@ depth_limit(depth-2) :-
 depth_limit(depth-3) :-
 	call_with_depth_limit(dl_det(10), 9, depth_limit_exceeded).
 depth_limit(ndet-1) :-
-	findall(X, 
+	findall(X,
 		call_with_depth_limit(dl_ndet(5), 10, X),
 		L),
 	L = [5, depth_limit_exceeded].
@@ -1316,7 +1331,7 @@ record(erase-2) :-
 		 *******************************/
 
 %	compiler
-%	
+%
 %	This suite tests whether all data-types can be compiled properly
 %	and handled by the decompiler.
 
@@ -1355,6 +1370,13 @@ compiler(assert-6) :-
 	T =.. [x|A],
 	assert(T),
 	T.
+compiler(assert-7) :-
+	Body = (a(A), A=[B|C], a(B), a(C)),
+	Clause = (at:-Body),
+	assert(Clause, Ref),
+	clause(_H, BC, Ref),
+	BC =@= Body,
+	erase(Ref).
 
 unify(X, X).
 
@@ -1543,6 +1565,7 @@ catchme :-
 
 undef :-
 	'this is not defined'.
+undef.					% avoid last-call optimization for context
 
 tcatch :-
 	catch(ierror, E, throw(ok(E))).
@@ -1564,13 +1587,12 @@ exception(call-4) :-
 	X = a.
 exception(call-5) :-
 	catch(throwit, _, catchme).
-
 exception(context-1) :-
 	catch(functor(_,_,_), E, true),
-	error_pred(E, functor/3).
+	error_context(E, functor/3).
 exception(context-2) :-
 	catch(undef, E, true),
-	error_pred(E, undef/0).
+	error_context(E, undef/0).
 exception(catch-gc) :-
 	catch(tcatch, E, true),
 	subsumes_chk(ok(error(_,_)), E).
@@ -1597,7 +1619,7 @@ string_overflow(StringList) :-
 string_overflow2([H|T]) :-		% Causes PL_throw() overflow
 	format(string(H), '~txx~1000000|', []),
 	string_overflow2(T).
-		
+
 
 resource(stack-1) :-
 	catch(local_overflow, E, true),
@@ -1615,6 +1637,9 @@ resource(stack-4) :-
 	E1 = error(resource_error(stack), global),
 	catch(global_overflow(_), E2, true),
 	E2 = error(resource_error(stack), global).
+resource(stack-5) :-
+	catch(length(_L,10000000), E, true),
+	E = error(resource_error(stack), global).
 
 
 		 /*******************************
@@ -1690,13 +1715,18 @@ intoverflow(syntax-2) :-
 	name(X, Chars),
 	float(X),
 	X =:= Float.
-	
+
 
 		 /*******************************
 		 *	ATTRIBUTED VARIABLES	*
 		 *******************************/
 
+:- dynamic
+	woken/2.
+
 test:attr_unify_hook(_Att, _Val).
+woken:attr_unify_hook(Att, Var) :-
+	assert(woken(Att, Var)).
 
 u_predarg(predarg).
 u_termarg(f(termarg)).
@@ -1830,6 +1860,11 @@ avar(order-1) :-			% attributes do not change standard
 	;   put_attr(A, test, x),
 	    A @> B
 	).
+avar(nowake-1) :-
+	retractall(woken(_,_)),
+	put_attr(V, woken, 10),
+	\+ (V-a = 10-b),
+	\+ woken(_,_).
 
 
 		 /*******************************
@@ -1932,7 +1967,7 @@ copy_term(av-4) :-
 	X = ok,
 	Done == true,
 	get_attr(Y, freeze, Att),
-	Att = (user:true, user:(D2=true)),
+	Att = '$and'(user:true, user:(D2=true)),
 	var(D2),
 	Y = ok,
 	D2  == true.
@@ -1958,16 +1993,11 @@ copy_term(nat-2) :-			% cyclic term
 
 
 		 /*******************************
-		 *	     HASH-TERM		*
+		 *	     TERM-HASH		*
 		 *******************************/
 
-%	NOTE: these numbers are for 32-bit platforms. Different value of
-%	max_tagged_integer is main reason why the result is different on
-%	64-bit platforms. What to do? Use the   32-bit  limit also on 64
-%	bit platforms?  For now the test is disabled on 64-bit machines.
-
 term_hash(simple-1) :-
-	term_hash(aap, 8246445).
+	term_hash(aap, 9270206).
 term_hash(simple-2) :-			% small int
 	term_hash(42, X),
 	memberchk(X, [ 12280004,	% little endian
@@ -1988,12 +2018,12 @@ term_hash(simple-5) :-
 	string_to_list(S, "hello world"),
 	term_hash(S, 13985775).
 term_hash(compound-1) :-
-	term_hash(hello(world), 2391568). 
+	term_hash(hello(world), 12599352).
 term_hash(compound-2) :-
 	X = x(a),
-	term_hash(hello(X, X), 4126440). 
+	term_hash(hello(X, X), 5826661).
 term_hash(compound-3) :-
-	term_hash(hello(x(a), x(a)), 4126440). 
+	term_hash(hello(x(a), x(a)), 5826661).
 
 
 		 /*******************************
@@ -2089,7 +2119,7 @@ popen(cat-1) :-
 	;   File = 'pltest.txt',
 	    Text = 'Hello World',
 	    Cmd = cat,
-	    concat_atom([Cmd, ' > ', File], Command),
+	    atomic_list_concat([Cmd, ' > ', File], Command),
 	    open(pipe(Command), write, Fd),
 	    format(Fd, '~w', [Text]),
 	    close(Fd),
@@ -2390,6 +2420,8 @@ th_do_something :-
 th_check_done :-
 	findall(X, retract(th_data(X)), [1,2,3,4,5]).
 
+thread(queue-1) :-			% crash if no queue exists?
+	findall(Q, message_queue_property(Q,_), _).
 thread(join-1) :-
 	thread_create(th_do_something, Id, []),
 	thread_join(Id, Exit),
@@ -2421,7 +2453,7 @@ thread(status-1) :-
 thread(create_error-1) :-
 	catch(thread_create(true, _, [local(a)]), E, true),
 	E = error(domain_error(thread_option, local(a)), _).
-	
+
 
 		 /*******************************
 		 *	 MUTEX HANDLING		*
@@ -2432,7 +2464,8 @@ mutex(trylock-1) :-
 	thread_self(Main),
 	thread_create(( mutex_lock(Mutex),
 			thread_send_message(Main, locked),
-			thread_get_message(_)),
+			thread_get_message(_),
+			mutex_unlock(Mutex)),
 		      Id, []),
 	thread_get_message(locked),
 	\+ mutex_trylock(Mutex),
@@ -2485,7 +2518,7 @@ run_test_script(Script) :-
 
 run_test_scripts(Directory) :-
 	(   script_dir(ScriptDir),
-	    concat_atom([ScriptDir, /, Directory], Dir),
+	    atomic_list_concat([ScriptDir, /, Directory], Dir),
 	    exists_directory(Dir)
 	->  true
 	;   Dir = Directory
@@ -2585,7 +2618,7 @@ testset(mutex) :-
 testset(resource).
 
 %	unicode_file_locale/0
-%	
+%
 %	True if out filesystem can   handle Unicode filenames. Difficult
 %	to have a good test.
 
@@ -2599,7 +2632,7 @@ unicode_file_locale :-
 	).
 
 %	wide_character_types
-%	
+%
 %	True if the  character  classification   routines  work  on wide
 %	characters. Hard to say when this is  the case. On some machines
 %	the wide character versions always work,  on others only for the
@@ -2609,7 +2642,7 @@ wide_character_types :-
 	current_prolog_flag(encoding, utf8), !.
 
 %	testdir(Dir)
-%	
+%
 %	Enumerate directories holding tests.
 
 testdir('Tests/core').
@@ -2677,7 +2710,7 @@ runtest(Name) :-
 	fail.
 runtest(_) :-
 	format(' done.~n').
-	
+
 nth_clause_head(Head, R) :-
 	nth_clause(Head, _N, R),
 	clause(Head, _, R).
@@ -2719,19 +2752,22 @@ blocked(Reason) :-
 
 
 %	error(+Exception, +Expected)
-%	
+%
 %	Check whether the correct exception  is thrown, disregarding the
 %	2nd context argument.
 
 error(error(Ex, _Ctx), Expected) :-
-	Ex =@= Expected, !.
+	subsumes_chk(Expected, Ex), !.
 error(error(Ex, _Ctx), Expected) :-
 	format('~NWrong exception: ~p (expected ~p)~n', [Ex, Expected]),
 	fail.
 
-error_pred(error(_, context(Pred, _)), Pred).
-error_pred(error(_, context(Module:Pred, _)), Pred) :-
-	hidden_module(Module).
+error_context(error(_, context(Pred, _)), Pred) :- !.
+error_context(error(_, context(Module:Pred, _)), Pred) :-
+	hidden_module(Module), !.
+error_context(Error, _Pred) :-
+	format('Wrong error context: ~q~n', [Error]),
+	fail.
 
 hidden_module(user) :- !.
 hidden_module(system) :- !.
